function zj(e,t){for(var r=0;r<t.length;r++){const a=t[r];if(typeof a!="string"&&!Array.isArray(a)){for(const u in a)if(u!=="default"&&!(u in e)){const o=Object.getOwnPropertyDescriptor(a,u);o&&Object.defineProperty(e,u,o.get?o:{enumerable:!0,get:()=>a[u]})}}}return Object.freeze(Object.defineProperty(e,Symbol.toStringTag,{value:"Module"}))}(function(){const t=document.createElement("link").relList;if(t&&t.supports&&t.supports("modulepreload"))return;for(const u of document.querySelectorAll('link[rel="modulepreload"]'))a(u);new MutationObserver(u=>{for(const o of u)if(o.type==="childList")for(const s of o.addedNodes)s.tagName==="LINK"&&s.rel==="modulepreload"&&a(s)}).observe(document,{childList:!0,subtree:!0});function r(u){const o={};return u.integrity&&(o.integrity=u.integrity),u.referrerPolicy&&(o.referrerPolicy=u.referrerPolicy),u.crossOrigin==="use-credentials"?o.credentials="include":u.crossOrigin==="anonymous"?o.credentials="omit":o.credentials="same-origin",o}function a(u){if(u.ep)return;u.ep=!0;const o=r(u);fetch(u.href,o)}})();function Wn(e){return e&&e.__esModule&&Object.prototype.hasOwnProperty.call(e,"default")?e.default:e}var Id={exports:{}},eu={};/**
 * @license React
 * react-jsx-runtime.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var qb;function Bj(){if(qb)return eu;qb=1;var e=Symbol.for("react.transitional.element"),t=Symbol.for("react.fragment");function r(a,u,o){var s=null;if(o!==void 0&&(s=""+o),u.key!==void 0&&(s=""+u.key),"key"in u){o={};for(var f in u)f!=="key"&&(o[f]=u[f])}else o=u;return u=o.ref,{$$typeof:e,type:a,key:s,ref:u!==void 0?u:null,props:o}}return eu.Fragment=t,eu.jsx=r,eu.jsxs=r,eu}var $b;function Lj(){return $b||($b=1,Id.exports=Bj()),Id.exports}var _=Lj(),Xd={exports:{}},we={};/**
 * @license React
 * react.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Hb;function Uj(){if(Hb)return we;Hb=1;var e=Symbol.for("react.transitional.element"),t=Symbol.for("react.portal"),r=Symbol.for("react.fragment"),a=Symbol.for("react.strict_mode"),u=Symbol.for("react.profiler"),o=Symbol.for("react.consumer"),s=Symbol.for("react.context"),f=Symbol.for("react.forward_ref"),v=Symbol.for("react.suspense"),h=Symbol.for("react.memo"),m=Symbol.for("react.lazy"),y=Symbol.iterator;function g(D){return D===null||typeof D!="object"?null:(D=y&&D[y]||D["@@iterator"],typeof D=="function"?D:null)}var b={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},w=Object.assign,S={};function O(D,W,Z){this.props=D,this.context=W,this.refs=S,this.updater=Z||b}O.prototype.isReactComponent={},O.prototype.setState=function(D,W){if(typeof D!="object"&&typeof D!="function"&&D!=null)throw Error("takes an object of state variables to update or a function which returns an object of state variables.");this.updater.enqueueSetState(this,D,W,"setState")},O.prototype.forceUpdate=function(D){this.updater.enqueueForceUpdate(this,D,"forceUpdate")};function T(){}T.prototype=O.prototype;function M(D,W,Z){this.props=D,this.context=W,this.refs=S,this.updater=Z||b}var j=M.prototype=new T;j.constructor=M,w(j,O.prototype),j.isPureReactComponent=!0;var k=Array.isArray,N={H:null,A:null,T:null,S:null,V:null},P=Object.prototype.hasOwnProperty;function L(D,W,Z,ae,fe,Se){return Z=Se.ref,{$$typeof:e,type:D,key:W,ref:Z!==void 0?Z:null,props:Se}}function G(D,W){return L(D.type,W,void 0,void 0,void 0,D.props)}function ee(D){return typeof D=="object"&&D!==null&&D.$$typeof===e}function ne(D){var W={"=":"=0",":":"=2"};return"$"+D.replace(/[=:]/g,function(Z){return W[Z]})}var J=/\/+/g;function re(D,W){return typeof D=="object"&&D!==null&&D.key!=null?ne(""+D.key):W.toString(36)}function he(){}function ce(D){switch(D.status){case"fulfilled":return D.value;case"rejected":throw D.reason;default:switch(typeof D.status=="string"?D.then(he,he):(D.status="pending",D.then(function(W){D.status==="pending"&&(D.status="fulfilled",D.value=W)},function(W){D.status==="pending"&&(D.status="rejected",D.reason=W)})),D.status){case"fulfilled":return D.value;case"rejected":throw D.reason}}throw D}function se(D,W,Z,ae,fe){var Se=typeof D;(Se==="undefined"||Se==="boolean")&&(D=null);var ye=!1;if(D===null)ye=!0;else switch(Se){case"bigint":case"string":case"number":ye=!0;break;case"object":switch(D.$$typeof){case e:case t:ye=!0;break;case m:return ye=D._init,se(ye(D._payload),W,Z,ae,fe)}}if(ye)return fe=fe(D),ye=ae===""?"."+re(D,0):ae,k(fe)?(Z="",ye!=null&&(Z=ye.replace(J,"$&/")+"/"),se(fe,W,Z,"",function(V){return V})):fe!=null&&(ee(fe)&&(fe=G(fe,Z+(fe.key==null||D&&D.key===fe.key?"":(""+fe.key).replace(J,"$&/")+"/")+ye)),W.push(fe)),1;ye=0;var ft=ae===""?".":ae+":";if(k(D))for(var ze=0;ze<D.length;ze++)ae=D[ze],Se=ft+re(ae,ze),ye+=se(ae,W,Z,Se,fe);else if(ze=g(D),typeof ze=="function")for(D=ze.call(D),ze=0;!(ae=D.next()).done;)ae=ae.value,Se=ft+re(ae,ze++),ye+=se(ae,W,Z,Se,fe);else if(Se==="object"){if(typeof D.then=="function")return se(ce(D),W,Z,ae,fe);throw W=String(D),Error("Objects are not valid as a React child (found: "+(W==="[object Object]"?"object with keys {"+Object.keys(D).join(", ")+"}":W)+"). If you meant to render a collection of children, use an array instead.")}return ye}function B(D,W,Z){if(D==null)return D;var ae=[],fe=0;return se(D,ae,"","",function(Se){return W.call(Z,Se,fe++)}),ae}function I(D){if(D._status===-1){var W=D._result;W=W(),W.then(function(Z){(D._status===0||D._status===-1)&&(D._status=1,D._result=Z)},function(Z){(D._status===0||D._status===-1)&&(D._status=2,D._result=Z)}),D._status===-1&&(D._status=0,D._result=W)}if(D._status===1)return D._result.default;throw D._result}var F=typeof reportError=="function"?reportError:function(D){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var W=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof D=="object"&&D!==null&&typeof D.message=="string"?String(D.message):String(D),error:D});if(!window.dispatchEvent(W))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",D);return}console.error(D)};function oe(){}return we.Children={map:B,forEach:function(D,W,Z){B(D,function(){W.apply(this,arguments)},Z)},count:function(D){var W=0;return B(D,function(){W++}),W},toArray:function(D){return B(D,function(W){return W})||[]},only:function(D){if(!ee(D))throw Error("React.Children.only expected to receive a single React element child.");return D}},we.Component=O,we.Fragment=r,we.Profiler=u,we.PureComponent=M,we.StrictMode=a,we.Suspense=v,we.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=N,we.__COMPILER_RUNTIME={__proto__:null,c:function(D){return N.H.useMemoCache(D)}},we.cache=function(D){return function(){return D.apply(null,arguments)}},we.cloneElement=function(D,W,Z){if(D==null)throw Error("The argument must be a React element, but you passed "+D+".");var ae=w({},D.props),fe=D.key,Se=void 0;if(W!=null)for(ye in W.ref!==void 0&&(Se=void 0),W.key!==void 0&&(fe=""+W.key),W)!P.call(W,ye)||ye==="key"||ye==="__self"||ye==="__source"||ye==="ref"&&W.ref===void 0||(ae[ye]=W[ye]);var ye=arguments.length-2;if(ye===1)ae.children=Z;else if(1<ye){for(var ft=Array(ye),ze=0;ze<ye;ze++)ft[ze]=arguments[ze+2];ae.children=ft}return L(D.type,fe,void 0,void 0,Se,ae)},we.createContext=function(D){return D={$$typeof:s,_currentValue:D,_currentValue2:D,_threadCount:0,Provider:null,Consumer:null},D.Provider=D,D.Consumer={$$typeof:o,_context:D},D},we.createElement=function(D,W,Z){var ae,fe={},Se=null;if(W!=null)for(ae in W.key!==void 0&&(Se=""+W.key),W)P.call(W,ae)&&ae!=="key"&&ae!=="__self"&&ae!=="__source"&&(fe[ae]=W[ae]);var ye=arguments.length-2;if(ye===1)fe.children=Z;else if(1<ye){for(var ft=Array(ye),ze=0;ze<ye;ze++)ft[ze]=arguments[ze+2];fe.children=ft}if(D&&D.defaultProps)for(ae in ye=D.defaultProps,ye)fe[ae]===void 0&&(fe[ae]=ye[ae]);return L(D,Se,void 0,void 0,null,fe)},we.createRef=function(){return{current:null}},we.forwardRef=function(D){return{$$typeof:f,render:D}},we.isValidElement=ee,we.lazy=function(D){return{$$typeof:m,_payload:{_status:-1,_result:D},_init:I}},we.memo=function(D,W){return{$$typeof:h,type:D,compare:W===void 0?null:W}},we.startTransition=function(D){var W=N.T,Z={};N.T=Z;try{var ae=D(),fe=N.S;fe!==null&&fe(Z,ae),typeof ae=="object"&&ae!==null&&typeof ae.then=="function"&&ae.then(oe,F)}catch(Se){F(Se)}finally{N.T=W}},we.unstable_useCacheRefresh=function(){return N.H.useCacheRefresh()},we.use=function(D){return N.H.use(D)},we.useActionState=function(D,W,Z){return N.H.useActionState(D,W,Z)},we.useCallback=function(D,W){return N.H.useCallback(D,W)},we.useContext=function(D){return N.H.useContext(D)},we.useDebugValue=function(){},we.useDeferredValue=function(D,W){return N.H.useDeferredValue(D,W)},we.useEffect=function(D,W,Z){var ae=N.H;if(typeof Z=="function")throw Error("useEffect CRUD overload is not enabled in this build of React.");return ae.useEffect(D,W)},we.useId=function(){return N.H.useId()},we.useImperativeHandle=function(D,W,Z){return N.H.useImperativeHandle(D,W,Z)},we.useInsertionEffect=function(D,W){return N.H.useInsertionEffect(D,W)},we.useLayoutEffect=function(D,W){return N.H.useLayoutEffect(D,W)},we.useMemo=function(D,W){return N.H.useMemo(D,W)},we.useOptimistic=function(D,W){return N.H.useOptimistic(D,W)},we.useReducer=function(D,W,Z){return N.H.useReducer(D,W,Z)},we.useRef=function(D){return N.H.useRef(D)},we.useState=function(D){return N.H.useState(D)},we.useSyncExternalStore=function(D,W,Z){return N.H.useSyncExternalStore(D,W,Z)},we.useTransition=function(){return N.H.useTransition()},we.version="19.1.1",we}var Yb;function Ii(){return Yb||(Yb=1,Xd.exports=Uj()),Xd.exports}var A=Ii();const Hw=Wn(A),qj=zj({__proto__:null,default:Hw},[A]);var Vd={exports:{}},tu={},Zd={exports:{}},Qd={};/**
 * @license React
 * scheduler.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Gb;function $j(){return Gb||(Gb=1,(function(e){function t(B,I){var F=B.length;B.push(I);e:for(;0<F;){var oe=F-1>>>1,D=B[oe];if(0<u(D,I))B[oe]=I,B[F]=D,F=oe;else break e}}function r(B){return B.length===0?null:B[0]}function a(B){if(B.length===0)return null;var I=B[0],F=B.pop();if(F!==I){B[0]=F;e:for(var oe=0,D=B.length,W=D>>>1;oe<W;){var Z=2*(oe+1)-1,ae=B[Z],fe=Z+1,Se=B[fe];if(0>u(ae,F))fe<D&&0>u(Se,ae)?(B[oe]=Se,B[fe]=F,oe=fe):(B[oe]=ae,B[Z]=F,oe=Z);else if(fe<D&&0>u(Se,F))B[oe]=Se,B[fe]=F,oe=fe;else break e}}return I}function u(B,I){var F=B.sortIndex-I.sortIndex;return F!==0?F:B.id-I.id}if(e.unstable_now=void 0,typeof performance=="object"&&typeof performance.now=="function"){var o=performance;e.unstable_now=function(){return o.now()}}else{var s=Date,f=s.now();e.unstable_now=function(){return s.now()-f}}var v=[],h=[],m=1,y=null,g=3,b=!1,w=!1,S=!1,O=!1,T=typeof setTimeout=="function"?setTimeout:null,M=typeof clearTimeout=="function"?clearTimeout:null,j=typeof setImmediate<"u"?setImmediate:null;function k(B){for(var I=r(h);I!==null;){if(I.callback===null)a(h);else if(I.startTime<=B)a(h),I.sortIndex=I.expirationTime,t(v,I);else break;I=r(h)}}function N(B){if(S=!1,k(B),!w)if(r(v)!==null)w=!0,P||(P=!0,re());else{var I=r(h);I!==null&&se(N,I.startTime-B)}}var P=!1,L=-1,G=5,ee=-1;function ne(){return O?!0:!(e.unstable_now()-ee<G)}function J(){if(O=!1,P){var B=e.unstable_now();ee=B;var I=!0;try{e:{w=!1,S&&(S=!1,M(L),L=-1),b=!0;var F=g;try{t:{for(k(B),y=r(v);y!==null&&!(y.expirationTime>B&&ne());){var oe=y.callback;if(typeof oe=="function"){y.callback=null,g=y.priorityLevel;var D=oe(y.expirationTime<=B);if(B=e.unstable_now(),typeof D=="function"){y.callback=D,k(B),I=!0;break t}y===r(v)&&a(v),k(B)}else a(v);y=r(v)}if(y!==null)I=!0;else{var W=r(h);W!==null&&se(N,W.startTime-B),I=!1}}break e}finally{y=null,g=F,b=!1}I=void 0}}finally{I?re():P=!1}}}var re;if(typeof j=="function")re=function(){j(J)};else if(typeof MessageChannel<"u"){var he=new MessageChannel,ce=he.port2;he.port1.onmessage=J,re=function(){ce.postMessage(null)}}else re=function(){T(J,0)};function se(B,I){L=T(function(){B(e.unstable_now())},I)}e.unstable_IdlePriority=5,e.unstable_ImmediatePriority=1,e.unstable_LowPriority=4,e.unstable_NormalPriority=3,e.unstable_Profiling=null,e.unstable_UserBlockingPriority=2,e.unstable_cancelCallback=function(B){B.callback=null},e.unstable_forceFrameRate=function(B){0>B||125<B?console.error("forceFrameRate takes a positive int between 0 and 125, forcing frame rates higher than 125 fps is not supported"):G=0<B?Math.floor(1e3/B):5},e.unstable_getCurrentPriorityLevel=function(){return g},e.unstable_next=function(B){switch(g){case 1:case 2:case 3:var I=3;break;default:I=g}var F=g;g=I;try{return B()}finally{g=F}},e.unstable_requestPaint=function(){O=!0},e.unstable_runWithPriority=function(B,I){switch(B){case 1:case 2:case 3:case 4:case 5:break;default:B=3}var F=g;g=B;try{return I()}finally{g=F}},e.unstable_scheduleCallback=function(B,I,F){var oe=e.unstable_now();switch(typeof F=="object"&&F!==null?(F=F.delay,F=typeof F=="number"&&0<F?oe+F:oe):F=oe,B){case 1:var D=-1;break;case 2:D=250;break;case 5:D=1073741823;break;case 4:D=1e4;break;default:D=5e3}return D=F+D,B={id:m++,callback:I,priorityLevel:B,startTime:F,expirationTime:D,sortIndex:-1},F>oe?(B.sortIndex=F,t(h,B),r(v)===null&&B===r(h)&&(S?(M(L),L=-1):S=!0,se(N,F-oe))):(B.sortIndex=D,t(v,B),w||b||(w=!0,P||(P=!0,re()))),B},e.unstable_shouldYield=ne,e.unstable_wrapCallback=function(B){var I=g;return function(){var F=g;g=I;try{return B.apply(this,arguments)}finally{g=F}}}})(Qd)),Qd}var Kb;function Hj(){return Kb||(Kb=1,Zd.exports=$j()),Zd.exports}var Wd={exports:{}},Nt={};/**
 * @license React
 * react-dom.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Ib;function Yj(){if(Ib)return Nt;Ib=1;var e=Ii();function t(v){var h="https://react.dev/errors/"+v;if(1<arguments.length){h+="?args[]="+encodeURIComponent(arguments[1]);for(var m=2;m<arguments.length;m++)h+="&args[]="+encodeURIComponent(arguments[m])}return"Minified React error #"+v+"; visit "+h+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function r(){}var a={d:{f:r,r:function(){throw Error(t(522))},D:r,C:r,L:r,m:r,X:r,S:r,M:r},p:0,findDOMNode:null},u=Symbol.for("react.portal");function o(v,h,m){var y=3<arguments.length&&arguments[3]!==void 0?arguments[3]:null;return{$$typeof:u,key:y==null?null:""+y,children:v,containerInfo:h,implementation:m}}var s=e.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE;function f(v,h){if(v==="font")return"";if(typeof h=="string")return h==="use-credentials"?h:""}return Nt.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE=a,Nt.createPortal=function(v,h){var m=2<arguments.length&&arguments[2]!==void 0?arguments[2]:null;if(!h||h.nodeType!==1&&h.nodeType!==9&&h.nodeType!==11)throw Error(t(299));return o(v,h,null,m)},Nt.flushSync=function(v){var h=s.T,m=a.p;try{if(s.T=null,a.p=2,v)return v()}finally{s.T=h,a.p=m,a.d.f()}},Nt.preconnect=function(v,h){typeof v=="string"&&(h?(h=h.crossOrigin,h=typeof h=="string"?h==="use-credentials"?h:"":void 0):h=null,a.d.C(v,h))},Nt.prefetchDNS=function(v){typeof v=="string"&&a.d.D(v)},Nt.preinit=function(v,h){if(typeof v=="string"&&h&&typeof h.as=="string"){var m=h.as,y=f(m,h.crossOrigin),g=typeof h.integrity=="string"?h.integrity:void 0,b=typeof h.fetchPriority=="string"?h.fetchPriority:void 0;m==="style"?a.d.S(v,typeof h.precedence=="string"?h.precedence:void 0,{crossOrigin:y,integrity:g,fetchPriority:b}):m==="script"&&a.d.X(v,{crossOrigin:y,integrity:g,fetchPriority:b,nonce:typeof h.nonce=="string"?h.nonce:void 0})}},Nt.preinitModule=function(v,h){if(typeof v=="string")if(typeof h=="object"&&h!==null){if(h.as==null||h.as==="script"){var m=f(h.as,h.crossOrigin);a.d.M(v,{crossOrigin:m,integrity:typeof h.integrity=="string"?h.integrity:void 0,nonce:typeof h.nonce=="string"?h.nonce:void 0})}}else h==null&&a.d.M(v)},Nt.preload=function(v,h){if(typeof v=="string"&&typeof h=="object"&&h!==null&&typeof h.as=="string"){var m=h.as,y=f(m,h.crossOrigin);a.d.L(v,m,{crossOrigin:y,integrity:typeof h.integrity=="string"?h.integrity:void 0,nonce:typeof h.nonce=="string"?h.nonce:void 0,type:typeof h.type=="string"?h.type:void 0,fetchPriority:typeof h.fetchPriority=="string"?h.fetchPriority:void 0,referrerPolicy:typeof h.referrerPolicy=="string"?h.referrerPolicy:void 0,imageSrcSet:typeof h.imageSrcSet=="string"?h.imageSrcSet:void 0,imageSizes:typeof h.imageSizes=="string"?h.imageSizes:void 0,media:typeof h.media=="string"?h.media:void 0})}},Nt.preloadModule=function(v,h){if(typeof v=="string")if(h){var m=f(h.as,h.crossOrigin);a.d.m(v,{as:typeof h.as=="string"&&h.as!=="script"?h.as:void 0,crossOrigin:m,integrity:typeof h.integrity=="string"?h.integrity:void 0})}else a.d.m(v)},Nt.requestFormReset=function(v){a.d.r(v)},Nt.unstable_batchedUpdates=function(v,h){return v(h)},Nt.useFormState=function(v,h,m){return s.H.useFormState(v,h,m)},Nt.useFormStatus=function(){return s.H.useHostTransitionStatus()},Nt.version="19.1.1",Nt}var Xb;function Yw(){if(Xb)return Wd.exports;Xb=1;function e(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(t){console.error(t)}}return e(),Wd.exports=Yj(),Wd.exports}/**
 * @license React
 * react-dom-client.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Vb;function Gj(){if(Vb)return tu;Vb=1;var e=Hj(),t=Ii(),r=Yw();function a(n){var i="https://react.dev/errors/"+n;if(1<arguments.length){i+="?args[]="+encodeURIComponent(arguments[1]);for(var l=2;l<arguments.length;l++)i+="&args[]="+encodeURIComponent(arguments[l])}return"Minified React error #"+n+"; visit "+i+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}function u(n){return!(!n||n.nodeType!==1&&n.nodeType!==9&&n.nodeType!==11)}function o(n){var i=n,l=n;if(n.alternate)for(;i.return;)i=i.return;else{n=i;do i=n,(i.flags&4098)!==0&&(l=i.return),n=i.return;while(n)}return i.tag===3?l:null}function s(n){if(n.tag===13){var i=n.memoizedState;if(i===null&&(n=n.alternate,n!==null&&(i=n.memoizedState)),i!==null)return i.dehydrated}return null}function f(n){if(o(n)!==n)throw Error(a(188))}function v(n){var i=n.alternate;if(!i){if(i=o(n),i===null)throw Error(a(188));return i!==n?null:n}for(var l=n,c=i;;){var d=l.return;if(d===null)break;var p=d.alternate;if(p===null){if(c=d.return,c!==null){l=c;continue}break}if(d.child===p.child){for(p=d.child;p;){if(p===l)return f(d),n;if(p===c)return f(d),i;p=p.sibling}throw Error(a(188))}if(l.return!==c.return)l=d,c=p;else{for(var x=!1,E=d.child;E;){if(E===l){x=!0,l=d,c=p;break}if(E===c){x=!0,c=d,l=p;break}E=E.sibling}if(!x){for(E=p.child;E;){if(E===l){x=!0,l=p,c=d;break}if(E===c){x=!0,c=p,l=d;break}E=E.sibling}if(!x)throw Error(a(189))}}if(l.alternate!==c)throw Error(a(190))}if(l.tag!==3)throw Error(a(188));return l.stateNode.current===l?n:i}function h(n){var i=n.tag;if(i===5||i===26||i===27||i===6)return n;for(n=n.child;n!==null;){if(i=h(n),i!==null)return i;n=n.sibling}return null}var m=Object.assign,y=Symbol.for("react.element"),g=Symbol.for("react.transitional.element"),b=Symbol.for("react.portal"),w=Symbol.for("react.fragment"),S=Symbol.for("react.strict_mode"),O=Symbol.for("react.profiler"),T=Symbol.for("react.provider"),M=Symbol.for("react.consumer"),j=Symbol.for("react.context"),k=Symbol.for("react.forward_ref"),N=Symbol.for("react.suspense"),P=Symbol.for("react.suspense_list"),L=Symbol.for("react.memo"),G=Symbol.for("react.lazy"),ee=Symbol.for("react.activity"),ne=Symbol.for("react.memo_cache_sentinel"),J=Symbol.iterator;function re(n){return n===null||typeof n!="object"?null:(n=J&&n[J]||n["@@iterator"],typeof n=="function"?n:null)}var he=Symbol.for("react.client.reference");function ce(n){if(n==null)return null;if(typeof n=="function")return n.$$typeof===he?null:n.displayName||n.name||null;if(typeof n=="string")return n;switch(n){case w:return"Fragment";case O:return"Profiler";case S:return"StrictMode";case N:return"Suspense";case P:return"SuspenseList";case ee:return"Activity"}if(typeof n=="object")switch(n.$$typeof){case b:return"Portal";case j:return(n.displayName||"Context")+".Provider";case M:return(n._context.displayName||"Context")+".Consumer";case k:var i=n.render;return n=n.displayName,n||(n=i.displayName||i.name||"",n=n!==""?"ForwardRef("+n+")":"ForwardRef"),n;case L:return i=n.displayName||null,i!==null?i:ce(n.type)||"Memo";case G:i=n._payload,n=n._init;try{return ce(n(i))}catch{}}return null}var se=Array.isArray,B=t.__CLIENT_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,I=r.__DOM_INTERNALS_DO_NOT_USE_OR_WARN_USERS_THEY_CANNOT_UPGRADE,F={pending:!1,data:null,method:null,action:null},oe=[],D=-1;function W(n){return{current:n}}function Z(n){0>D||(n.current=oe[D],oe[D]=null,D--)}function ae(n,i){D++,oe[D]=n.current,n.current=i}var fe=W(null),Se=W(null),ye=W(null),ft=W(null);function ze(n,i){switch(ae(ye,i),ae(Se,n),ae(fe,null),i.nodeType){case 9:case 11:n=(n=i.documentElement)&&(n=n.namespaceURI)?vb(n):0;break;default:if(n=i.tagName,i=i.namespaceURI)i=vb(i),n=mb(i,n);else switch(n){case"svg":n=1;break;case"math":n=2;break;default:n=0}}Z(fe),ae(fe,n)}function V(){Z(fe),Z(Se),Z(ye)}function ue(n){n.memoizedState!==null&&ae(ft,n);var i=fe.current,l=mb(i,n.type);i!==l&&(ae(Se,n),ae(fe,l))}function ge(n){Se.current===n&&(Z(fe),Z(Se)),ft.current===n&&(Z(ft),Zl._currentValue=F)}var te=Object.prototype.hasOwnProperty,Fe=e.unstable_scheduleCallback,je=e.unstable_cancelCallback,kt=e.unstable_shouldYield,zt=e.unstable_requestPaint,dt=e.unstable_now,Uu=e.unstable_getCurrentPriorityLevel,Vp=e.unstable_ImmediatePriority,Zp=e.unstable_UserBlockingPriority,qu=e.unstable_NormalPriority,y_=e.unstable_LowPriority,Qp=e.unstable_IdlePriority,g_=e.log,b_=e.unstable_setDisableYieldValue,rl=null,Wt=null;function Mr(n){if(typeof g_=="function"&&b_(n),Wt&&typeof Wt.setStrictMode=="function")try{Wt.setStrictMode(rl,n)}catch{}}var Ft=Math.clz32?Math.clz32:w_,x_=Math.log,S_=Math.LN2;function w_(n){return n>>>=0,n===0?32:31-(x_(n)/S_|0)|0}var $u=256,Hu=4194304;function oa(n){var i=n&42;if(i!==0)return i;switch(n&-n){case 1:return 1;case 2:return 2;case 4:return 4;case 8:return 8;case 16:return 16;case 32:return 32;case 64:return 64;case 128:return 128;case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return n&4194048;case 4194304:case 8388608:case 16777216:case 33554432:return n&62914560;case 67108864:return 67108864;case 134217728:return 134217728;case 268435456:return 268435456;case 536870912:return 536870912;case 1073741824:return 0;default:return n}}function Yu(n,i,l){var c=n.pendingLanes;if(c===0)return 0;var d=0,p=n.suspendedLanes,x=n.pingedLanes;n=n.warmLanes;var E=c&134217727;return E!==0?(c=E&~p,c!==0?d=oa(c):(x&=E,x!==0?d=oa(x):l||(l=E&~n,l!==0&&(d=oa(l))))):(E=c&~p,E!==0?d=oa(E):x!==0?d=oa(x):l||(l=c&~n,l!==0&&(d=oa(l)))),d===0?0:i!==0&&i!==d&&(i&p)===0&&(p=d&-d,l=i&-i,p>=l||p===32&&(l&4194048)!==0)?i:d}function al(n,i){return(n.pendingLanes&~(n.suspendedLanes&~n.pingedLanes)&i)===0}function O_(n,i){switch(n){case 1:case 2:case 4:case 8:case 64:return i+250;case 16:case 32:case 128:case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:return i+5e3;case 4194304:case 8388608:case 16777216:case 33554432:return-1;case 67108864:case 134217728:case 268435456:case 536870912:case 1073741824:return-1;default:return-1}}function Wp(){var n=$u;return $u<<=1,($u&4194048)===0&&($u=256),n}function Fp(){var n=Hu;return Hu<<=1,(Hu&62914560)===0&&(Hu=4194304),n}function Rs(n){for(var i=[],l=0;31>l;l++)i.push(n);return i}function il(n,i){n.pendingLanes|=i,i!==268435456&&(n.suspendedLanes=0,n.pingedLanes=0,n.warmLanes=0)}function A_(n,i,l,c,d,p){var x=n.pendingLanes;n.pendingLanes=l,n.suspendedLanes=0,n.pingedLanes=0,n.warmLanes=0,n.expiredLanes&=l,n.entangledLanes&=l,n.errorRecoveryDisabledLanes&=l,n.shellSuspendCounter=0;var E=n.entanglements,C=n.expirationTimes,q=n.hiddenUpdates;for(l=x&~l;0<l;){var K=31-Ft(l),Q=1<<K;E[K]=0,C[K]=-1;var $=q[K];if($!==null)for(q[K]=null,K=0;K<$.length;K++){var H=$[K];H!==null&&(H.lane&=-536870913)}l&=~Q}c!==0&&Jp(n,c,0),p!==0&&d===0&&n.tag!==0&&(n.suspendedLanes|=p&~(x&~i))}function Jp(n,i,l){n.pendingLanes|=i,n.suspendedLanes&=~i;var c=31-Ft(i);n.entangledLanes|=i,n.entanglements[c]=n.entanglements[c]|1073741824|l&4194090}function ey(n,i){var l=n.entangledLanes|=i;for(n=n.entanglements;l;){var c=31-Ft(l),d=1<<c;d&i|n[c]&i&&(n[c]|=i),l&=~d}}function ks(n){switch(n){case 2:n=1;break;case 8:n=4;break;case 32:n=16;break;case 256:case 512:case 1024:case 2048:case 4096:case 8192:case 16384:case 32768:case 65536:case 131072:case 262144:case 524288:case 1048576:case 2097152:case 4194304:case 8388608:case 16777216:case 33554432:n=128;break;case 268435456:n=134217728;break;default:n=0}return n}function zs(n){return n&=-n,2<n?8<n?(n&134217727)!==0?32:268435456:8:2}function ty(){var n=I.p;return n!==0?n:(n=window.event,n===void 0?32:Rb(n.type))}function E_(n,i){var l=I.p;try{return I.p=n,i()}finally{I.p=l}}var Cr=Math.random().toString(36).slice(2),Ct="__reactFiber$"+Cr,Ht="__reactProps$"+Cr,Ia="__reactContainer$"+Cr,Bs="__reactEvents$"+Cr,__="__reactListeners$"+Cr,T_="__reactHandles$"+Cr,ny="__reactResources$"+Cr,ll="__reactMarker$"+Cr;function Ls(n){delete n[Ct],delete n[Ht],delete n[Bs],delete n[__],delete n[T_]}function Xa(n){var i=n[Ct];if(i)return i;for(var l=n.parentNode;l;){if(i=l[Ia]||l[Ct]){if(l=i.alternate,i.child!==null||l!==null&&l.child!==null)for(n=bb(n);n!==null;){if(l=n[Ct])return l;n=bb(n)}return i}n=l,l=n.parentNode}return null}function Va(n){if(n=n[Ct]||n[Ia]){var i=n.tag;if(i===5||i===6||i===13||i===26||i===27||i===3)return n}return null}function ul(n){var i=n.tag;if(i===5||i===26||i===27||i===6)return n.stateNode;throw Error(a(33))}function Za(n){var i=n[ny];return i||(i=n[ny]={hoistableStyles:new Map,hoistableScripts:new Map}),i}function ht(n){n[ll]=!0}var ry=new Set,ay={};function ca(n,i){Qa(n,i),Qa(n+"Capture",i)}function Qa(n,i){for(ay[n]=i,n=0;n<i.length;n++)ry.add(i[n])}var j_=RegExp("^[:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD][:A-Z_a-z\\u00C0-\\u00D6\\u00D8-\\u00F6\\u00F8-\\u02FF\\u0370-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD\\-.0-9\\u00B7\\u0300-\\u036F\\u203F-\\u2040]*$"),iy={},ly={};function M_(n){return te.call(ly,n)?!0:te.call(iy,n)?!1:j_.test(n)?ly[n]=!0:(iy[n]=!0,!1)}function Gu(n,i,l){if(M_(i))if(l===null)n.removeAttribute(i);else{switch(typeof l){case"undefined":case"function":case"symbol":n.removeAttribute(i);return;case"boolean":var c=i.toLowerCase().slice(0,5);if(c!=="data-"&&c!=="aria-"){n.removeAttribute(i);return}}n.setAttribute(i,""+l)}}function Ku(n,i,l){if(l===null)n.removeAttribute(i);else{switch(typeof l){case"undefined":case"function":case"symbol":case"boolean":n.removeAttribute(i);return}n.setAttribute(i,""+l)}}function Fn(n,i,l,c){if(c===null)n.removeAttribute(l);else{switch(typeof c){case"undefined":case"function":case"symbol":case"boolean":n.removeAttribute(l);return}n.setAttributeNS(i,l,""+c)}}var Us,uy;function Wa(n){if(Us===void 0)try{throw Error()}catch(l){var i=l.stack.trim().match(/\n( *(at )?)/);Us=i&&i[1]||"",uy=-1<l.stack.indexOf(`
    at`)?" (<anonymous>)":-1<l.stack.indexOf("@")?"@unknown:0:0":""}return`
`+Us+n+uy}var qs=!1;function $s(n,i){if(!n||qs)return"";qs=!0;var l=Error.prepareStackTrace;Error.prepareStackTrace=void 0;try{var c={DetermineComponentFrameRoot:function(){try{if(i){var Q=function(){throw Error()};if(Object.defineProperty(Q.prototype,"props",{set:function(){throw Error()}}),typeof Reflect=="object"&&Reflect.construct){try{Reflect.construct(Q,[])}catch(H){var $=H}Reflect.construct(n,[],Q)}else{try{Q.call()}catch(H){$=H}n.call(Q.prototype)}}else{try{throw Error()}catch(H){$=H}(Q=n())&&typeof Q.catch=="function"&&Q.catch(function(){})}}catch(H){if(H&&$&&typeof H.stack=="string")return[H.stack,$.stack]}return[null,null]}};c.DetermineComponentFrameRoot.displayName="DetermineComponentFrameRoot";var d=Object.getOwnPropertyDescriptor(c.DetermineComponentFrameRoot,"name");d&&d.configurable&&Object.defineProperty(c.DetermineComponentFrameRoot,"name",{value:"DetermineComponentFrameRoot"});var p=c.DetermineComponentFrameRoot(),x=p[0],E=p[1];if(x&&E){var C=x.split(`
`),q=E.split(`
`);for(d=c=0;c<C.length&&!C[c].includes("DetermineComponentFrameRoot");)c++;for(;d<q.length&&!q[d].includes("DetermineComponentFrameRoot");)d++;if(c===C.length||d===q.length)for(c=C.length-1,d=q.length-1;1<=c&&0<=d&&C[c]!==q[d];)d--;for(;1<=c&&0<=d;c--,d--)if(C[c]!==q[d]){if(c!==1||d!==1)do if(c--,d--,0>d||C[c]!==q[d]){var K=`
`+C[c].replace(" at new "," at ");return n.displayName&&K.includes("<anonymous>")&&(K=K.replace("<anonymous>",n.displayName)),K}while(1<=c&&0<=d);break}}}finally{qs=!1,Error.prepareStackTrace=l}return(l=n?n.displayName||n.name:"")?Wa(l):""}function C_(n){switch(n.tag){case 26:case 27:case 5:return Wa(n.type);case 16:return Wa("Lazy");case 13:return Wa("Suspense");case 19:return Wa("SuspenseList");case 0:case 15:return $s(n.type,!1);case 11:return $s(n.type.render,!1);case 1:return $s(n.type,!0);case 31:return Wa("Activity");default:return""}}function oy(n){try{var i="";do i+=C_(n),n=n.return;while(n);return i}catch(l){return`
Error generating stack: `+l.message+`
`+l.stack}}function pn(n){switch(typeof n){case"bigint":case"boolean":case"number":case"string":case"undefined":return n;case"object":return n;default:return""}}function cy(n){var i=n.type;return(n=n.nodeName)&&n.toLowerCase()==="input"&&(i==="checkbox"||i==="radio")}function D_(n){var i=cy(n)?"checked":"value",l=Object.getOwnPropertyDescriptor(n.constructor.prototype,i),c=""+n[i];if(!n.hasOwnProperty(i)&&typeof l<"u"&&typeof l.get=="function"&&typeof l.set=="function"){var d=l.get,p=l.set;return Object.defineProperty(n,i,{configurable:!0,get:function(){return d.call(this)},set:function(x){c=""+x,p.call(this,x)}}),Object.defineProperty(n,i,{enumerable:l.enumerable}),{getValue:function(){return c},setValue:function(x){c=""+x},stopTracking:function(){n._valueTracker=null,delete n[i]}}}}function Iu(n){n._valueTracker||(n._valueTracker=D_(n))}function sy(n){if(!n)return!1;var i=n._valueTracker;if(!i)return!0;var l=i.getValue(),c="";return n&&(c=cy(n)?n.checked?"true":"false":n.value),n=c,n!==l?(i.setValue(n),!0):!1}function Xu(n){if(n=n||(typeof document<"u"?document:void 0),typeof n>"u")return null;try{return n.activeElement||n.body}catch{return n.body}}var N_=/[\n"\\]/g;function yn(n){return n.replace(N_,function(i){return"\\"+i.charCodeAt(0).toString(16)+" "})}function Hs(n,i,l,c,d,p,x,E){n.name="",x!=null&&typeof x!="function"&&typeof x!="symbol"&&typeof x!="boolean"?n.type=x:n.removeAttribute("type"),i!=null?x==="number"?(i===0&&n.value===""||n.value!=i)&&(n.value=""+pn(i)):n.value!==""+pn(i)&&(n.value=""+pn(i)):x!=="submit"&&x!=="reset"||n.removeAttribute("value"),i!=null?Ys(n,x,pn(i)):l!=null?Ys(n,x,pn(l)):c!=null&&n.removeAttribute("value"),d==null&&p!=null&&(n.defaultChecked=!!p),d!=null&&(n.checked=d&&typeof d!="function"&&typeof d!="symbol"),E!=null&&typeof E!="function"&&typeof E!="symbol"&&typeof E!="boolean"?n.name=""+pn(E):n.removeAttribute("name")}function fy(n,i,l,c,d,p,x,E){if(p!=null&&typeof p!="function"&&typeof p!="symbol"&&typeof p!="boolean"&&(n.type=p),i!=null||l!=null){if(!(p!=="submit"&&p!=="reset"||i!=null))return;l=l!=null?""+pn(l):"",i=i!=null?""+pn(i):l,E||i===n.value||(n.value=i),n.defaultValue=i}c=c??d,c=typeof c!="function"&&typeof c!="symbol"&&!!c,n.checked=E?n.checked:!!c,n.defaultChecked=!!c,x!=null&&typeof x!="function"&&typeof x!="symbol"&&typeof x!="boolean"&&(n.name=x)}function Ys(n,i,l){i==="number"&&Xu(n.ownerDocument)===n||n.defaultValue===""+l||(n.defaultValue=""+l)}function Fa(n,i,l,c){if(n=n.options,i){i={};for(var d=0;d<l.length;d++)i["$"+l[d]]=!0;for(l=0;l<n.length;l++)d=i.hasOwnProperty("$"+n[l].value),n[l].selected!==d&&(n[l].selected=d),d&&c&&(n[l].defaultSelected=!0)}else{for(l=""+pn(l),i=null,d=0;d<n.length;d++){if(n[d].value===l){n[d].selected=!0,c&&(n[d].defaultSelected=!0);return}i!==null||n[d].disabled||(i=n[d])}i!==null&&(i.selected=!0)}}function dy(n,i,l){if(i!=null&&(i=""+pn(i),i!==n.value&&(n.value=i),l==null)){n.defaultValue!==i&&(n.defaultValue=i);return}n.defaultValue=l!=null?""+pn(l):""}function hy(n,i,l,c){if(i==null){if(c!=null){if(l!=null)throw Error(a(92));if(se(c)){if(1<c.length)throw Error(a(93));c=c[0]}l=c}l==null&&(l=""),i=l}l=pn(i),n.defaultValue=l,c=n.textContent,c===l&&c!==""&&c!==null&&(n.value=c)}function Ja(n,i){if(i){var l=n.firstChild;if(l&&l===n.lastChild&&l.nodeType===3){l.nodeValue=i;return}}n.textContent=i}var P_=new Set("animationIterationCount aspectRatio borderImageOutset borderImageSlice borderImageWidth boxFlex boxFlexGroup boxOrdinalGroup columnCount columns flex flexGrow flexPositive flexShrink flexNegative flexOrder gridArea gridRow gridRowEnd gridRowSpan gridRowStart gridColumn gridColumnEnd gridColumnSpan gridColumnStart fontWeight lineClamp lineHeight opacity order orphans scale tabSize widows zIndex zoom fillOpacity floodOpacity stopOpacity strokeDasharray strokeDashoffset strokeMiterlimit strokeOpacity strokeWidth MozAnimationIterationCount MozBoxFlex MozBoxFlexGroup MozLineClamp msAnimationIterationCount msFlex msZoom msFlexGrow msFlexNegative msFlexOrder msFlexPositive msFlexShrink msGridColumn msGridColumnSpan msGridRow msGridRowSpan WebkitAnimationIterationCount WebkitBoxFlex WebKitBoxFlexGroup WebkitBoxOrdinalGroup WebkitColumnCount WebkitColumns WebkitFlex WebkitFlexGrow WebkitFlexPositive WebkitFlexShrink WebkitLineClamp".split(" "));function vy(n,i,l){var c=i.indexOf("--")===0;l==null||typeof l=="boolean"||l===""?c?n.setProperty(i,""):i==="float"?n.cssFloat="":n[i]="":c?n.setProperty(i,l):typeof l!="number"||l===0||P_.has(i)?i==="float"?n.cssFloat=l:n[i]=(""+l).trim():n[i]=l+"px"}function my(n,i,l){if(i!=null&&typeof i!="object")throw Error(a(62));if(n=n.style,l!=null){for(var c in l)!l.hasOwnProperty(c)||i!=null&&i.hasOwnProperty(c)||(c.indexOf("--")===0?n.setProperty(c,""):c==="float"?n.cssFloat="":n[c]="");for(var d in i)c=i[d],i.hasOwnProperty(d)&&l[d]!==c&&vy(n,d,c)}else for(var p in i)i.hasOwnProperty(p)&&vy(n,p,i[p])}function Gs(n){if(n.indexOf("-")===-1)return!1;switch(n){case"annotation-xml":case"color-profile":case"font-face":case"font-face-src":case"font-face-uri":case"font-face-format":case"font-face-name":case"missing-glyph":return!1;default:return!0}}var R_=new Map([["acceptCharset","accept-charset"],["htmlFor","for"],["httpEquiv","http-equiv"],["crossOrigin","crossorigin"],["accentHeight","accent-height"],["alignmentBaseline","alignment-baseline"],["arabicForm","arabic-form"],["baselineShift","baseline-shift"],["capHeight","cap-height"],["clipPath","clip-path"],["clipRule","clip-rule"],["colorInterpolation","color-interpolation"],["colorInterpolationFilters","color-interpolation-filters"],["colorProfile","color-profile"],["colorRendering","color-rendering"],["dominantBaseline","dominant-baseline"],["enableBackground","enable-background"],["fillOpacity","fill-opacity"],["fillRule","fill-rule"],["floodColor","flood-color"],["floodOpacity","flood-opacity"],["fontFamily","font-family"],["fontSize","font-size"],["fontSizeAdjust","font-size-adjust"],["fontStretch","font-stretch"],["fontStyle","font-style"],["fontVariant","font-variant"],["fontWeight","font-weight"],["glyphName","glyph-name"],["glyphOrientationHorizontal","glyph-orientation-horizontal"],["glyphOrientationVertical","glyph-orientation-vertical"],["horizAdvX","horiz-adv-x"],["horizOriginX","horiz-origin-x"],["imageRendering","image-rendering"],["letterSpacing","letter-spacing"],["lightingColor","lighting-color"],["markerEnd","marker-end"],["markerMid","marker-mid"],["markerStart","marker-start"],["overlinePosition","overline-position"],["overlineThickness","overline-thickness"],["paintOrder","paint-order"],["panose-1","panose-1"],["pointerEvents","pointer-events"],["renderingIntent","rendering-intent"],["shapeRendering","shape-rendering"],["stopColor","stop-color"],["stopOpacity","stop-opacity"],["strikethroughPosition","strikethrough-position"],["strikethroughThickness","strikethrough-thickness"],["strokeDasharray","stroke-dasharray"],["strokeDashoffset","stroke-dashoffset"],["strokeLinecap","stroke-linecap"],["strokeLinejoin","stroke-linejoin"],["strokeMiterlimit","stroke-miterlimit"],["strokeOpacity","stroke-opacity"],["strokeWidth","stroke-width"],["textAnchor","text-anchor"],["textDecoration","text-decoration"],["textRendering","text-rendering"],["transformOrigin","transform-origin"],["underlinePosition","underline-position"],["underlineThickness","underline-thickness"],["unicodeBidi","unicode-bidi"],["unicodeRange","unicode-range"],["unitsPerEm","units-per-em"],["vAlphabetic","v-alphabetic"],["vHanging","v-hanging"],["vIdeographic","v-ideographic"],["vMathematical","v-mathematical"],["vectorEffect","vector-effect"],["vertAdvY","vert-adv-y"],["vertOriginX","vert-origin-x"],["vertOriginY","vert-origin-y"],["wordSpacing","word-spacing"],["writingMode","writing-mode"],["xmlnsXlink","xmlns:xlink"],["xHeight","x-height"]]),k_=/^[\u0000-\u001F ]*j[\r\n\t]*a[\r\n\t]*v[\r\n\t]*a[\r\n\t]*s[\r\n\t]*c[\r\n\t]*r[\r\n\t]*i[\r\n\t]*p[\r\n\t]*t[\r\n\t]*:/i;function Vu(n){return k_.test(""+n)?"javascript:throw new Error('React has blocked a javascript: URL as a security precaution.')":n}var Ks=null;function Is(n){return n=n.target||n.srcElement||window,n.correspondingUseElement&&(n=n.correspondingUseElement),n.nodeType===3?n.parentNode:n}var ei=null,ti=null;function py(n){var i=Va(n);if(i&&(n=i.stateNode)){var l=n[Ht]||null;e:switch(n=i.stateNode,i.type){case"input":if(Hs(n,l.value,l.defaultValue,l.defaultValue,l.checked,l.defaultChecked,l.type,l.name),i=l.name,l.type==="radio"&&i!=null){for(l=n;l.parentNode;)l=l.parentNode;for(l=l.querySelectorAll('input[name="'+yn(""+i)+'"][type="radio"]'),i=0;i<l.length;i++){var c=l[i];if(c!==n&&c.form===n.form){var d=c[Ht]||null;if(!d)throw Error(a(90));Hs(c,d.value,d.defaultValue,d.defaultValue,d.checked,d.defaultChecked,d.type,d.name)}}for(i=0;i<l.length;i++)c=l[i],c.form===n.form&&sy(c)}break e;case"textarea":dy(n,l.value,l.defaultValue);break e;case"select":i=l.value,i!=null&&Fa(n,!!l.multiple,i,!1)}}}var Xs=!1;function yy(n,i,l){if(Xs)return n(i,l);Xs=!0;try{var c=n(i);return c}finally{if(Xs=!1,(ei!==null||ti!==null)&&(Ro(),ei&&(i=ei,n=ti,ti=ei=null,py(i),n)))for(i=0;i<n.length;i++)py(n[i])}}function ol(n,i){var l=n.stateNode;if(l===null)return null;var c=l[Ht]||null;if(c===null)return null;l=c[i];e:switch(i){case"onClick":case"onClickCapture":case"onDoubleClick":case"onDoubleClickCapture":case"onMouseDown":case"onMouseDownCapture":case"onMouseMove":case"onMouseMoveCapture":case"onMouseUp":case"onMouseUpCapture":case"onMouseEnter":(c=!c.disabled)||(n=n.type,c=!(n==="button"||n==="input"||n==="select"||n==="textarea")),n=!c;break e;default:n=!1}if(n)return null;if(l&&typeof l!="function")throw Error(a(231,i,typeof l));return l}var Jn=!(typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"),Vs=!1;if(Jn)try{var cl={};Object.defineProperty(cl,"passive",{get:function(){Vs=!0}}),window.addEventListener("test",cl,cl),window.removeEventListener("test",cl,cl)}catch{Vs=!1}var Dr=null,Zs=null,Zu=null;function gy(){if(Zu)return Zu;var n,i=Zs,l=i.length,c,d="value"in Dr?Dr.value:Dr.textContent,p=d.length;for(n=0;n<l&&i[n]===d[n];n++);var x=l-n;for(c=1;c<=x&&i[l-c]===d[p-c];c++);return Zu=d.slice(n,1<c?1-c:void 0)}function Qu(n){var i=n.keyCode;return"charCode"in n?(n=n.charCode,n===0&&i===13&&(n=13)):n=i,n===10&&(n=13),32<=n||n===13?n:0}function Wu(){return!0}function by(){return!1}function Yt(n){function i(l,c,d,p,x){this._reactName=l,this._targetInst=d,this.type=c,this.nativeEvent=p,this.target=x,this.currentTarget=null;for(var E in n)n.hasOwnProperty(E)&&(l=n[E],this[E]=l?l(p):p[E]);return this.isDefaultPrevented=(p.defaultPrevented!=null?p.defaultPrevented:p.returnValue===!1)?Wu:by,this.isPropagationStopped=by,this}return m(i.prototype,{preventDefault:function(){this.defaultPrevented=!0;var l=this.nativeEvent;l&&(l.preventDefault?l.preventDefault():typeof l.returnValue!="unknown"&&(l.returnValue=!1),this.isDefaultPrevented=Wu)},stopPropagation:function(){var l=this.nativeEvent;l&&(l.stopPropagation?l.stopPropagation():typeof l.cancelBubble!="unknown"&&(l.cancelBubble=!0),this.isPropagationStopped=Wu)},persist:function(){},isPersistent:Wu}),i}var sa={eventPhase:0,bubbles:0,cancelable:0,timeStamp:function(n){return n.timeStamp||Date.now()},defaultPrevented:0,isTrusted:0},Fu=Yt(sa),sl=m({},sa,{view:0,detail:0}),z_=Yt(sl),Qs,Ws,fl,Ju=m({},sl,{screenX:0,screenY:0,clientX:0,clientY:0,pageX:0,pageY:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,getModifierState:Js,button:0,buttons:0,relatedTarget:function(n){return n.relatedTarget===void 0?n.fromElement===n.srcElement?n.toElement:n.fromElement:n.relatedTarget},movementX:function(n){return"movementX"in n?n.movementX:(n!==fl&&(fl&&n.type==="mousemove"?(Qs=n.screenX-fl.screenX,Ws=n.screenY-fl.screenY):Ws=Qs=0,fl=n),Qs)},movementY:function(n){return"movementY"in n?n.movementY:Ws}}),xy=Yt(Ju),B_=m({},Ju,{dataTransfer:0}),L_=Yt(B_),U_=m({},sl,{relatedTarget:0}),Fs=Yt(U_),q_=m({},sa,{animationName:0,elapsedTime:0,pseudoElement:0}),$_=Yt(q_),H_=m({},sa,{clipboardData:function(n){return"clipboardData"in n?n.clipboardData:window.clipboardData}}),Y_=Yt(H_),G_=m({},sa,{data:0}),Sy=Yt(G_),K_={Esc:"Escape",Spacebar:" ",Left:"ArrowLeft",Up:"ArrowUp",Right:"ArrowRight",Down:"ArrowDown",Del:"Delete",Win:"OS",Menu:"ContextMenu",Apps:"ContextMenu",Scroll:"ScrollLock",MozPrintableKey:"Unidentified"},I_={8:"Backspace",9:"Tab",12:"Clear",13:"Enter",16:"Shift",17:"Control",18:"Alt",19:"Pause",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",45:"Insert",46:"Delete",112:"F1",113:"F2",114:"F3",115:"F4",116:"F5",117:"F6",118:"F7",119:"F8",120:"F9",121:"F10",122:"F11",123:"F12",144:"NumLock",145:"ScrollLock",224:"Meta"},X_={Alt:"altKey",Control:"ctrlKey",Meta:"metaKey",Shift:"shiftKey"};function V_(n){var i=this.nativeEvent;return i.getModifierState?i.getModifierState(n):(n=X_[n])?!!i[n]:!1}function Js(){return V_}var Z_=m({},sl,{key:function(n){if(n.key){var i=K_[n.key]||n.key;if(i!=="Unidentified")return i}return n.type==="keypress"?(n=Qu(n),n===13?"Enter":String.fromCharCode(n)):n.type==="keydown"||n.type==="keyup"?I_[n.keyCode]||"Unidentified":""},code:0,location:0,ctrlKey:0,shiftKey:0,altKey:0,metaKey:0,repeat:0,locale:0,getModifierState:Js,charCode:function(n){return n.type==="keypress"?Qu(n):0},keyCode:function(n){return n.type==="keydown"||n.type==="keyup"?n.keyCode:0},which:function(n){return n.type==="keypress"?Qu(n):n.type==="keydown"||n.type==="keyup"?n.keyCode:0}}),Q_=Yt(Z_),W_=m({},Ju,{pointerId:0,width:0,height:0,pressure:0,tangentialPressure:0,tiltX:0,tiltY:0,twist:0,pointerType:0,isPrimary:0}),wy=Yt(W_),F_=m({},sl,{touches:0,targetTouches:0,changedTouches:0,altKey:0,metaKey:0,ctrlKey:0,shiftKey:0,getModifierState:Js}),J_=Yt(F_),eT=m({},sa,{propertyName:0,elapsedTime:0,pseudoElement:0}),tT=Yt(eT),nT=m({},Ju,{deltaX:function(n){return"deltaX"in n?n.deltaX:"wheelDeltaX"in n?-n.wheelDeltaX:0},deltaY:function(n){return"deltaY"in n?n.deltaY:"wheelDeltaY"in n?-n.wheelDeltaY:"wheelDelta"in n?-n.wheelDelta:0},deltaZ:0,deltaMode:0}),rT=Yt(nT),aT=m({},sa,{newState:0,oldState:0}),iT=Yt(aT),lT=[9,13,27,32],ef=Jn&&"CompositionEvent"in window,dl=null;Jn&&"documentMode"in document&&(dl=document.documentMode);var uT=Jn&&"TextEvent"in window&&!dl,Oy=Jn&&(!ef||dl&&8<dl&&11>=dl),Ay=" ",Ey=!1;function _y(n,i){switch(n){case"keyup":return lT.indexOf(i.keyCode)!==-1;case"keydown":return i.keyCode!==229;case"keypress":case"mousedown":case"focusout":return!0;default:return!1}}function Ty(n){return n=n.detail,typeof n=="object"&&"data"in n?n.data:null}var ni=!1;function oT(n,i){switch(n){case"compositionend":return Ty(i);case"keypress":return i.which!==32?null:(Ey=!0,Ay);case"textInput":return n=i.data,n===Ay&&Ey?null:n;default:return null}}function cT(n,i){if(ni)return n==="compositionend"||!ef&&_y(n,i)?(n=gy(),Zu=Zs=Dr=null,ni=!1,n):null;switch(n){case"paste":return null;case"keypress":if(!(i.ctrlKey||i.altKey||i.metaKey)||i.ctrlKey&&i.altKey){if(i.char&&1<i.char.length)return i.char;if(i.which)return String.fromCharCode(i.which)}return null;case"compositionend":return Oy&&i.locale!=="ko"?null:i.data;default:return null}}var sT={color:!0,date:!0,datetime:!0,"datetime-local":!0,email:!0,month:!0,number:!0,password:!0,range:!0,search:!0,tel:!0,text:!0,time:!0,url:!0,week:!0};function jy(n){var i=n&&n.nodeName&&n.nodeName.toLowerCase();return i==="input"?!!sT[n.type]:i==="textarea"}function My(n,i,l,c){ei?ti?ti.push(c):ti=[c]:ei=c,i=qo(i,"onChange"),0<i.length&&(l=new Fu("onChange","change",null,l,c),n.push({event:l,listeners:i}))}var hl=null,vl=null;function fT(n){cb(n,0)}function eo(n){var i=ul(n);if(sy(i))return n}function Cy(n,i){if(n==="change")return i}var Dy=!1;if(Jn){var tf;if(Jn){var nf="oninput"in document;if(!nf){var Ny=document.createElement("div");Ny.setAttribute("oninput","return;"),nf=typeof Ny.oninput=="function"}tf=nf}else tf=!1;Dy=tf&&(!document.documentMode||9<document.documentMode)}function Py(){hl&&(hl.detachEvent("onpropertychange",Ry),vl=hl=null)}function Ry(n){if(n.propertyName==="value"&&eo(vl)){var i=[];My(i,vl,n,Is(n)),yy(fT,i)}}function dT(n,i,l){n==="focusin"?(Py(),hl=i,vl=l,hl.attachEvent("onpropertychange",Ry)):n==="focusout"&&Py()}function hT(n){if(n==="selectionchange"||n==="keyup"||n==="keydown")return eo(vl)}function vT(n,i){if(n==="click")return eo(i)}function mT(n,i){if(n==="input"||n==="change")return eo(i)}function pT(n,i){return n===i&&(n!==0||1/n===1/i)||n!==n&&i!==i}var Jt=typeof Object.is=="function"?Object.is:pT;function ml(n,i){if(Jt(n,i))return!0;if(typeof n!="object"||n===null||typeof i!="object"||i===null)return!1;var l=Object.keys(n),c=Object.keys(i);if(l.length!==c.length)return!1;for(c=0;c<l.length;c++){var d=l[c];if(!te.call(i,d)||!Jt(n[d],i[d]))return!1}return!0}function ky(n){for(;n&&n.firstChild;)n=n.firstChild;return n}function zy(n,i){var l=ky(n);n=0;for(var c;l;){if(l.nodeType===3){if(c=n+l.textContent.length,n<=i&&c>=i)return{node:l,offset:i-n};n=c}e:{for(;l;){if(l.nextSibling){l=l.nextSibling;break e}l=l.parentNode}l=void 0}l=ky(l)}}function By(n,i){return n&&i?n===i?!0:n&&n.nodeType===3?!1:i&&i.nodeType===3?By(n,i.parentNode):"contains"in n?n.contains(i):n.compareDocumentPosition?!!(n.compareDocumentPosition(i)&16):!1:!1}function Ly(n){n=n!=null&&n.ownerDocument!=null&&n.ownerDocument.defaultView!=null?n.ownerDocument.defaultView:window;for(var i=Xu(n.document);i instanceof n.HTMLIFrameElement;){try{var l=typeof i.contentWindow.location.href=="string"}catch{l=!1}if(l)n=i.contentWindow;else break;i=Xu(n.document)}return i}function rf(n){var i=n&&n.nodeName&&n.nodeName.toLowerCase();return i&&(i==="input"&&(n.type==="text"||n.type==="search"||n.type==="tel"||n.type==="url"||n.type==="password")||i==="textarea"||n.contentEditable==="true")}var yT=Jn&&"documentMode"in document&&11>=document.documentMode,ri=null,af=null,pl=null,lf=!1;function Uy(n,i,l){var c=l.window===l?l.document:l.nodeType===9?l:l.ownerDocument;lf||ri==null||ri!==Xu(c)||(c=ri,"selectionStart"in c&&rf(c)?c={start:c.selectionStart,end:c.selectionEnd}:(c=(c.ownerDocument&&c.ownerDocument.defaultView||window).getSelection(),c={anchorNode:c.anchorNode,anchorOffset:c.anchorOffset,focusNode:c.focusNode,focusOffset:c.focusOffset}),pl&&ml(pl,c)||(pl=c,c=qo(af,"onSelect"),0<c.length&&(i=new Fu("onSelect","select",null,i,l),n.push({event:i,listeners:c}),i.target=ri)))}function fa(n,i){var l={};return l[n.toLowerCase()]=i.toLowerCase(),l["Webkit"+n]="webkit"+i,l["Moz"+n]="moz"+i,l}var ai={animationend:fa("Animation","AnimationEnd"),animationiteration:fa("Animation","AnimationIteration"),animationstart:fa("Animation","AnimationStart"),transitionrun:fa("Transition","TransitionRun"),transitionstart:fa("Transition","TransitionStart"),transitioncancel:fa("Transition","TransitionCancel"),transitionend:fa("Transition","TransitionEnd")},uf={},qy={};Jn&&(qy=document.createElement("div").style,"AnimationEvent"in window||(delete ai.animationend.animation,delete ai.animationiteration.animation,delete ai.animationstart.animation),"TransitionEvent"in window||delete ai.transitionend.transition);function da(n){if(uf[n])return uf[n];if(!ai[n])return n;var i=ai[n],l;for(l in i)if(i.hasOwnProperty(l)&&l in qy)return uf[n]=i[l];return n}var $y=da("animationend"),Hy=da("animationiteration"),Yy=da("animationstart"),gT=da("transitionrun"),bT=da("transitionstart"),xT=da("transitioncancel"),Gy=da("transitionend"),Ky=new Map,of="abort auxClick beforeToggle cancel canPlay canPlayThrough click close contextMenu copy cut drag dragEnd dragEnter dragExit dragLeave dragOver dragStart drop durationChange emptied encrypted ended error gotPointerCapture input invalid keyDown keyPress keyUp load loadedData loadedMetadata loadStart lostPointerCapture mouseDown mouseMove mouseOut mouseOver mouseUp paste pause play playing pointerCancel pointerDown pointerMove pointerOut pointerOver pointerUp progress rateChange reset resize seeked seeking stalled submit suspend timeUpdate touchCancel touchEnd touchStart volumeChange scroll toggle touchMove waiting wheel".split(" ");of.push("scrollEnd");function kn(n,i){Ky.set(n,i),ca(i,[n])}var Iy=new WeakMap;function gn(n,i){if(typeof n=="object"&&n!==null){var l=Iy.get(n);return l!==void 0?l:(i={value:n,source:i,stack:oy(i)},Iy.set(n,i),i)}return{value:n,source:i,stack:oy(i)}}var bn=[],ii=0,cf=0;function to(){for(var n=ii,i=cf=ii=0;i<n;){var l=bn[i];bn[i++]=null;var c=bn[i];bn[i++]=null;var d=bn[i];bn[i++]=null;var p=bn[i];if(bn[i++]=null,c!==null&&d!==null){var x=c.pending;x===null?d.next=d:(d.next=x.next,x.next=d),c.pending=d}p!==0&&Xy(l,d,p)}}function no(n,i,l,c){bn[ii++]=n,bn[ii++]=i,bn[ii++]=l,bn[ii++]=c,cf|=c,n.lanes|=c,n=n.alternate,n!==null&&(n.lanes|=c)}function sf(n,i,l,c){return no(n,i,l,c),ro(n)}function li(n,i){return no(n,null,null,i),ro(n)}function Xy(n,i,l){n.lanes|=l;var c=n.alternate;c!==null&&(c.lanes|=l);for(var d=!1,p=n.return;p!==null;)p.childLanes|=l,c=p.alternate,c!==null&&(c.childLanes|=l),p.tag===22&&(n=p.stateNode,n===null||n._visibility&1||(d=!0)),n=p,p=p.return;return n.tag===3?(p=n.stateNode,d&&i!==null&&(d=31-Ft(l),n=p.hiddenUpdates,c=n[d],c===null?n[d]=[i]:c.push(i),i.lane=l|536870912),p):null}function ro(n){if(50<$l)throw $l=0,pd=null,Error(a(185));for(var i=n.return;i!==null;)n=i,i=n.return;return n.tag===3?n.stateNode:null}var ui={};function ST(n,i,l,c){this.tag=n,this.key=l,this.sibling=this.child=this.return=this.stateNode=this.type=this.elementType=null,this.index=0,this.refCleanup=this.ref=null,this.pendingProps=i,this.dependencies=this.memoizedState=this.updateQueue=this.memoizedProps=null,this.mode=c,this.subtreeFlags=this.flags=0,this.deletions=null,this.childLanes=this.lanes=0,this.alternate=null}function en(n,i,l,c){return new ST(n,i,l,c)}function ff(n){return n=n.prototype,!(!n||!n.isReactComponent)}function er(n,i){var l=n.alternate;return l===null?(l=en(n.tag,i,n.key,n.mode),l.elementType=n.elementType,l.type=n.type,l.stateNode=n.stateNode,l.alternate=n,n.alternate=l):(l.pendingProps=i,l.type=n.type,l.flags=0,l.subtreeFlags=0,l.deletions=null),l.flags=n.flags&65011712,l.childLanes=n.childLanes,l.lanes=n.lanes,l.child=n.child,l.memoizedProps=n.memoizedProps,l.memoizedState=n.memoizedState,l.updateQueue=n.updateQueue,i=n.dependencies,l.dependencies=i===null?null:{lanes:i.lanes,firstContext:i.firstContext},l.sibling=n.sibling,l.index=n.index,l.ref=n.ref,l.refCleanup=n.refCleanup,l}function Vy(n,i){n.flags&=65011714;var l=n.alternate;return l===null?(n.childLanes=0,n.lanes=i,n.child=null,n.subtreeFlags=0,n.memoizedProps=null,n.memoizedState=null,n.updateQueue=null,n.dependencies=null,n.stateNode=null):(n.childLanes=l.childLanes,n.lanes=l.lanes,n.child=l.child,n.subtreeFlags=0,n.deletions=null,n.memoizedProps=l.memoizedProps,n.memoizedState=l.memoizedState,n.updateQueue=l.updateQueue,n.type=l.type,i=l.dependencies,n.dependencies=i===null?null:{lanes:i.lanes,firstContext:i.firstContext}),n}function ao(n,i,l,c,d,p){var x=0;if(c=n,typeof n=="function")ff(n)&&(x=1);else if(typeof n=="string")x=Oj(n,l,fe.current)?26:n==="html"||n==="head"||n==="body"?27:5;else e:switch(n){case ee:return n=en(31,l,i,d),n.elementType=ee,n.lanes=p,n;case w:return ha(l.children,d,p,i);case S:x=8,d|=24;break;case O:return n=en(12,l,i,d|2),n.elementType=O,n.lanes=p,n;case N:return n=en(13,l,i,d),n.elementType=N,n.lanes=p,n;case P:return n=en(19,l,i,d),n.elementType=P,n.lanes=p,n;default:if(typeof n=="object"&&n!==null)switch(n.$$typeof){case T:case j:x=10;break e;case M:x=9;break e;case k:x=11;break e;case L:x=14;break e;case G:x=16,c=null;break e}x=29,l=Error(a(130,n===null?"null":typeof n,"")),c=null}return i=en(x,l,i,d),i.elementType=n,i.type=c,i.lanes=p,i}function ha(n,i,l,c){return n=en(7,n,c,i),n.lanes=l,n}function df(n,i,l){return n=en(6,n,null,i),n.lanes=l,n}function hf(n,i,l){return i=en(4,n.children!==null?n.children:[],n.key,i),i.lanes=l,i.stateNode={containerInfo:n.containerInfo,pendingChildren:null,implementation:n.implementation},i}var oi=[],ci=0,io=null,lo=0,xn=[],Sn=0,va=null,tr=1,nr="";function ma(n,i){oi[ci++]=lo,oi[ci++]=io,io=n,lo=i}function Zy(n,i,l){xn[Sn++]=tr,xn[Sn++]=nr,xn[Sn++]=va,va=n;var c=tr;n=nr;var d=32-Ft(c)-1;c&=~(1<<d),l+=1;var p=32-Ft(i)+d;if(30<p){var x=d-d%5;p=(c&(1<<x)-1).toString(32),c>>=x,d-=x,tr=1<<32-Ft(i)+d|l<<d|c,nr=p+n}else tr=1<<p|l<<d|c,nr=n}function vf(n){n.return!==null&&(ma(n,1),Zy(n,1,0))}function mf(n){for(;n===io;)io=oi[--ci],oi[ci]=null,lo=oi[--ci],oi[ci]=null;for(;n===va;)va=xn[--Sn],xn[Sn]=null,nr=xn[--Sn],xn[Sn]=null,tr=xn[--Sn],xn[Sn]=null}var Bt=null,Qe=null,De=!1,pa=null,Hn=!1,pf=Error(a(519));function ya(n){var i=Error(a(418,""));throw bl(gn(i,n)),pf}function Qy(n){var i=n.stateNode,l=n.type,c=n.memoizedProps;switch(i[Ct]=n,i[Ht]=c,l){case"dialog":_e("cancel",i),_e("close",i);break;case"iframe":case"object":case"embed":_e("load",i);break;case"video":case"audio":for(l=0;l<Yl.length;l++)_e(Yl[l],i);break;case"source":_e("error",i);break;case"img":case"image":case"link":_e("error",i),_e("load",i);break;case"details":_e("toggle",i);break;case"input":_e("invalid",i),fy(i,c.value,c.defaultValue,c.checked,c.defaultChecked,c.type,c.name,!0),Iu(i);break;case"select":_e("invalid",i);break;case"textarea":_e("invalid",i),hy(i,c.value,c.defaultValue,c.children),Iu(i)}l=c.children,typeof l!="string"&&typeof l!="number"&&typeof l!="bigint"||i.textContent===""+l||c.suppressHydrationWarning===!0||hb(i.textContent,l)?(c.popover!=null&&(_e("beforetoggle",i),_e("toggle",i)),c.onScroll!=null&&_e("scroll",i),c.onScrollEnd!=null&&_e("scrollend",i),c.onClick!=null&&(i.onclick=$o),i=!0):i=!1,i||ya(n)}function Wy(n){for(Bt=n.return;Bt;)switch(Bt.tag){case 5:case 13:Hn=!1;return;case 27:case 3:Hn=!0;return;default:Bt=Bt.return}}function yl(n){if(n!==Bt)return!1;if(!De)return Wy(n),De=!0,!1;var i=n.tag,l;if((l=i!==3&&i!==27)&&((l=i===5)&&(l=n.type,l=!(l!=="form"&&l!=="button")||Nd(n.type,n.memoizedProps)),l=!l),l&&Qe&&ya(n),Wy(n),i===13){if(n=n.memoizedState,n=n!==null?n.dehydrated:null,!n)throw Error(a(317));e:{for(n=n.nextSibling,i=0;n;){if(n.nodeType===8)if(l=n.data,l==="/$"){if(i===0){Qe=Bn(n.nextSibling);break e}i--}else l!=="$"&&l!=="$!"&&l!=="$?"||i++;n=n.nextSibling}Qe=null}}else i===27?(i=Qe,Xr(n.type)?(n=zd,zd=null,Qe=n):Qe=i):Qe=Bt?Bn(n.stateNode.nextSibling):null;return!0}function gl(){Qe=Bt=null,De=!1}function Fy(){var n=pa;return n!==null&&(It===null?It=n:It.push.apply(It,n),pa=null),n}function bl(n){pa===null?pa=[n]:pa.push(n)}var yf=W(null),ga=null,rr=null;function Nr(n,i,l){ae(yf,i._currentValue),i._currentValue=l}function ar(n){n._currentValue=yf.current,Z(yf)}function gf(n,i,l){for(;n!==null;){var c=n.alternate;if((n.childLanes&i)!==i?(n.childLanes|=i,c!==null&&(c.childLanes|=i)):c!==null&&(c.childLanes&i)!==i&&(c.childLanes|=i),n===l)break;n=n.return}}function bf(n,i,l,c){var d=n.child;for(d!==null&&(d.return=n);d!==null;){var p=d.dependencies;if(p!==null){var x=d.child;p=p.firstContext;e:for(;p!==null;){var E=p;p=d;for(var C=0;C<i.length;C++)if(E.context===i[C]){p.lanes|=l,E=p.alternate,E!==null&&(E.lanes|=l),gf(p.return,l,n),c||(x=null);break e}p=E.next}}else if(d.tag===18){if(x=d.return,x===null)throw Error(a(341));x.lanes|=l,p=x.alternate,p!==null&&(p.lanes|=l),gf(x,l,n),x=null}else x=d.child;if(x!==null)x.return=d;else for(x=d;x!==null;){if(x===n){x=null;break}if(d=x.sibling,d!==null){d.return=x.return,x=d;break}x=x.return}d=x}}function xl(n,i,l,c){n=null;for(var d=i,p=!1;d!==null;){if(!p){if((d.flags&524288)!==0)p=!0;else if((d.flags&262144)!==0)break}if(d.tag===10){var x=d.alternate;if(x===null)throw Error(a(387));if(x=x.memoizedProps,x!==null){var E=d.type;Jt(d.pendingProps.value,x.value)||(n!==null?n.push(E):n=[E])}}else if(d===ft.current){if(x=d.alternate,x===null)throw Error(a(387));x.memoizedState.memoizedState!==d.memoizedState.memoizedState&&(n!==null?n.push(Zl):n=[Zl])}d=d.return}n!==null&&bf(i,n,l,c),i.flags|=262144}function uo(n){for(n=n.firstContext;n!==null;){if(!Jt(n.context._currentValue,n.memoizedValue))return!0;n=n.next}return!1}function ba(n){ga=n,rr=null,n=n.dependencies,n!==null&&(n.firstContext=null)}function Dt(n){return Jy(ga,n)}function oo(n,i){return ga===null&&ba(n),Jy(n,i)}function Jy(n,i){var l=i._currentValue;if(i={context:i,memoizedValue:l,next:null},rr===null){if(n===null)throw Error(a(308));rr=i,n.dependencies={lanes:0,firstContext:i},n.flags|=524288}else rr=rr.next=i;return l}var wT=typeof AbortController<"u"?AbortController:function(){var n=[],i=this.signal={aborted:!1,addEventListener:function(l,c){n.push(c)}};this.abort=function(){i.aborted=!0,n.forEach(function(l){return l()})}},OT=e.unstable_scheduleCallback,AT=e.unstable_NormalPriority,lt={$$typeof:j,Consumer:null,Provider:null,_currentValue:null,_currentValue2:null,_threadCount:0};function xf(){return{controller:new wT,data:new Map,refCount:0}}function Sl(n){n.refCount--,n.refCount===0&&OT(AT,function(){n.controller.abort()})}var wl=null,Sf=0,si=0,fi=null;function ET(n,i){if(wl===null){var l=wl=[];Sf=0,si=Od(),fi={status:"pending",value:void 0,then:function(c){l.push(c)}}}return Sf++,i.then(eg,eg),i}function eg(){if(--Sf===0&&wl!==null){fi!==null&&(fi.status="fulfilled");var n=wl;wl=null,si=0,fi=null;for(var i=0;i<n.length;i++)(0,n[i])()}}function _T(n,i){var l=[],c={status:"pending",value:null,reason:null,then:function(d){l.push(d)}};return n.then(function(){c.status="fulfilled",c.value=i;for(var d=0;d<l.length;d++)(0,l[d])(i)},function(d){for(c.status="rejected",c.reason=d,d=0;d<l.length;d++)(0,l[d])(void 0)}),c}var tg=B.S;B.S=function(n,i){typeof i=="object"&&i!==null&&typeof i.then=="function"&&ET(n,i),tg!==null&&tg(n,i)};var xa=W(null);function wf(){var n=xa.current;return n!==null?n:Ye.pooledCache}function co(n,i){i===null?ae(xa,xa.current):ae(xa,i.pool)}function ng(){var n=wf();return n===null?null:{parent:lt._currentValue,pool:n}}var Ol=Error(a(460)),rg=Error(a(474)),so=Error(a(542)),Of={then:function(){}};function ag(n){return n=n.status,n==="fulfilled"||n==="rejected"}function fo(){}function ig(n,i,l){switch(l=n[l],l===void 0?n.push(i):l!==i&&(i.then(fo,fo),i=l),i.status){case"fulfilled":return i.value;case"rejected":throw n=i.reason,ug(n),n;default:if(typeof i.status=="string")i.then(fo,fo);else{if(n=Ye,n!==null&&100<n.shellSuspendCounter)throw Error(a(482));n=i,n.status="pending",n.then(function(c){if(i.status==="pending"){var d=i;d.status="fulfilled",d.value=c}},function(c){if(i.status==="pending"){var d=i;d.status="rejected",d.reason=c}})}switch(i.status){case"fulfilled":return i.value;case"rejected":throw n=i.reason,ug(n),n}throw Al=i,Ol}}var Al=null;function lg(){if(Al===null)throw Error(a(459));var n=Al;return Al=null,n}function ug(n){if(n===Ol||n===so)throw Error(a(483))}var Pr=!1;function Af(n){n.updateQueue={baseState:n.memoizedState,firstBaseUpdate:null,lastBaseUpdate:null,shared:{pending:null,lanes:0,hiddenCallbacks:null},callbacks:null}}function Ef(n,i){n=n.updateQueue,i.updateQueue===n&&(i.updateQueue={baseState:n.baseState,firstBaseUpdate:n.firstBaseUpdate,lastBaseUpdate:n.lastBaseUpdate,shared:n.shared,callbacks:null})}function Rr(n){return{lane:n,tag:0,payload:null,callback:null,next:null}}function kr(n,i,l){var c=n.updateQueue;if(c===null)return null;if(c=c.shared,(Ne&2)!==0){var d=c.pending;return d===null?i.next=i:(i.next=d.next,d.next=i),c.pending=i,i=ro(n),Xy(n,null,l),i}return no(n,c,i,l),ro(n)}function El(n,i,l){if(i=i.updateQueue,i!==null&&(i=i.shared,(l&4194048)!==0)){var c=i.lanes;c&=n.pendingLanes,l|=c,i.lanes=l,ey(n,l)}}function _f(n,i){var l=n.updateQueue,c=n.alternate;if(c!==null&&(c=c.updateQueue,l===c)){var d=null,p=null;if(l=l.firstBaseUpdate,l!==null){do{var x={lane:l.lane,tag:l.tag,payload:l.payload,callback:null,next:null};p===null?d=p=x:p=p.next=x,l=l.next}while(l!==null);p===null?d=p=i:p=p.next=i}else d=p=i;l={baseState:c.baseState,firstBaseUpdate:d,lastBaseUpdate:p,shared:c.shared,callbacks:c.callbacks},n.updateQueue=l;return}n=l.lastBaseUpdate,n===null?l.firstBaseUpdate=i:n.next=i,l.lastBaseUpdate=i}var Tf=!1;function _l(){if(Tf){var n=fi;if(n!==null)throw n}}function Tl(n,i,l,c){Tf=!1;var d=n.updateQueue;Pr=!1;var p=d.firstBaseUpdate,x=d.lastBaseUpdate,E=d.shared.pending;if(E!==null){d.shared.pending=null;var C=E,q=C.next;C.next=null,x===null?p=q:x.next=q,x=C;var K=n.alternate;K!==null&&(K=K.updateQueue,E=K.lastBaseUpdate,E!==x&&(E===null?K.firstBaseUpdate=q:E.next=q,K.lastBaseUpdate=C))}if(p!==null){var Q=d.baseState;x=0,K=q=C=null,E=p;do{var $=E.lane&-536870913,H=$!==E.lane;if(H?(Te&$)===$:(c&$)===$){$!==0&&$===si&&(Tf=!0),K!==null&&(K=K.next={lane:0,tag:E.tag,payload:E.payload,callback:null,next:null});e:{var xe=n,pe=E;$=i;var Ue=l;switch(pe.tag){case 1:if(xe=pe.payload,typeof xe=="function"){Q=xe.call(Ue,Q,$);break e}Q=xe;break e;case 3:xe.flags=xe.flags&-65537|128;case 0:if(xe=pe.payload,$=typeof xe=="function"?xe.call(Ue,Q,$):xe,$==null)break e;Q=m({},Q,$);break e;case 2:Pr=!0}}$=E.callback,$!==null&&(n.flags|=64,H&&(n.flags|=8192),H=d.callbacks,H===null?d.callbacks=[$]:H.push($))}else H={lane:$,tag:E.tag,payload:E.payload,callback:E.callback,next:null},K===null?(q=K=H,C=Q):K=K.next=H,x|=$;if(E=E.next,E===null){if(E=d.shared.pending,E===null)break;H=E,E=H.next,H.next=null,d.lastBaseUpdate=H,d.shared.pending=null}}while(!0);K===null&&(C=Q),d.baseState=C,d.firstBaseUpdate=q,d.lastBaseUpdate=K,p===null&&(d.shared.lanes=0),Yr|=x,n.lanes=x,n.memoizedState=Q}}function og(n,i){if(typeof n!="function")throw Error(a(191,n));n.call(i)}function cg(n,i){var l=n.callbacks;if(l!==null)for(n.callbacks=null,n=0;n<l.length;n++)og(l[n],i)}var di=W(null),ho=W(0);function sg(n,i){n=fr,ae(ho,n),ae(di,i),fr=n|i.baseLanes}function jf(){ae(ho,fr),ae(di,di.current)}function Mf(){fr=ho.current,Z(di),Z(ho)}var zr=0,Oe=null,Be=null,nt=null,vo=!1,hi=!1,Sa=!1,mo=0,jl=0,vi=null,TT=0;function Je(){throw Error(a(321))}function Cf(n,i){if(i===null)return!1;for(var l=0;l<i.length&&l<n.length;l++)if(!Jt(n[l],i[l]))return!1;return!0}function Df(n,i,l,c,d,p){return zr=p,Oe=i,i.memoizedState=null,i.updateQueue=null,i.lanes=0,B.H=n===null||n.memoizedState===null?Xg:Vg,Sa=!1,p=l(c,d),Sa=!1,hi&&(p=dg(i,l,c,d)),fg(n),p}function fg(n){B.H=So;var i=Be!==null&&Be.next!==null;if(zr=0,nt=Be=Oe=null,vo=!1,jl=0,vi=null,i)throw Error(a(300));n===null||vt||(n=n.dependencies,n!==null&&uo(n)&&(vt=!0))}function dg(n,i,l,c){Oe=n;var d=0;do{if(hi&&(vi=null),jl=0,hi=!1,25<=d)throw Error(a(301));if(d+=1,nt=Be=null,n.updateQueue!=null){var p=n.updateQueue;p.lastEffect=null,p.events=null,p.stores=null,p.memoCache!=null&&(p.memoCache.index=0)}B.H=RT,p=i(l,c)}while(hi);return p}function jT(){var n=B.H,i=n.useState()[0];return i=typeof i.then=="function"?Ml(i):i,n=n.useState()[0],(Be!==null?Be.memoizedState:null)!==n&&(Oe.flags|=1024),i}function Nf(){var n=mo!==0;return mo=0,n}function Pf(n,i,l){i.updateQueue=n.updateQueue,i.flags&=-2053,n.lanes&=~l}function Rf(n){if(vo){for(n=n.memoizedState;n!==null;){var i=n.queue;i!==null&&(i.pending=null),n=n.next}vo=!1}zr=0,nt=Be=Oe=null,hi=!1,jl=mo=0,vi=null}function Gt(){var n={memoizedState:null,baseState:null,baseQueue:null,queue:null,next:null};return nt===null?Oe.memoizedState=nt=n:nt=nt.next=n,nt}function rt(){if(Be===null){var n=Oe.alternate;n=n!==null?n.memoizedState:null}else n=Be.next;var i=nt===null?Oe.memoizedState:nt.next;if(i!==null)nt=i,Be=n;else{if(n===null)throw Oe.alternate===null?Error(a(467)):Error(a(310));Be=n,n={memoizedState:Be.memoizedState,baseState:Be.baseState,baseQueue:Be.baseQueue,queue:Be.queue,next:null},nt===null?Oe.memoizedState=nt=n:nt=nt.next=n}return nt}function kf(){return{lastEffect:null,events:null,stores:null,memoCache:null}}function Ml(n){var i=jl;return jl+=1,vi===null&&(vi=[]),n=ig(vi,n,i),i=Oe,(nt===null?i.memoizedState:nt.next)===null&&(i=i.alternate,B.H=i===null||i.memoizedState===null?Xg:Vg),n}function po(n){if(n!==null&&typeof n=="object"){if(typeof n.then=="function")return Ml(n);if(n.$$typeof===j)return Dt(n)}throw Error(a(438,String(n)))}function zf(n){var i=null,l=Oe.updateQueue;if(l!==null&&(i=l.memoCache),i==null){var c=Oe.alternate;c!==null&&(c=c.updateQueue,c!==null&&(c=c.memoCache,c!=null&&(i={data:c.data.map(function(d){return d.slice()}),index:0})))}if(i==null&&(i={data:[],index:0}),l===null&&(l=kf(),Oe.updateQueue=l),l.memoCache=i,l=i.data[i.index],l===void 0)for(l=i.data[i.index]=Array(n),c=0;c<n;c++)l[c]=ne;return i.index++,l}function ir(n,i){return typeof i=="function"?i(n):i}function yo(n){var i=rt();return Bf(i,Be,n)}function Bf(n,i,l){var c=n.queue;if(c===null)throw Error(a(311));c.lastRenderedReducer=l;var d=n.baseQueue,p=c.pending;if(p!==null){if(d!==null){var x=d.next;d.next=p.next,p.next=x}i.baseQueue=d=p,c.pending=null}if(p=n.baseState,d===null)n.memoizedState=p;else{i=d.next;var E=x=null,C=null,q=i,K=!1;do{var Q=q.lane&-536870913;if(Q!==q.lane?(Te&Q)===Q:(zr&Q)===Q){var $=q.revertLane;if($===0)C!==null&&(C=C.next={lane:0,revertLane:0,action:q.action,hasEagerState:q.hasEagerState,eagerState:q.eagerState,next:null}),Q===si&&(K=!0);else if((zr&$)===$){q=q.next,$===si&&(K=!0);continue}else Q={lane:0,revertLane:q.revertLane,action:q.action,hasEagerState:q.hasEagerState,eagerState:q.eagerState,next:null},C===null?(E=C=Q,x=p):C=C.next=Q,Oe.lanes|=$,Yr|=$;Q=q.action,Sa&&l(p,Q),p=q.hasEagerState?q.eagerState:l(p,Q)}else $={lane:Q,revertLane:q.revertLane,action:q.action,hasEagerState:q.hasEagerState,eagerState:q.eagerState,next:null},C===null?(E=C=$,x=p):C=C.next=$,Oe.lanes|=Q,Yr|=Q;q=q.next}while(q!==null&&q!==i);if(C===null?x=p:C.next=E,!Jt(p,n.memoizedState)&&(vt=!0,K&&(l=fi,l!==null)))throw l;n.memoizedState=p,n.baseState=x,n.baseQueue=C,c.lastRenderedState=p}return d===null&&(c.lanes=0),[n.memoizedState,c.dispatch]}function Lf(n){var i=rt(),l=i.queue;if(l===null)throw Error(a(311));l.lastRenderedReducer=n;var c=l.dispatch,d=l.pending,p=i.memoizedState;if(d!==null){l.pending=null;var x=d=d.next;do p=n(p,x.action),x=x.next;while(x!==d);Jt(p,i.memoizedState)||(vt=!0),i.memoizedState=p,i.baseQueue===null&&(i.baseState=p),l.lastRenderedState=p}return[p,c]}function hg(n,i,l){var c=Oe,d=rt(),p=De;if(p){if(l===void 0)throw Error(a(407));l=l()}else l=i();var x=!Jt((Be||d).memoizedState,l);x&&(d.memoizedState=l,vt=!0),d=d.queue;var E=pg.bind(null,c,d,n);if(Cl(2048,8,E,[n]),d.getSnapshot!==i||x||nt!==null&&nt.memoizedState.tag&1){if(c.flags|=2048,mi(9,go(),mg.bind(null,c,d,l,i),null),Ye===null)throw Error(a(349));p||(zr&124)!==0||vg(c,i,l)}return l}function vg(n,i,l){n.flags|=16384,n={getSnapshot:i,value:l},i=Oe.updateQueue,i===null?(i=kf(),Oe.updateQueue=i,i.stores=[n]):(l=i.stores,l===null?i.stores=[n]:l.push(n))}function mg(n,i,l,c){i.value=l,i.getSnapshot=c,yg(i)&&gg(n)}function pg(n,i,l){return l(function(){yg(i)&&gg(n)})}function yg(n){var i=n.getSnapshot;n=n.value;try{var l=i();return!Jt(n,l)}catch{return!0}}function gg(n){var i=li(n,2);i!==null&&ln(i,n,2)}function Uf(n){var i=Gt();if(typeof n=="function"){var l=n;if(n=l(),Sa){Mr(!0);try{l()}finally{Mr(!1)}}}return i.memoizedState=i.baseState=n,i.queue={pending:null,lanes:0,dispatch:null,lastRenderedReducer:ir,lastRenderedState:n},i}function bg(n,i,l,c){return n.baseState=l,Bf(n,Be,typeof c=="function"?c:ir)}function MT(n,i,l,c,d){if(xo(n))throw Error(a(485));if(n=i.action,n!==null){var p={payload:d,action:n,next:null,isTransition:!0,status:"pending",value:null,reason:null,listeners:[],then:function(x){p.listeners.push(x)}};B.T!==null?l(!0):p.isTransition=!1,c(p),l=i.pending,l===null?(p.next=i.pending=p,xg(i,p)):(p.next=l.next,i.pending=l.next=p)}}function xg(n,i){var l=i.action,c=i.payload,d=n.state;if(i.isTransition){var p=B.T,x={};B.T=x;try{var E=l(d,c),C=B.S;C!==null&&C(x,E),Sg(n,i,E)}catch(q){qf(n,i,q)}finally{B.T=p}}else try{p=l(d,c),Sg(n,i,p)}catch(q){qf(n,i,q)}}function Sg(n,i,l){l!==null&&typeof l=="object"&&typeof l.then=="function"?l.then(function(c){wg(n,i,c)},function(c){return qf(n,i,c)}):wg(n,i,l)}function wg(n,i,l){i.status="fulfilled",i.value=l,Og(i),n.state=l,i=n.pending,i!==null&&(l=i.next,l===i?n.pending=null:(l=l.next,i.next=l,xg(n,l)))}function qf(n,i,l){var c=n.pending;if(n.pending=null,c!==null){c=c.next;do i.status="rejected",i.reason=l,Og(i),i=i.next;while(i!==c)}n.action=null}function Og(n){n=n.listeners;for(var i=0;i<n.length;i++)(0,n[i])()}function Ag(n,i){return i}function Eg(n,i){if(De){var l=Ye.formState;if(l!==null){e:{var c=Oe;if(De){if(Qe){t:{for(var d=Qe,p=Hn;d.nodeType!==8;){if(!p){d=null;break t}if(d=Bn(d.nextSibling),d===null){d=null;break t}}p=d.data,d=p==="F!"||p==="F"?d:null}if(d){Qe=Bn(d.nextSibling),c=d.data==="F!";break e}}ya(c)}c=!1}c&&(i=l[0])}}return l=Gt(),l.memoizedState=l.baseState=i,c={pending:null,lanes:0,dispatch:null,lastRenderedReducer:Ag,lastRenderedState:i},l.queue=c,l=Gg.bind(null,Oe,c),c.dispatch=l,c=Uf(!1),p=Kf.bind(null,Oe,!1,c.queue),c=Gt(),d={state:i,dispatch:null,action:n,pending:null},c.queue=d,l=MT.bind(null,Oe,d,p,l),d.dispatch=l,c.memoizedState=n,[i,l,!1]}function _g(n){var i=rt();return Tg(i,Be,n)}function Tg(n,i,l){if(i=Bf(n,i,Ag)[0],n=yo(ir)[0],typeof i=="object"&&i!==null&&typeof i.then=="function")try{var c=Ml(i)}catch(x){throw x===Ol?so:x}else c=i;i=rt();var d=i.queue,p=d.dispatch;return l!==i.memoizedState&&(Oe.flags|=2048,mi(9,go(),CT.bind(null,d,l),null)),[c,p,n]}function CT(n,i){n.action=i}function jg(n){var i=rt(),l=Be;if(l!==null)return Tg(i,l,n);rt(),i=i.memoizedState,l=rt();var c=l.queue.dispatch;return l.memoizedState=n,[i,c,!1]}function mi(n,i,l,c){return n={tag:n,create:l,deps:c,inst:i,next:null},i=Oe.updateQueue,i===null&&(i=kf(),Oe.updateQueue=i),l=i.lastEffect,l===null?i.lastEffect=n.next=n:(c=l.next,l.next=n,n.next=c,i.lastEffect=n),n}function go(){return{destroy:void 0,resource:void 0}}function Mg(){return rt().memoizedState}function bo(n,i,l,c){var d=Gt();c=c===void 0?null:c,Oe.flags|=n,d.memoizedState=mi(1|i,go(),l,c)}function Cl(n,i,l,c){var d=rt();c=c===void 0?null:c;var p=d.memoizedState.inst;Be!==null&&c!==null&&Cf(c,Be.memoizedState.deps)?d.memoizedState=mi(i,p,l,c):(Oe.flags|=n,d.memoizedState=mi(1|i,p,l,c))}function Cg(n,i){bo(8390656,8,n,i)}function Dg(n,i){Cl(2048,8,n,i)}function Ng(n,i){return Cl(4,2,n,i)}function Pg(n,i){return Cl(4,4,n,i)}function Rg(n,i){if(typeof i=="function"){n=n();var l=i(n);return function(){typeof l=="function"?l():i(null)}}if(i!=null)return n=n(),i.current=n,function(){i.current=null}}function kg(n,i,l){l=l!=null?l.concat([n]):null,Cl(4,4,Rg.bind(null,i,n),l)}function $f(){}function zg(n,i){var l=rt();i=i===void 0?null:i;var c=l.memoizedState;return i!==null&&Cf(i,c[1])?c[0]:(l.memoizedState=[n,i],n)}function Bg(n,i){var l=rt();i=i===void 0?null:i;var c=l.memoizedState;if(i!==null&&Cf(i,c[1]))return c[0];if(c=n(),Sa){Mr(!0);try{n()}finally{Mr(!1)}}return l.memoizedState=[c,i],c}function Hf(n,i,l){return l===void 0||(zr&1073741824)!==0?n.memoizedState=i:(n.memoizedState=l,n=q0(),Oe.lanes|=n,Yr|=n,l)}function Lg(n,i,l,c){return Jt(l,i)?l:di.current!==null?(n=Hf(n,l,c),Jt(n,i)||(vt=!0),n):(zr&42)===0?(vt=!0,n.memoizedState=l):(n=q0(),Oe.lanes|=n,Yr|=n,i)}function Ug(n,i,l,c,d){var p=I.p;I.p=p!==0&&8>p?p:8;var x=B.T,E={};B.T=E,Kf(n,!1,i,l);try{var C=d(),q=B.S;if(q!==null&&q(E,C),C!==null&&typeof C=="object"&&typeof C.then=="function"){var K=_T(C,c);Dl(n,i,K,an(n))}else Dl(n,i,c,an(n))}catch(Q){Dl(n,i,{then:function(){},status:"rejected",reason:Q},an())}finally{I.p=p,B.T=x}}function DT(){}function Yf(n,i,l,c){if(n.tag!==5)throw Error(a(476));var d=qg(n).queue;Ug(n,d,i,F,l===null?DT:function(){return $g(n),l(c)})}function qg(n){var i=n.memoizedState;if(i!==null)return i;i={memoizedState:F,baseState:F,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:ir,lastRenderedState:F},next:null};var l={};return i.next={memoizedState:l,baseState:l,baseQueue:null,queue:{pending:null,lanes:0,dispatch:null,lastRenderedReducer:ir,lastRenderedState:l},next:null},n.memoizedState=i,n=n.alternate,n!==null&&(n.memoizedState=i),i}function $g(n){var i=qg(n).next.queue;Dl(n,i,{},an())}function Gf(){return Dt(Zl)}function Hg(){return rt().memoizedState}function Yg(){return rt().memoizedState}function NT(n){for(var i=n.return;i!==null;){switch(i.tag){case 24:case 3:var l=an();n=Rr(l);var c=kr(i,n,l);c!==null&&(ln(c,i,l),El(c,i,l)),i={cache:xf()},n.payload=i;return}i=i.return}}function PT(n,i,l){var c=an();l={lane:c,revertLane:0,action:l,hasEagerState:!1,eagerState:null,next:null},xo(n)?Kg(i,l):(l=sf(n,i,l,c),l!==null&&(ln(l,n,c),Ig(l,i,c)))}function Gg(n,i,l){var c=an();Dl(n,i,l,c)}function Dl(n,i,l,c){var d={lane:c,revertLane:0,action:l,hasEagerState:!1,eagerState:null,next:null};if(xo(n))Kg(i,d);else{var p=n.alternate;if(n.lanes===0&&(p===null||p.lanes===0)&&(p=i.lastRenderedReducer,p!==null))try{var x=i.lastRenderedState,E=p(x,l);if(d.hasEagerState=!0,d.eagerState=E,Jt(E,x))return no(n,i,d,0),Ye===null&&to(),!1}catch{}finally{}if(l=sf(n,i,d,c),l!==null)return ln(l,n,c),Ig(l,i,c),!0}return!1}function Kf(n,i,l,c){if(c={lane:2,revertLane:Od(),action:c,hasEagerState:!1,eagerState:null,next:null},xo(n)){if(i)throw Error(a(479))}else i=sf(n,l,c,2),i!==null&&ln(i,n,2)}function xo(n){var i=n.alternate;return n===Oe||i!==null&&i===Oe}function Kg(n,i){hi=vo=!0;var l=n.pending;l===null?i.next=i:(i.next=l.next,l.next=i),n.pending=i}function Ig(n,i,l){if((l&4194048)!==0){var c=i.lanes;c&=n.pendingLanes,l|=c,i.lanes=l,ey(n,l)}}var So={readContext:Dt,use:po,useCallback:Je,useContext:Je,useEffect:Je,useImperativeHandle:Je,useLayoutEffect:Je,useInsertionEffect:Je,useMemo:Je,useReducer:Je,useRef:Je,useState:Je,useDebugValue:Je,useDeferredValue:Je,useTransition:Je,useSyncExternalStore:Je,useId:Je,useHostTransitionStatus:Je,useFormState:Je,useActionState:Je,useOptimistic:Je,useMemoCache:Je,useCacheRefresh:Je},Xg={readContext:Dt,use:po,useCallback:function(n,i){return Gt().memoizedState=[n,i===void 0?null:i],n},useContext:Dt,useEffect:Cg,useImperativeHandle:function(n,i,l){l=l!=null?l.concat([n]):null,bo(4194308,4,Rg.bind(null,i,n),l)},useLayoutEffect:function(n,i){return bo(4194308,4,n,i)},useInsertionEffect:function(n,i){bo(4,2,n,i)},useMemo:function(n,i){var l=Gt();i=i===void 0?null:i;var c=n();if(Sa){Mr(!0);try{n()}finally{Mr(!1)}}return l.memoizedState=[c,i],c},useReducer:function(n,i,l){var c=Gt();if(l!==void 0){var d=l(i);if(Sa){Mr(!0);try{l(i)}finally{Mr(!1)}}}else d=i;return c.memoizedState=c.baseState=d,n={pending:null,lanes:0,dispatch:null,lastRenderedReducer:n,lastRenderedState:d},c.queue=n,n=n.dispatch=PT.bind(null,Oe,n),[c.memoizedState,n]},useRef:function(n){var i=Gt();return n={current:n},i.memoizedState=n},useState:function(n){n=Uf(n);var i=n.queue,l=Gg.bind(null,Oe,i);return i.dispatch=l,[n.memoizedState,l]},useDebugValue:$f,useDeferredValue:function(n,i){var l=Gt();return Hf(l,n,i)},useTransition:function(){var n=Uf(!1);return n=Ug.bind(null,Oe,n.queue,!0,!1),Gt().memoizedState=n,[!1,n]},useSyncExternalStore:function(n,i,l){var c=Oe,d=Gt();if(De){if(l===void 0)throw Error(a(407));l=l()}else{if(l=i(),Ye===null)throw Error(a(349));(Te&124)!==0||vg(c,i,l)}d.memoizedState=l;var p={value:l,getSnapshot:i};return d.queue=p,Cg(pg.bind(null,c,p,n),[n]),c.flags|=2048,mi(9,go(),mg.bind(null,c,p,l,i),null),l},useId:function(){var n=Gt(),i=Ye.identifierPrefix;if(De){var l=nr,c=tr;l=(c&~(1<<32-Ft(c)-1)).toString(32)+l,i="«"+i+"R"+l,l=mo++,0<l&&(i+="H"+l.toString(32)),i+="»"}else l=TT++,i="«"+i+"r"+l.toString(32)+"»";return n.memoizedState=i},useHostTransitionStatus:Gf,useFormState:Eg,useActionState:Eg,useOptimistic:function(n){var i=Gt();i.memoizedState=i.baseState=n;var l={pending:null,lanes:0,dispatch:null,lastRenderedReducer:null,lastRenderedState:null};return i.queue=l,i=Kf.bind(null,Oe,!0,l),l.dispatch=i,[n,i]},useMemoCache:zf,useCacheRefresh:function(){return Gt().memoizedState=NT.bind(null,Oe)}},Vg={readContext:Dt,use:po,useCallback:zg,useContext:Dt,useEffect:Dg,useImperativeHandle:kg,useInsertionEffect:Ng,useLayoutEffect:Pg,useMemo:Bg,useReducer:yo,useRef:Mg,useState:function(){return yo(ir)},useDebugValue:$f,useDeferredValue:function(n,i){var l=rt();return Lg(l,Be.memoizedState,n,i)},useTransition:function(){var n=yo(ir)[0],i=rt().memoizedState;return[typeof n=="boolean"?n:Ml(n),i]},useSyncExternalStore:hg,useId:Hg,useHostTransitionStatus:Gf,useFormState:_g,useActionState:_g,useOptimistic:function(n,i){var l=rt();return bg(l,Be,n,i)},useMemoCache:zf,useCacheRefresh:Yg},RT={readContext:Dt,use:po,useCallback:zg,useContext:Dt,useEffect:Dg,useImperativeHandle:kg,useInsertionEffect:Ng,useLayoutEffect:Pg,useMemo:Bg,useReducer:Lf,useRef:Mg,useState:function(){return Lf(ir)},useDebugValue:$f,useDeferredValue:function(n,i){var l=rt();return Be===null?Hf(l,n,i):Lg(l,Be.memoizedState,n,i)},useTransition:function(){var n=Lf(ir)[0],i=rt().memoizedState;return[typeof n=="boolean"?n:Ml(n),i]},useSyncExternalStore:hg,useId:Hg,useHostTransitionStatus:Gf,useFormState:jg,useActionState:jg,useOptimistic:function(n,i){var l=rt();return Be!==null?bg(l,Be,n,i):(l.baseState=n,[n,l.queue.dispatch])},useMemoCache:zf,useCacheRefresh:Yg},pi=null,Nl=0;function wo(n){var i=Nl;return Nl+=1,pi===null&&(pi=[]),ig(pi,n,i)}function Pl(n,i){i=i.props.ref,n.ref=i!==void 0?i:null}function Oo(n,i){throw i.$$typeof===y?Error(a(525)):(n=Object.prototype.toString.call(i),Error(a(31,n==="[object Object]"?"object with keys {"+Object.keys(i).join(", ")+"}":n)))}function Zg(n){var i=n._init;return i(n._payload)}function Qg(n){function i(z,R){if(n){var U=z.deletions;U===null?(z.deletions=[R],z.flags|=16):U.push(R)}}function l(z,R){if(!n)return null;for(;R!==null;)i(z,R),R=R.sibling;return null}function c(z){for(var R=new Map;z!==null;)z.key!==null?R.set(z.key,z):R.set(z.index,z),z=z.sibling;return R}function d(z,R){return z=er(z,R),z.index=0,z.sibling=null,z}function p(z,R,U){return z.index=U,n?(U=z.alternate,U!==null?(U=U.index,U<R?(z.flags|=67108866,R):U):(z.flags|=67108866,R)):(z.flags|=1048576,R)}function x(z){return n&&z.alternate===null&&(z.flags|=67108866),z}function E(z,R,U,X){return R===null||R.tag!==6?(R=df(U,z.mode,X),R.return=z,R):(R=d(R,U),R.return=z,R)}function C(z,R,U,X){var le=U.type;return le===w?K(z,R,U.props.children,X,U.key):R!==null&&(R.elementType===le||typeof le=="object"&&le!==null&&le.$$typeof===G&&Zg(le)===R.type)?(R=d(R,U.props),Pl(R,U),R.return=z,R):(R=ao(U.type,U.key,U.props,null,z.mode,X),Pl(R,U),R.return=z,R)}function q(z,R,U,X){return R===null||R.tag!==4||R.stateNode.containerInfo!==U.containerInfo||R.stateNode.implementation!==U.implementation?(R=hf(U,z.mode,X),R.return=z,R):(R=d(R,U.children||[]),R.return=z,R)}function K(z,R,U,X,le){return R===null||R.tag!==7?(R=ha(U,z.mode,X,le),R.return=z,R):(R=d(R,U),R.return=z,R)}function Q(z,R,U){if(typeof R=="string"&&R!==""||typeof R=="number"||typeof R=="bigint")return R=df(""+R,z.mode,U),R.return=z,R;if(typeof R=="object"&&R!==null){switch(R.$$typeof){case g:return U=ao(R.type,R.key,R.props,null,z.mode,U),Pl(U,R),U.return=z,U;case b:return R=hf(R,z.mode,U),R.return=z,R;case G:var X=R._init;return R=X(R._payload),Q(z,R,U)}if(se(R)||re(R))return R=ha(R,z.mode,U,null),R.return=z,R;if(typeof R.then=="function")return Q(z,wo(R),U);if(R.$$typeof===j)return Q(z,oo(z,R),U);Oo(z,R)}return null}function $(z,R,U,X){var le=R!==null?R.key:null;if(typeof U=="string"&&U!==""||typeof U=="number"||typeof U=="bigint")return le!==null?null:E(z,R,""+U,X);if(typeof U=="object"&&U!==null){switch(U.$$typeof){case g:return U.key===le?C(z,R,U,X):null;case b:return U.key===le?q(z,R,U,X):null;case G:return le=U._init,U=le(U._payload),$(z,R,U,X)}if(se(U)||re(U))return le!==null?null:K(z,R,U,X,null);if(typeof U.then=="function")return $(z,R,wo(U),X);if(U.$$typeof===j)return $(z,R,oo(z,U),X);Oo(z,U)}return null}function H(z,R,U,X,le){if(typeof X=="string"&&X!==""||typeof X=="number"||typeof X=="bigint")return z=z.get(U)||null,E(R,z,""+X,le);if(typeof X=="object"&&X!==null){switch(X.$$typeof){case g:return z=z.get(X.key===null?U:X.key)||null,C(R,z,X,le);case b:return z=z.get(X.key===null?U:X.key)||null,q(R,z,X,le);case G:var Ae=X._init;return X=Ae(X._payload),H(z,R,U,X,le)}if(se(X)||re(X))return z=z.get(U)||null,K(R,z,X,le,null);if(typeof X.then=="function")return H(z,R,U,wo(X),le);if(X.$$typeof===j)return H(z,R,U,oo(R,X),le);Oo(R,X)}return null}function xe(z,R,U,X){for(var le=null,Ae=null,de=R,be=R=0,pt=null;de!==null&&be<U.length;be++){de.index>be?(pt=de,de=null):pt=de.sibling;var Me=$(z,de,U[be],X);if(Me===null){de===null&&(de=pt);break}n&&de&&Me.alternate===null&&i(z,de),R=p(Me,R,be),Ae===null?le=Me:Ae.sibling=Me,Ae=Me,de=pt}if(be===U.length)return l(z,de),De&&ma(z,be),le;if(de===null){for(;be<U.length;be++)de=Q(z,U[be],X),de!==null&&(R=p(de,R,be),Ae===null?le=de:Ae.sibling=de,Ae=de);return De&&ma(z,be),le}for(de=c(de);be<U.length;be++)pt=H(de,z,be,U[be],X),pt!==null&&(n&&pt.alternate!==null&&de.delete(pt.key===null?be:pt.key),R=p(pt,R,be),Ae===null?le=pt:Ae.sibling=pt,Ae=pt);return n&&de.forEach(function(Fr){return i(z,Fr)}),De&&ma(z,be),le}function pe(z,R,U,X){if(U==null)throw Error(a(151));for(var le=null,Ae=null,de=R,be=R=0,pt=null,Me=U.next();de!==null&&!Me.done;be++,Me=U.next()){de.index>be?(pt=de,de=null):pt=de.sibling;var Fr=$(z,de,Me.value,X);if(Fr===null){de===null&&(de=pt);break}n&&de&&Fr.alternate===null&&i(z,de),R=p(Fr,R,be),Ae===null?le=Fr:Ae.sibling=Fr,Ae=Fr,de=pt}if(Me.done)return l(z,de),De&&ma(z,be),le;if(de===null){for(;!Me.done;be++,Me=U.next())Me=Q(z,Me.value,X),Me!==null&&(R=p(Me,R,be),Ae===null?le=Me:Ae.sibling=Me,Ae=Me);return De&&ma(z,be),le}for(de=c(de);!Me.done;be++,Me=U.next())Me=H(de,z,be,Me.value,X),Me!==null&&(n&&Me.alternate!==null&&de.delete(Me.key===null?be:Me.key),R=p(Me,R,be),Ae===null?le=Me:Ae.sibling=Me,Ae=Me);return n&&de.forEach(function(kj){return i(z,kj)}),De&&ma(z,be),le}function Ue(z,R,U,X){if(typeof U=="object"&&U!==null&&U.type===w&&U.key===null&&(U=U.props.children),typeof U=="object"&&U!==null){switch(U.$$typeof){case g:e:{for(var le=U.key;R!==null;){if(R.key===le){if(le=U.type,le===w){if(R.tag===7){l(z,R.sibling),X=d(R,U.props.children),X.return=z,z=X;break e}}else if(R.elementType===le||typeof le=="object"&&le!==null&&le.$$typeof===G&&Zg(le)===R.type){l(z,R.sibling),X=d(R,U.props),Pl(X,U),X.return=z,z=X;break e}l(z,R);break}else i(z,R);R=R.sibling}U.type===w?(X=ha(U.props.children,z.mode,X,U.key),X.return=z,z=X):(X=ao(U.type,U.key,U.props,null,z.mode,X),Pl(X,U),X.return=z,z=X)}return x(z);case b:e:{for(le=U.key;R!==null;){if(R.key===le)if(R.tag===4&&R.stateNode.containerInfo===U.containerInfo&&R.stateNode.implementation===U.implementation){l(z,R.sibling),X=d(R,U.children||[]),X.return=z,z=X;break e}else{l(z,R);break}else i(z,R);R=R.sibling}X=hf(U,z.mode,X),X.return=z,z=X}return x(z);case G:return le=U._init,U=le(U._payload),Ue(z,R,U,X)}if(se(U))return xe(z,R,U,X);if(re(U)){if(le=re(U),typeof le!="function")throw Error(a(150));return U=le.call(U),pe(z,R,U,X)}if(typeof U.then=="function")return Ue(z,R,wo(U),X);if(U.$$typeof===j)return Ue(z,R,oo(z,U),X);Oo(z,U)}return typeof U=="string"&&U!==""||typeof U=="number"||typeof U=="bigint"?(U=""+U,R!==null&&R.tag===6?(l(z,R.sibling),X=d(R,U),X.return=z,z=X):(l(z,R),X=df(U,z.mode,X),X.return=z,z=X),x(z)):l(z,R)}return function(z,R,U,X){try{Nl=0;var le=Ue(z,R,U,X);return pi=null,le}catch(de){if(de===Ol||de===so)throw de;var Ae=en(29,de,null,z.mode);return Ae.lanes=X,Ae.return=z,Ae}finally{}}}var yi=Qg(!0),Wg=Qg(!1),wn=W(null),Yn=null;function Br(n){var i=n.alternate;ae(ut,ut.current&1),ae(wn,n),Yn===null&&(i===null||di.current!==null||i.memoizedState!==null)&&(Yn=n)}function Fg(n){if(n.tag===22){if(ae(ut,ut.current),ae(wn,n),Yn===null){var i=n.alternate;i!==null&&i.memoizedState!==null&&(Yn=n)}}else Lr()}function Lr(){ae(ut,ut.current),ae(wn,wn.current)}function lr(n){Z(wn),Yn===n&&(Yn=null),Z(ut)}var ut=W(0);function Ao(n){for(var i=n;i!==null;){if(i.tag===13){var l=i.memoizedState;if(l!==null&&(l=l.dehydrated,l===null||l.data==="$?"||kd(l)))return i}else if(i.tag===19&&i.memoizedProps.revealOrder!==void 0){if((i.flags&128)!==0)return i}else if(i.child!==null){i.child.return=i,i=i.child;continue}if(i===n)break;for(;i.sibling===null;){if(i.return===null||i.return===n)return null;i=i.return}i.sibling.return=i.return,i=i.sibling}return null}function If(n,i,l,c){i=n.memoizedState,l=l(c,i),l=l==null?i:m({},i,l),n.memoizedState=l,n.lanes===0&&(n.updateQueue.baseState=l)}var Xf={enqueueSetState:function(n,i,l){n=n._reactInternals;var c=an(),d=Rr(c);d.payload=i,l!=null&&(d.callback=l),i=kr(n,d,c),i!==null&&(ln(i,n,c),El(i,n,c))},enqueueReplaceState:function(n,i,l){n=n._reactInternals;var c=an(),d=Rr(c);d.tag=1,d.payload=i,l!=null&&(d.callback=l),i=kr(n,d,c),i!==null&&(ln(i,n,c),El(i,n,c))},enqueueForceUpdate:function(n,i){n=n._reactInternals;var l=an(),c=Rr(l);c.tag=2,i!=null&&(c.callback=i),i=kr(n,c,l),i!==null&&(ln(i,n,l),El(i,n,l))}};function Jg(n,i,l,c,d,p,x){return n=n.stateNode,typeof n.shouldComponentUpdate=="function"?n.shouldComponentUpdate(c,p,x):i.prototype&&i.prototype.isPureReactComponent?!ml(l,c)||!ml(d,p):!0}function e0(n,i,l,c){n=i.state,typeof i.componentWillReceiveProps=="function"&&i.componentWillReceiveProps(l,c),typeof i.UNSAFE_componentWillReceiveProps=="function"&&i.UNSAFE_componentWillReceiveProps(l,c),i.state!==n&&Xf.enqueueReplaceState(i,i.state,null)}function wa(n,i){var l=i;if("ref"in i){l={};for(var c in i)c!=="ref"&&(l[c]=i[c])}if(n=n.defaultProps){l===i&&(l=m({},l));for(var d in n)l[d]===void 0&&(l[d]=n[d])}return l}var Eo=typeof reportError=="function"?reportError:function(n){if(typeof window=="object"&&typeof window.ErrorEvent=="function"){var i=new window.ErrorEvent("error",{bubbles:!0,cancelable:!0,message:typeof n=="object"&&n!==null&&typeof n.message=="string"?String(n.message):String(n),error:n});if(!window.dispatchEvent(i))return}else if(typeof process=="object"&&typeof process.emit=="function"){process.emit("uncaughtException",n);return}console.error(n)};function t0(n){Eo(n)}function n0(n){console.error(n)}function r0(n){Eo(n)}function _o(n,i){try{var l=n.onUncaughtError;l(i.value,{componentStack:i.stack})}catch(c){setTimeout(function(){throw c})}}function a0(n,i,l){try{var c=n.onCaughtError;c(l.value,{componentStack:l.stack,errorBoundary:i.tag===1?i.stateNode:null})}catch(d){setTimeout(function(){throw d})}}function Vf(n,i,l){return l=Rr(l),l.tag=3,l.payload={element:null},l.callback=function(){_o(n,i)},l}function i0(n){return n=Rr(n),n.tag=3,n}function l0(n,i,l,c){var d=l.type.getDerivedStateFromError;if(typeof d=="function"){var p=c.value;n.payload=function(){return d(p)},n.callback=function(){a0(i,l,c)}}var x=l.stateNode;x!==null&&typeof x.componentDidCatch=="function"&&(n.callback=function(){a0(i,l,c),typeof d!="function"&&(Gr===null?Gr=new Set([this]):Gr.add(this));var E=c.stack;this.componentDidCatch(c.value,{componentStack:E!==null?E:""})})}function kT(n,i,l,c,d){if(l.flags|=32768,c!==null&&typeof c=="object"&&typeof c.then=="function"){if(i=l.alternate,i!==null&&xl(i,l,d,!0),l=wn.current,l!==null){switch(l.tag){case 13:return Yn===null?gd():l.alternate===null&&We===0&&(We=3),l.flags&=-257,l.flags|=65536,l.lanes=d,c===Of?l.flags|=16384:(i=l.updateQueue,i===null?l.updateQueue=new Set([c]):i.add(c),xd(n,c,d)),!1;case 22:return l.flags|=65536,c===Of?l.flags|=16384:(i=l.updateQueue,i===null?(i={transitions:null,markerInstances:null,retryQueue:new Set([c])},l.updateQueue=i):(l=i.retryQueue,l===null?i.retryQueue=new Set([c]):l.add(c)),xd(n,c,d)),!1}throw Error(a(435,l.tag))}return xd(n,c,d),gd(),!1}if(De)return i=wn.current,i!==null?((i.flags&65536)===0&&(i.flags|=256),i.flags|=65536,i.lanes=d,c!==pf&&(n=Error(a(422),{cause:c}),bl(gn(n,l)))):(c!==pf&&(i=Error(a(423),{cause:c}),bl(gn(i,l))),n=n.current.alternate,n.flags|=65536,d&=-d,n.lanes|=d,c=gn(c,l),d=Vf(n.stateNode,c,d),_f(n,d),We!==4&&(We=2)),!1;var p=Error(a(520),{cause:c});if(p=gn(p,l),ql===null?ql=[p]:ql.push(p),We!==4&&(We=2),i===null)return!0;c=gn(c,l),l=i;do{switch(l.tag){case 3:return l.flags|=65536,n=d&-d,l.lanes|=n,n=Vf(l.stateNode,c,n),_f(l,n),!1;case 1:if(i=l.type,p=l.stateNode,(l.flags&128)===0&&(typeof i.getDerivedStateFromError=="function"||p!==null&&typeof p.componentDidCatch=="function"&&(Gr===null||!Gr.has(p))))return l.flags|=65536,d&=-d,l.lanes|=d,d=i0(d),l0(d,n,l,c),_f(l,d),!1}l=l.return}while(l!==null);return!1}var u0=Error(a(461)),vt=!1;function Ot(n,i,l,c){i.child=n===null?Wg(i,null,l,c):yi(i,n.child,l,c)}function o0(n,i,l,c,d){l=l.render;var p=i.ref;if("ref"in c){var x={};for(var E in c)E!=="ref"&&(x[E]=c[E])}else x=c;return ba(i),c=Df(n,i,l,x,p,d),E=Nf(),n!==null&&!vt?(Pf(n,i,d),ur(n,i,d)):(De&&E&&vf(i),i.flags|=1,Ot(n,i,c,d),i.child)}function c0(n,i,l,c,d){if(n===null){var p=l.type;return typeof p=="function"&&!ff(p)&&p.defaultProps===void 0&&l.compare===null?(i.tag=15,i.type=p,s0(n,i,p,c,d)):(n=ao(l.type,null,c,i,i.mode,d),n.ref=i.ref,n.return=i,i.child=n)}if(p=n.child,!nd(n,d)){var x=p.memoizedProps;if(l=l.compare,l=l!==null?l:ml,l(x,c)&&n.ref===i.ref)return ur(n,i,d)}return i.flags|=1,n=er(p,c),n.ref=i.ref,n.return=i,i.child=n}function s0(n,i,l,c,d){if(n!==null){var p=n.memoizedProps;if(ml(p,c)&&n.ref===i.ref)if(vt=!1,i.pendingProps=c=p,nd(n,d))(n.flags&131072)!==0&&(vt=!0);else return i.lanes=n.lanes,ur(n,i,d)}return Zf(n,i,l,c,d)}function f0(n,i,l){var c=i.pendingProps,d=c.children,p=n!==null?n.memoizedState:null;if(c.mode==="hidden"){if((i.flags&128)!==0){if(c=p!==null?p.baseLanes|l:l,n!==null){for(d=i.child=n.child,p=0;d!==null;)p=p|d.lanes|d.childLanes,d=d.sibling;i.childLanes=p&~c}else i.childLanes=0,i.child=null;return d0(n,i,c,l)}if((l&536870912)!==0)i.memoizedState={baseLanes:0,cachePool:null},n!==null&&co(i,p!==null?p.cachePool:null),p!==null?sg(i,p):jf(),Fg(i);else return i.lanes=i.childLanes=536870912,d0(n,i,p!==null?p.baseLanes|l:l,l)}else p!==null?(co(i,p.cachePool),sg(i,p),Lr(),i.memoizedState=null):(n!==null&&co(i,null),jf(),Lr());return Ot(n,i,d,l),i.child}function d0(n,i,l,c){var d=wf();return d=d===null?null:{parent:lt._currentValue,pool:d},i.memoizedState={baseLanes:l,cachePool:d},n!==null&&co(i,null),jf(),Fg(i),n!==null&&xl(n,i,c,!0),null}function To(n,i){var l=i.ref;if(l===null)n!==null&&n.ref!==null&&(i.flags|=4194816);else{if(typeof l!="function"&&typeof l!="object")throw Error(a(284));(n===null||n.ref!==l)&&(i.flags|=4194816)}}function Zf(n,i,l,c,d){return ba(i),l=Df(n,i,l,c,void 0,d),c=Nf(),n!==null&&!vt?(Pf(n,i,d),ur(n,i,d)):(De&&c&&vf(i),i.flags|=1,Ot(n,i,l,d),i.child)}function h0(n,i,l,c,d,p){return ba(i),i.updateQueue=null,l=dg(i,c,l,d),fg(n),c=Nf(),n!==null&&!vt?(Pf(n,i,p),ur(n,i,p)):(De&&c&&vf(i),i.flags|=1,Ot(n,i,l,p),i.child)}function v0(n,i,l,c,d){if(ba(i),i.stateNode===null){var p=ui,x=l.contextType;typeof x=="object"&&x!==null&&(p=Dt(x)),p=new l(c,p),i.memoizedState=p.state!==null&&p.state!==void 0?p.state:null,p.updater=Xf,i.stateNode=p,p._reactInternals=i,p=i.stateNode,p.props=c,p.state=i.memoizedState,p.refs={},Af(i),x=l.contextType,p.context=typeof x=="object"&&x!==null?Dt(x):ui,p.state=i.memoizedState,x=l.getDerivedStateFromProps,typeof x=="function"&&(If(i,l,x,c),p.state=i.memoizedState),typeof l.getDerivedStateFromProps=="function"||typeof p.getSnapshotBeforeUpdate=="function"||typeof p.UNSAFE_componentWillMount!="function"&&typeof p.componentWillMount!="function"||(x=p.state,typeof p.componentWillMount=="function"&&p.componentWillMount(),typeof p.UNSAFE_componentWillMount=="function"&&p.UNSAFE_componentWillMount(),x!==p.state&&Xf.enqueueReplaceState(p,p.state,null),Tl(i,c,p,d),_l(),p.state=i.memoizedState),typeof p.componentDidMount=="function"&&(i.flags|=4194308),c=!0}else if(n===null){p=i.stateNode;var E=i.memoizedProps,C=wa(l,E);p.props=C;var q=p.context,K=l.contextType;x=ui,typeof K=="object"&&K!==null&&(x=Dt(K));var Q=l.getDerivedStateFromProps;K=typeof Q=="function"||typeof p.getSnapshotBeforeUpdate=="function",E=i.pendingProps!==E,K||typeof p.UNSAFE_componentWillReceiveProps!="function"&&typeof p.componentWillReceiveProps!="function"||(E||q!==x)&&e0(i,p,c,x),Pr=!1;var $=i.memoizedState;p.state=$,Tl(i,c,p,d),_l(),q=i.memoizedState,E||$!==q||Pr?(typeof Q=="function"&&(If(i,l,Q,c),q=i.memoizedState),(C=Pr||Jg(i,l,C,c,$,q,x))?(K||typeof p.UNSAFE_componentWillMount!="function"&&typeof p.componentWillMount!="function"||(typeof p.componentWillMount=="function"&&p.componentWillMount(),typeof p.UNSAFE_componentWillMount=="function"&&p.UNSAFE_componentWillMount()),typeof p.componentDidMount=="function"&&(i.flags|=4194308)):(typeof p.componentDidMount=="function"&&(i.flags|=4194308),i.memoizedProps=c,i.memoizedState=q),p.props=c,p.state=q,p.context=x,c=C):(typeof p.componentDidMount=="function"&&(i.flags|=4194308),c=!1)}else{p=i.stateNode,Ef(n,i),x=i.memoizedProps,K=wa(l,x),p.props=K,Q=i.pendingProps,$=p.context,q=l.contextType,C=ui,typeof q=="object"&&q!==null&&(C=Dt(q)),E=l.getDerivedStateFromProps,(q=typeof E=="function"||typeof p.getSnapshotBeforeUpdate=="function")||typeof p.UNSAFE_componentWillReceiveProps!="function"&&typeof p.componentWillReceiveProps!="function"||(x!==Q||$!==C)&&e0(i,p,c,C),Pr=!1,$=i.memoizedState,p.state=$,Tl(i,c,p,d),_l();var H=i.memoizedState;x!==Q||$!==H||Pr||n!==null&&n.dependencies!==null&&uo(n.dependencies)?(typeof E=="function"&&(If(i,l,E,c),H=i.memoizedState),(K=Pr||Jg(i,l,K,c,$,H,C)||n!==null&&n.dependencies!==null&&uo(n.dependencies))?(q||typeof p.UNSAFE_componentWillUpdate!="function"&&typeof p.componentWillUpdate!="function"||(typeof p.componentWillUpdate=="function"&&p.componentWillUpdate(c,H,C),typeof p.UNSAFE_componentWillUpdate=="function"&&p.UNSAFE_componentWillUpdate(c,H,C)),typeof p.componentDidUpdate=="function"&&(i.flags|=4),typeof p.getSnapshotBeforeUpdate=="function"&&(i.flags|=1024)):(typeof p.componentDidUpdate!="function"||x===n.memoizedProps&&$===n.memoizedState||(i.flags|=4),typeof p.getSnapshotBeforeUpdate!="function"||x===n.memoizedProps&&$===n.memoizedState||(i.flags|=1024),i.memoizedProps=c,i.memoizedState=H),p.props=c,p.state=H,p.context=C,c=K):(typeof p.componentDidUpdate!="function"||x===n.memoizedProps&&$===n.memoizedState||(i.flags|=4),typeof p.getSnapshotBeforeUpdate!="function"||x===n.memoizedProps&&$===n.memoizedState||(i.flags|=1024),c=!1)}return p=c,To(n,i),c=(i.flags&128)!==0,p||c?(p=i.stateNode,l=c&&typeof l.getDerivedStateFromError!="function"?null:p.render(),i.flags|=1,n!==null&&c?(i.child=yi(i,n.child,null,d),i.child=yi(i,null,l,d)):Ot(n,i,l,d),i.memoizedState=p.state,n=i.child):n=ur(n,i,d),n}function m0(n,i,l,c){return gl(),i.flags|=256,Ot(n,i,l,c),i.child}var Qf={dehydrated:null,treeContext:null,retryLane:0,hydrationErrors:null};function Wf(n){return{baseLanes:n,cachePool:ng()}}function Ff(n,i,l){return n=n!==null?n.childLanes&~l:0,i&&(n|=On),n}function p0(n,i,l){var c=i.pendingProps,d=!1,p=(i.flags&128)!==0,x;if((x=p)||(x=n!==null&&n.memoizedState===null?!1:(ut.current&2)!==0),x&&(d=!0,i.flags&=-129),x=(i.flags&32)!==0,i.flags&=-33,n===null){if(De){if(d?Br(i):Lr(),De){var E=Qe,C;if(C=E){e:{for(C=E,E=Hn;C.nodeType!==8;){if(!E){E=null;break e}if(C=Bn(C.nextSibling),C===null){E=null;break e}}E=C}E!==null?(i.memoizedState={dehydrated:E,treeContext:va!==null?{id:tr,overflow:nr}:null,retryLane:536870912,hydrationErrors:null},C=en(18,null,null,0),C.stateNode=E,C.return=i,i.child=C,Bt=i,Qe=null,C=!0):C=!1}C||ya(i)}if(E=i.memoizedState,E!==null&&(E=E.dehydrated,E!==null))return kd(E)?i.lanes=32:i.lanes=536870912,null;lr(i)}return E=c.children,c=c.fallback,d?(Lr(),d=i.mode,E=jo({mode:"hidden",children:E},d),c=ha(c,d,l,null),E.return=i,c.return=i,E.sibling=c,i.child=E,d=i.child,d.memoizedState=Wf(l),d.childLanes=Ff(n,x,l),i.memoizedState=Qf,c):(Br(i),Jf(i,E))}if(C=n.memoizedState,C!==null&&(E=C.dehydrated,E!==null)){if(p)i.flags&256?(Br(i),i.flags&=-257,i=ed(n,i,l)):i.memoizedState!==null?(Lr(),i.child=n.child,i.flags|=128,i=null):(Lr(),d=c.fallback,E=i.mode,c=jo({mode:"visible",children:c.children},E),d=ha(d,E,l,null),d.flags|=2,c.return=i,d.return=i,c.sibling=d,i.child=c,yi(i,n.child,null,l),c=i.child,c.memoizedState=Wf(l),c.childLanes=Ff(n,x,l),i.memoizedState=Qf,i=d);else if(Br(i),kd(E)){if(x=E.nextSibling&&E.nextSibling.dataset,x)var q=x.dgst;x=q,c=Error(a(419)),c.stack="",c.digest=x,bl({value:c,source:null,stack:null}),i=ed(n,i,l)}else if(vt||xl(n,i,l,!1),x=(l&n.childLanes)!==0,vt||x){if(x=Ye,x!==null&&(c=l&-l,c=(c&42)!==0?1:ks(c),c=(c&(x.suspendedLanes|l))!==0?0:c,c!==0&&c!==C.retryLane))throw C.retryLane=c,li(n,c),ln(x,n,c),u0;E.data==="$?"||gd(),i=ed(n,i,l)}else E.data==="$?"?(i.flags|=192,i.child=n.child,i=null):(n=C.treeContext,Qe=Bn(E.nextSibling),Bt=i,De=!0,pa=null,Hn=!1,n!==null&&(xn[Sn++]=tr,xn[Sn++]=nr,xn[Sn++]=va,tr=n.id,nr=n.overflow,va=i),i=Jf(i,c.children),i.flags|=4096);return i}return d?(Lr(),d=c.fallback,E=i.mode,C=n.child,q=C.sibling,c=er(C,{mode:"hidden",children:c.children}),c.subtreeFlags=C.subtreeFlags&65011712,q!==null?d=er(q,d):(d=ha(d,E,l,null),d.flags|=2),d.return=i,c.return=i,c.sibling=d,i.child=c,c=d,d=i.child,E=n.child.memoizedState,E===null?E=Wf(l):(C=E.cachePool,C!==null?(q=lt._currentValue,C=C.parent!==q?{parent:q,pool:q}:C):C=ng(),E={baseLanes:E.baseLanes|l,cachePool:C}),d.memoizedState=E,d.childLanes=Ff(n,x,l),i.memoizedState=Qf,c):(Br(i),l=n.child,n=l.sibling,l=er(l,{mode:"visible",children:c.children}),l.return=i,l.sibling=null,n!==null&&(x=i.deletions,x===null?(i.deletions=[n],i.flags|=16):x.push(n)),i.child=l,i.memoizedState=null,l)}function Jf(n,i){return i=jo({mode:"visible",children:i},n.mode),i.return=n,n.child=i}function jo(n,i){return n=en(22,n,null,i),n.lanes=0,n.stateNode={_visibility:1,_pendingMarkers:null,_retryCache:null,_transitions:null},n}function ed(n,i,l){return yi(i,n.child,null,l),n=Jf(i,i.pendingProps.children),n.flags|=2,i.memoizedState=null,n}function y0(n,i,l){n.lanes|=i;var c=n.alternate;c!==null&&(c.lanes|=i),gf(n.return,i,l)}function td(n,i,l,c,d){var p=n.memoizedState;p===null?n.memoizedState={isBackwards:i,rendering:null,renderingStartTime:0,last:c,tail:l,tailMode:d}:(p.isBackwards=i,p.rendering=null,p.renderingStartTime=0,p.last=c,p.tail=l,p.tailMode=d)}function g0(n,i,l){var c=i.pendingProps,d=c.revealOrder,p=c.tail;if(Ot(n,i,c.children,l),c=ut.current,(c&2)!==0)c=c&1|2,i.flags|=128;else{if(n!==null&&(n.flags&128)!==0)e:for(n=i.child;n!==null;){if(n.tag===13)n.memoizedState!==null&&y0(n,l,i);else if(n.tag===19)y0(n,l,i);else if(n.child!==null){n.child.return=n,n=n.child;continue}if(n===i)break e;for(;n.sibling===null;){if(n.return===null||n.return===i)break e;n=n.return}n.sibling.return=n.return,n=n.sibling}c&=1}switch(ae(ut,c),d){case"forwards":for(l=i.child,d=null;l!==null;)n=l.alternate,n!==null&&Ao(n)===null&&(d=l),l=l.sibling;l=d,l===null?(d=i.child,i.child=null):(d=l.sibling,l.sibling=null),td(i,!1,d,l,p);break;case"backwards":for(l=null,d=i.child,i.child=null;d!==null;){if(n=d.alternate,n!==null&&Ao(n)===null){i.child=d;break}n=d.sibling,d.sibling=l,l=d,d=n}td(i,!0,l,null,p);break;case"together":td(i,!1,null,null,void 0);break;default:i.memoizedState=null}return i.child}function ur(n,i,l){if(n!==null&&(i.dependencies=n.dependencies),Yr|=i.lanes,(l&i.childLanes)===0)if(n!==null){if(xl(n,i,l,!1),(l&i.childLanes)===0)return null}else return null;if(n!==null&&i.child!==n.child)throw Error(a(153));if(i.child!==null){for(n=i.child,l=er(n,n.pendingProps),i.child=l,l.return=i;n.sibling!==null;)n=n.sibling,l=l.sibling=er(n,n.pendingProps),l.return=i;l.sibling=null}return i.child}function nd(n,i){return(n.lanes&i)!==0?!0:(n=n.dependencies,!!(n!==null&&uo(n)))}function zT(n,i,l){switch(i.tag){case 3:ze(i,i.stateNode.containerInfo),Nr(i,lt,n.memoizedState.cache),gl();break;case 27:case 5:ue(i);break;case 4:ze(i,i.stateNode.containerInfo);break;case 10:Nr(i,i.type,i.memoizedProps.value);break;case 13:var c=i.memoizedState;if(c!==null)return c.dehydrated!==null?(Br(i),i.flags|=128,null):(l&i.child.childLanes)!==0?p0(n,i,l):(Br(i),n=ur(n,i,l),n!==null?n.sibling:null);Br(i);break;case 19:var d=(n.flags&128)!==0;if(c=(l&i.childLanes)!==0,c||(xl(n,i,l,!1),c=(l&i.childLanes)!==0),d){if(c)return g0(n,i,l);i.flags|=128}if(d=i.memoizedState,d!==null&&(d.rendering=null,d.tail=null,d.lastEffect=null),ae(ut,ut.current),c)break;return null;case 22:case 23:return i.lanes=0,f0(n,i,l);case 24:Nr(i,lt,n.memoizedState.cache)}return ur(n,i,l)}function b0(n,i,l){if(n!==null)if(n.memoizedProps!==i.pendingProps)vt=!0;else{if(!nd(n,l)&&(i.flags&128)===0)return vt=!1,zT(n,i,l);vt=(n.flags&131072)!==0}else vt=!1,De&&(i.flags&1048576)!==0&&Zy(i,lo,i.index);switch(i.lanes=0,i.tag){case 16:e:{n=i.pendingProps;var c=i.elementType,d=c._init;if(c=d(c._payload),i.type=c,typeof c=="function")ff(c)?(n=wa(c,n),i.tag=1,i=v0(null,i,c,n,l)):(i.tag=0,i=Zf(null,i,c,n,l));else{if(c!=null){if(d=c.$$typeof,d===k){i.tag=11,i=o0(null,i,c,n,l);break e}else if(d===L){i.tag=14,i=c0(null,i,c,n,l);break e}}throw i=ce(c)||c,Error(a(306,i,""))}}return i;case 0:return Zf(n,i,i.type,i.pendingProps,l);case 1:return c=i.type,d=wa(c,i.pendingProps),v0(n,i,c,d,l);case 3:e:{if(ze(i,i.stateNode.containerInfo),n===null)throw Error(a(387));c=i.pendingProps;var p=i.memoizedState;d=p.element,Ef(n,i),Tl(i,c,null,l);var x=i.memoizedState;if(c=x.cache,Nr(i,lt,c),c!==p.cache&&bf(i,[lt],l,!0),_l(),c=x.element,p.isDehydrated)if(p={element:c,isDehydrated:!1,cache:x.cache},i.updateQueue.baseState=p,i.memoizedState=p,i.flags&256){i=m0(n,i,c,l);break e}else if(c!==d){d=gn(Error(a(424)),i),bl(d),i=m0(n,i,c,l);break e}else{switch(n=i.stateNode.containerInfo,n.nodeType){case 9:n=n.body;break;default:n=n.nodeName==="HTML"?n.ownerDocument.body:n}for(Qe=Bn(n.firstChild),Bt=i,De=!0,pa=null,Hn=!0,l=Wg(i,null,c,l),i.child=l;l;)l.flags=l.flags&-3|4096,l=l.sibling}else{if(gl(),c===d){i=ur(n,i,l);break e}Ot(n,i,c,l)}i=i.child}return i;case 26:return To(n,i),n===null?(l=Ob(i.type,null,i.pendingProps,null))?i.memoizedState=l:De||(l=i.type,n=i.pendingProps,c=Ho(ye.current).createElement(l),c[Ct]=i,c[Ht]=n,Et(c,l,n),ht(c),i.stateNode=c):i.memoizedState=Ob(i.type,n.memoizedProps,i.pendingProps,n.memoizedState),null;case 27:return ue(i),n===null&&De&&(c=i.stateNode=xb(i.type,i.pendingProps,ye.current),Bt=i,Hn=!0,d=Qe,Xr(i.type)?(zd=d,Qe=Bn(c.firstChild)):Qe=d),Ot(n,i,i.pendingProps.children,l),To(n,i),n===null&&(i.flags|=4194304),i.child;case 5:return n===null&&De&&((d=c=Qe)&&(c=sj(c,i.type,i.pendingProps,Hn),c!==null?(i.stateNode=c,Bt=i,Qe=Bn(c.firstChild),Hn=!1,d=!0):d=!1),d||ya(i)),ue(i),d=i.type,p=i.pendingProps,x=n!==null?n.memoizedProps:null,c=p.children,Nd(d,p)?c=null:x!==null&&Nd(d,x)&&(i.flags|=32),i.memoizedState!==null&&(d=Df(n,i,jT,null,null,l),Zl._currentValue=d),To(n,i),Ot(n,i,c,l),i.child;case 6:return n===null&&De&&((n=l=Qe)&&(l=fj(l,i.pendingProps,Hn),l!==null?(i.stateNode=l,Bt=i,Qe=null,n=!0):n=!1),n||ya(i)),null;case 13:return p0(n,i,l);case 4:return ze(i,i.stateNode.containerInfo),c=i.pendingProps,n===null?i.child=yi(i,null,c,l):Ot(n,i,c,l),i.child;case 11:return o0(n,i,i.type,i.pendingProps,l);case 7:return Ot(n,i,i.pendingProps,l),i.child;case 8:return Ot(n,i,i.pendingProps.children,l),i.child;case 12:return Ot(n,i,i.pendingProps.children,l),i.child;case 10:return c=i.pendingProps,Nr(i,i.type,c.value),Ot(n,i,c.children,l),i.child;case 9:return d=i.type._context,c=i.pendingProps.children,ba(i),d=Dt(d),c=c(d),i.flags|=1,Ot(n,i,c,l),i.child;case 14:return c0(n,i,i.type,i.pendingProps,l);case 15:return s0(n,i,i.type,i.pendingProps,l);case 19:return g0(n,i,l);case 31:return c=i.pendingProps,l=i.mode,c={mode:c.mode,children:c.children},n===null?(l=jo(c,l),l.ref=i.ref,i.child=l,l.return=i,i=l):(l=er(n.child,c),l.ref=i.ref,i.child=l,l.return=i,i=l),i;case 22:return f0(n,i,l);case 24:return ba(i),c=Dt(lt),n===null?(d=wf(),d===null&&(d=Ye,p=xf(),d.pooledCache=p,p.refCount++,p!==null&&(d.pooledCacheLanes|=l),d=p),i.memoizedState={parent:c,cache:d},Af(i),Nr(i,lt,d)):((n.lanes&l)!==0&&(Ef(n,i),Tl(i,null,null,l),_l()),d=n.memoizedState,p=i.memoizedState,d.parent!==c?(d={parent:c,cache:c},i.memoizedState=d,i.lanes===0&&(i.memoizedState=i.updateQueue.baseState=d),Nr(i,lt,c)):(c=p.cache,Nr(i,lt,c),c!==d.cache&&bf(i,[lt],l,!0))),Ot(n,i,i.pendingProps.children,l),i.child;case 29:throw i.pendingProps}throw Error(a(156,i.tag))}function or(n){n.flags|=4}function x0(n,i){if(i.type!=="stylesheet"||(i.state.loading&4)!==0)n.flags&=-16777217;else if(n.flags|=16777216,!jb(i)){if(i=wn.current,i!==null&&((Te&4194048)===Te?Yn!==null:(Te&62914560)!==Te&&(Te&536870912)===0||i!==Yn))throw Al=Of,rg;n.flags|=8192}}function Mo(n,i){i!==null&&(n.flags|=4),n.flags&16384&&(i=n.tag!==22?Fp():536870912,n.lanes|=i,Si|=i)}function Rl(n,i){if(!De)switch(n.tailMode){case"hidden":i=n.tail;for(var l=null;i!==null;)i.alternate!==null&&(l=i),i=i.sibling;l===null?n.tail=null:l.sibling=null;break;case"collapsed":l=n.tail;for(var c=null;l!==null;)l.alternate!==null&&(c=l),l=l.sibling;c===null?i||n.tail===null?n.tail=null:n.tail.sibling=null:c.sibling=null}}function Ze(n){var i=n.alternate!==null&&n.alternate.child===n.child,l=0,c=0;if(i)for(var d=n.child;d!==null;)l|=d.lanes|d.childLanes,c|=d.subtreeFlags&65011712,c|=d.flags&65011712,d.return=n,d=d.sibling;else for(d=n.child;d!==null;)l|=d.lanes|d.childLanes,c|=d.subtreeFlags,c|=d.flags,d.return=n,d=d.sibling;return n.subtreeFlags|=c,n.childLanes=l,i}function BT(n,i,l){var c=i.pendingProps;switch(mf(i),i.tag){case 31:case 16:case 15:case 0:case 11:case 7:case 8:case 12:case 9:case 14:return Ze(i),null;case 1:return Ze(i),null;case 3:return l=i.stateNode,c=null,n!==null&&(c=n.memoizedState.cache),i.memoizedState.cache!==c&&(i.flags|=2048),ar(lt),V(),l.pendingContext&&(l.context=l.pendingContext,l.pendingContext=null),(n===null||n.child===null)&&(yl(i)?or(i):n===null||n.memoizedState.isDehydrated&&(i.flags&256)===0||(i.flags|=1024,Fy())),Ze(i),null;case 26:return l=i.memoizedState,n===null?(or(i),l!==null?(Ze(i),x0(i,l)):(Ze(i),i.flags&=-16777217)):l?l!==n.memoizedState?(or(i),Ze(i),x0(i,l)):(Ze(i),i.flags&=-16777217):(n.memoizedProps!==c&&or(i),Ze(i),i.flags&=-16777217),null;case 27:ge(i),l=ye.current;var d=i.type;if(n!==null&&i.stateNode!=null)n.memoizedProps!==c&&or(i);else{if(!c){if(i.stateNode===null)throw Error(a(166));return Ze(i),null}n=fe.current,yl(i)?Qy(i):(n=xb(d,c,l),i.stateNode=n,or(i))}return Ze(i),null;case 5:if(ge(i),l=i.type,n!==null&&i.stateNode!=null)n.memoizedProps!==c&&or(i);else{if(!c){if(i.stateNode===null)throw Error(a(166));return Ze(i),null}if(n=fe.current,yl(i))Qy(i);else{switch(d=Ho(ye.current),n){case 1:n=d.createElementNS("http://www.w3.org/2000/svg",l);break;case 2:n=d.createElementNS("http://www.w3.org/1998/Math/MathML",l);break;default:switch(l){case"svg":n=d.createElementNS("http://www.w3.org/2000/svg",l);break;case"math":n=d.createElementNS("http://www.w3.org/1998/Math/MathML",l);break;case"script":n=d.createElement("div"),n.innerHTML="<script><\/script>",n=n.removeChild(n.firstChild);break;case"select":n=typeof c.is=="string"?d.createElement("select",{is:c.is}):d.createElement("select"),c.multiple?n.multiple=!0:c.size&&(n.size=c.size);break;default:n=typeof c.is=="string"?d.createElement(l,{is:c.is}):d.createElement(l)}}n[Ct]=i,n[Ht]=c;e:for(d=i.child;d!==null;){if(d.tag===5||d.tag===6)n.appendChild(d.stateNode);else if(d.tag!==4&&d.tag!==27&&d.child!==null){d.child.return=d,d=d.child;continue}if(d===i)break e;for(;d.sibling===null;){if(d.return===null||d.return===i)break e;d=d.return}d.sibling.return=d.return,d=d.sibling}i.stateNode=n;e:switch(Et(n,l,c),l){case"button":case"input":case"select":case"textarea":n=!!c.autoFocus;break e;case"img":n=!0;break e;default:n=!1}n&&or(i)}}return Ze(i),i.flags&=-16777217,null;case 6:if(n&&i.stateNode!=null)n.memoizedProps!==c&&or(i);else{if(typeof c!="string"&&i.stateNode===null)throw Error(a(166));if(n=ye.current,yl(i)){if(n=i.stateNode,l=i.memoizedProps,c=null,d=Bt,d!==null)switch(d.tag){case 27:case 5:c=d.memoizedProps}n[Ct]=i,n=!!(n.nodeValue===l||c!==null&&c.suppressHydrationWarning===!0||hb(n.nodeValue,l)),n||ya(i)}else n=Ho(n).createTextNode(c),n[Ct]=i,i.stateNode=n}return Ze(i),null;case 13:if(c=i.memoizedState,n===null||n.memoizedState!==null&&n.memoizedState.dehydrated!==null){if(d=yl(i),c!==null&&c.dehydrated!==null){if(n===null){if(!d)throw Error(a(318));if(d=i.memoizedState,d=d!==null?d.dehydrated:null,!d)throw Error(a(317));d[Ct]=i}else gl(),(i.flags&128)===0&&(i.memoizedState=null),i.flags|=4;Ze(i),d=!1}else d=Fy(),n!==null&&n.memoizedState!==null&&(n.memoizedState.hydrationErrors=d),d=!0;if(!d)return i.flags&256?(lr(i),i):(lr(i),null)}if(lr(i),(i.flags&128)!==0)return i.lanes=l,i;if(l=c!==null,n=n!==null&&n.memoizedState!==null,l){c=i.child,d=null,c.alternate!==null&&c.alternate.memoizedState!==null&&c.alternate.memoizedState.cachePool!==null&&(d=c.alternate.memoizedState.cachePool.pool);var p=null;c.memoizedState!==null&&c.memoizedState.cachePool!==null&&(p=c.memoizedState.cachePool.pool),p!==d&&(c.flags|=2048)}return l!==n&&l&&(i.child.flags|=8192),Mo(i,i.updateQueue),Ze(i),null;case 4:return V(),n===null&&Td(i.stateNode.containerInfo),Ze(i),null;case 10:return ar(i.type),Ze(i),null;case 19:if(Z(ut),d=i.memoizedState,d===null)return Ze(i),null;if(c=(i.flags&128)!==0,p=d.rendering,p===null)if(c)Rl(d,!1);else{if(We!==0||n!==null&&(n.flags&128)!==0)for(n=i.child;n!==null;){if(p=Ao(n),p!==null){for(i.flags|=128,Rl(d,!1),n=p.updateQueue,i.updateQueue=n,Mo(i,n),i.subtreeFlags=0,n=l,l=i.child;l!==null;)Vy(l,n),l=l.sibling;return ae(ut,ut.current&1|2),i.child}n=n.sibling}d.tail!==null&&dt()>No&&(i.flags|=128,c=!0,Rl(d,!1),i.lanes=4194304)}else{if(!c)if(n=Ao(p),n!==null){if(i.flags|=128,c=!0,n=n.updateQueue,i.updateQueue=n,Mo(i,n),Rl(d,!0),d.tail===null&&d.tailMode==="hidden"&&!p.alternate&&!De)return Ze(i),null}else 2*dt()-d.renderingStartTime>No&&l!==536870912&&(i.flags|=128,c=!0,Rl(d,!1),i.lanes=4194304);d.isBackwards?(p.sibling=i.child,i.child=p):(n=d.last,n!==null?n.sibling=p:i.child=p,d.last=p)}return d.tail!==null?(i=d.tail,d.rendering=i,d.tail=i.sibling,d.renderingStartTime=dt(),i.sibling=null,n=ut.current,ae(ut,c?n&1|2:n&1),i):(Ze(i),null);case 22:case 23:return lr(i),Mf(),c=i.memoizedState!==null,n!==null?n.memoizedState!==null!==c&&(i.flags|=8192):c&&(i.flags|=8192),c?(l&536870912)!==0&&(i.flags&128)===0&&(Ze(i),i.subtreeFlags&6&&(i.flags|=8192)):Ze(i),l=i.updateQueue,l!==null&&Mo(i,l.retryQueue),l=null,n!==null&&n.memoizedState!==null&&n.memoizedState.cachePool!==null&&(l=n.memoizedState.cachePool.pool),c=null,i.memoizedState!==null&&i.memoizedState.cachePool!==null&&(c=i.memoizedState.cachePool.pool),c!==l&&(i.flags|=2048),n!==null&&Z(xa),null;case 24:return l=null,n!==null&&(l=n.memoizedState.cache),i.memoizedState.cache!==l&&(i.flags|=2048),ar(lt),Ze(i),null;case 25:return null;case 30:return null}throw Error(a(156,i.tag))}function LT(n,i){switch(mf(i),i.tag){case 1:return n=i.flags,n&65536?(i.flags=n&-65537|128,i):null;case 3:return ar(lt),V(),n=i.flags,(n&65536)!==0&&(n&128)===0?(i.flags=n&-65537|128,i):null;case 26:case 27:case 5:return ge(i),null;case 13:if(lr(i),n=i.memoizedState,n!==null&&n.dehydrated!==null){if(i.alternate===null)throw Error(a(340));gl()}return n=i.flags,n&65536?(i.flags=n&-65537|128,i):null;case 19:return Z(ut),null;case 4:return V(),null;case 10:return ar(i.type),null;case 22:case 23:return lr(i),Mf(),n!==null&&Z(xa),n=i.flags,n&65536?(i.flags=n&-65537|128,i):null;case 24:return ar(lt),null;case 25:return null;default:return null}}function S0(n,i){switch(mf(i),i.tag){case 3:ar(lt),V();break;case 26:case 27:case 5:ge(i);break;case 4:V();break;case 13:lr(i);break;case 19:Z(ut);break;case 10:ar(i.type);break;case 22:case 23:lr(i),Mf(),n!==null&&Z(xa);break;case 24:ar(lt)}}function kl(n,i){try{var l=i.updateQueue,c=l!==null?l.lastEffect:null;if(c!==null){var d=c.next;l=d;do{if((l.tag&n)===n){c=void 0;var p=l.create,x=l.inst;c=p(),x.destroy=c}l=l.next}while(l!==d)}}catch(E){$e(i,i.return,E)}}function Ur(n,i,l){try{var c=i.updateQueue,d=c!==null?c.lastEffect:null;if(d!==null){var p=d.next;c=p;do{if((c.tag&n)===n){var x=c.inst,E=x.destroy;if(E!==void 0){x.destroy=void 0,d=i;var C=l,q=E;try{q()}catch(K){$e(d,C,K)}}}c=c.next}while(c!==p)}}catch(K){$e(i,i.return,K)}}function w0(n){var i=n.updateQueue;if(i!==null){var l=n.stateNode;try{cg(i,l)}catch(c){$e(n,n.return,c)}}}function O0(n,i,l){l.props=wa(n.type,n.memoizedProps),l.state=n.memoizedState;try{l.componentWillUnmount()}catch(c){$e(n,i,c)}}function zl(n,i){try{var l=n.ref;if(l!==null){switch(n.tag){case 26:case 27:case 5:var c=n.stateNode;break;case 30:c=n.stateNode;break;default:c=n.stateNode}typeof l=="function"?n.refCleanup=l(c):l.current=c}}catch(d){$e(n,i,d)}}function Gn(n,i){var l=n.ref,c=n.refCleanup;if(l!==null)if(typeof c=="function")try{c()}catch(d){$e(n,i,d)}finally{n.refCleanup=null,n=n.alternate,n!=null&&(n.refCleanup=null)}else if(typeof l=="function")try{l(null)}catch(d){$e(n,i,d)}else l.current=null}function A0(n){var i=n.type,l=n.memoizedProps,c=n.stateNode;try{e:switch(i){case"button":case"input":case"select":case"textarea":l.autoFocus&&c.focus();break e;case"img":l.src?c.src=l.src:l.srcSet&&(c.srcset=l.srcSet)}}catch(d){$e(n,n.return,d)}}function rd(n,i,l){try{var c=n.stateNode;ij(c,n.type,l,i),c[Ht]=i}catch(d){$e(n,n.return,d)}}function E0(n){return n.tag===5||n.tag===3||n.tag===26||n.tag===27&&Xr(n.type)||n.tag===4}function ad(n){e:for(;;){for(;n.sibling===null;){if(n.return===null||E0(n.return))return null;n=n.return}for(n.sibling.return=n.return,n=n.sibling;n.tag!==5&&n.tag!==6&&n.tag!==18;){if(n.tag===27&&Xr(n.type)||n.flags&2||n.child===null||n.tag===4)continue e;n.child.return=n,n=n.child}if(!(n.flags&2))return n.stateNode}}function id(n,i,l){var c=n.tag;if(c===5||c===6)n=n.stateNode,i?(l.nodeType===9?l.body:l.nodeName==="HTML"?l.ownerDocument.body:l).insertBefore(n,i):(i=l.nodeType===9?l.body:l.nodeName==="HTML"?l.ownerDocument.body:l,i.appendChild(n),l=l._reactRootContainer,l!=null||i.onclick!==null||(i.onclick=$o));else if(c!==4&&(c===27&&Xr(n.type)&&(l=n.stateNode,i=null),n=n.child,n!==null))for(id(n,i,l),n=n.sibling;n!==null;)id(n,i,l),n=n.sibling}function Co(n,i,l){var c=n.tag;if(c===5||c===6)n=n.stateNode,i?l.insertBefore(n,i):l.appendChild(n);else if(c!==4&&(c===27&&Xr(n.type)&&(l=n.stateNode),n=n.child,n!==null))for(Co(n,i,l),n=n.sibling;n!==null;)Co(n,i,l),n=n.sibling}function _0(n){var i=n.stateNode,l=n.memoizedProps;try{for(var c=n.type,d=i.attributes;d.length;)i.removeAttributeNode(d[0]);Et(i,c,l),i[Ct]=n,i[Ht]=l}catch(p){$e(n,n.return,p)}}var cr=!1,et=!1,ld=!1,T0=typeof WeakSet=="function"?WeakSet:Set,mt=null;function UT(n,i){if(n=n.containerInfo,Cd=Vo,n=Ly(n),rf(n)){if("selectionStart"in n)var l={start:n.selectionStart,end:n.selectionEnd};else e:{l=(l=n.ownerDocument)&&l.defaultView||window;var c=l.getSelection&&l.getSelection();if(c&&c.rangeCount!==0){l=c.anchorNode;var d=c.anchorOffset,p=c.focusNode;c=c.focusOffset;try{l.nodeType,p.nodeType}catch{l=null;break e}var x=0,E=-1,C=-1,q=0,K=0,Q=n,$=null;t:for(;;){for(var H;Q!==l||d!==0&&Q.nodeType!==3||(E=x+d),Q!==p||c!==0&&Q.nodeType!==3||(C=x+c),Q.nodeType===3&&(x+=Q.nodeValue.length),(H=Q.firstChild)!==null;)$=Q,Q=H;for(;;){if(Q===n)break t;if($===l&&++q===d&&(E=x),$===p&&++K===c&&(C=x),(H=Q.nextSibling)!==null)break;Q=$,$=Q.parentNode}Q=H}l=E===-1||C===-1?null:{start:E,end:C}}else l=null}l=l||{start:0,end:0}}else l=null;for(Dd={focusedElem:n,selectionRange:l},Vo=!1,mt=i;mt!==null;)if(i=mt,n=i.child,(i.subtreeFlags&1024)!==0&&n!==null)n.return=i,mt=n;else for(;mt!==null;){switch(i=mt,p=i.alternate,n=i.flags,i.tag){case 0:break;case 11:case 15:break;case 1:if((n&1024)!==0&&p!==null){n=void 0,l=i,d=p.memoizedProps,p=p.memoizedState,c=l.stateNode;try{var xe=wa(l.type,d,l.elementType===l.type);n=c.getSnapshotBeforeUpdate(xe,p),c.__reactInternalSnapshotBeforeUpdate=n}catch(pe){$e(l,l.return,pe)}}break;case 3:if((n&1024)!==0){if(n=i.stateNode.containerInfo,l=n.nodeType,l===9)Rd(n);else if(l===1)switch(n.nodeName){case"HEAD":case"HTML":case"BODY":Rd(n);break;default:n.textContent=""}}break;case 5:case 26:case 27:case 6:case 4:case 17:break;default:if((n&1024)!==0)throw Error(a(163))}if(n=i.sibling,n!==null){n.return=i.return,mt=n;break}mt=i.return}}function j0(n,i,l){var c=l.flags;switch(l.tag){case 0:case 11:case 15:qr(n,l),c&4&&kl(5,l);break;case 1:if(qr(n,l),c&4)if(n=l.stateNode,i===null)try{n.componentDidMount()}catch(x){$e(l,l.return,x)}else{var d=wa(l.type,i.memoizedProps);i=i.memoizedState;try{n.componentDidUpdate(d,i,n.__reactInternalSnapshotBeforeUpdate)}catch(x){$e(l,l.return,x)}}c&64&&w0(l),c&512&&zl(l,l.return);break;case 3:if(qr(n,l),c&64&&(n=l.updateQueue,n!==null)){if(i=null,l.child!==null)switch(l.child.tag){case 27:case 5:i=l.child.stateNode;break;case 1:i=l.child.stateNode}try{cg(n,i)}catch(x){$e(l,l.return,x)}}break;case 27:i===null&&c&4&&_0(l);case 26:case 5:qr(n,l),i===null&&c&4&&A0(l),c&512&&zl(l,l.return);break;case 12:qr(n,l);break;case 13:qr(n,l),c&4&&D0(n,l),c&64&&(n=l.memoizedState,n!==null&&(n=n.dehydrated,n!==null&&(l=VT.bind(null,l),dj(n,l))));break;case 22:if(c=l.memoizedState!==null||cr,!c){i=i!==null&&i.memoizedState!==null||et,d=cr;var p=et;cr=c,(et=i)&&!p?$r(n,l,(l.subtreeFlags&8772)!==0):qr(n,l),cr=d,et=p}break;case 30:break;default:qr(n,l)}}function M0(n){var i=n.alternate;i!==null&&(n.alternate=null,M0(i)),n.child=null,n.deletions=null,n.sibling=null,n.tag===5&&(i=n.stateNode,i!==null&&Ls(i)),n.stateNode=null,n.return=null,n.dependencies=null,n.memoizedProps=null,n.memoizedState=null,n.pendingProps=null,n.stateNode=null,n.updateQueue=null}var Xe=null,Kt=!1;function sr(n,i,l){for(l=l.child;l!==null;)C0(n,i,l),l=l.sibling}function C0(n,i,l){if(Wt&&typeof Wt.onCommitFiberUnmount=="function")try{Wt.onCommitFiberUnmount(rl,l)}catch{}switch(l.tag){case 26:et||Gn(l,i),sr(n,i,l),l.memoizedState?l.memoizedState.count--:l.stateNode&&(l=l.stateNode,l.parentNode.removeChild(l));break;case 27:et||Gn(l,i);var c=Xe,d=Kt;Xr(l.type)&&(Xe=l.stateNode,Kt=!1),sr(n,i,l),Kl(l.stateNode),Xe=c,Kt=d;break;case 5:et||Gn(l,i);case 6:if(c=Xe,d=Kt,Xe=null,sr(n,i,l),Xe=c,Kt=d,Xe!==null)if(Kt)try{(Xe.nodeType===9?Xe.body:Xe.nodeName==="HTML"?Xe.ownerDocument.body:Xe).removeChild(l.stateNode)}catch(p){$e(l,i,p)}else try{Xe.removeChild(l.stateNode)}catch(p){$e(l,i,p)}break;case 18:Xe!==null&&(Kt?(n=Xe,gb(n.nodeType===9?n.body:n.nodeName==="HTML"?n.ownerDocument.body:n,l.stateNode),Jl(n)):gb(Xe,l.stateNode));break;case 4:c=Xe,d=Kt,Xe=l.stateNode.containerInfo,Kt=!0,sr(n,i,l),Xe=c,Kt=d;break;case 0:case 11:case 14:case 15:et||Ur(2,l,i),et||Ur(4,l,i),sr(n,i,l);break;case 1:et||(Gn(l,i),c=l.stateNode,typeof c.componentWillUnmount=="function"&&O0(l,i,c)),sr(n,i,l);break;case 21:sr(n,i,l);break;case 22:et=(c=et)||l.memoizedState!==null,sr(n,i,l),et=c;break;default:sr(n,i,l)}}function D0(n,i){if(i.memoizedState===null&&(n=i.alternate,n!==null&&(n=n.memoizedState,n!==null&&(n=n.dehydrated,n!==null))))try{Jl(n)}catch(l){$e(i,i.return,l)}}function qT(n){switch(n.tag){case 13:case 19:var i=n.stateNode;return i===null&&(i=n.stateNode=new T0),i;case 22:return n=n.stateNode,i=n._retryCache,i===null&&(i=n._retryCache=new T0),i;default:throw Error(a(435,n.tag))}}function ud(n,i){var l=qT(n);i.forEach(function(c){var d=ZT.bind(null,n,c);l.has(c)||(l.add(c),c.then(d,d))})}function tn(n,i){var l=i.deletions;if(l!==null)for(var c=0;c<l.length;c++){var d=l[c],p=n,x=i,E=x;e:for(;E!==null;){switch(E.tag){case 27:if(Xr(E.type)){Xe=E.stateNode,Kt=!1;break e}break;case 5:Xe=E.stateNode,Kt=!1;break e;case 3:case 4:Xe=E.stateNode.containerInfo,Kt=!0;break e}E=E.return}if(Xe===null)throw Error(a(160));C0(p,x,d),Xe=null,Kt=!1,p=d.alternate,p!==null&&(p.return=null),d.return=null}if(i.subtreeFlags&13878)for(i=i.child;i!==null;)N0(i,n),i=i.sibling}var zn=null;function N0(n,i){var l=n.alternate,c=n.flags;switch(n.tag){case 0:case 11:case 14:case 15:tn(i,n),nn(n),c&4&&(Ur(3,n,n.return),kl(3,n),Ur(5,n,n.return));break;case 1:tn(i,n),nn(n),c&512&&(et||l===null||Gn(l,l.return)),c&64&&cr&&(n=n.updateQueue,n!==null&&(c=n.callbacks,c!==null&&(l=n.shared.hiddenCallbacks,n.shared.hiddenCallbacks=l===null?c:l.concat(c))));break;case 26:var d=zn;if(tn(i,n),nn(n),c&512&&(et||l===null||Gn(l,l.return)),c&4){var p=l!==null?l.memoizedState:null;if(c=n.memoizedState,l===null)if(c===null)if(n.stateNode===null){e:{c=n.type,l=n.memoizedProps,d=d.ownerDocument||d;t:switch(c){case"title":p=d.getElementsByTagName("title")[0],(!p||p[ll]||p[Ct]||p.namespaceURI==="http://www.w3.org/2000/svg"||p.hasAttribute("itemprop"))&&(p=d.createElement(c),d.head.insertBefore(p,d.querySelector("head > title"))),Et(p,c,l),p[Ct]=n,ht(p),c=p;break e;case"link":var x=_b("link","href",d).get(c+(l.href||""));if(x){for(var E=0;E<x.length;E++)if(p=x[E],p.getAttribute("href")===(l.href==null||l.href===""?null:l.href)&&p.getAttribute("rel")===(l.rel==null?null:l.rel)&&p.getAttribute("title")===(l.title==null?null:l.title)&&p.getAttribute("crossorigin")===(l.crossOrigin==null?null:l.crossOrigin)){x.splice(E,1);break t}}p=d.createElement(c),Et(p,c,l),d.head.appendChild(p);break;case"meta":if(x=_b("meta","content",d).get(c+(l.content||""))){for(E=0;E<x.length;E++)if(p=x[E],p.getAttribute("content")===(l.content==null?null:""+l.content)&&p.getAttribute("name")===(l.name==null?null:l.name)&&p.getAttribute("property")===(l.property==null?null:l.property)&&p.getAttribute("http-equiv")===(l.httpEquiv==null?null:l.httpEquiv)&&p.getAttribute("charset")===(l.charSet==null?null:l.charSet)){x.splice(E,1);break t}}p=d.createElement(c),Et(p,c,l),d.head.appendChild(p);break;default:throw Error(a(468,c))}p[Ct]=n,ht(p),c=p}n.stateNode=c}else Tb(d,n.type,n.stateNode);else n.stateNode=Eb(d,c,n.memoizedProps);else p!==c?(p===null?l.stateNode!==null&&(l=l.stateNode,l.parentNode.removeChild(l)):p.count--,c===null?Tb(d,n.type,n.stateNode):Eb(d,c,n.memoizedProps)):c===null&&n.stateNode!==null&&rd(n,n.memoizedProps,l.memoizedProps)}break;case 27:tn(i,n),nn(n),c&512&&(et||l===null||Gn(l,l.return)),l!==null&&c&4&&rd(n,n.memoizedProps,l.memoizedProps);break;case 5:if(tn(i,n),nn(n),c&512&&(et||l===null||Gn(l,l.return)),n.flags&32){d=n.stateNode;try{Ja(d,"")}catch(H){$e(n,n.return,H)}}c&4&&n.stateNode!=null&&(d=n.memoizedProps,rd(n,d,l!==null?l.memoizedProps:d)),c&1024&&(ld=!0);break;case 6:if(tn(i,n),nn(n),c&4){if(n.stateNode===null)throw Error(a(162));c=n.memoizedProps,l=n.stateNode;try{l.nodeValue=c}catch(H){$e(n,n.return,H)}}break;case 3:if(Ko=null,d=zn,zn=Yo(i.containerInfo),tn(i,n),zn=d,nn(n),c&4&&l!==null&&l.memoizedState.isDehydrated)try{Jl(i.containerInfo)}catch(H){$e(n,n.return,H)}ld&&(ld=!1,P0(n));break;case 4:c=zn,zn=Yo(n.stateNode.containerInfo),tn(i,n),nn(n),zn=c;break;case 12:tn(i,n),nn(n);break;case 13:tn(i,n),nn(n),n.child.flags&8192&&n.memoizedState!==null!=(l!==null&&l.memoizedState!==null)&&(hd=dt()),c&4&&(c=n.updateQueue,c!==null&&(n.updateQueue=null,ud(n,c)));break;case 22:d=n.memoizedState!==null;var C=l!==null&&l.memoizedState!==null,q=cr,K=et;if(cr=q||d,et=K||C,tn(i,n),et=K,cr=q,nn(n),c&8192)e:for(i=n.stateNode,i._visibility=d?i._visibility&-2:i._visibility|1,d&&(l===null||C||cr||et||Oa(n)),l=null,i=n;;){if(i.tag===5||i.tag===26){if(l===null){C=l=i;try{if(p=C.stateNode,d)x=p.style,typeof x.setProperty=="function"?x.setProperty("display","none","important"):x.display="none";else{E=C.stateNode;var Q=C.memoizedProps.style,$=Q!=null&&Q.hasOwnProperty("display")?Q.display:null;E.style.display=$==null||typeof $=="boolean"?"":(""+$).trim()}}catch(H){$e(C,C.return,H)}}}else if(i.tag===6){if(l===null){C=i;try{C.stateNode.nodeValue=d?"":C.memoizedProps}catch(H){$e(C,C.return,H)}}}else if((i.tag!==22&&i.tag!==23||i.memoizedState===null||i===n)&&i.child!==null){i.child.return=i,i=i.child;continue}if(i===n)break e;for(;i.sibling===null;){if(i.return===null||i.return===n)break e;l===i&&(l=null),i=i.return}l===i&&(l=null),i.sibling.return=i.return,i=i.sibling}c&4&&(c=n.updateQueue,c!==null&&(l=c.retryQueue,l!==null&&(c.retryQueue=null,ud(n,l))));break;case 19:tn(i,n),nn(n),c&4&&(c=n.updateQueue,c!==null&&(n.updateQueue=null,ud(n,c)));break;case 30:break;case 21:break;default:tn(i,n),nn(n)}}function nn(n){var i=n.flags;if(i&2){try{for(var l,c=n.return;c!==null;){if(E0(c)){l=c;break}c=c.return}if(l==null)throw Error(a(160));switch(l.tag){case 27:var d=l.stateNode,p=ad(n);Co(n,p,d);break;case 5:var x=l.stateNode;l.flags&32&&(Ja(x,""),l.flags&=-33);var E=ad(n);Co(n,E,x);break;case 3:case 4:var C=l.stateNode.containerInfo,q=ad(n);id(n,q,C);break;default:throw Error(a(161))}}catch(K){$e(n,n.return,K)}n.flags&=-3}i&4096&&(n.flags&=-4097)}function P0(n){if(n.subtreeFlags&1024)for(n=n.child;n!==null;){var i=n;P0(i),i.tag===5&&i.flags&1024&&i.stateNode.reset(),n=n.sibling}}function qr(n,i){if(i.subtreeFlags&8772)for(i=i.child;i!==null;)j0(n,i.alternate,i),i=i.sibling}function Oa(n){for(n=n.child;n!==null;){var i=n;switch(i.tag){case 0:case 11:case 14:case 15:Ur(4,i,i.return),Oa(i);break;case 1:Gn(i,i.return);var l=i.stateNode;typeof l.componentWillUnmount=="function"&&O0(i,i.return,l),Oa(i);break;case 27:Kl(i.stateNode);case 26:case 5:Gn(i,i.return),Oa(i);break;case 22:i.memoizedState===null&&Oa(i);break;case 30:Oa(i);break;default:Oa(i)}n=n.sibling}}function $r(n,i,l){for(l=l&&(i.subtreeFlags&8772)!==0,i=i.child;i!==null;){var c=i.alternate,d=n,p=i,x=p.flags;switch(p.tag){case 0:case 11:case 15:$r(d,p,l),kl(4,p);break;case 1:if($r(d,p,l),c=p,d=c.stateNode,typeof d.componentDidMount=="function")try{d.componentDidMount()}catch(q){$e(c,c.return,q)}if(c=p,d=c.updateQueue,d!==null){var E=c.stateNode;try{var C=d.shared.hiddenCallbacks;if(C!==null)for(d.shared.hiddenCallbacks=null,d=0;d<C.length;d++)og(C[d],E)}catch(q){$e(c,c.return,q)}}l&&x&64&&w0(p),zl(p,p.return);break;case 27:_0(p);case 26:case 5:$r(d,p,l),l&&c===null&&x&4&&A0(p),zl(p,p.return);break;case 12:$r(d,p,l);break;case 13:$r(d,p,l),l&&x&4&&D0(d,p);break;case 22:p.memoizedState===null&&$r(d,p,l),zl(p,p.return);break;case 30:break;default:$r(d,p,l)}i=i.sibling}}function od(n,i){var l=null;n!==null&&n.memoizedState!==null&&n.memoizedState.cachePool!==null&&(l=n.memoizedState.cachePool.pool),n=null,i.memoizedState!==null&&i.memoizedState.cachePool!==null&&(n=i.memoizedState.cachePool.pool),n!==l&&(n!=null&&n.refCount++,l!=null&&Sl(l))}function cd(n,i){n=null,i.alternate!==null&&(n=i.alternate.memoizedState.cache),i=i.memoizedState.cache,i!==n&&(i.refCount++,n!=null&&Sl(n))}function Kn(n,i,l,c){if(i.subtreeFlags&10256)for(i=i.child;i!==null;)R0(n,i,l,c),i=i.sibling}function R0(n,i,l,c){var d=i.flags;switch(i.tag){case 0:case 11:case 15:Kn(n,i,l,c),d&2048&&kl(9,i);break;case 1:Kn(n,i,l,c);break;case 3:Kn(n,i,l,c),d&2048&&(n=null,i.alternate!==null&&(n=i.alternate.memoizedState.cache),i=i.memoizedState.cache,i!==n&&(i.refCount++,n!=null&&Sl(n)));break;case 12:if(d&2048){Kn(n,i,l,c),n=i.stateNode;try{var p=i.memoizedProps,x=p.id,E=p.onPostCommit;typeof E=="function"&&E(x,i.alternate===null?"mount":"update",n.passiveEffectDuration,-0)}catch(C){$e(i,i.return,C)}}else Kn(n,i,l,c);break;case 13:Kn(n,i,l,c);break;case 23:break;case 22:p=i.stateNode,x=i.alternate,i.memoizedState!==null?p._visibility&2?Kn(n,i,l,c):Bl(n,i):p._visibility&2?Kn(n,i,l,c):(p._visibility|=2,gi(n,i,l,c,(i.subtreeFlags&10256)!==0)),d&2048&&od(x,i);break;case 24:Kn(n,i,l,c),d&2048&&cd(i.alternate,i);break;default:Kn(n,i,l,c)}}function gi(n,i,l,c,d){for(d=d&&(i.subtreeFlags&10256)!==0,i=i.child;i!==null;){var p=n,x=i,E=l,C=c,q=x.flags;switch(x.tag){case 0:case 11:case 15:gi(p,x,E,C,d),kl(8,x);break;case 23:break;case 22:var K=x.stateNode;x.memoizedState!==null?K._visibility&2?gi(p,x,E,C,d):Bl(p,x):(K._visibility|=2,gi(p,x,E,C,d)),d&&q&2048&&od(x.alternate,x);break;case 24:gi(p,x,E,C,d),d&&q&2048&&cd(x.alternate,x);break;default:gi(p,x,E,C,d)}i=i.sibling}}function Bl(n,i){if(i.subtreeFlags&10256)for(i=i.child;i!==null;){var l=n,c=i,d=c.flags;switch(c.tag){case 22:Bl(l,c),d&2048&&od(c.alternate,c);break;case 24:Bl(l,c),d&2048&&cd(c.alternate,c);break;default:Bl(l,c)}i=i.sibling}}var Ll=8192;function bi(n){if(n.subtreeFlags&Ll)for(n=n.child;n!==null;)k0(n),n=n.sibling}function k0(n){switch(n.tag){case 26:bi(n),n.flags&Ll&&n.memoizedState!==null&&Ej(zn,n.memoizedState,n.memoizedProps);break;case 5:bi(n);break;case 3:case 4:var i=zn;zn=Yo(n.stateNode.containerInfo),bi(n),zn=i;break;case 22:n.memoizedState===null&&(i=n.alternate,i!==null&&i.memoizedState!==null?(i=Ll,Ll=16777216,bi(n),Ll=i):bi(n));break;default:bi(n)}}function z0(n){var i=n.alternate;if(i!==null&&(n=i.child,n!==null)){i.child=null;do i=n.sibling,n.sibling=null,n=i;while(n!==null)}}function Ul(n){var i=n.deletions;if((n.flags&16)!==0){if(i!==null)for(var l=0;l<i.length;l++){var c=i[l];mt=c,L0(c,n)}z0(n)}if(n.subtreeFlags&10256)for(n=n.child;n!==null;)B0(n),n=n.sibling}function B0(n){switch(n.tag){case 0:case 11:case 15:Ul(n),n.flags&2048&&Ur(9,n,n.return);break;case 3:Ul(n);break;case 12:Ul(n);break;case 22:var i=n.stateNode;n.memoizedState!==null&&i._visibility&2&&(n.return===null||n.return.tag!==13)?(i._visibility&=-3,Do(n)):Ul(n);break;default:Ul(n)}}function Do(n){var i=n.deletions;if((n.flags&16)!==0){if(i!==null)for(var l=0;l<i.length;l++){var c=i[l];mt=c,L0(c,n)}z0(n)}for(n=n.child;n!==null;){switch(i=n,i.tag){case 0:case 11:case 15:Ur(8,i,i.return),Do(i);break;case 22:l=i.stateNode,l._visibility&2&&(l._visibility&=-3,Do(i));break;default:Do(i)}n=n.sibling}}function L0(n,i){for(;mt!==null;){var l=mt;switch(l.tag){case 0:case 11:case 15:Ur(8,l,i);break;case 23:case 22:if(l.memoizedState!==null&&l.memoizedState.cachePool!==null){var c=l.memoizedState.cachePool.pool;c!=null&&c.refCount++}break;case 24:Sl(l.memoizedState.cache)}if(c=l.child,c!==null)c.return=l,mt=c;else e:for(l=n;mt!==null;){c=mt;var d=c.sibling,p=c.return;if(M0(c),c===l){mt=null;break e}if(d!==null){d.return=p,mt=d;break e}mt=p}}}var $T={getCacheForType:function(n){var i=Dt(lt),l=i.data.get(n);return l===void 0&&(l=n(),i.data.set(n,l)),l}},HT=typeof WeakMap=="function"?WeakMap:Map,Ne=0,Ye=null,Ee=null,Te=0,Pe=0,rn=null,Hr=!1,xi=!1,sd=!1,fr=0,We=0,Yr=0,Aa=0,fd=0,On=0,Si=0,ql=null,It=null,dd=!1,hd=0,No=1/0,Po=null,Gr=null,At=0,Kr=null,wi=null,Oi=0,vd=0,md=null,U0=null,$l=0,pd=null;function an(){if((Ne&2)!==0&&Te!==0)return Te&-Te;if(B.T!==null){var n=si;return n!==0?n:Od()}return ty()}function q0(){On===0&&(On=(Te&536870912)===0||De?Wp():536870912);var n=wn.current;return n!==null&&(n.flags|=32),On}function ln(n,i,l){(n===Ye&&(Pe===2||Pe===9)||n.cancelPendingCommit!==null)&&(Ai(n,0),Ir(n,Te,On,!1)),il(n,l),((Ne&2)===0||n!==Ye)&&(n===Ye&&((Ne&2)===0&&(Aa|=l),We===4&&Ir(n,Te,On,!1)),In(n))}function $0(n,i,l){if((Ne&6)!==0)throw Error(a(327));var c=!l&&(i&124)===0&&(i&n.expiredLanes)===0||al(n,i),d=c?KT(n,i):bd(n,i,!0),p=c;do{if(d===0){xi&&!c&&Ir(n,i,0,!1);break}else{if(l=n.current.alternate,p&&!YT(l)){d=bd(n,i,!1),p=!1;continue}if(d===2){if(p=i,n.errorRecoveryDisabledLanes&p)var x=0;else x=n.pendingLanes&-536870913,x=x!==0?x:x&536870912?536870912:0;if(x!==0){i=x;e:{var E=n;d=ql;var C=E.current.memoizedState.isDehydrated;if(C&&(Ai(E,x).flags|=256),x=bd(E,x,!1),x!==2){if(sd&&!C){E.errorRecoveryDisabledLanes|=p,Aa|=p,d=4;break e}p=It,It=d,p!==null&&(It===null?It=p:It.push.apply(It,p))}d=x}if(p=!1,d!==2)continue}}if(d===1){Ai(n,0),Ir(n,i,0,!0);break}e:{switch(c=n,p=d,p){case 0:case 1:throw Error(a(345));case 4:if((i&4194048)!==i)break;case 6:Ir(c,i,On,!Hr);break e;case 2:It=null;break;case 3:case 5:break;default:throw Error(a(329))}if((i&62914560)===i&&(d=hd+300-dt(),10<d)){if(Ir(c,i,On,!Hr),Yu(c,0,!0)!==0)break e;c.timeoutHandle=pb(H0.bind(null,c,l,It,Po,dd,i,On,Aa,Si,Hr,p,2,-0,0),d);break e}H0(c,l,It,Po,dd,i,On,Aa,Si,Hr,p,0,-0,0)}}break}while(!0);In(n)}function H0(n,i,l,c,d,p,x,E,C,q,K,Q,$,H){if(n.timeoutHandle=-1,Q=i.subtreeFlags,(Q&8192||(Q&16785408)===16785408)&&(Vl={stylesheets:null,count:0,unsuspend:Aj},k0(i),Q=_j(),Q!==null)){n.cancelPendingCommit=Q(Z0.bind(null,n,i,p,l,c,d,x,E,C,K,1,$,H)),Ir(n,p,x,!q);return}Z0(n,i,p,l,c,d,x,E,C)}function YT(n){for(var i=n;;){var l=i.tag;if((l===0||l===11||l===15)&&i.flags&16384&&(l=i.updateQueue,l!==null&&(l=l.stores,l!==null)))for(var c=0;c<l.length;c++){var d=l[c],p=d.getSnapshot;d=d.value;try{if(!Jt(p(),d))return!1}catch{return!1}}if(l=i.child,i.subtreeFlags&16384&&l!==null)l.return=i,i=l;else{if(i===n)break;for(;i.sibling===null;){if(i.return===null||i.return===n)return!0;i=i.return}i.sibling.return=i.return,i=i.sibling}}return!0}function Ir(n,i,l,c){i&=~fd,i&=~Aa,n.suspendedLanes|=i,n.pingedLanes&=~i,c&&(n.warmLanes|=i),c=n.expirationTimes;for(var d=i;0<d;){var p=31-Ft(d),x=1<<p;c[p]=-1,d&=~x}l!==0&&Jp(n,l,i)}function Ro(){return(Ne&6)===0?(Hl(0),!1):!0}function yd(){if(Ee!==null){if(Pe===0)var n=Ee.return;else n=Ee,rr=ga=null,Rf(n),pi=null,Nl=0,n=Ee;for(;n!==null;)S0(n.alternate,n),n=n.return;Ee=null}}function Ai(n,i){var l=n.timeoutHandle;l!==-1&&(n.timeoutHandle=-1,uj(l)),l=n.cancelPendingCommit,l!==null&&(n.cancelPendingCommit=null,l()),yd(),Ye=n,Ee=l=er(n.current,null),Te=i,Pe=0,rn=null,Hr=!1,xi=al(n,i),sd=!1,Si=On=fd=Aa=Yr=We=0,It=ql=null,dd=!1,(i&8)!==0&&(i|=i&32);var c=n.entangledLanes;if(c!==0)for(n=n.entanglements,c&=i;0<c;){var d=31-Ft(c),p=1<<d;i|=n[d],c&=~p}return fr=i,to(),l}function Y0(n,i){Oe=null,B.H=So,i===Ol||i===so?(i=lg(),Pe=3):i===rg?(i=lg(),Pe=4):Pe=i===u0?8:i!==null&&typeof i=="object"&&typeof i.then=="function"?6:1,rn=i,Ee===null&&(We=1,_o(n,gn(i,n.current)))}function G0(){var n=B.H;return B.H=So,n===null?So:n}function K0(){var n=B.A;return B.A=$T,n}function gd(){We=4,Hr||(Te&4194048)!==Te&&wn.current!==null||(xi=!0),(Yr&134217727)===0&&(Aa&134217727)===0||Ye===null||Ir(Ye,Te,On,!1)}function bd(n,i,l){var c=Ne;Ne|=2;var d=G0(),p=K0();(Ye!==n||Te!==i)&&(Po=null,Ai(n,i)),i=!1;var x=We;e:do try{if(Pe!==0&&Ee!==null){var E=Ee,C=rn;switch(Pe){case 8:yd(),x=6;break e;case 3:case 2:case 9:case 6:wn.current===null&&(i=!0);var q=Pe;if(Pe=0,rn=null,Ei(n,E,C,q),l&&xi){x=0;break e}break;default:q=Pe,Pe=0,rn=null,Ei(n,E,C,q)}}GT(),x=We;break}catch(K){Y0(n,K)}while(!0);return i&&n.shellSuspendCounter++,rr=ga=null,Ne=c,B.H=d,B.A=p,Ee===null&&(Ye=null,Te=0,to()),x}function GT(){for(;Ee!==null;)I0(Ee)}function KT(n,i){var l=Ne;Ne|=2;var c=G0(),d=K0();Ye!==n||Te!==i?(Po=null,No=dt()+500,Ai(n,i)):xi=al(n,i);e:do try{if(Pe!==0&&Ee!==null){i=Ee;var p=rn;t:switch(Pe){case 1:Pe=0,rn=null,Ei(n,i,p,1);break;case 2:case 9:if(ag(p)){Pe=0,rn=null,X0(i);break}i=function(){Pe!==2&&Pe!==9||Ye!==n||(Pe=7),In(n)},p.then(i,i);break e;case 3:Pe=7;break e;case 4:Pe=5;break e;case 7:ag(p)?(Pe=0,rn=null,X0(i)):(Pe=0,rn=null,Ei(n,i,p,7));break;case 5:var x=null;switch(Ee.tag){case 26:x=Ee.memoizedState;case 5:case 27:var E=Ee;if(!x||jb(x)){Pe=0,rn=null;var C=E.sibling;if(C!==null)Ee=C;else{var q=E.return;q!==null?(Ee=q,ko(q)):Ee=null}break t}}Pe=0,rn=null,Ei(n,i,p,5);break;case 6:Pe=0,rn=null,Ei(n,i,p,6);break;case 8:yd(),We=6;break e;default:throw Error(a(462))}}IT();break}catch(K){Y0(n,K)}while(!0);return rr=ga=null,B.H=c,B.A=d,Ne=l,Ee!==null?0:(Ye=null,Te=0,to(),We)}function IT(){for(;Ee!==null&&!kt();)I0(Ee)}function I0(n){var i=b0(n.alternate,n,fr);n.memoizedProps=n.pendingProps,i===null?ko(n):Ee=i}function X0(n){var i=n,l=i.alternate;switch(i.tag){case 15:case 0:i=h0(l,i,i.pendingProps,i.type,void 0,Te);break;case 11:i=h0(l,i,i.pendingProps,i.type.render,i.ref,Te);break;case 5:Rf(i);default:S0(l,i),i=Ee=Vy(i,fr),i=b0(l,i,fr)}n.memoizedProps=n.pendingProps,i===null?ko(n):Ee=i}function Ei(n,i,l,c){rr=ga=null,Rf(i),pi=null,Nl=0;var d=i.return;try{if(kT(n,d,i,l,Te)){We=1,_o(n,gn(l,n.current)),Ee=null;return}}catch(p){if(d!==null)throw Ee=d,p;We=1,_o(n,gn(l,n.current)),Ee=null;return}i.flags&32768?(De||c===1?n=!0:xi||(Te&536870912)!==0?n=!1:(Hr=n=!0,(c===2||c===9||c===3||c===6)&&(c=wn.current,c!==null&&c.tag===13&&(c.flags|=16384))),V0(i,n)):ko(i)}function ko(n){var i=n;do{if((i.flags&32768)!==0){V0(i,Hr);return}n=i.return;var l=BT(i.alternate,i,fr);if(l!==null){Ee=l;return}if(i=i.sibling,i!==null){Ee=i;return}Ee=i=n}while(i!==null);We===0&&(We=5)}function V0(n,i){do{var l=LT(n.alternate,n);if(l!==null){l.flags&=32767,Ee=l;return}if(l=n.return,l!==null&&(l.flags|=32768,l.subtreeFlags=0,l.deletions=null),!i&&(n=n.sibling,n!==null)){Ee=n;return}Ee=n=l}while(n!==null);We=6,Ee=null}function Z0(n,i,l,c,d,p,x,E,C){n.cancelPendingCommit=null;do zo();while(At!==0);if((Ne&6)!==0)throw Error(a(327));if(i!==null){if(i===n.current)throw Error(a(177));if(p=i.lanes|i.childLanes,p|=cf,A_(n,l,p,x,E,C),n===Ye&&(Ee=Ye=null,Te=0),wi=i,Kr=n,Oi=l,vd=p,md=d,U0=c,(i.subtreeFlags&10256)!==0||(i.flags&10256)!==0?(n.callbackNode=null,n.callbackPriority=0,QT(qu,function(){return eb(),null})):(n.callbackNode=null,n.callbackPriority=0),c=(i.flags&13878)!==0,(i.subtreeFlags&13878)!==0||c){c=B.T,B.T=null,d=I.p,I.p=2,x=Ne,Ne|=4;try{UT(n,i,l)}finally{Ne=x,I.p=d,B.T=c}}At=1,Q0(),W0(),F0()}}function Q0(){if(At===1){At=0;var n=Kr,i=wi,l=(i.flags&13878)!==0;if((i.subtreeFlags&13878)!==0||l){l=B.T,B.T=null;var c=I.p;I.p=2;var d=Ne;Ne|=4;try{N0(i,n);var p=Dd,x=Ly(n.containerInfo),E=p.focusedElem,C=p.selectionRange;if(x!==E&&E&&E.ownerDocument&&By(E.ownerDocument.documentElement,E)){if(C!==null&&rf(E)){var q=C.start,K=C.end;if(K===void 0&&(K=q),"selectionStart"in E)E.selectionStart=q,E.selectionEnd=Math.min(K,E.value.length);else{var Q=E.ownerDocument||document,$=Q&&Q.defaultView||window;if($.getSelection){var H=$.getSelection(),xe=E.textContent.length,pe=Math.min(C.start,xe),Ue=C.end===void 0?pe:Math.min(C.end,xe);!H.extend&&pe>Ue&&(x=Ue,Ue=pe,pe=x);var z=zy(E,pe),R=zy(E,Ue);if(z&&R&&(H.rangeCount!==1||H.anchorNode!==z.node||H.anchorOffset!==z.offset||H.focusNode!==R.node||H.focusOffset!==R.offset)){var U=Q.createRange();U.setStart(z.node,z.offset),H.removeAllRanges(),pe>Ue?(H.addRange(U),H.extend(R.node,R.offset)):(U.setEnd(R.node,R.offset),H.addRange(U))}}}}for(Q=[],H=E;H=H.parentNode;)H.nodeType===1&&Q.push({element:H,left:H.scrollLeft,top:H.scrollTop});for(typeof E.focus=="function"&&E.focus(),E=0;E<Q.length;E++){var X=Q[E];X.element.scrollLeft=X.left,X.element.scrollTop=X.top}}Vo=!!Cd,Dd=Cd=null}finally{Ne=d,I.p=c,B.T=l}}n.current=i,At=2}}function W0(){if(At===2){At=0;var n=Kr,i=wi,l=(i.flags&8772)!==0;if((i.subtreeFlags&8772)!==0||l){l=B.T,B.T=null;var c=I.p;I.p=2;var d=Ne;Ne|=4;try{j0(n,i.alternate,i)}finally{Ne=d,I.p=c,B.T=l}}At=3}}function F0(){if(At===4||At===3){At=0,zt();var n=Kr,i=wi,l=Oi,c=U0;(i.subtreeFlags&10256)!==0||(i.flags&10256)!==0?At=5:(At=0,wi=Kr=null,J0(n,n.pendingLanes));var d=n.pendingLanes;if(d===0&&(Gr=null),zs(l),i=i.stateNode,Wt&&typeof Wt.onCommitFiberRoot=="function")try{Wt.onCommitFiberRoot(rl,i,void 0,(i.current.flags&128)===128)}catch{}if(c!==null){i=B.T,d=I.p,I.p=2,B.T=null;try{for(var p=n.onRecoverableError,x=0;x<c.length;x++){var E=c[x];p(E.value,{componentStack:E.stack})}}finally{B.T=i,I.p=d}}(Oi&3)!==0&&zo(),In(n),d=n.pendingLanes,(l&4194090)!==0&&(d&42)!==0?n===pd?$l++:($l=0,pd=n):$l=0,Hl(0)}}function J0(n,i){(n.pooledCacheLanes&=i)===0&&(i=n.pooledCache,i!=null&&(n.pooledCache=null,Sl(i)))}function zo(n){return Q0(),W0(),F0(),eb()}function eb(){if(At!==5)return!1;var n=Kr,i=vd;vd=0;var l=zs(Oi),c=B.T,d=I.p;try{I.p=32>l?32:l,B.T=null,l=md,md=null;var p=Kr,x=Oi;if(At=0,wi=Kr=null,Oi=0,(Ne&6)!==0)throw Error(a(331));var E=Ne;if(Ne|=4,B0(p.current),R0(p,p.current,x,l),Ne=E,Hl(0,!1),Wt&&typeof Wt.onPostCommitFiberRoot=="function")try{Wt.onPostCommitFiberRoot(rl,p)}catch{}return!0}finally{I.p=d,B.T=c,J0(n,i)}}function tb(n,i,l){i=gn(l,i),i=Vf(n.stateNode,i,2),n=kr(n,i,2),n!==null&&(il(n,2),In(n))}function $e(n,i,l){if(n.tag===3)tb(n,n,l);else for(;i!==null;){if(i.tag===3){tb(i,n,l);break}else if(i.tag===1){var c=i.stateNode;if(typeof i.type.getDerivedStateFromError=="function"||typeof c.componentDidCatch=="function"&&(Gr===null||!Gr.has(c))){n=gn(l,n),l=i0(2),c=kr(i,l,2),c!==null&&(l0(l,c,i,n),il(c,2),In(c));break}}i=i.return}}function xd(n,i,l){var c=n.pingCache;if(c===null){c=n.pingCache=new HT;var d=new Set;c.set(i,d)}else d=c.get(i),d===void 0&&(d=new Set,c.set(i,d));d.has(l)||(sd=!0,d.add(l),n=XT.bind(null,n,i,l),i.then(n,n))}function XT(n,i,l){var c=n.pingCache;c!==null&&c.delete(i),n.pingedLanes|=n.suspendedLanes&l,n.warmLanes&=~l,Ye===n&&(Te&l)===l&&(We===4||We===3&&(Te&62914560)===Te&&300>dt()-hd?(Ne&2)===0&&Ai(n,0):fd|=l,Si===Te&&(Si=0)),In(n)}function nb(n,i){i===0&&(i=Fp()),n=li(n,i),n!==null&&(il(n,i),In(n))}function VT(n){var i=n.memoizedState,l=0;i!==null&&(l=i.retryLane),nb(n,l)}function ZT(n,i){var l=0;switch(n.tag){case 13:var c=n.stateNode,d=n.memoizedState;d!==null&&(l=d.retryLane);break;case 19:c=n.stateNode;break;case 22:c=n.stateNode._retryCache;break;default:throw Error(a(314))}c!==null&&c.delete(i),nb(n,l)}function QT(n,i){return Fe(n,i)}var Bo=null,_i=null,Sd=!1,Lo=!1,wd=!1,Ea=0;function In(n){n!==_i&&n.next===null&&(_i===null?Bo=_i=n:_i=_i.next=n),Lo=!0,Sd||(Sd=!0,FT())}function Hl(n,i){if(!wd&&Lo){wd=!0;do for(var l=!1,c=Bo;c!==null;){if(n!==0){var d=c.pendingLanes;if(d===0)var p=0;else{var x=c.suspendedLanes,E=c.pingedLanes;p=(1<<31-Ft(42|n)+1)-1,p&=d&~(x&~E),p=p&201326741?p&201326741|1:p?p|2:0}p!==0&&(l=!0,lb(c,p))}else p=Te,p=Yu(c,c===Ye?p:0,c.cancelPendingCommit!==null||c.timeoutHandle!==-1),(p&3)===0||al(c,p)||(l=!0,lb(c,p));c=c.next}while(l);wd=!1}}function WT(){rb()}function rb(){Lo=Sd=!1;var n=0;Ea!==0&&(lj()&&(n=Ea),Ea=0);for(var i=dt(),l=null,c=Bo;c!==null;){var d=c.next,p=ab(c,i);p===0?(c.next=null,l===null?Bo=d:l.next=d,d===null&&(_i=l)):(l=c,(n!==0||(p&3)!==0)&&(Lo=!0)),c=d}Hl(n)}function ab(n,i){for(var l=n.suspendedLanes,c=n.pingedLanes,d=n.expirationTimes,p=n.pendingLanes&-62914561;0<p;){var x=31-Ft(p),E=1<<x,C=d[x];C===-1?((E&l)===0||(E&c)!==0)&&(d[x]=O_(E,i)):C<=i&&(n.expiredLanes|=E),p&=~E}if(i=Ye,l=Te,l=Yu(n,n===i?l:0,n.cancelPendingCommit!==null||n.timeoutHandle!==-1),c=n.callbackNode,l===0||n===i&&(Pe===2||Pe===9)||n.cancelPendingCommit!==null)return c!==null&&c!==null&&je(c),n.callbackNode=null,n.callbackPriority=0;if((l&3)===0||al(n,l)){if(i=l&-l,i===n.callbackPriority)return i;switch(c!==null&&je(c),zs(l)){case 2:case 8:l=Zp;break;case 32:l=qu;break;case 268435456:l=Qp;break;default:l=qu}return c=ib.bind(null,n),l=Fe(l,c),n.callbackPriority=i,n.callbackNode=l,i}return c!==null&&c!==null&&je(c),n.callbackPriority=2,n.callbackNode=null,2}function ib(n,i){if(At!==0&&At!==5)return n.callbackNode=null,n.callbackPriority=0,null;var l=n.callbackNode;if(zo()&&n.callbackNode!==l)return null;var c=Te;return c=Yu(n,n===Ye?c:0,n.cancelPendingCommit!==null||n.timeoutHandle!==-1),c===0?null:($0(n,c,i),ab(n,dt()),n.callbackNode!=null&&n.callbackNode===l?ib.bind(null,n):null)}function lb(n,i){if(zo())return null;$0(n,i,!0)}function FT(){oj(function(){(Ne&6)!==0?Fe(Vp,WT):rb()})}function Od(){return Ea===0&&(Ea=Wp()),Ea}function ub(n){return n==null||typeof n=="symbol"||typeof n=="boolean"?null:typeof n=="function"?n:Vu(""+n)}function ob(n,i){var l=i.ownerDocument.createElement("input");return l.name=i.name,l.value=i.value,n.id&&l.setAttribute("form",n.id),i.parentNode.insertBefore(l,i),n=new FormData(n),l.parentNode.removeChild(l),n}function JT(n,i,l,c,d){if(i==="submit"&&l&&l.stateNode===d){var p=ub((d[Ht]||null).action),x=c.submitter;x&&(i=(i=x[Ht]||null)?ub(i.formAction):x.getAttribute("formAction"),i!==null&&(p=i,x=null));var E=new Fu("action","action",null,c,d);n.push({event:E,listeners:[{instance:null,listener:function(){if(c.defaultPrevented){if(Ea!==0){var C=x?ob(d,x):new FormData(d);Yf(l,{pending:!0,data:C,method:d.method,action:p},null,C)}}else typeof p=="function"&&(E.preventDefault(),C=x?ob(d,x):new FormData(d),Yf(l,{pending:!0,data:C,method:d.method,action:p},p,C))},currentTarget:d}]})}}for(var Ad=0;Ad<of.length;Ad++){var Ed=of[Ad],ej=Ed.toLowerCase(),tj=Ed[0].toUpperCase()+Ed.slice(1);kn(ej,"on"+tj)}kn($y,"onAnimationEnd"),kn(Hy,"onAnimationIteration"),kn(Yy,"onAnimationStart"),kn("dblclick","onDoubleClick"),kn("focusin","onFocus"),kn("focusout","onBlur"),kn(gT,"onTransitionRun"),kn(bT,"onTransitionStart"),kn(xT,"onTransitionCancel"),kn(Gy,"onTransitionEnd"),Qa("onMouseEnter",["mouseout","mouseover"]),Qa("onMouseLeave",["mouseout","mouseover"]),Qa("onPointerEnter",["pointerout","pointerover"]),Qa("onPointerLeave",["pointerout","pointerover"]),ca("onChange","change click focusin focusout input keydown keyup selectionchange".split(" ")),ca("onSelect","focusout contextmenu dragend focusin keydown keyup mousedown mouseup selectionchange".split(" ")),ca("onBeforeInput",["compositionend","keypress","textInput","paste"]),ca("onCompositionEnd","compositionend focusout keydown keypress keyup mousedown".split(" ")),ca("onCompositionStart","compositionstart focusout keydown keypress keyup mousedown".split(" ")),ca("onCompositionUpdate","compositionupdate focusout keydown keypress keyup mousedown".split(" "));var Yl="abort canplay canplaythrough durationchange emptied encrypted ended error loadeddata loadedmetadata loadstart pause play playing progress ratechange resize seeked seeking stalled suspend timeupdate volumechange waiting".split(" "),nj=new Set("beforetoggle cancel close invalid load scroll scrollend toggle".split(" ").concat(Yl));function cb(n,i){i=(i&4)!==0;for(var l=0;l<n.length;l++){var c=n[l],d=c.event;c=c.listeners;e:{var p=void 0;if(i)for(var x=c.length-1;0<=x;x--){var E=c[x],C=E.instance,q=E.currentTarget;if(E=E.listener,C!==p&&d.isPropagationStopped())break e;p=E,d.currentTarget=q;try{p(d)}catch(K){Eo(K)}d.currentTarget=null,p=C}else for(x=0;x<c.length;x++){if(E=c[x],C=E.instance,q=E.currentTarget,E=E.listener,C!==p&&d.isPropagationStopped())break e;p=E,d.currentTarget=q;try{p(d)}catch(K){Eo(K)}d.currentTarget=null,p=C}}}}function _e(n,i){var l=i[Bs];l===void 0&&(l=i[Bs]=new Set);var c=n+"__bubble";l.has(c)||(sb(i,n,2,!1),l.add(c))}function _d(n,i,l){var c=0;i&&(c|=4),sb(l,n,c,i)}var Uo="_reactListening"+Math.random().toString(36).slice(2);function Td(n){if(!n[Uo]){n[Uo]=!0,ry.forEach(function(l){l!=="selectionchange"&&(nj.has(l)||_d(l,!1,n),_d(l,!0,n))});var i=n.nodeType===9?n:n.ownerDocument;i===null||i[Uo]||(i[Uo]=!0,_d("selectionchange",!1,i))}}function sb(n,i,l,c){switch(Rb(i)){case 2:var d=Mj;break;case 8:d=Cj;break;default:d=$d}l=d.bind(null,i,l,n),d=void 0,!Vs||i!=="touchstart"&&i!=="touchmove"&&i!=="wheel"||(d=!0),c?d!==void 0?n.addEventListener(i,l,{capture:!0,passive:d}):n.addEventListener(i,l,!0):d!==void 0?n.addEventListener(i,l,{passive:d}):n.addEventListener(i,l,!1)}function jd(n,i,l,c,d){var p=c;if((i&1)===0&&(i&2)===0&&c!==null)e:for(;;){if(c===null)return;var x=c.tag;if(x===3||x===4){var E=c.stateNode.containerInfo;if(E===d)break;if(x===4)for(x=c.return;x!==null;){var C=x.tag;if((C===3||C===4)&&x.stateNode.containerInfo===d)return;x=x.return}for(;E!==null;){if(x=Xa(E),x===null)return;if(C=x.tag,C===5||C===6||C===26||C===27){c=p=x;continue e}E=E.parentNode}}c=c.return}yy(function(){var q=p,K=Is(l),Q=[];e:{var $=Ky.get(n);if($!==void 0){var H=Fu,xe=n;switch(n){case"keypress":if(Qu(l)===0)break e;case"keydown":case"keyup":H=Q_;break;case"focusin":xe="focus",H=Fs;break;case"focusout":xe="blur",H=Fs;break;case"beforeblur":case"afterblur":H=Fs;break;case"click":if(l.button===2)break e;case"auxclick":case"dblclick":case"mousedown":case"mousemove":case"mouseup":case"mouseout":case"mouseover":case"contextmenu":H=xy;break;case"drag":case"dragend":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"dragstart":case"drop":H=L_;break;case"touchcancel":case"touchend":case"touchmove":case"touchstart":H=J_;break;case $y:case Hy:case Yy:H=$_;break;case Gy:H=tT;break;case"scroll":case"scrollend":H=z_;break;case"wheel":H=rT;break;case"copy":case"cut":case"paste":H=Y_;break;case"gotpointercapture":case"lostpointercapture":case"pointercancel":case"pointerdown":case"pointermove":case"pointerout":case"pointerover":case"pointerup":H=wy;break;case"toggle":case"beforetoggle":H=iT}var pe=(i&4)!==0,Ue=!pe&&(n==="scroll"||n==="scrollend"),z=pe?$!==null?$+"Capture":null:$;pe=[];for(var R=q,U;R!==null;){var X=R;if(U=X.stateNode,X=X.tag,X!==5&&X!==26&&X!==27||U===null||z===null||(X=ol(R,z),X!=null&&pe.push(Gl(R,X,U))),Ue)break;R=R.return}0<pe.length&&($=new H($,xe,null,l,K),Q.push({event:$,listeners:pe}))}}if((i&7)===0){e:{if($=n==="mouseover"||n==="pointerover",H=n==="mouseout"||n==="pointerout",$&&l!==Ks&&(xe=l.relatedTarget||l.fromElement)&&(Xa(xe)||xe[Ia]))break e;if((H||$)&&($=K.window===K?K:($=K.ownerDocument)?$.defaultView||$.parentWindow:window,H?(xe=l.relatedTarget||l.toElement,H=q,xe=xe?Xa(xe):null,xe!==null&&(Ue=o(xe),pe=xe.tag,xe!==Ue||pe!==5&&pe!==27&&pe!==6)&&(xe=null)):(H=null,xe=q),H!==xe)){if(pe=xy,X="onMouseLeave",z="onMouseEnter",R="mouse",(n==="pointerout"||n==="pointerover")&&(pe=wy,X="onPointerLeave",z="onPointerEnter",R="pointer"),Ue=H==null?$:ul(H),U=xe==null?$:ul(xe),$=new pe(X,R+"leave",H,l,K),$.target=Ue,$.relatedTarget=U,X=null,Xa(K)===q&&(pe=new pe(z,R+"enter",xe,l,K),pe.target=U,pe.relatedTarget=Ue,X=pe),Ue=X,H&&xe)t:{for(pe=H,z=xe,R=0,U=pe;U;U=Ti(U))R++;for(U=0,X=z;X;X=Ti(X))U++;for(;0<R-U;)pe=Ti(pe),R--;for(;0<U-R;)z=Ti(z),U--;for(;R--;){if(pe===z||z!==null&&pe===z.alternate)break t;pe=Ti(pe),z=Ti(z)}pe=null}else pe=null;H!==null&&fb(Q,$,H,pe,!1),xe!==null&&Ue!==null&&fb(Q,Ue,xe,pe,!0)}}e:{if($=q?ul(q):window,H=$.nodeName&&$.nodeName.toLowerCase(),H==="select"||H==="input"&&$.type==="file")var le=Cy;else if(jy($))if(Dy)le=mT;else{le=hT;var Ae=dT}else H=$.nodeName,!H||H.toLowerCase()!=="input"||$.type!=="checkbox"&&$.type!=="radio"?q&&Gs(q.elementType)&&(le=Cy):le=vT;if(le&&(le=le(n,q))){My(Q,le,l,K);break e}Ae&&Ae(n,$,q),n==="focusout"&&q&&$.type==="number"&&q.memoizedProps.value!=null&&Ys($,"number",$.value)}switch(Ae=q?ul(q):window,n){case"focusin":(jy(Ae)||Ae.contentEditable==="true")&&(ri=Ae,af=q,pl=null);break;case"focusout":pl=af=ri=null;break;case"mousedown":lf=!0;break;case"contextmenu":case"mouseup":case"dragend":lf=!1,Uy(Q,l,K);break;case"selectionchange":if(yT)break;case"keydown":case"keyup":Uy(Q,l,K)}var de;if(ef)e:{switch(n){case"compositionstart":var be="onCompositionStart";break e;case"compositionend":be="onCompositionEnd";break e;case"compositionupdate":be="onCompositionUpdate";break e}be=void 0}else ni?_y(n,l)&&(be="onCompositionEnd"):n==="keydown"&&l.keyCode===229&&(be="onCompositionStart");be&&(Oy&&l.locale!=="ko"&&(ni||be!=="onCompositionStart"?be==="onCompositionEnd"&&ni&&(de=gy()):(Dr=K,Zs="value"in Dr?Dr.value:Dr.textContent,ni=!0)),Ae=qo(q,be),0<Ae.length&&(be=new Sy(be,n,null,l,K),Q.push({event:be,listeners:Ae}),de?be.data=de:(de=Ty(l),de!==null&&(be.data=de)))),(de=uT?oT(n,l):cT(n,l))&&(be=qo(q,"onBeforeInput"),0<be.length&&(Ae=new Sy("onBeforeInput","beforeinput",null,l,K),Q.push({event:Ae,listeners:be}),Ae.data=de)),JT(Q,n,q,l,K)}cb(Q,i)})}function Gl(n,i,l){return{instance:n,listener:i,currentTarget:l}}function qo(n,i){for(var l=i+"Capture",c=[];n!==null;){var d=n,p=d.stateNode;if(d=d.tag,d!==5&&d!==26&&d!==27||p===null||(d=ol(n,l),d!=null&&c.unshift(Gl(n,d,p)),d=ol(n,i),d!=null&&c.push(Gl(n,d,p))),n.tag===3)return c;n=n.return}return[]}function Ti(n){if(n===null)return null;do n=n.return;while(n&&n.tag!==5&&n.tag!==27);return n||null}function fb(n,i,l,c,d){for(var p=i._reactName,x=[];l!==null&&l!==c;){var E=l,C=E.alternate,q=E.stateNode;if(E=E.tag,C!==null&&C===c)break;E!==5&&E!==26&&E!==27||q===null||(C=q,d?(q=ol(l,p),q!=null&&x.unshift(Gl(l,q,C))):d||(q=ol(l,p),q!=null&&x.push(Gl(l,q,C)))),l=l.return}x.length!==0&&n.push({event:i,listeners:x})}var rj=/\r\n?/g,aj=/\u0000|\uFFFD/g;function db(n){return(typeof n=="string"?n:""+n).replace(rj,`
`).replace(aj,"")}function hb(n,i){return i=db(i),db(n)===i}function $o(){}function Le(n,i,l,c,d,p){switch(l){case"children":typeof c=="string"?i==="body"||i==="textarea"&&c===""||Ja(n,c):(typeof c=="number"||typeof c=="bigint")&&i!=="body"&&Ja(n,""+c);break;case"className":Ku(n,"class",c);break;case"tabIndex":Ku(n,"tabindex",c);break;case"dir":case"role":case"viewBox":case"width":case"height":Ku(n,l,c);break;case"style":my(n,c,p);break;case"data":if(i!=="object"){Ku(n,"data",c);break}case"src":case"href":if(c===""&&(i!=="a"||l!=="href")){n.removeAttribute(l);break}if(c==null||typeof c=="function"||typeof c=="symbol"||typeof c=="boolean"){n.removeAttribute(l);break}c=Vu(""+c),n.setAttribute(l,c);break;case"action":case"formAction":if(typeof c=="function"){n.setAttribute(l,"javascript:throw new Error('A React form was unexpectedly submitted. If you called form.submit() manually, consider using form.requestSubmit() instead. If you\\'re trying to use event.stopPropagation() in a submit event handler, consider also calling event.preventDefault().')");break}else typeof p=="function"&&(l==="formAction"?(i!=="input"&&Le(n,i,"name",d.name,d,null),Le(n,i,"formEncType",d.formEncType,d,null),Le(n,i,"formMethod",d.formMethod,d,null),Le(n,i,"formTarget",d.formTarget,d,null)):(Le(n,i,"encType",d.encType,d,null),Le(n,i,"method",d.method,d,null),Le(n,i,"target",d.target,d,null)));if(c==null||typeof c=="symbol"||typeof c=="boolean"){n.removeAttribute(l);break}c=Vu(""+c),n.setAttribute(l,c);break;case"onClick":c!=null&&(n.onclick=$o);break;case"onScroll":c!=null&&_e("scroll",n);break;case"onScrollEnd":c!=null&&_e("scrollend",n);break;case"dangerouslySetInnerHTML":if(c!=null){if(typeof c!="object"||!("__html"in c))throw Error(a(61));if(l=c.__html,l!=null){if(d.children!=null)throw Error(a(60));n.innerHTML=l}}break;case"multiple":n.multiple=c&&typeof c!="function"&&typeof c!="symbol";break;case"muted":n.muted=c&&typeof c!="function"&&typeof c!="symbol";break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"defaultValue":case"defaultChecked":case"innerHTML":case"ref":break;case"autoFocus":break;case"xlinkHref":if(c==null||typeof c=="function"||typeof c=="boolean"||typeof c=="symbol"){n.removeAttribute("xlink:href");break}l=Vu(""+c),n.setAttributeNS("http://www.w3.org/1999/xlink","xlink:href",l);break;case"contentEditable":case"spellCheck":case"draggable":case"value":case"autoReverse":case"externalResourcesRequired":case"focusable":case"preserveAlpha":c!=null&&typeof c!="function"&&typeof c!="symbol"?n.setAttribute(l,""+c):n.removeAttribute(l);break;case"inert":case"allowFullScreen":case"async":case"autoPlay":case"controls":case"default":case"defer":case"disabled":case"disablePictureInPicture":case"disableRemotePlayback":case"formNoValidate":case"hidden":case"loop":case"noModule":case"noValidate":case"open":case"playsInline":case"readOnly":case"required":case"reversed":case"scoped":case"seamless":case"itemScope":c&&typeof c!="function"&&typeof c!="symbol"?n.setAttribute(l,""):n.removeAttribute(l);break;case"capture":case"download":c===!0?n.setAttribute(l,""):c!==!1&&c!=null&&typeof c!="function"&&typeof c!="symbol"?n.setAttribute(l,c):n.removeAttribute(l);break;case"cols":case"rows":case"size":case"span":c!=null&&typeof c!="function"&&typeof c!="symbol"&&!isNaN(c)&&1<=c?n.setAttribute(l,c):n.removeAttribute(l);break;case"rowSpan":case"start":c==null||typeof c=="function"||typeof c=="symbol"||isNaN(c)?n.removeAttribute(l):n.setAttribute(l,c);break;case"popover":_e("beforetoggle",n),_e("toggle",n),Gu(n,"popover",c);break;case"xlinkActuate":Fn(n,"http://www.w3.org/1999/xlink","xlink:actuate",c);break;case"xlinkArcrole":Fn(n,"http://www.w3.org/1999/xlink","xlink:arcrole",c);break;case"xlinkRole":Fn(n,"http://www.w3.org/1999/xlink","xlink:role",c);break;case"xlinkShow":Fn(n,"http://www.w3.org/1999/xlink","xlink:show",c);break;case"xlinkTitle":Fn(n,"http://www.w3.org/1999/xlink","xlink:title",c);break;case"xlinkType":Fn(n,"http://www.w3.org/1999/xlink","xlink:type",c);break;case"xmlBase":Fn(n,"http://www.w3.org/XML/1998/namespace","xml:base",c);break;case"xmlLang":Fn(n,"http://www.w3.org/XML/1998/namespace","xml:lang",c);break;case"xmlSpace":Fn(n,"http://www.w3.org/XML/1998/namespace","xml:space",c);break;case"is":Gu(n,"is",c);break;case"innerText":case"textContent":break;default:(!(2<l.length)||l[0]!=="o"&&l[0]!=="O"||l[1]!=="n"&&l[1]!=="N")&&(l=R_.get(l)||l,Gu(n,l,c))}}function Md(n,i,l,c,d,p){switch(l){case"style":my(n,c,p);break;case"dangerouslySetInnerHTML":if(c!=null){if(typeof c!="object"||!("__html"in c))throw Error(a(61));if(l=c.__html,l!=null){if(d.children!=null)throw Error(a(60));n.innerHTML=l}}break;case"children":typeof c=="string"?Ja(n,c):(typeof c=="number"||typeof c=="bigint")&&Ja(n,""+c);break;case"onScroll":c!=null&&_e("scroll",n);break;case"onScrollEnd":c!=null&&_e("scrollend",n);break;case"onClick":c!=null&&(n.onclick=$o);break;case"suppressContentEditableWarning":case"suppressHydrationWarning":case"innerHTML":case"ref":break;case"innerText":case"textContent":break;default:if(!ay.hasOwnProperty(l))e:{if(l[0]==="o"&&l[1]==="n"&&(d=l.endsWith("Capture"),i=l.slice(2,d?l.length-7:void 0),p=n[Ht]||null,p=p!=null?p[l]:null,typeof p=="function"&&n.removeEventListener(i,p,d),typeof c=="function")){typeof p!="function"&&p!==null&&(l in n?n[l]=null:n.hasAttribute(l)&&n.removeAttribute(l)),n.addEventListener(i,c,d);break e}l in n?n[l]=c:c===!0?n.setAttribute(l,""):Gu(n,l,c)}}}function Et(n,i,l){switch(i){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"img":_e("error",n),_e("load",n);var c=!1,d=!1,p;for(p in l)if(l.hasOwnProperty(p)){var x=l[p];if(x!=null)switch(p){case"src":c=!0;break;case"srcSet":d=!0;break;case"children":case"dangerouslySetInnerHTML":throw Error(a(137,i));default:Le(n,i,p,x,l,null)}}d&&Le(n,i,"srcSet",l.srcSet,l,null),c&&Le(n,i,"src",l.src,l,null);return;case"input":_e("invalid",n);var E=p=x=d=null,C=null,q=null;for(c in l)if(l.hasOwnProperty(c)){var K=l[c];if(K!=null)switch(c){case"name":d=K;break;case"type":x=K;break;case"checked":C=K;break;case"defaultChecked":q=K;break;case"value":p=K;break;case"defaultValue":E=K;break;case"children":case"dangerouslySetInnerHTML":if(K!=null)throw Error(a(137,i));break;default:Le(n,i,c,K,l,null)}}fy(n,p,E,C,q,x,d,!1),Iu(n);return;case"select":_e("invalid",n),c=x=p=null;for(d in l)if(l.hasOwnProperty(d)&&(E=l[d],E!=null))switch(d){case"value":p=E;break;case"defaultValue":x=E;break;case"multiple":c=E;default:Le(n,i,d,E,l,null)}i=p,l=x,n.multiple=!!c,i!=null?Fa(n,!!c,i,!1):l!=null&&Fa(n,!!c,l,!0);return;case"textarea":_e("invalid",n),p=d=c=null;for(x in l)if(l.hasOwnProperty(x)&&(E=l[x],E!=null))switch(x){case"value":c=E;break;case"defaultValue":d=E;break;case"children":p=E;break;case"dangerouslySetInnerHTML":if(E!=null)throw Error(a(91));break;default:Le(n,i,x,E,l,null)}hy(n,c,d,p),Iu(n);return;case"option":for(C in l)if(l.hasOwnProperty(C)&&(c=l[C],c!=null))switch(C){case"selected":n.selected=c&&typeof c!="function"&&typeof c!="symbol";break;default:Le(n,i,C,c,l,null)}return;case"dialog":_e("beforetoggle",n),_e("toggle",n),_e("cancel",n),_e("close",n);break;case"iframe":case"object":_e("load",n);break;case"video":case"audio":for(c=0;c<Yl.length;c++)_e(Yl[c],n);break;case"image":_e("error",n),_e("load",n);break;case"details":_e("toggle",n);break;case"embed":case"source":case"link":_e("error",n),_e("load",n);case"area":case"base":case"br":case"col":case"hr":case"keygen":case"meta":case"param":case"track":case"wbr":case"menuitem":for(q in l)if(l.hasOwnProperty(q)&&(c=l[q],c!=null))switch(q){case"children":case"dangerouslySetInnerHTML":throw Error(a(137,i));default:Le(n,i,q,c,l,null)}return;default:if(Gs(i)){for(K in l)l.hasOwnProperty(K)&&(c=l[K],c!==void 0&&Md(n,i,K,c,l,void 0));return}}for(E in l)l.hasOwnProperty(E)&&(c=l[E],c!=null&&Le(n,i,E,c,l,null))}function ij(n,i,l,c){switch(i){case"div":case"span":case"svg":case"path":case"a":case"g":case"p":case"li":break;case"input":var d=null,p=null,x=null,E=null,C=null,q=null,K=null;for(H in l){var Q=l[H];if(l.hasOwnProperty(H)&&Q!=null)switch(H){case"checked":break;case"value":break;case"defaultValue":C=Q;default:c.hasOwnProperty(H)||Le(n,i,H,null,c,Q)}}for(var $ in c){var H=c[$];if(Q=l[$],c.hasOwnProperty($)&&(H!=null||Q!=null))switch($){case"type":p=H;break;case"name":d=H;break;case"checked":q=H;break;case"defaultChecked":K=H;break;case"value":x=H;break;case"defaultValue":E=H;break;case"children":case"dangerouslySetInnerHTML":if(H!=null)throw Error(a(137,i));break;default:H!==Q&&Le(n,i,$,H,c,Q)}}Hs(n,x,E,C,q,K,p,d);return;case"select":H=x=E=$=null;for(p in l)if(C=l[p],l.hasOwnProperty(p)&&C!=null)switch(p){case"value":break;case"multiple":H=C;default:c.hasOwnProperty(p)||Le(n,i,p,null,c,C)}for(d in c)if(p=c[d],C=l[d],c.hasOwnProperty(d)&&(p!=null||C!=null))switch(d){case"value":$=p;break;case"defaultValue":E=p;break;case"multiple":x=p;default:p!==C&&Le(n,i,d,p,c,C)}i=E,l=x,c=H,$!=null?Fa(n,!!l,$,!1):!!c!=!!l&&(i!=null?Fa(n,!!l,i,!0):Fa(n,!!l,l?[]:"",!1));return;case"textarea":H=$=null;for(E in l)if(d=l[E],l.hasOwnProperty(E)&&d!=null&&!c.hasOwnProperty(E))switch(E){case"value":break;case"children":break;default:Le(n,i,E,null,c,d)}for(x in c)if(d=c[x],p=l[x],c.hasOwnProperty(x)&&(d!=null||p!=null))switch(x){case"value":$=d;break;case"defaultValue":H=d;break;case"children":break;case"dangerouslySetInnerHTML":if(d!=null)throw Error(a(91));break;default:d!==p&&Le(n,i,x,d,c,p)}dy(n,$,H);return;case"option":for(var xe in l)if($=l[xe],l.hasOwnProperty(xe)&&$!=null&&!c.hasOwnProperty(xe))switch(xe){case"selected":n.selected=!1;break;default:Le(n,i,xe,null,c,$)}for(C in c)if($=c[C],H=l[C],c.hasOwnProperty(C)&&$!==H&&($!=null||H!=null))switch(C){case"selected":n.selected=$&&typeof $!="function"&&typeof $!="symbol";break;default:Le(n,i,C,$,c,H)}return;case"img":case"link":case"area":case"base":case"br":case"col":case"embed":case"hr":case"keygen":case"meta":case"param":case"source":case"track":case"wbr":case"menuitem":for(var pe in l)$=l[pe],l.hasOwnProperty(pe)&&$!=null&&!c.hasOwnProperty(pe)&&Le(n,i,pe,null,c,$);for(q in c)if($=c[q],H=l[q],c.hasOwnProperty(q)&&$!==H&&($!=null||H!=null))switch(q){case"children":case"dangerouslySetInnerHTML":if($!=null)throw Error(a(137,i));break;default:Le(n,i,q,$,c,H)}return;default:if(Gs(i)){for(var Ue in l)$=l[Ue],l.hasOwnProperty(Ue)&&$!==void 0&&!c.hasOwnProperty(Ue)&&Md(n,i,Ue,void 0,c,$);for(K in c)$=c[K],H=l[K],!c.hasOwnProperty(K)||$===H||$===void 0&&H===void 0||Md(n,i,K,$,c,H);return}}for(var z in l)$=l[z],l.hasOwnProperty(z)&&$!=null&&!c.hasOwnProperty(z)&&Le(n,i,z,null,c,$);for(Q in c)$=c[Q],H=l[Q],!c.hasOwnProperty(Q)||$===H||$==null&&H==null||Le(n,i,Q,$,c,H)}var Cd=null,Dd=null;function Ho(n){return n.nodeType===9?n:n.ownerDocument}function vb(n){switch(n){case"http://www.w3.org/2000/svg":return 1;case"http://www.w3.org/1998/Math/MathML":return 2;default:return 0}}function mb(n,i){if(n===0)switch(i){case"svg":return 1;case"math":return 2;default:return 0}return n===1&&i==="foreignObject"?0:n}function Nd(n,i){return n==="textarea"||n==="noscript"||typeof i.children=="string"||typeof i.children=="number"||typeof i.children=="bigint"||typeof i.dangerouslySetInnerHTML=="object"&&i.dangerouslySetInnerHTML!==null&&i.dangerouslySetInnerHTML.__html!=null}var Pd=null;function lj(){var n=window.event;return n&&n.type==="popstate"?n===Pd?!1:(Pd=n,!0):(Pd=null,!1)}var pb=typeof setTimeout=="function"?setTimeout:void 0,uj=typeof clearTimeout=="function"?clearTimeout:void 0,yb=typeof Promise=="function"?Promise:void 0,oj=typeof queueMicrotask=="function"?queueMicrotask:typeof yb<"u"?function(n){return yb.resolve(null).then(n).catch(cj)}:pb;function cj(n){setTimeout(function(){throw n})}function Xr(n){return n==="head"}function gb(n,i){var l=i,c=0,d=0;do{var p=l.nextSibling;if(n.removeChild(l),p&&p.nodeType===8)if(l=p.data,l==="/$"){if(0<c&&8>c){l=c;var x=n.ownerDocument;if(l&1&&Kl(x.documentElement),l&2&&Kl(x.body),l&4)for(l=x.head,Kl(l),x=l.firstChild;x;){var E=x.nextSibling,C=x.nodeName;x[ll]||C==="SCRIPT"||C==="STYLE"||C==="LINK"&&x.rel.toLowerCase()==="stylesheet"||l.removeChild(x),x=E}}if(d===0){n.removeChild(p),Jl(i);return}d--}else l==="$"||l==="$?"||l==="$!"?d++:c=l.charCodeAt(0)-48;else c=0;l=p}while(l);Jl(i)}function Rd(n){var i=n.firstChild;for(i&&i.nodeType===10&&(i=i.nextSibling);i;){var l=i;switch(i=i.nextSibling,l.nodeName){case"HTML":case"HEAD":case"BODY":Rd(l),Ls(l);continue;case"SCRIPT":case"STYLE":continue;case"LINK":if(l.rel.toLowerCase()==="stylesheet")continue}n.removeChild(l)}}function sj(n,i,l,c){for(;n.nodeType===1;){var d=l;if(n.nodeName.toLowerCase()!==i.toLowerCase()){if(!c&&(n.nodeName!=="INPUT"||n.type!=="hidden"))break}else if(c){if(!n[ll])switch(i){case"meta":if(!n.hasAttribute("itemprop"))break;return n;case"link":if(p=n.getAttribute("rel"),p==="stylesheet"&&n.hasAttribute("data-precedence"))break;if(p!==d.rel||n.getAttribute("href")!==(d.href==null||d.href===""?null:d.href)||n.getAttribute("crossorigin")!==(d.crossOrigin==null?null:d.crossOrigin)||n.getAttribute("title")!==(d.title==null?null:d.title))break;return n;case"style":if(n.hasAttribute("data-precedence"))break;return n;case"script":if(p=n.getAttribute("src"),(p!==(d.src==null?null:d.src)||n.getAttribute("type")!==(d.type==null?null:d.type)||n.getAttribute("crossorigin")!==(d.crossOrigin==null?null:d.crossOrigin))&&p&&n.hasAttribute("async")&&!n.hasAttribute("itemprop"))break;return n;default:return n}}else if(i==="input"&&n.type==="hidden"){var p=d.name==null?null:""+d.name;if(d.type==="hidden"&&n.getAttribute("name")===p)return n}else return n;if(n=Bn(n.nextSibling),n===null)break}return null}function fj(n,i,l){if(i==="")return null;for(;n.nodeType!==3;)if((n.nodeType!==1||n.nodeName!=="INPUT"||n.type!=="hidden")&&!l||(n=Bn(n.nextSibling),n===null))return null;return n}function kd(n){return n.data==="$!"||n.data==="$?"&&n.ownerDocument.readyState==="complete"}function dj(n,i){var l=n.ownerDocument;if(n.data!=="$?"||l.readyState==="complete")i();else{var c=function(){i(),l.removeEventListener("DOMContentLoaded",c)};l.addEventListener("DOMContentLoaded",c),n._reactRetry=c}}function Bn(n){for(;n!=null;n=n.nextSibling){var i=n.nodeType;if(i===1||i===3)break;if(i===8){if(i=n.data,i==="$"||i==="$!"||i==="$?"||i==="F!"||i==="F")break;if(i==="/$")return null}}return n}var zd=null;function bb(n){n=n.previousSibling;for(var i=0;n;){if(n.nodeType===8){var l=n.data;if(l==="$"||l==="$!"||l==="$?"){if(i===0)return n;i--}else l==="/$"&&i++}n=n.previousSibling}return null}function xb(n,i,l){switch(i=Ho(l),n){case"html":if(n=i.documentElement,!n)throw Error(a(452));return n;case"head":if(n=i.head,!n)throw Error(a(453));return n;case"body":if(n=i.body,!n)throw Error(a(454));return n;default:throw Error(a(451))}}function Kl(n){for(var i=n.attributes;i.length;)n.removeAttributeNode(i[0]);Ls(n)}var An=new Map,Sb=new Set;function Yo(n){return typeof n.getRootNode=="function"?n.getRootNode():n.nodeType===9?n:n.ownerDocument}var dr=I.d;I.d={f:hj,r:vj,D:mj,C:pj,L:yj,m:gj,X:xj,S:bj,M:Sj};function hj(){var n=dr.f(),i=Ro();return n||i}function vj(n){var i=Va(n);i!==null&&i.tag===5&&i.type==="form"?$g(i):dr.r(n)}var ji=typeof document>"u"?null:document;function wb(n,i,l){var c=ji;if(c&&typeof i=="string"&&i){var d=yn(i);d='link[rel="'+n+'"][href="'+d+'"]',typeof l=="string"&&(d+='[crossorigin="'+l+'"]'),Sb.has(d)||(Sb.add(d),n={rel:n,crossOrigin:l,href:i},c.querySelector(d)===null&&(i=c.createElement("link"),Et(i,"link",n),ht(i),c.head.appendChild(i)))}}function mj(n){dr.D(n),wb("dns-prefetch",n,null)}function pj(n,i){dr.C(n,i),wb("preconnect",n,i)}function yj(n,i,l){dr.L(n,i,l);var c=ji;if(c&&n&&i){var d='link[rel="preload"][as="'+yn(i)+'"]';i==="image"&&l&&l.imageSrcSet?(d+='[imagesrcset="'+yn(l.imageSrcSet)+'"]',typeof l.imageSizes=="string"&&(d+='[imagesizes="'+yn(l.imageSizes)+'"]')):d+='[href="'+yn(n)+'"]';var p=d;switch(i){case"style":p=Mi(n);break;case"script":p=Ci(n)}An.has(p)||(n=m({rel:"preload",href:i==="image"&&l&&l.imageSrcSet?void 0:n,as:i},l),An.set(p,n),c.querySelector(d)!==null||i==="style"&&c.querySelector(Il(p))||i==="script"&&c.querySelector(Xl(p))||(i=c.createElement("link"),Et(i,"link",n),ht(i),c.head.appendChild(i)))}}function gj(n,i){dr.m(n,i);var l=ji;if(l&&n){var c=i&&typeof i.as=="string"?i.as:"script",d='link[rel="modulepreload"][as="'+yn(c)+'"][href="'+yn(n)+'"]',p=d;switch(c){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":p=Ci(n)}if(!An.has(p)&&(n=m({rel:"modulepreload",href:n},i),An.set(p,n),l.querySelector(d)===null)){switch(c){case"audioworklet":case"paintworklet":case"serviceworker":case"sharedworker":case"worker":case"script":if(l.querySelector(Xl(p)))return}c=l.createElement("link"),Et(c,"link",n),ht(c),l.head.appendChild(c)}}}function bj(n,i,l){dr.S(n,i,l);var c=ji;if(c&&n){var d=Za(c).hoistableStyles,p=Mi(n);i=i||"default";var x=d.get(p);if(!x){var E={loading:0,preload:null};if(x=c.querySelector(Il(p)))E.loading=5;else{n=m({rel:"stylesheet",href:n,"data-precedence":i},l),(l=An.get(p))&&Bd(n,l);var C=x=c.createElement("link");ht(C),Et(C,"link",n),C._p=new Promise(function(q,K){C.onload=q,C.onerror=K}),C.addEventListener("load",function(){E.loading|=1}),C.addEventListener("error",function(){E.loading|=2}),E.loading|=4,Go(x,i,c)}x={type:"stylesheet",instance:x,count:1,state:E},d.set(p,x)}}}function xj(n,i){dr.X(n,i);var l=ji;if(l&&n){var c=Za(l).hoistableScripts,d=Ci(n),p=c.get(d);p||(p=l.querySelector(Xl(d)),p||(n=m({src:n,async:!0},i),(i=An.get(d))&&Ld(n,i),p=l.createElement("script"),ht(p),Et(p,"link",n),l.head.appendChild(p)),p={type:"script",instance:p,count:1,state:null},c.set(d,p))}}function Sj(n,i){dr.M(n,i);var l=ji;if(l&&n){var c=Za(l).hoistableScripts,d=Ci(n),p=c.get(d);p||(p=l.querySelector(Xl(d)),p||(n=m({src:n,async:!0,type:"module"},i),(i=An.get(d))&&Ld(n,i),p=l.createElement("script"),ht(p),Et(p,"link",n),l.head.appendChild(p)),p={type:"script",instance:p,count:1,state:null},c.set(d,p))}}function Ob(n,i,l,c){var d=(d=ye.current)?Yo(d):null;if(!d)throw Error(a(446));switch(n){case"meta":case"title":return null;case"style":return typeof l.precedence=="string"&&typeof l.href=="string"?(i=Mi(l.href),l=Za(d).hoistableStyles,c=l.get(i),c||(c={type:"style",instance:null,count:0,state:null},l.set(i,c)),c):{type:"void",instance:null,count:0,state:null};case"link":if(l.rel==="stylesheet"&&typeof l.href=="string"&&typeof l.precedence=="string"){n=Mi(l.href);var p=Za(d).hoistableStyles,x=p.get(n);if(x||(d=d.ownerDocument||d,x={type:"stylesheet",instance:null,count:0,state:{loading:0,preload:null}},p.set(n,x),(p=d.querySelector(Il(n)))&&!p._p&&(x.instance=p,x.state.loading=5),An.has(n)||(l={rel:"preload",as:"style",href:l.href,crossOrigin:l.crossOrigin,integrity:l.integrity,media:l.media,hrefLang:l.hrefLang,referrerPolicy:l.referrerPolicy},An.set(n,l),p||wj(d,n,l,x.state))),i&&c===null)throw Error(a(528,""));return x}if(i&&c!==null)throw Error(a(529,""));return null;case"script":return i=l.async,l=l.src,typeof l=="string"&&i&&typeof i!="function"&&typeof i!="symbol"?(i=Ci(l),l=Za(d).hoistableScripts,c=l.get(i),c||(c={type:"script",instance:null,count:0,state:null},l.set(i,c)),c):{type:"void",instance:null,count:0,state:null};default:throw Error(a(444,n))}}function Mi(n){return'href="'+yn(n)+'"'}function Il(n){return'link[rel="stylesheet"]['+n+"]"}function Ab(n){return m({},n,{"data-precedence":n.precedence,precedence:null})}function wj(n,i,l,c){n.querySelector('link[rel="preload"][as="style"]['+i+"]")?c.loading=1:(i=n.createElement("link"),c.preload=i,i.addEventListener("load",function(){return c.loading|=1}),i.addEventListener("error",function(){return c.loading|=2}),Et(i,"link",l),ht(i),n.head.appendChild(i))}function Ci(n){return'[src="'+yn(n)+'"]'}function Xl(n){return"script[async]"+n}function Eb(n,i,l){if(i.count++,i.instance===null)switch(i.type){case"style":var c=n.querySelector('style[data-href~="'+yn(l.href)+'"]');if(c)return i.instance=c,ht(c),c;var d=m({},l,{"data-href":l.href,"data-precedence":l.precedence,href:null,precedence:null});return c=(n.ownerDocument||n).createElement("style"),ht(c),Et(c,"style",d),Go(c,l.precedence,n),i.instance=c;case"stylesheet":d=Mi(l.href);var p=n.querySelector(Il(d));if(p)return i.state.loading|=4,i.instance=p,ht(p),p;c=Ab(l),(d=An.get(d))&&Bd(c,d),p=(n.ownerDocument||n).createElement("link"),ht(p);var x=p;return x._p=new Promise(function(E,C){x.onload=E,x.onerror=C}),Et(p,"link",c),i.state.loading|=4,Go(p,l.precedence,n),i.instance=p;case"script":return p=Ci(l.src),(d=n.querySelector(Xl(p)))?(i.instance=d,ht(d),d):(c=l,(d=An.get(p))&&(c=m({},l),Ld(c,d)),n=n.ownerDocument||n,d=n.createElement("script"),ht(d),Et(d,"link",c),n.head.appendChild(d),i.instance=d);case"void":return null;default:throw Error(a(443,i.type))}else i.type==="stylesheet"&&(i.state.loading&4)===0&&(c=i.instance,i.state.loading|=4,Go(c,l.precedence,n));return i.instance}function Go(n,i,l){for(var c=l.querySelectorAll('link[rel="stylesheet"][data-precedence],style[data-precedence]'),d=c.length?c[c.length-1]:null,p=d,x=0;x<c.length;x++){var E=c[x];if(E.dataset.precedence===i)p=E;else if(p!==d)break}p?p.parentNode.insertBefore(n,p.nextSibling):(i=l.nodeType===9?l.head:l,i.insertBefore(n,i.firstChild))}function Bd(n,i){n.crossOrigin==null&&(n.crossOrigin=i.crossOrigin),n.referrerPolicy==null&&(n.referrerPolicy=i.referrerPolicy),n.title==null&&(n.title=i.title)}function Ld(n,i){n.crossOrigin==null&&(n.crossOrigin=i.crossOrigin),n.referrerPolicy==null&&(n.referrerPolicy=i.referrerPolicy),n.integrity==null&&(n.integrity=i.integrity)}var Ko=null;function _b(n,i,l){if(Ko===null){var c=new Map,d=Ko=new Map;d.set(l,c)}else d=Ko,c=d.get(l),c||(c=new Map,d.set(l,c));if(c.has(n))return c;for(c.set(n,null),l=l.getElementsByTagName(n),d=0;d<l.length;d++){var p=l[d];if(!(p[ll]||p[Ct]||n==="link"&&p.getAttribute("rel")==="stylesheet")&&p.namespaceURI!=="http://www.w3.org/2000/svg"){var x=p.getAttribute(i)||"";x=n+x;var E=c.get(x);E?E.push(p):c.set(x,[p])}}return c}function Tb(n,i,l){n=n.ownerDocument||n,n.head.insertBefore(l,i==="title"?n.querySelector("head > title"):null)}function Oj(n,i,l){if(l===1||i.itemProp!=null)return!1;switch(n){case"meta":case"title":return!0;case"style":if(typeof i.precedence!="string"||typeof i.href!="string"||i.href==="")break;return!0;case"link":if(typeof i.rel!="string"||typeof i.href!="string"||i.href===""||i.onLoad||i.onError)break;switch(i.rel){case"stylesheet":return n=i.disabled,typeof i.precedence=="string"&&n==null;default:return!0}case"script":if(i.async&&typeof i.async!="function"&&typeof i.async!="symbol"&&!i.onLoad&&!i.onError&&i.src&&typeof i.src=="string")return!0}return!1}function jb(n){return!(n.type==="stylesheet"&&(n.state.loading&3)===0)}var Vl=null;function Aj(){}function Ej(n,i,l){if(Vl===null)throw Error(a(475));var c=Vl;if(i.type==="stylesheet"&&(typeof l.media!="string"||matchMedia(l.media).matches!==!1)&&(i.state.loading&4)===0){if(i.instance===null){var d=Mi(l.href),p=n.querySelector(Il(d));if(p){n=p._p,n!==null&&typeof n=="object"&&typeof n.then=="function"&&(c.count++,c=Io.bind(c),n.then(c,c)),i.state.loading|=4,i.instance=p,ht(p);return}p=n.ownerDocument||n,l=Ab(l),(d=An.get(d))&&Bd(l,d),p=p.createElement("link"),ht(p);var x=p;x._p=new Promise(function(E,C){x.onload=E,x.onerror=C}),Et(p,"link",l),i.instance=p}c.stylesheets===null&&(c.stylesheets=new Map),c.stylesheets.set(i,n),(n=i.state.preload)&&(i.state.loading&3)===0&&(c.count++,i=Io.bind(c),n.addEventListener("load",i),n.addEventListener("error",i))}}function _j(){if(Vl===null)throw Error(a(475));var n=Vl;return n.stylesheets&&n.count===0&&Ud(n,n.stylesheets),0<n.count?function(i){var l=setTimeout(function(){if(n.stylesheets&&Ud(n,n.stylesheets),n.unsuspend){var c=n.unsuspend;n.unsuspend=null,c()}},6e4);return n.unsuspend=i,function(){n.unsuspend=null,clearTimeout(l)}}:null}function Io(){if(this.count--,this.count===0){if(this.stylesheets)Ud(this,this.stylesheets);else if(this.unsuspend){var n=this.unsuspend;this.unsuspend=null,n()}}}var Xo=null;function Ud(n,i){n.stylesheets=null,n.unsuspend!==null&&(n.count++,Xo=new Map,i.forEach(Tj,n),Xo=null,Io.call(n))}function Tj(n,i){if(!(i.state.loading&4)){var l=Xo.get(n);if(l)var c=l.get(null);else{l=new Map,Xo.set(n,l);for(var d=n.querySelectorAll("link[data-precedence],style[data-precedence]"),p=0;p<d.length;p++){var x=d[p];(x.nodeName==="LINK"||x.getAttribute("media")!=="not all")&&(l.set(x.dataset.precedence,x),c=x)}c&&l.set(null,c)}d=i.instance,x=d.getAttribute("data-precedence"),p=l.get(x)||c,p===c&&l.set(null,d),l.set(x,d),this.count++,c=Io.bind(this),d.addEventListener("load",c),d.addEventListener("error",c),p?p.parentNode.insertBefore(d,p.nextSibling):(n=n.nodeType===9?n.head:n,n.insertBefore(d,n.firstChild)),i.state.loading|=4}}var Zl={$$typeof:j,Provider:null,Consumer:null,_currentValue:F,_currentValue2:F,_threadCount:0};function jj(n,i,l,c,d,p,x,E){this.tag=1,this.containerInfo=n,this.pingCache=this.current=this.pendingChildren=null,this.timeoutHandle=-1,this.callbackNode=this.next=this.pendingContext=this.context=this.cancelPendingCommit=null,this.callbackPriority=0,this.expirationTimes=Rs(-1),this.entangledLanes=this.shellSuspendCounter=this.errorRecoveryDisabledLanes=this.expiredLanes=this.warmLanes=this.pingedLanes=this.suspendedLanes=this.pendingLanes=0,this.entanglements=Rs(0),this.hiddenUpdates=Rs(null),this.identifierPrefix=c,this.onUncaughtError=d,this.onCaughtError=p,this.onRecoverableError=x,this.pooledCache=null,this.pooledCacheLanes=0,this.formState=E,this.incompleteTransitions=new Map}function Mb(n,i,l,c,d,p,x,E,C,q,K,Q){return n=new jj(n,i,l,x,E,C,q,Q),i=1,p===!0&&(i|=24),p=en(3,null,null,i),n.current=p,p.stateNode=n,i=xf(),i.refCount++,n.pooledCache=i,i.refCount++,p.memoizedState={element:c,isDehydrated:l,cache:i},Af(p),n}function Cb(n){return n?(n=ui,n):ui}function Db(n,i,l,c,d,p){d=Cb(d),c.context===null?c.context=d:c.pendingContext=d,c=Rr(i),c.payload={element:l},p=p===void 0?null:p,p!==null&&(c.callback=p),l=kr(n,c,i),l!==null&&(ln(l,n,i),El(l,n,i))}function Nb(n,i){if(n=n.memoizedState,n!==null&&n.dehydrated!==null){var l=n.retryLane;n.retryLane=l!==0&&l<i?l:i}}function qd(n,i){Nb(n,i),(n=n.alternate)&&Nb(n,i)}function Pb(n){if(n.tag===13){var i=li(n,67108864);i!==null&&ln(i,n,67108864),qd(n,67108864)}}var Vo=!0;function Mj(n,i,l,c){var d=B.T;B.T=null;var p=I.p;try{I.p=2,$d(n,i,l,c)}finally{I.p=p,B.T=d}}function Cj(n,i,l,c){var d=B.T;B.T=null;var p=I.p;try{I.p=8,$d(n,i,l,c)}finally{I.p=p,B.T=d}}function $d(n,i,l,c){if(Vo){var d=Hd(c);if(d===null)jd(n,i,c,Zo,l),kb(n,c);else if(Nj(d,n,i,l,c))c.stopPropagation();else if(kb(n,c),i&4&&-1<Dj.indexOf(n)){for(;d!==null;){var p=Va(d);if(p!==null)switch(p.tag){case 3:if(p=p.stateNode,p.current.memoizedState.isDehydrated){var x=oa(p.pendingLanes);if(x!==0){var E=p;for(E.pendingLanes|=2,E.entangledLanes|=2;x;){var C=1<<31-Ft(x);E.entanglements[1]|=C,x&=~C}In(p),(Ne&6)===0&&(No=dt()+500,Hl(0))}}break;case 13:E=li(p,2),E!==null&&ln(E,p,2),Ro(),qd(p,2)}if(p=Hd(c),p===null&&jd(n,i,c,Zo,l),p===d)break;d=p}d!==null&&c.stopPropagation()}else jd(n,i,c,null,l)}}function Hd(n){return n=Is(n),Yd(n)}var Zo=null;function Yd(n){if(Zo=null,n=Xa(n),n!==null){var i=o(n);if(i===null)n=null;else{var l=i.tag;if(l===13){if(n=s(i),n!==null)return n;n=null}else if(l===3){if(i.stateNode.current.memoizedState.isDehydrated)return i.tag===3?i.stateNode.containerInfo:null;n=null}else i!==n&&(n=null)}}return Zo=n,null}function Rb(n){switch(n){case"beforetoggle":case"cancel":case"click":case"close":case"contextmenu":case"copy":case"cut":case"auxclick":case"dblclick":case"dragend":case"dragstart":case"drop":case"focusin":case"focusout":case"input":case"invalid":case"keydown":case"keypress":case"keyup":case"mousedown":case"mouseup":case"paste":case"pause":case"play":case"pointercancel":case"pointerdown":case"pointerup":case"ratechange":case"reset":case"resize":case"seeked":case"submit":case"toggle":case"touchcancel":case"touchend":case"touchstart":case"volumechange":case"change":case"selectionchange":case"textInput":case"compositionstart":case"compositionend":case"compositionupdate":case"beforeblur":case"afterblur":case"beforeinput":case"blur":case"fullscreenchange":case"focus":case"hashchange":case"popstate":case"select":case"selectstart":return 2;case"drag":case"dragenter":case"dragexit":case"dragleave":case"dragover":case"mousemove":case"mouseout":case"mouseover":case"pointermove":case"pointerout":case"pointerover":case"scroll":case"touchmove":case"wheel":case"mouseenter":case"mouseleave":case"pointerenter":case"pointerleave":return 8;case"message":switch(Uu()){case Vp:return 2;case Zp:return 8;case qu:case y_:return 32;case Qp:return 268435456;default:return 32}default:return 32}}var Gd=!1,Vr=null,Zr=null,Qr=null,Ql=new Map,Wl=new Map,Wr=[],Dj="mousedown mouseup touchcancel touchend touchstart auxclick dblclick pointercancel pointerdown pointerup dragend dragstart drop compositionend compositionstart keydown keypress keyup input textInput copy cut paste click change contextmenu reset".split(" ");function kb(n,i){switch(n){case"focusin":case"focusout":Vr=null;break;case"dragenter":case"dragleave":Zr=null;break;case"mouseover":case"mouseout":Qr=null;break;case"pointerover":case"pointerout":Ql.delete(i.pointerId);break;case"gotpointercapture":case"lostpointercapture":Wl.delete(i.pointerId)}}function Fl(n,i,l,c,d,p){return n===null||n.nativeEvent!==p?(n={blockedOn:i,domEventName:l,eventSystemFlags:c,nativeEvent:p,targetContainers:[d]},i!==null&&(i=Va(i),i!==null&&Pb(i)),n):(n.eventSystemFlags|=c,i=n.targetContainers,d!==null&&i.indexOf(d)===-1&&i.push(d),n)}function Nj(n,i,l,c,d){switch(i){case"focusin":return Vr=Fl(Vr,n,i,l,c,d),!0;case"dragenter":return Zr=Fl(Zr,n,i,l,c,d),!0;case"mouseover":return Qr=Fl(Qr,n,i,l,c,d),!0;case"pointerover":var p=d.pointerId;return Ql.set(p,Fl(Ql.get(p)||null,n,i,l,c,d)),!0;case"gotpointercapture":return p=d.pointerId,Wl.set(p,Fl(Wl.get(p)||null,n,i,l,c,d)),!0}return!1}function zb(n){var i=Xa(n.target);if(i!==null){var l=o(i);if(l!==null){if(i=l.tag,i===13){if(i=s(l),i!==null){n.blockedOn=i,E_(n.priority,function(){if(l.tag===13){var c=an();c=ks(c);var d=li(l,c);d!==null&&ln(d,l,c),qd(l,c)}});return}}else if(i===3&&l.stateNode.current.memoizedState.isDehydrated){n.blockedOn=l.tag===3?l.stateNode.containerInfo:null;return}}}n.blockedOn=null}function Qo(n){if(n.blockedOn!==null)return!1;for(var i=n.targetContainers;0<i.length;){var l=Hd(n.nativeEvent);if(l===null){l=n.nativeEvent;var c=new l.constructor(l.type,l);Ks=c,l.target.dispatchEvent(c),Ks=null}else return i=Va(l),i!==null&&Pb(i),n.blockedOn=l,!1;i.shift()}return!0}function Bb(n,i,l){Qo(n)&&l.delete(i)}function Pj(){Gd=!1,Vr!==null&&Qo(Vr)&&(Vr=null),Zr!==null&&Qo(Zr)&&(Zr=null),Qr!==null&&Qo(Qr)&&(Qr=null),Ql.forEach(Bb),Wl.forEach(Bb)}function Wo(n,i){n.blockedOn===i&&(n.blockedOn=null,Gd||(Gd=!0,e.unstable_scheduleCallback(e.unstable_NormalPriority,Pj)))}var Fo=null;function Lb(n){Fo!==n&&(Fo=n,e.unstable_scheduleCallback(e.unstable_NormalPriority,function(){Fo===n&&(Fo=null);for(var i=0;i<n.length;i+=3){var l=n[i],c=n[i+1],d=n[i+2];if(typeof c!="function"){if(Yd(c||l)===null)continue;break}var p=Va(l);p!==null&&(n.splice(i,3),i-=3,Yf(p,{pending:!0,data:d,method:l.method,action:c},c,d))}}))}function Jl(n){function i(C){return Wo(C,n)}Vr!==null&&Wo(Vr,n),Zr!==null&&Wo(Zr,n),Qr!==null&&Wo(Qr,n),Ql.forEach(i),Wl.forEach(i);for(var l=0;l<Wr.length;l++){var c=Wr[l];c.blockedOn===n&&(c.blockedOn=null)}for(;0<Wr.length&&(l=Wr[0],l.blockedOn===null);)zb(l),l.blockedOn===null&&Wr.shift();if(l=(n.ownerDocument||n).$$reactFormReplay,l!=null)for(c=0;c<l.length;c+=3){var d=l[c],p=l[c+1],x=d[Ht]||null;if(typeof p=="function")x||Lb(l);else if(x){var E=null;if(p&&p.hasAttribute("formAction")){if(d=p,x=p[Ht]||null)E=x.formAction;else if(Yd(d)!==null)continue}else E=x.action;typeof E=="function"?l[c+1]=E:(l.splice(c,3),c-=3),Lb(l)}}}function Kd(n){this._internalRoot=n}Jo.prototype.render=Kd.prototype.render=function(n){var i=this._internalRoot;if(i===null)throw Error(a(409));var l=i.current,c=an();Db(l,c,n,i,null,null)},Jo.prototype.unmount=Kd.prototype.unmount=function(){var n=this._internalRoot;if(n!==null){this._internalRoot=null;var i=n.containerInfo;Db(n.current,2,null,n,null,null),Ro(),i[Ia]=null}};function Jo(n){this._internalRoot=n}Jo.prototype.unstable_scheduleHydration=function(n){if(n){var i=ty();n={blockedOn:null,target:n,priority:i};for(var l=0;l<Wr.length&&i!==0&&i<Wr[l].priority;l++);Wr.splice(l,0,n),l===0&&zb(n)}};var Ub=t.version;if(Ub!=="19.1.1")throw Error(a(527,Ub,"19.1.1"));I.findDOMNode=function(n){var i=n._reactInternals;if(i===void 0)throw typeof n.render=="function"?Error(a(188)):(n=Object.keys(n).join(","),Error(a(268,n)));return n=v(i),n=n!==null?h(n):null,n=n===null?null:n.stateNode,n};var Rj={bundleType:0,version:"19.1.1",rendererPackageName:"react-dom",currentDispatcherRef:B,reconcilerVersion:"19.1.1"};if(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__<"u"){var ec=__REACT_DEVTOOLS_GLOBAL_HOOK__;if(!ec.isDisabled&&ec.supportsFiber)try{rl=ec.inject(Rj),Wt=ec}catch{}}return tu.createRoot=function(n,i){if(!u(n))throw Error(a(299));var l=!1,c="",d=t0,p=n0,x=r0,E=null;return i!=null&&(i.unstable_strictMode===!0&&(l=!0),i.identifierPrefix!==void 0&&(c=i.identifierPrefix),i.onUncaughtError!==void 0&&(d=i.onUncaughtError),i.onCaughtError!==void 0&&(p=i.onCaughtError),i.onRecoverableError!==void 0&&(x=i.onRecoverableError),i.unstable_transitionCallbacks!==void 0&&(E=i.unstable_transitionCallbacks)),i=Mb(n,1,!1,null,null,l,c,d,p,x,E,null),n[Ia]=i.current,Td(n),new Kd(i)},tu.hydrateRoot=function(n,i,l){if(!u(n))throw Error(a(299));var c=!1,d="",p=t0,x=n0,E=r0,C=null,q=null;return l!=null&&(l.unstable_strictMode===!0&&(c=!0),l.identifierPrefix!==void 0&&(d=l.identifierPrefix),l.onUncaughtError!==void 0&&(p=l.onUncaughtError),l.onCaughtError!==void 0&&(x=l.onCaughtError),l.onRecoverableError!==void 0&&(E=l.onRecoverableError),l.unstable_transitionCallbacks!==void 0&&(C=l.unstable_transitionCallbacks),l.formState!==void 0&&(q=l.formState)),i=Mb(n,1,!0,i,l??null,c,d,p,x,E,C,q),i.context=Cb(null),l=i.current,c=an(),c=ks(c),d=Rr(c),d.callback=null,kr(l,d,c),l=c,i.current.lanes=l,il(i,l),In(i),n[Ia]=i.current,Td(n),new Jo(i)},tu.version="19.1.1",tu}var Zb;function Kj(){if(Zb)return Vd.exports;Zb=1;function e(){if(!(typeof __REACT_DEVTOOLS_GLOBAL_HOOK__>"u"||typeof __REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE!="function"))try{__REACT_DEVTOOLS_GLOBAL_HOOK__.checkDCE(e)}catch(t){console.error(t)}}return e(),Vd.exports=Gj(),Vd.exports}var Ij=Kj();const Xj=Wn(Ij);async function Vj(e){const t=await fetch(`/api/benchmarks/${e}/stats`);if(!t.ok)throw new Error(`Failed to fetch benchmark stats: ${t.statusText}`);return t.json()}async function Zj(){const e=await fetch("/api/games");if(!e.ok)throw new Error(`Failed to fetch games: ${e.statusText}`);return e.json()}async function Qj(e){const t=await fetch(`/api/benchmarks/${e}/run`,{method:"POST"});if(!t.ok){const r=await t.json();throw new Error(r.detail||"Failed to start benchmark")}return t.json()}async function Wj(e){const t=await fetch(`/api/benchmarks/${e}/results`,{method:"DELETE"});if(!t.ok){const r=await t.json();throw new Error(r.detail||"Failed to delete results")}return t.json()}async function Fj(e,t){const r=await fetch(`/api/benchmarks/${e}/add-players`,{method:"POST",headers:{"Content-Type":"application/json"},body:JSON.stringify({players:t})});if(!r.ok){const a=await r.json();throw new Error(a.detail||"Failed to add players")}return r.json()}function Jj({data:e,title:t,className:r=""}){if(!e||e.length===0)return _.jsxs("div",{className:`${r}`,children:[t&&_.jsx("h3",{className:"text-lg font-semibold mb-4 text-gray-800",children:t}),_.jsx("div",{className:"text-center py-8 text-gray-500",children:"No data available"})]});const a=[...e].sort((u,o)=>o.value-u.value);return _.jsxs("div",{className:`${r}`,children:[t&&_.jsx("h3",{className:"text-lg font-semibold mb-4 text-gray-800",children:t}),_.jsx("div",{className:"overflow-x-auto",children:_.jsxs("table",{className:"min-w-full bg-white border border-gray-200 rounded-lg shadow-sm",children:[_.jsx("thead",{className:"bg-gray-50",children:_.jsxs("tr",{children:[_.jsx("th",{className:"px-6 py-3 text-left text-xs font-medium text-gray-500 uppercase tracking-wider",children:"Rank"}),_.jsx("th",{className:"px-6 py-3 text-left text-xs font-medium text-gray-500 uppercase tracking-wider",children:"Player"}),_.jsx("th",{className:"px-6 py-3 text-left text-xs font-medium text-gray-500 uppercase tracking-wider",children:"Score"})]})}),_.jsx("tbody",{className:"divide-y divide-gray-200",children:a.map((u,o)=>_.jsxs("tr",{className:o%2===0?"bg-white":"bg-gray-50",children:[_.jsx("td",{className:"px-6 py-4 whitespace-nowrap text-sm text-gray-900",children:_.jsxs("span",{className:`font-semibold ${o===0?"text-yellow-600":o===1?"text-gray-400":o===2?"text-amber-600":"text-gray-600"}`,children:["#",o+1]})}),_.jsx("td",{className:"px-6 py-4 whitespace-nowrap text-sm font-medium text-gray-900",children:u.name}),_.jsx("td",{className:"px-6 py-4 whitespace-nowrap text-sm text-gray-900",children:_.jsx("span",{className:"font-mono",children:u.value.toFixed(2)})})]},u.name))})]})})]})}function Gw(e){var t,r,a="";if(typeof e=="string"||typeof e=="number")a+=e;else if(typeof e=="object")if(Array.isArray(e)){var u=e.length;for(t=0;t<u;t++)e[t]&&(r=Gw(e[t]))&&(a&&(a+=" "),a+=r)}else for(r in e)e[r]&&(a&&(a+=" "),a+=r);return a}function Ie(){for(var e,t,r=0,a="",u=arguments.length;r<u;r++)(e=arguments[r])&&(t=Gw(e))&&(a&&(a+=" "),a+=t);return a}var Fd={},Jd={},Qb;function eM(){return Qb||(Qb=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return r==="__proto__"}e.isUnsafeProperty=t})(Jd)),Jd}var eh={},Wb;function Kw(){return Wb||(Wb=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){switch(typeof r){case"number":case"symbol":return!1;case"string":return r.includes(".")||r.includes("[")||r.includes("]")}}e.isDeepKey=t})(eh)),eh}var th={},Fb;function Iw(){return Fb||(Fb=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return typeof r=="string"||typeof r=="symbol"?r:Object.is(r?.valueOf?.(),-0)?"-0":String(r)}e.toKey=t})(th)),th}var nh={},Jb;function xm(){return Jb||(Jb=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){const a=[],u=r.length;if(u===0)return a;let o=0,s="",f="",v=!1;for(r.charCodeAt(0)===46&&(a.push(""),o++);o<u;){const h=r[o];f?h==="\\"&&o+1<u?(o++,s+=r[o]):h===f?f="":s+=h:v?h==='"'||h==="'"?f=h:h==="]"?(v=!1,a.push(s),s=""):s+=h:h==="["?(v=!0,s&&(a.push(s),s="")):h==="."?s&&(a.push(s),s=""):s+=h,o++}return s&&a.push(s),a}e.toPath=t})(nh)),nh}var ex;function Sm(){return ex||(ex=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=eM(),r=Kw(),a=Iw(),u=xm();function o(f,v,h){if(f==null)return h;switch(typeof v){case"string":{if(t.isUnsafeProperty(v))return h;const m=f[v];return m===void 0?r.isDeepKey(v)?o(f,u.toPath(v),h):h:m}case"number":case"symbol":{typeof v=="number"&&(v=a.toKey(v));const m=f[v];return m===void 0?h:m}default:{if(Array.isArray(v))return s(f,v,h);if(Object.is(v?.valueOf(),-0)?v="-0":v=String(v),t.isUnsafeProperty(v))return h;const m=f[v];return m===void 0?h:m}}}function s(f,v,h){if(v.length===0)return h;let m=f;for(let y=0;y<v.length;y++){if(m==null||t.isUnsafeProperty(v[y]))return h;m=m[v[y]]}return m===void 0?h:m}e.get=o})(Fd)),Fd}var rh,tx;function tM(){return tx||(tx=1,rh=Sm().get),rh}var nM=tM();const ki=Wn(nM);var ah={exports:{}},qe={};/**
 * @license React
 * react-is.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var nx;function rM(){if(nx)return qe;nx=1;var e=Symbol.for("react.transitional.element"),t=Symbol.for("react.portal"),r=Symbol.for("react.fragment"),a=Symbol.for("react.strict_mode"),u=Symbol.for("react.profiler"),o=Symbol.for("react.consumer"),s=Symbol.for("react.context"),f=Symbol.for("react.forward_ref"),v=Symbol.for("react.suspense"),h=Symbol.for("react.suspense_list"),m=Symbol.for("react.memo"),y=Symbol.for("react.lazy"),g=Symbol.for("react.view_transition"),b=Symbol.for("react.client.reference");function w(S){if(typeof S=="object"&&S!==null){var O=S.$$typeof;switch(O){case e:switch(S=S.type,S){case r:case u:case a:case v:case h:case g:return S;default:switch(S=S&&S.$$typeof,S){case s:case f:case y:case m:return S;case o:return S;default:return O}}case t:return O}}}return qe.ContextConsumer=o,qe.ContextProvider=s,qe.Element=e,qe.ForwardRef=f,qe.Fragment=r,qe.Lazy=y,qe.Memo=m,qe.Portal=t,qe.Profiler=u,qe.StrictMode=a,qe.Suspense=v,qe.SuspenseList=h,qe.isContextConsumer=function(S){return w(S)===o},qe.isContextProvider=function(S){return w(S)===s},qe.isElement=function(S){return typeof S=="object"&&S!==null&&S.$$typeof===e},qe.isForwardRef=function(S){return w(S)===f},qe.isFragment=function(S){return w(S)===r},qe.isLazy=function(S){return w(S)===y},qe.isMemo=function(S){return w(S)===m},qe.isPortal=function(S){return w(S)===t},qe.isProfiler=function(S){return w(S)===u},qe.isStrictMode=function(S){return w(S)===a},qe.isSuspense=function(S){return w(S)===v},qe.isSuspenseList=function(S){return w(S)===h},qe.isValidElementType=function(S){return typeof S=="string"||typeof S=="function"||S===r||S===u||S===a||S===v||S===h||typeof S=="object"&&S!==null&&(S.$$typeof===y||S.$$typeof===m||S.$$typeof===s||S.$$typeof===o||S.$$typeof===f||S.$$typeof===b||S.getModuleId!==void 0)},qe.typeOf=w,qe}var rx;function aM(){return rx||(rx=1,ah.exports=rM()),ah.exports}var iM=aM(),Vt=e=>e===0?0:e>0?1:-1,dn=e=>typeof e=="number"&&e!=+e,Ca=e=>typeof e=="string"&&e.indexOf("%")===e.length-1,ve=e=>(typeof e=="number"||e instanceof Number)&&!dn(e),Zn=e=>ve(e)||typeof e=="string",lM=0,vu=e=>{var t=++lM;return"".concat(e||"").concat(t)},qn=function(t,r){var a=arguments.length>2&&arguments[2]!==void 0?arguments[2]:0,u=arguments.length>3&&arguments[3]!==void 0?arguments[3]:!1;if(!ve(t)&&typeof t!="string")return a;var o;if(Ca(t)){if(r==null)return a;var s=t.indexOf("%");o=r*parseFloat(t.slice(0,s))/100}else o=+t;return dn(o)&&(o=a),u&&r!=null&&o>r&&(o=r),o},Xw=e=>{if(!Array.isArray(e))return!1;for(var t=e.length,r={},a=0;a<t;a++)if(!r[e[a]])r[e[a]]=!0;else return!0;return!1};function Lt(e,t,r){return ve(e)&&ve(t)?e+r*(t-e):t}function uM(e,t,r){if(!(!e||!e.length))return e.find(a=>a&&(typeof t=="function"?t(a):ki(a,t))===r)}var gt=e=>e===null||typeof e>"u",Au=e=>gt(e)?e:"".concat(e.charAt(0).toUpperCase()).concat(e.slice(1)),oM=["dangerouslySetInnerHTML","onCopy","onCopyCapture","onCut","onCutCapture","onPaste","onPasteCapture","onCompositionEnd","onCompositionEndCapture","onCompositionStart","onCompositionStartCapture","onCompositionUpdate","onCompositionUpdateCapture","onFocus","onFocusCapture","onBlur","onBlurCapture","onChange","onChangeCapture","onBeforeInput","onBeforeInputCapture","onInput","onInputCapture","onReset","onResetCapture","onSubmit","onSubmitCapture","onInvalid","onInvalidCapture","onLoad","onLoadCapture","onError","onErrorCapture","onKeyDown","onKeyDownCapture","onKeyPress","onKeyPressCapture","onKeyUp","onKeyUpCapture","onAbort","onAbortCapture","onCanPlay","onCanPlayCapture","onCanPlayThrough","onCanPlayThroughCapture","onDurationChange","onDurationChangeCapture","onEmptied","onEmptiedCapture","onEncrypted","onEncryptedCapture","onEnded","onEndedCapture","onLoadedData","onLoadedDataCapture","onLoadedMetadata","onLoadedMetadataCapture","onLoadStart","onLoadStartCapture","onPause","onPauseCapture","onPlay","onPlayCapture","onPlaying","onPlayingCapture","onProgress","onProgressCapture","onRateChange","onRateChangeCapture","onSeeked","onSeekedCapture","onSeeking","onSeekingCapture","onStalled","onStalledCapture","onSuspend","onSuspendCapture","onTimeUpdate","onTimeUpdateCapture","onVolumeChange","onVolumeChangeCapture","onWaiting","onWaitingCapture","onAuxClick","onAuxClickCapture","onClick","onClickCapture","onContextMenu","onContextMenuCapture","onDoubleClick","onDoubleClickCapture","onDrag","onDragCapture","onDragEnd","onDragEndCapture","onDragEnter","onDragEnterCapture","onDragExit","onDragExitCapture","onDragLeave","onDragLeaveCapture","onDragOver","onDragOverCapture","onDragStart","onDragStartCapture","onDrop","onDropCapture","onMouseDown","onMouseDownCapture","onMouseEnter","onMouseLeave","onMouseMove","onMouseMoveCapture","onMouseOut","onMouseOutCapture","onMouseOver","onMouseOverCapture","onMouseUp","onMouseUpCapture","onSelect","onSelectCapture","onTouchCancel","onTouchCancelCapture","onTouchEnd","onTouchEndCapture","onTouchMove","onTouchMoveCapture","onTouchStart","onTouchStartCapture","onPointerDown","onPointerDownCapture","onPointerMove","onPointerMoveCapture","onPointerUp","onPointerUpCapture","onPointerCancel","onPointerCancelCapture","onPointerEnter","onPointerEnterCapture","onPointerLeave","onPointerLeaveCapture","onPointerOver","onPointerOverCapture","onPointerOut","onPointerOutCapture","onGotPointerCapture","onGotPointerCaptureCapture","onLostPointerCapture","onLostPointerCaptureCapture","onScroll","onScrollCapture","onWheel","onWheelCapture","onAnimationStart","onAnimationStartCapture","onAnimationEnd","onAnimationEndCapture","onAnimationIteration","onAnimationIterationCapture","onTransitionEnd","onTransitionEndCapture"];function wm(e){if(typeof e!="string")return!1;var t=oM;return t.includes(e)}var cM=["viewBox","children"],ax=["points","pathLength"],ih={svg:cM,polygon:ax,polyline:ax},sM=(e,t)=>{if(!e||typeof e=="function"||typeof e=="boolean")return null;var r=e;if(A.isValidElement(e)&&(r=e.props),typeof r!="object"&&typeof r!="function")return null;var a={};return Object.keys(r).forEach(u=>{wm(u)&&(a[u]=(o=>r[u](r,o)))}),a},fM=(e,t,r)=>a=>(e(t,r,a),null),Om=(e,t,r)=>{if(e===null||typeof e!="object"&&typeof e!="function")return null;var a=null;return Object.keys(e).forEach(u=>{var o=e[u];wm(u)&&typeof o=="function"&&(a||(a={}),a[u]=fM(o,t,r))}),a},dM=["aria-activedescendant","aria-atomic","aria-autocomplete","aria-busy","aria-checked","aria-colcount","aria-colindex","aria-colspan","aria-controls","aria-current","aria-describedby","aria-details","aria-disabled","aria-errormessage","aria-expanded","aria-flowto","aria-haspopup","aria-hidden","aria-invalid","aria-keyshortcuts","aria-label","aria-labelledby","aria-level","aria-live","aria-modal","aria-multiline","aria-multiselectable","aria-orientation","aria-owns","aria-placeholder","aria-posinset","aria-pressed","aria-readonly","aria-relevant","aria-required","aria-roledescription","aria-rowcount","aria-rowindex","aria-rowspan","aria-selected","aria-setsize","aria-sort","aria-valuemax","aria-valuemin","aria-valuenow","aria-valuetext","className","color","height","id","lang","max","media","method","min","name","style","target","width","role","tabIndex","accentHeight","accumulate","additive","alignmentBaseline","allowReorder","alphabetic","amplitude","arabicForm","ascent","attributeName","attributeType","autoReverse","azimuth","baseFrequency","baselineShift","baseProfile","bbox","begin","bias","by","calcMode","capHeight","clip","clipPath","clipPathUnits","clipRule","colorInterpolation","colorInterpolationFilters","colorProfile","colorRendering","contentScriptType","contentStyleType","cursor","cx","cy","d","decelerate","descent","diffuseConstant","direction","display","divisor","dominantBaseline","dur","dx","dy","edgeMode","elevation","enableBackground","end","exponent","externalResourcesRequired","fill","fillOpacity","fillRule","filter","filterRes","filterUnits","floodColor","floodOpacity","focusable","fontFamily","fontSize","fontSizeAdjust","fontStretch","fontStyle","fontVariant","fontWeight","format","from","fx","fy","g1","g2","glyphName","glyphOrientationHorizontal","glyphOrientationVertical","glyphRef","gradientTransform","gradientUnits","hanging","horizAdvX","horizOriginX","href","ideographic","imageRendering","in2","in","intercept","k1","k2","k3","k4","k","kernelMatrix","kernelUnitLength","kerning","keyPoints","keySplines","keyTimes","lengthAdjust","letterSpacing","lightingColor","limitingConeAngle","local","markerEnd","markerHeight","markerMid","markerStart","markerUnits","markerWidth","mask","maskContentUnits","maskUnits","mathematical","mode","numOctaves","offset","opacity","operator","order","orient","orientation","origin","overflow","overlinePosition","overlineThickness","paintOrder","panose1","pathLength","patternContentUnits","patternTransform","patternUnits","pointerEvents","pointsAtX","pointsAtY","pointsAtZ","preserveAlpha","preserveAspectRatio","primitiveUnits","r","radius","refX","refY","renderingIntent","repeatCount","repeatDur","requiredExtensions","requiredFeatures","restart","result","rotate","rx","ry","seed","shapeRendering","slope","spacing","specularConstant","specularExponent","speed","spreadMethod","startOffset","stdDeviation","stemh","stemv","stitchTiles","stopColor","stopOpacity","strikethroughPosition","strikethroughThickness","string","stroke","strokeDasharray","strokeDashoffset","strokeLinecap","strokeLinejoin","strokeMiterlimit","strokeOpacity","strokeWidth","surfaceScale","systemLanguage","tableValues","targetX","targetY","textAnchor","textDecoration","textLength","textRendering","to","transform","u1","u2","underlinePosition","underlineThickness","unicode","unicodeBidi","unicodeRange","unitsPerEm","vAlphabetic","values","vectorEffect","version","vertAdvY","vertOriginX","vertOriginY","vHanging","vIdeographic","viewTarget","visibility","vMathematical","widths","wordSpacing","writingMode","x1","x2","x","xChannelSelector","xHeight","xlinkActuate","xlinkArcrole","xlinkHref","xlinkRole","xlinkShow","xlinkTitle","xlinkType","xmlBase","xmlLang","xmlns","xmlnsXlink","xmlSpace","y1","y2","y","yChannelSelector","z","zoomAndPan","ref","key","angle"];function Vw(e){if(typeof e!="string")return!1;var t=dM;return t.includes(e)}function Xi(e){var t=Object.entries(e).filter(r=>{var[a]=r;return Vw(a)});return Object.fromEntries(t)}var ix=e=>typeof e=="string"?e:e?e.displayName||e.name||"Component":"",lx=null,lh=null,Zw=e=>{if(e===lx&&Array.isArray(lh))return lh;var t=[];return A.Children.forEach(e,r=>{gt(r)||(iM.isFragment(r)?t=t.concat(Zw(r.props.children)):t.push(r))}),lh=t,lx=e,t};function hM(e,t){var r=[],a=[];return Array.isArray(t)?a=t.map(u=>ix(u)):a=[ix(t)],Zw(e).forEach(u=>{var o=ki(u,"type.displayName")||ki(u,"type.name");a.indexOf(o)!==-1&&r.push(u)}),r}var vM=(e,t,r,a)=>{var u;if(typeof t=="symbol"||typeof t=="number")return!0;var o=(u=a&&ih?.[a])!==null&&u!==void 0?u:[],s=t.startsWith("data-"),f=typeof e!="function"&&(!!a&&o.includes(t)||Vw(t)),v=!!r&&wm(t);return s||f||v},Mt=(e,t,r)=>{if(!e||typeof e=="function"||typeof e=="boolean")return null;var a=e;if(A.isValidElement(e)&&(a=e.props),typeof a!="object"&&typeof a!="function")return null;var u={};return Object.keys(a).forEach(o=>{var s;vM((s=a)===null||s===void 0?void 0:s[o],o,t,r)&&(u[o]=a[o])}),u},mM=["children","width","height","viewBox","className","style","title","desc"];function Dv(){return Dv=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},Dv.apply(null,arguments)}function pM(e,t){if(e==null)return{};var r,a,u=yM(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function yM(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}var Qw=A.forwardRef((e,t)=>{var{children:r,width:a,height:u,viewBox:o,className:s,style:f,title:v,desc:h}=e,m=pM(e,mM),y=o||{width:a,height:u,x:0,y:0},g=Ie("recharts-surface",s);return A.createElement("svg",Dv({},Mt(m,!0,"svg"),{className:g,width:a,height:u,style:f,viewBox:"".concat(y.x," ").concat(y.y," ").concat(y.width," ").concat(y.height),ref:t}),A.createElement("title",null,v),A.createElement("desc",null,h),r)}),gM=["children","className"];function Nv(){return Nv=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},Nv.apply(null,arguments)}function bM(e,t){if(e==null)return{};var r,a,u=xM(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function xM(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}var ra=A.forwardRef((e,t)=>{var{children:r,className:a}=e,u=bM(e,gM),o=Ie("recharts-layer",a);return A.createElement("g",Nv({className:o},Mt(u,!0),{ref:t}),r)}),SM=Yw(),wM=A.createContext(null);function Ge(e){return function(){return e}}const Ww=Math.cos,bc=Math.sin,$n=Math.sqrt,xc=Math.PI,as=2*xc,Pv=Math.PI,Rv=2*Pv,Ta=1e-6,OM=Rv-Ta;function Fw(e){this._+=e[0];for(let t=1,r=e.length;t<r;++t)this._+=arguments[t]+e[t]}function AM(e){let t=Math.floor(e);if(!(t>=0))throw new Error(`invalid digits: ${e}`);if(t>15)return Fw;const r=10**t;return function(a){this._+=a[0];for(let u=1,o=a.length;u<o;++u)this._+=Math.round(arguments[u]*r)/r+a[u]}}class EM{constructor(t){this._x0=this._y0=this._x1=this._y1=null,this._="",this._append=t==null?Fw:AM(t)}moveTo(t,r){this._append`M${this._x0=this._x1=+t},${this._y0=this._y1=+r}`}closePath(){this._x1!==null&&(this._x1=this._x0,this._y1=this._y0,this._append`Z`)}lineTo(t,r){this._append`L${this._x1=+t},${this._y1=+r}`}quadraticCurveTo(t,r,a,u){this._append`Q${+t},${+r},${this._x1=+a},${this._y1=+u}`}bezierCurveTo(t,r,a,u,o,s){this._append`C${+t},${+r},${+a},${+u},${this._x1=+o},${this._y1=+s}`}arcTo(t,r,a,u,o){if(t=+t,r=+r,a=+a,u=+u,o=+o,o<0)throw new Error(`negative radius: ${o}`);let s=this._x1,f=this._y1,v=a-t,h=u-r,m=s-t,y=f-r,g=m*m+y*y;if(this._x1===null)this._append`M${this._x1=t},${this._y1=r}`;else if(g>Ta)if(!(Math.abs(y*v-h*m)>Ta)||!o)this._append`L${this._x1=t},${this._y1=r}`;else{let b=a-s,w=u-f,S=v*v+h*h,O=b*b+w*w,T=Math.sqrt(S),M=Math.sqrt(g),j=o*Math.tan((Pv-Math.acos((S+g-O)/(2*T*M)))/2),k=j/M,N=j/T;Math.abs(k-1)>Ta&&this._append`L${t+k*m},${r+k*y}`,this._append`A${o},${o},0,0,${+(y*b>m*w)},${this._x1=t+N*v},${this._y1=r+N*h}`}}arc(t,r,a,u,o,s){if(t=+t,r=+r,a=+a,s=!!s,a<0)throw new Error(`negative radius: ${a}`);let f=a*Math.cos(u),v=a*Math.sin(u),h=t+f,m=r+v,y=1^s,g=s?u-o:o-u;this._x1===null?this._append`M${h},${m}`:(Math.abs(this._x1-h)>Ta||Math.abs(this._y1-m)>Ta)&&this._append`L${h},${m}`,a&&(g<0&&(g=g%Rv+Rv),g>OM?this._append`A${a},${a},0,1,${y},${t-f},${r-v}A${a},${a},0,1,${y},${this._x1=h},${this._y1=m}`:g>Ta&&this._append`A${a},${a},0,${+(g>=Pv)},${y},${this._x1=t+a*Math.cos(o)},${this._y1=r+a*Math.sin(o)}`)}rect(t,r,a,u){this._append`M${this._x0=this._x1=+t},${this._y0=this._y1=+r}h${a=+a}v${+u}h${-a}Z`}toString(){return this._}}function Am(e){let t=3;return e.digits=function(r){if(!arguments.length)return t;if(r==null)t=null;else{const a=Math.floor(r);if(!(a>=0))throw new RangeError(`invalid digits: ${r}`);t=a}return e},()=>new EM(t)}function Em(e){return typeof e=="object"&&"length"in e?e:Array.from(e)}function Jw(e){this._context=e}Jw.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._point=0},lineEnd:function(){(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;default:this._context.lineTo(e,t);break}}};function is(e){return new Jw(e)}function eO(e){return e[0]}function tO(e){return e[1]}function nO(e,t){var r=Ge(!0),a=null,u=is,o=null,s=Am(f);e=typeof e=="function"?e:e===void 0?eO:Ge(e),t=typeof t=="function"?t:t===void 0?tO:Ge(t);function f(v){var h,m=(v=Em(v)).length,y,g=!1,b;for(a==null&&(o=u(b=s())),h=0;h<=m;++h)!(h<m&&r(y=v[h],h,v))===g&&((g=!g)?o.lineStart():o.lineEnd()),g&&o.point(+e(y,h,v),+t(y,h,v));if(b)return o=null,b+""||null}return f.x=function(v){return arguments.length?(e=typeof v=="function"?v:Ge(+v),f):e},f.y=function(v){return arguments.length?(t=typeof v=="function"?v:Ge(+v),f):t},f.defined=function(v){return arguments.length?(r=typeof v=="function"?v:Ge(!!v),f):r},f.curve=function(v){return arguments.length?(u=v,a!=null&&(o=u(a)),f):u},f.context=function(v){return arguments.length?(v==null?a=o=null:o=u(a=v),f):a},f}function tc(e,t,r){var a=null,u=Ge(!0),o=null,s=is,f=null,v=Am(h);e=typeof e=="function"?e:e===void 0?eO:Ge(+e),t=typeof t=="function"?t:Ge(t===void 0?0:+t),r=typeof r=="function"?r:r===void 0?tO:Ge(+r);function h(y){var g,b,w,S=(y=Em(y)).length,O,T=!1,M,j=new Array(S),k=new Array(S);for(o==null&&(f=s(M=v())),g=0;g<=S;++g){if(!(g<S&&u(O=y[g],g,y))===T)if(T=!T)b=g,f.areaStart(),f.lineStart();else{for(f.lineEnd(),f.lineStart(),w=g-1;w>=b;--w)f.point(j[w],k[w]);f.lineEnd(),f.areaEnd()}T&&(j[g]=+e(O,g,y),k[g]=+t(O,g,y),f.point(a?+a(O,g,y):j[g],r?+r(O,g,y):k[g]))}if(M)return f=null,M+""||null}function m(){return nO().defined(u).curve(s).context(o)}return h.x=function(y){return arguments.length?(e=typeof y=="function"?y:Ge(+y),a=null,h):e},h.x0=function(y){return arguments.length?(e=typeof y=="function"?y:Ge(+y),h):e},h.x1=function(y){return arguments.length?(a=y==null?null:typeof y=="function"?y:Ge(+y),h):a},h.y=function(y){return arguments.length?(t=typeof y=="function"?y:Ge(+y),r=null,h):t},h.y0=function(y){return arguments.length?(t=typeof y=="function"?y:Ge(+y),h):t},h.y1=function(y){return arguments.length?(r=y==null?null:typeof y=="function"?y:Ge(+y),h):r},h.lineX0=h.lineY0=function(){return m().x(e).y(t)},h.lineY1=function(){return m().x(e).y(r)},h.lineX1=function(){return m().x(a).y(t)},h.defined=function(y){return arguments.length?(u=typeof y=="function"?y:Ge(!!y),h):u},h.curve=function(y){return arguments.length?(s=y,o!=null&&(f=s(o)),h):s},h.context=function(y){return arguments.length?(y==null?o=f=null:f=s(o=y),h):o},h}class rO{constructor(t,r){this._context=t,this._x=r}areaStart(){this._line=0}areaEnd(){this._line=NaN}lineStart(){this._point=0}lineEnd(){(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line=1-this._line}point(t,r){switch(t=+t,r=+r,this._point){case 0:{this._point=1,this._line?this._context.lineTo(t,r):this._context.moveTo(t,r);break}case 1:this._point=2;default:{this._x?this._context.bezierCurveTo(this._x0=(this._x0+t)/2,this._y0,this._x0,r,t,r):this._context.bezierCurveTo(this._x0,this._y0=(this._y0+r)/2,t,this._y0,t,r);break}}this._x0=t,this._y0=r}}function _M(e){return new rO(e,!0)}function TM(e){return new rO(e,!1)}const _m={draw(e,t){const r=$n(t/xc);e.moveTo(r,0),e.arc(0,0,r,0,as)}},jM={draw(e,t){const r=$n(t/5)/2;e.moveTo(-3*r,-r),e.lineTo(-r,-r),e.lineTo(-r,-3*r),e.lineTo(r,-3*r),e.lineTo(r,-r),e.lineTo(3*r,-r),e.lineTo(3*r,r),e.lineTo(r,r),e.lineTo(r,3*r),e.lineTo(-r,3*r),e.lineTo(-r,r),e.lineTo(-3*r,r),e.closePath()}},aO=$n(1/3),MM=aO*2,CM={draw(e,t){const r=$n(t/MM),a=r*aO;e.moveTo(0,-r),e.lineTo(a,0),e.lineTo(0,r),e.lineTo(-a,0),e.closePath()}},DM={draw(e,t){const r=$n(t),a=-r/2;e.rect(a,a,r,r)}},NM=.8908130915292852,iO=bc(xc/10)/bc(7*xc/10),PM=bc(as/10)*iO,RM=-Ww(as/10)*iO,kM={draw(e,t){const r=$n(t*NM),a=PM*r,u=RM*r;e.moveTo(0,-r),e.lineTo(a,u);for(let o=1;o<5;++o){const s=as*o/5,f=Ww(s),v=bc(s);e.lineTo(v*r,-f*r),e.lineTo(f*a-v*u,v*a+f*u)}e.closePath()}},uh=$n(3),zM={draw(e,t){const r=-$n(t/(uh*3));e.moveTo(0,r*2),e.lineTo(-uh*r,-r),e.lineTo(uh*r,-r),e.closePath()}},En=-.5,_n=$n(3)/2,kv=1/$n(12),BM=(kv/2+1)*3,LM={draw(e,t){const r=$n(t/BM),a=r/2,u=r*kv,o=a,s=r*kv+r,f=-o,v=s;e.moveTo(a,u),e.lineTo(o,s),e.lineTo(f,v),e.lineTo(En*a-_n*u,_n*a+En*u),e.lineTo(En*o-_n*s,_n*o+En*s),e.lineTo(En*f-_n*v,_n*f+En*v),e.lineTo(En*a+_n*u,En*u-_n*a),e.lineTo(En*o+_n*s,En*s-_n*o),e.lineTo(En*f+_n*v,En*v-_n*f),e.closePath()}};function UM(e,t){let r=null,a=Am(u);e=typeof e=="function"?e:Ge(e||_m),t=typeof t=="function"?t:Ge(t===void 0?64:+t);function u(){let o;if(r||(r=o=a()),e.apply(this,arguments).draw(r,+t.apply(this,arguments)),o)return r=null,o+""||null}return u.type=function(o){return arguments.length?(e=typeof o=="function"?o:Ge(o),u):e},u.size=function(o){return arguments.length?(t=typeof o=="function"?o:Ge(+o),u):t},u.context=function(o){return arguments.length?(r=o??null,u):r},u}function Sc(){}function wc(e,t,r){e._context.bezierCurveTo((2*e._x0+e._x1)/3,(2*e._y0+e._y1)/3,(e._x0+2*e._x1)/3,(e._y0+2*e._y1)/3,(e._x0+4*e._x1+t)/6,(e._y0+4*e._y1+r)/6)}function lO(e){this._context=e}lO.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){switch(this._point){case 3:wc(this,this._x1,this._y1);case 2:this._context.lineTo(this._x1,this._y1);break}(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;break;case 2:this._point=3,this._context.lineTo((5*this._x0+this._x1)/6,(5*this._y0+this._y1)/6);default:wc(this,e,t);break}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t}};function qM(e){return new lO(e)}function uO(e){this._context=e}uO.prototype={areaStart:Sc,areaEnd:Sc,lineStart:function(){this._x0=this._x1=this._x2=this._x3=this._x4=this._y0=this._y1=this._y2=this._y3=this._y4=NaN,this._point=0},lineEnd:function(){switch(this._point){case 1:{this._context.moveTo(this._x2,this._y2),this._context.closePath();break}case 2:{this._context.moveTo((this._x2+2*this._x3)/3,(this._y2+2*this._y3)/3),this._context.lineTo((this._x3+2*this._x2)/3,(this._y3+2*this._y2)/3),this._context.closePath();break}case 3:{this.point(this._x2,this._y2),this.point(this._x3,this._y3),this.point(this._x4,this._y4);break}}},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._x2=e,this._y2=t;break;case 1:this._point=2,this._x3=e,this._y3=t;break;case 2:this._point=3,this._x4=e,this._y4=t,this._context.moveTo((this._x0+4*this._x1+e)/6,(this._y0+4*this._y1+t)/6);break;default:wc(this,e,t);break}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t}};function $M(e){return new uO(e)}function oO(e){this._context=e}oO.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=NaN,this._point=0},lineEnd:function(){(this._line||this._line!==0&&this._point===3)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1;break;case 1:this._point=2;break;case 2:this._point=3;var r=(this._x0+4*this._x1+e)/6,a=(this._y0+4*this._y1+t)/6;this._line?this._context.lineTo(r,a):this._context.moveTo(r,a);break;case 3:this._point=4;default:wc(this,e,t);break}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t}};function HM(e){return new oO(e)}function cO(e){this._context=e}cO.prototype={areaStart:Sc,areaEnd:Sc,lineStart:function(){this._point=0},lineEnd:function(){this._point&&this._context.closePath()},point:function(e,t){e=+e,t=+t,this._point?this._context.lineTo(e,t):(this._point=1,this._context.moveTo(e,t))}};function YM(e){return new cO(e)}function ux(e){return e<0?-1:1}function ox(e,t,r){var a=e._x1-e._x0,u=t-e._x1,o=(e._y1-e._y0)/(a||u<0&&-0),s=(r-e._y1)/(u||a<0&&-0),f=(o*u+s*a)/(a+u);return(ux(o)+ux(s))*Math.min(Math.abs(o),Math.abs(s),.5*Math.abs(f))||0}function cx(e,t){var r=e._x1-e._x0;return r?(3*(e._y1-e._y0)/r-t)/2:t}function oh(e,t,r){var a=e._x0,u=e._y0,o=e._x1,s=e._y1,f=(o-a)/3;e._context.bezierCurveTo(a+f,u+f*t,o-f,s-f*r,o,s)}function Oc(e){this._context=e}Oc.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x0=this._x1=this._y0=this._y1=this._t0=NaN,this._point=0},lineEnd:function(){switch(this._point){case 2:this._context.lineTo(this._x1,this._y1);break;case 3:oh(this,this._t0,cx(this,this._t0));break}(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line=1-this._line},point:function(e,t){var r=NaN;if(e=+e,t=+t,!(e===this._x1&&t===this._y1)){switch(this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;break;case 2:this._point=3,oh(this,cx(this,r=ox(this,e,t)),r);break;default:oh(this,this._t0,r=ox(this,e,t));break}this._x0=this._x1,this._x1=e,this._y0=this._y1,this._y1=t,this._t0=r}}};function sO(e){this._context=new fO(e)}(sO.prototype=Object.create(Oc.prototype)).point=function(e,t){Oc.prototype.point.call(this,t,e)};function fO(e){this._context=e}fO.prototype={moveTo:function(e,t){this._context.moveTo(t,e)},closePath:function(){this._context.closePath()},lineTo:function(e,t){this._context.lineTo(t,e)},bezierCurveTo:function(e,t,r,a,u,o){this._context.bezierCurveTo(t,e,a,r,o,u)}};function GM(e){return new Oc(e)}function KM(e){return new sO(e)}function dO(e){this._context=e}dO.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=[],this._y=[]},lineEnd:function(){var e=this._x,t=this._y,r=e.length;if(r)if(this._line?this._context.lineTo(e[0],t[0]):this._context.moveTo(e[0],t[0]),r===2)this._context.lineTo(e[1],t[1]);else for(var a=sx(e),u=sx(t),o=0,s=1;s<r;++o,++s)this._context.bezierCurveTo(a[0][o],u[0][o],a[1][o],u[1][o],e[s],t[s]);(this._line||this._line!==0&&r===1)&&this._context.closePath(),this._line=1-this._line,this._x=this._y=null},point:function(e,t){this._x.push(+e),this._y.push(+t)}};function sx(e){var t,r=e.length-1,a,u=new Array(r),o=new Array(r),s=new Array(r);for(u[0]=0,o[0]=2,s[0]=e[0]+2*e[1],t=1;t<r-1;++t)u[t]=1,o[t]=4,s[t]=4*e[t]+2*e[t+1];for(u[r-1]=2,o[r-1]=7,s[r-1]=8*e[r-1]+e[r],t=1;t<r;++t)a=u[t]/o[t-1],o[t]-=a,s[t]-=a*s[t-1];for(u[r-1]=s[r-1]/o[r-1],t=r-2;t>=0;--t)u[t]=(s[t]-u[t+1])/o[t];for(o[r-1]=(e[r]+u[r-1])/2,t=0;t<r-1;++t)o[t]=2*e[t+1]-u[t+1];return[u,o]}function IM(e){return new dO(e)}function ls(e,t){this._context=e,this._t=t}ls.prototype={areaStart:function(){this._line=0},areaEnd:function(){this._line=NaN},lineStart:function(){this._x=this._y=NaN,this._point=0},lineEnd:function(){0<this._t&&this._t<1&&this._point===2&&this._context.lineTo(this._x,this._y),(this._line||this._line!==0&&this._point===1)&&this._context.closePath(),this._line>=0&&(this._t=1-this._t,this._line=1-this._line)},point:function(e,t){switch(e=+e,t=+t,this._point){case 0:this._point=1,this._line?this._context.lineTo(e,t):this._context.moveTo(e,t);break;case 1:this._point=2;default:{if(this._t<=0)this._context.lineTo(this._x,t),this._context.lineTo(e,t);else{var r=this._x*(1-this._t)+e*this._t;this._context.lineTo(r,this._y),this._context.lineTo(r,t)}break}}this._x=e,this._y=t}};function XM(e){return new ls(e,.5)}function VM(e){return new ls(e,0)}function ZM(e){return new ls(e,1)}function zi(e,t){if((s=e.length)>1)for(var r=1,a,u,o=e[t[0]],s,f=o.length;r<s;++r)for(u=o,o=e[t[r]],a=0;a<f;++a)o[a][1]+=o[a][0]=isNaN(u[a][1])?u[a][0]:u[a][1]}function zv(e){for(var t=e.length,r=new Array(t);--t>=0;)r[t]=t;return r}function QM(e,t){return e[t]}function WM(e){const t=[];return t.key=e,t}function FM(){var e=Ge([]),t=zv,r=zi,a=QM;function u(o){var s=Array.from(e.apply(this,arguments),WM),f,v=s.length,h=-1,m;for(const y of o)for(f=0,++h;f<v;++f)(s[f][h]=[0,+a(y,s[f].key,h,o)]).data=y;for(f=0,m=Em(t(s));f<v;++f)s[m[f]].index=f;return r(s,m),s}return u.keys=function(o){return arguments.length?(e=typeof o=="function"?o:Ge(Array.from(o)),u):e},u.value=function(o){return arguments.length?(a=typeof o=="function"?o:Ge(+o),u):a},u.order=function(o){return arguments.length?(t=o==null?zv:typeof o=="function"?o:Ge(Array.from(o)),u):t},u.offset=function(o){return arguments.length?(r=o??zi,u):r},u}function JM(e,t){if((a=e.length)>0){for(var r,a,u=0,o=e[0].length,s;u<o;++u){for(s=r=0;r<a;++r)s+=e[r][u][1]||0;if(s)for(r=0;r<a;++r)e[r][u][1]/=s}zi(e,t)}}function eC(e,t){if((u=e.length)>0){for(var r=0,a=e[t[0]],u,o=a.length;r<o;++r){for(var s=0,f=0;s<u;++s)f+=e[s][r][1]||0;a[r][1]+=a[r][0]=-f/2}zi(e,t)}}function tC(e,t){if(!(!((s=e.length)>0)||!((o=(u=e[t[0]]).length)>0))){for(var r=0,a=1,u,o,s;a<o;++a){for(var f=0,v=0,h=0;f<s;++f){for(var m=e[t[f]],y=m[a][1]||0,g=m[a-1][1]||0,b=(y-g)/2,w=0;w<f;++w){var S=e[t[w]],O=S[a][1]||0,T=S[a-1][1]||0;b+=O-T}v+=y,h+=b*y}u[a-1][1]+=u[a-1][0]=r,v&&(r-=h/v)}u[a-1][1]+=u[a-1][0]=r,zi(e,t)}}var nC=["type","size","sizeType"];function Bv(){return Bv=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},Bv.apply(null,arguments)}function fx(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function dx(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?fx(Object(r),!0).forEach(function(a){rC(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):fx(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function rC(e,t,r){return(t=aC(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function aC(e){var t=iC(e,"string");return typeof t=="symbol"?t:t+""}function iC(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function lC(e,t){if(e==null)return{};var r,a,u=uC(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function uC(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}var hO={symbolCircle:_m,symbolCross:jM,symbolDiamond:CM,symbolSquare:DM,symbolStar:kM,symbolTriangle:zM,symbolWye:LM},oC=Math.PI/180,cC=e=>{var t="symbol".concat(Au(e));return hO[t]||_m},sC=(e,t,r)=>{if(t==="area")return e;switch(r){case"cross":return 5*e*e/9;case"diamond":return .5*e*e/Math.sqrt(3);case"square":return e*e;case"star":{var a=18*oC;return 1.25*e*e*(Math.tan(a)-Math.tan(a*2)*Math.tan(a)**2)}case"triangle":return Math.sqrt(3)*e*e/4;case"wye":return(21-10*Math.sqrt(3))*e*e/8;default:return Math.PI*e*e/4}},fC=(e,t)=>{hO["symbol".concat(Au(e))]=t},vO=e=>{var{type:t="circle",size:r=64,sizeType:a="area"}=e,u=lC(e,nC),o=dx(dx({},u),{},{type:t,size:r,sizeType:a}),s="circle";typeof t=="string"&&(s=t);var f=()=>{var g=cC(s),b=UM().type(g).size(sC(r,a,s));return b()},{className:v,cx:h,cy:m}=o,y=Mt(o,!0);return h===+h&&m===+m&&r===+r?A.createElement("path",Bv({},y,{className:Ie("recharts-symbols",v),transform:"translate(".concat(h,", ").concat(m,")"),d:f()})):null};vO.registerSymbol=fC;var ch={},sh={},hx;function dC(){return hx||(hx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r,a){const u=new Map;for(let o=0;o<r.length;o++){const s=r[o],f=a(s);u.has(f)||u.set(f,s)}return Array.from(u.values())}e.uniqBy=t})(sh)),sh}var fh={},vx;function mO(){return vx||(vx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return r}e.identity=t})(fh)),fh}var dh={},hh={},vh={},mx;function hC(){return mx||(mx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return Number.isSafeInteger(r)&&r>=0}e.isLength=t})(vh)),vh}var px;function Tm(){return px||(px=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=hC();function r(a){return a!=null&&typeof a!="function"&&t.isLength(a.length)}e.isArrayLike=r})(hh)),hh}var mh={},yx;function vC(){return yx||(yx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return typeof r=="object"&&r!==null}e.isObjectLike=t})(mh)),mh}var gx;function mC(){return gx||(gx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Tm(),r=vC();function a(u){return r.isObjectLike(u)&&t.isArrayLike(u)}e.isArrayLikeObject=a})(dh)),dh}var ph={},yh={},bx;function pC(){return bx||(bx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Sm();function r(a){return function(u){return t.get(u,a)}}e.property=r})(yh)),yh}var gh={},bh={},xh={},Sh={},xx;function pO(){return xx||(xx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return r!==null&&(typeof r=="object"||typeof r=="function")}e.isObject=t})(Sh)),Sh}var wh={},Sx;function yO(){return Sx||(Sx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return r==null||typeof r!="object"&&typeof r!="function"}e.isPrimitive=t})(wh)),wh}var Oh={},wx;function gO(){return wx||(wx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r,a){return r===a||Number.isNaN(r)&&Number.isNaN(a)}e.eq=t})(Oh)),Oh}var Ox;function yC(){return Ox||(Ox=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=jm(),r=pO(),a=yO(),u=gO();function o(y,g,b){return typeof b!="function"?t.isMatch(y,g):s(y,g,function w(S,O,T,M,j,k){const N=b(S,O,T,M,j,k);return N!==void 0?!!N:s(S,O,w,k)},new Map)}function s(y,g,b,w){if(g===y)return!0;switch(typeof g){case"object":return f(y,g,b,w);case"function":return Object.keys(g).length>0?s(y,{...g},b,w):u.eq(y,g);default:return r.isObject(y)?typeof g=="string"?g==="":!0:u.eq(y,g)}}function f(y,g,b,w){if(g==null)return!0;if(Array.isArray(g))return h(y,g,b,w);if(g instanceof Map)return v(y,g,b,w);if(g instanceof Set)return m(y,g,b,w);const S=Object.keys(g);if(y==null)return S.length===0;if(S.length===0)return!0;if(w&&w.has(g))return w.get(g)===y;w&&w.set(g,y);try{for(let O=0;O<S.length;O++){const T=S[O];if(!a.isPrimitive(y)&&!(T in y)||g[T]===void 0&&y[T]!==void 0||g[T]===null&&y[T]!==null||!b(y[T],g[T],T,y,g,w))return!1}return!0}finally{w&&w.delete(g)}}function v(y,g,b,w){if(g.size===0)return!0;if(!(y instanceof Map))return!1;for(const[S,O]of g.entries()){const T=y.get(S);if(b(T,O,S,y,g,w)===!1)return!1}return!0}function h(y,g,b,w){if(g.length===0)return!0;if(!Array.isArray(y))return!1;const S=new Set;for(let O=0;O<g.length;O++){const T=g[O];let M=!1;for(let j=0;j<y.length;j++){if(S.has(j))continue;const k=y[j];let N=!1;if(b(k,T,O,y,g,w)&&(N=!0),N){S.add(j),M=!0;break}}if(!M)return!1}return!0}function m(y,g,b,w){return g.size===0?!0:y instanceof Set?h([...y],[...g],b,w):!1}e.isMatchWith=o,e.isSetMatch=m})(xh)),xh}var Ax;function jm(){return Ax||(Ax=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=yC();function r(a,u){return t.isMatchWith(a,u,()=>{})}e.isMatch=r})(bh)),bh}var Ah={},Eh={},_h={},Ex;function gC(){return Ex||(Ex=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return Object.getOwnPropertySymbols(r).filter(a=>Object.prototype.propertyIsEnumerable.call(r,a))}e.getSymbols=t})(_h)),_h}var Th={},_x;function bO(){return _x||(_x=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return r==null?r===void 0?"[object Undefined]":"[object Null]":Object.prototype.toString.call(r)}e.getTag=t})(Th)),Th}var jh={},Tx;function xO(){return Tx||(Tx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t="[object RegExp]",r="[object String]",a="[object Number]",u="[object Boolean]",o="[object Arguments]",s="[object Symbol]",f="[object Date]",v="[object Map]",h="[object Set]",m="[object Array]",y="[object Function]",g="[object ArrayBuffer]",b="[object Object]",w="[object Error]",S="[object DataView]",O="[object Uint8Array]",T="[object Uint8ClampedArray]",M="[object Uint16Array]",j="[object Uint32Array]",k="[object BigUint64Array]",N="[object Int8Array]",P="[object Int16Array]",L="[object Int32Array]",G="[object BigInt64Array]",ee="[object Float32Array]",ne="[object Float64Array]";e.argumentsTag=o,e.arrayBufferTag=g,e.arrayTag=m,e.bigInt64ArrayTag=G,e.bigUint64ArrayTag=k,e.booleanTag=u,e.dataViewTag=S,e.dateTag=f,e.errorTag=w,e.float32ArrayTag=ee,e.float64ArrayTag=ne,e.functionTag=y,e.int16ArrayTag=P,e.int32ArrayTag=L,e.int8ArrayTag=N,e.mapTag=v,e.numberTag=a,e.objectTag=b,e.regexpTag=t,e.setTag=h,e.stringTag=r,e.symbolTag=s,e.uint16ArrayTag=M,e.uint32ArrayTag=j,e.uint8ArrayTag=O,e.uint8ClampedArrayTag=T})(jh)),jh}var Mh={},jx;function bC(){return jx||(jx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return ArrayBuffer.isView(r)&&!(r instanceof DataView)}e.isTypedArray=t})(Mh)),Mh}var Mx;function SO(){return Mx||(Mx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=gC(),r=bO(),a=xO(),u=yO(),o=bC();function s(m,y){return f(m,void 0,m,new Map,y)}function f(m,y,g,b=new Map,w=void 0){const S=w?.(m,y,g,b);if(S!==void 0)return S;if(u.isPrimitive(m))return m;if(b.has(m))return b.get(m);if(Array.isArray(m)){const O=new Array(m.length);b.set(m,O);for(let T=0;T<m.length;T++)O[T]=f(m[T],T,g,b,w);return Object.hasOwn(m,"index")&&(O.index=m.index),Object.hasOwn(m,"input")&&(O.input=m.input),O}if(m instanceof Date)return new Date(m.getTime());if(m instanceof RegExp){const O=new RegExp(m.source,m.flags);return O.lastIndex=m.lastIndex,O}if(m instanceof Map){const O=new Map;b.set(m,O);for(const[T,M]of m)O.set(T,f(M,T,g,b,w));return O}if(m instanceof Set){const O=new Set;b.set(m,O);for(const T of m)O.add(f(T,void 0,g,b,w));return O}if(typeof Buffer<"u"&&Buffer.isBuffer(m))return m.subarray();if(o.isTypedArray(m)){const O=new(Object.getPrototypeOf(m)).constructor(m.length);b.set(m,O);for(let T=0;T<m.length;T++)O[T]=f(m[T],T,g,b,w);return O}if(m instanceof ArrayBuffer||typeof SharedArrayBuffer<"u"&&m instanceof SharedArrayBuffer)return m.slice(0);if(m instanceof DataView){const O=new DataView(m.buffer.slice(0),m.byteOffset,m.byteLength);return b.set(m,O),v(O,m,g,b,w),O}if(typeof File<"u"&&m instanceof File){const O=new File([m],m.name,{type:m.type});return b.set(m,O),v(O,m,g,b,w),O}if(m instanceof Blob){const O=new Blob([m],{type:m.type});return b.set(m,O),v(O,m,g,b,w),O}if(m instanceof Error){const O=new m.constructor;return b.set(m,O),O.message=m.message,O.name=m.name,O.stack=m.stack,O.cause=m.cause,v(O,m,g,b,w),O}if(typeof m=="object"&&h(m)){const O=Object.create(Object.getPrototypeOf(m));return b.set(m,O),v(O,m,g,b,w),O}return m}function v(m,y,g=m,b,w){const S=[...Object.keys(y),...t.getSymbols(y)];for(let O=0;O<S.length;O++){const T=S[O],M=Object.getOwnPropertyDescriptor(m,T);(M==null||M.writable)&&(m[T]=f(y[T],T,g,b,w))}}function h(m){switch(r.getTag(m)){case a.argumentsTag:case a.arrayTag:case a.arrayBufferTag:case a.dataViewTag:case a.booleanTag:case a.dateTag:case a.float32ArrayTag:case a.float64ArrayTag:case a.int8ArrayTag:case a.int16ArrayTag:case a.int32ArrayTag:case a.mapTag:case a.numberTag:case a.objectTag:case a.regexpTag:case a.setTag:case a.stringTag:case a.symbolTag:case a.uint8ArrayTag:case a.uint8ClampedArrayTag:case a.uint16ArrayTag:case a.uint32ArrayTag:return!0;default:return!1}}e.cloneDeepWith=s,e.cloneDeepWithImpl=f,e.copyProperties=v})(Eh)),Eh}var Cx;function xC(){return Cx||(Cx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=SO();function r(a){return t.cloneDeepWithImpl(a,void 0,a,new Map,void 0)}e.cloneDeep=r})(Ah)),Ah}var Dx;function SC(){return Dx||(Dx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=jm(),r=xC();function a(u){return u=r.cloneDeep(u),o=>t.isMatch(o,u)}e.matches=a})(gh)),gh}var Ch={},Dh={},Nh={},Nx;function wC(){return Nx||(Nx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=SO(),r=xO();function a(u,o){return t.cloneDeepWith(u,(s,f,v,h)=>{const m=o?.(s,f,v,h);if(m!==void 0)return m;if(typeof u=="object")switch(Object.prototype.toString.call(u)){case r.numberTag:case r.stringTag:case r.booleanTag:{const y=new u.constructor(u?.valueOf());return t.copyProperties(y,u),y}case r.argumentsTag:{const y={};return t.copyProperties(y,u),y.length=u.length,y[Symbol.iterator]=u[Symbol.iterator],y}default:return}})}e.cloneDeepWith=a})(Nh)),Nh}var Px;function OC(){return Px||(Px=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=wC();function r(a){return t.cloneDeepWith(a)}e.cloneDeep=r})(Dh)),Dh}var Ph={},Rh={},Rx;function wO(){return Rx||(Rx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=/^(?:0|[1-9]\d*)$/;function r(a,u=Number.MAX_SAFE_INTEGER){switch(typeof a){case"number":return Number.isInteger(a)&&a>=0&&a<u;case"symbol":return!1;case"string":return t.test(a)}}e.isIndex=r})(Rh)),Rh}var kh={},kx;function AC(){return kx||(kx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=bO();function r(a){return a!==null&&typeof a=="object"&&t.getTag(a)==="[object Arguments]"}e.isArguments=r})(kh)),kh}var zx;function EC(){return zx||(zx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Kw(),r=wO(),a=AC(),u=xm();function o(s,f){let v;if(Array.isArray(f)?v=f:typeof f=="string"&&t.isDeepKey(f)&&s?.[f]==null?v=u.toPath(f):v=[f],v.length===0)return!1;let h=s;for(let m=0;m<v.length;m++){const y=v[m];if((h==null||!Object.hasOwn(h,y))&&!((Array.isArray(h)||a.isArguments(h))&&r.isIndex(y)&&y<h.length))return!1;h=h[y]}return!0}e.has=o})(Ph)),Ph}var Bx;function _C(){return Bx||(Bx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=jm(),r=Iw(),a=OC(),u=Sm(),o=EC();function s(f,v){switch(typeof f){case"object":{Object.is(f?.valueOf(),-0)&&(f="-0");break}case"number":{f=r.toKey(f);break}}return v=a.cloneDeep(v),function(h){const m=u.get(h,f);return m===void 0?o.has(h,f):v===void 0?m===void 0:t.isMatch(m,v)}}e.matchesProperty=s})(Ch)),Ch}var Lx;function TC(){return Lx||(Lx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=mO(),r=pC(),a=SC(),u=_C();function o(s){if(s==null)return t.identity;switch(typeof s){case"function":return s;case"object":return Array.isArray(s)&&s.length===2?u.matchesProperty(s[0],s[1]):a.matches(s);case"string":case"symbol":case"number":return r.property(s)}}e.iteratee=o})(ph)),ph}var Ux;function jC(){return Ux||(Ux=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=dC(),r=mO(),a=mC(),u=TC();function o(s,f=r.identity){return a.isArrayLikeObject(s)?t.uniqBy(Array.from(s),u.iteratee(f)):[]}e.uniqBy=o})(ch)),ch}var zh,qx;function MC(){return qx||(qx=1,zh=jC().uniqBy),zh}var CC=MC();const $x=Wn(CC);function DC(e,t,r){return t===!0?$x(e,r):typeof t=="function"?$x(e,t):e}var Bh={exports:{}},Lh={},Uh={exports:{}},qh={};/**
 * @license React
 * use-sync-external-store-shim.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Hx;function NC(){if(Hx)return qh;Hx=1;var e=Ii();function t(y,g){return y===g&&(y!==0||1/y===1/g)||y!==y&&g!==g}var r=typeof Object.is=="function"?Object.is:t,a=e.useState,u=e.useEffect,o=e.useLayoutEffect,s=e.useDebugValue;function f(y,g){var b=g(),w=a({inst:{value:b,getSnapshot:g}}),S=w[0].inst,O=w[1];return o(function(){S.value=b,S.getSnapshot=g,v(S)&&O({inst:S})},[y,b,g]),u(function(){return v(S)&&O({inst:S}),y(function(){v(S)&&O({inst:S})})},[y]),s(b),b}function v(y){var g=y.getSnapshot;y=y.value;try{var b=g();return!r(y,b)}catch{return!0}}function h(y,g){return g()}var m=typeof window>"u"||typeof window.document>"u"||typeof window.document.createElement>"u"?h:f;return qh.useSyncExternalStore=e.useSyncExternalStore!==void 0?e.useSyncExternalStore:m,qh}var Yx;function PC(){return Yx||(Yx=1,Uh.exports=NC()),Uh.exports}/**
 * @license React
 * use-sync-external-store-shim/with-selector.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Gx;function RC(){if(Gx)return Lh;Gx=1;var e=Ii(),t=PC();function r(h,m){return h===m&&(h!==0||1/h===1/m)||h!==h&&m!==m}var a=typeof Object.is=="function"?Object.is:r,u=t.useSyncExternalStore,o=e.useRef,s=e.useEffect,f=e.useMemo,v=e.useDebugValue;return Lh.useSyncExternalStoreWithSelector=function(h,m,y,g,b){var w=o(null);if(w.current===null){var S={hasValue:!1,value:null};w.current=S}else S=w.current;w=f(function(){function T(P){if(!M){if(M=!0,j=P,P=g(P),b!==void 0&&S.hasValue){var L=S.value;if(b(L,P))return k=L}return k=P}if(L=k,a(j,P))return L;var G=g(P);return b!==void 0&&b(L,G)?(j=P,L):(j=P,k=G)}var M=!1,j,k,N=y===void 0?null:y;return[function(){return T(m())},N===null?void 0:function(){return T(N())}]},[m,y,g,b]);var O=u(h,w[0],w[1]);return s(function(){S.hasValue=!0,S.value=O},[O]),v(O),O},Lh}var Kx;function kC(){return Kx||(Kx=1,Bh.exports=RC()),Bh.exports}var zC=kC(),Mm=A.createContext(null),BC=e=>e,bt=()=>{var e=A.useContext(Mm);return e?e.store.dispatch:BC},yc=()=>{},LC=()=>yc,UC=(e,t)=>e===t;function me(e){var t=A.useContext(Mm);return zC.useSyncExternalStoreWithSelector(t?t.subscription.addNestedSub:LC,t?t.store.getState:yc,t?t.store.getState:yc,t?e:yc,UC)}function qC(e,t=`expected a function, instead received ${typeof e}`){if(typeof e!="function")throw new TypeError(t)}function $C(e,t=`expected an object, instead received ${typeof e}`){if(typeof e!="object")throw new TypeError(t)}function HC(e,t="expected all items to be functions, instead received the following types: "){if(!e.every(r=>typeof r=="function")){const r=e.map(a=>typeof a=="function"?`function ${a.name||"unnamed"}()`:typeof a).join(", ");throw new TypeError(`${t}[${r}]`)}}var Ix=e=>Array.isArray(e)?e:[e];function YC(e){const t=Array.isArray(e[0])?e[0]:e;return HC(t,"createSelector expects all input-selectors to be functions, but received the following types: "),t}function GC(e,t){const r=[],{length:a}=e;for(let u=0;u<a;u++)r.push(e[u].apply(null,t));return r}var KC=class{constructor(e){this.value=e}deref(){return this.value}},IC=typeof WeakRef<"u"?WeakRef:KC,XC=0,Xx=1;function nc(){return{s:XC,v:void 0,o:null,p:null}}function OO(e,t={}){let r=nc();const{resultEqualityCheck:a}=t;let u,o=0;function s(){let f=r;const{length:v}=arguments;for(let y=0,g=v;y<g;y++){const b=arguments[y];if(typeof b=="function"||typeof b=="object"&&b!==null){let w=f.o;w===null&&(f.o=w=new WeakMap);const S=w.get(b);S===void 0?(f=nc(),w.set(b,f)):f=S}else{let w=f.p;w===null&&(f.p=w=new Map);const S=w.get(b);S===void 0?(f=nc(),w.set(b,f)):f=S}}const h=f;let m;if(f.s===Xx)m=f.v;else if(m=e.apply(null,arguments),o++,a){const y=u?.deref?.()??u;y!=null&&a(y,m)&&(m=y,o!==0&&o--),u=typeof m=="object"&&m!==null||typeof m=="function"?new IC(m):m}return h.s=Xx,h.v=m,m}return s.clearCache=()=>{r=nc(),s.resetResultsCount()},s.resultsCount=()=>o,s.resetResultsCount=()=>{o=0},s}function VC(e,...t){const r=typeof e=="function"?{memoize:e,memoizeOptions:t}:e,a=(...u)=>{let o=0,s=0,f,v={},h=u.pop();typeof h=="object"&&(v=h,h=u.pop()),qC(h,`createSelector expects an output function after the inputs, but received: [${typeof h}]`);const m={...r,...v},{memoize:y,memoizeOptions:g=[],argsMemoize:b=OO,argsMemoizeOptions:w=[]}=m,S=Ix(g),O=Ix(w),T=YC(u),M=y(function(){return o++,h.apply(null,arguments)},...S),j=b(function(){s++;const N=GC(T,arguments);return f=M.apply(null,N),f},...O);return Object.assign(j,{resultFunc:h,memoizedResultFunc:M,dependencies:T,dependencyRecomputations:()=>s,resetDependencyRecomputations:()=>{s=0},lastResult:()=>f,recomputations:()=>o,resetRecomputations:()=>{o=0},memoize:y,argsMemoize:b})};return Object.assign(a,{withTypes:()=>a}),a}var Y=VC(OO),ZC=Object.assign((e,t=Y)=>{$C(e,`createStructuredSelector expects first argument to be an object where each property is a selector, instead received a ${typeof e}`);const r=Object.keys(e),a=r.map(o=>e[o]);return t(a,(...o)=>o.reduce((s,f,v)=>(s[r[v]]=f,s),{}))},{withTypes:()=>ZC}),$h={},Hh={},Yh={},Vx;function QC(){return Vx||(Vx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(a){return typeof a=="symbol"?1:a===null?2:a===void 0?3:a!==a?4:0}const r=(a,u,o)=>{if(a!==u){const s=t(a),f=t(u);if(s===f&&s===0){if(a<u)return o==="desc"?1:-1;if(a>u)return o==="desc"?-1:1}return o==="desc"?f-s:s-f}return 0};e.compareValues=r})(Yh)),Yh}var Gh={},Kh={},Zx;function AO(){return Zx||(Zx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return typeof r=="symbol"||r instanceof Symbol}e.isSymbol=t})(Kh)),Kh}var Qx;function WC(){return Qx||(Qx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=AO(),r=/\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,a=/^\w*$/;function u(o,s){return Array.isArray(o)?!1:typeof o=="number"||typeof o=="boolean"||o==null||t.isSymbol(o)?!0:typeof o=="string"&&(a.test(o)||!r.test(o))||s!=null&&Object.hasOwn(s,o)}e.isKey=u})(Gh)),Gh}var Wx;function FC(){return Wx||(Wx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=QC(),r=WC(),a=xm();function u(o,s,f,v){if(o==null)return[];f=v?void 0:f,Array.isArray(o)||(o=Object.values(o)),Array.isArray(s)||(s=s==null?[null]:[s]),s.length===0&&(s=[null]),Array.isArray(f)||(f=f==null?[]:[f]),f=f.map(b=>String(b));const h=(b,w)=>{let S=b;for(let O=0;O<w.length&&S!=null;++O)S=S[w[O]];return S},m=(b,w)=>w==null||b==null?w:typeof b=="object"&&"key"in b?Object.hasOwn(w,b.key)?w[b.key]:h(w,b.path):typeof b=="function"?b(w):Array.isArray(b)?h(w,b):typeof w=="object"?w[b]:w,y=s.map(b=>(Array.isArray(b)&&b.length===1&&(b=b[0]),b==null||typeof b=="function"||Array.isArray(b)||r.isKey(b)?b:{key:b,path:a.toPath(b)}));return o.map(b=>({original:b,criteria:y.map(w=>m(w,b))})).slice().sort((b,w)=>{for(let S=0;S<y.length;S++){const O=t.compareValues(b.criteria[S],w.criteria[S],f[S]);if(O!==0)return O}return 0}).map(b=>b.original)}e.orderBy=u})(Hh)),Hh}var Ih={},Fx;function JC(){return Fx||(Fx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r,a=1){const u=[],o=Math.floor(a),s=(f,v)=>{for(let h=0;h<f.length;h++){const m=f[h];Array.isArray(m)&&v<o?s(m,v+1):u.push(m)}};return s(r,0),u}e.flatten=t})(Ih)),Ih}var Xh={},Jx;function EO(){return Jx||(Jx=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=wO(),r=Tm(),a=pO(),u=gO();function o(s,f,v){return a.isObject(v)&&(typeof f=="number"&&r.isArrayLike(v)&&t.isIndex(f)&&f<v.length||typeof f=="string"&&f in v)?u.eq(v[f],s):!1}e.isIterateeCall=o})(Xh)),Xh}var e1;function eD(){return e1||(e1=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=FC(),r=JC(),a=EO();function u(o,...s){const f=s.length;return f>1&&a.isIterateeCall(o,s[0],s[1])?s=[]:f>2&&a.isIterateeCall(s[0],s[1],s[2])&&(s=[s[0]]),t.orderBy(o,r.flatten(s),["asc"])}e.sortBy=u})($h)),$h}var Vh,t1;function tD(){return t1||(t1=1,Vh=eD().sortBy),Vh}var nD=tD();const us=Wn(nD);var _O=e=>e.legend.settings,rD=e=>e.legend.size,aD=e=>e.legend.payload;Y([aD,_O],(e,t)=>{var{itemSorter:r}=t,a=e.flat(1);return r?us(a,r):a});var rc=1;function iD(){var e=arguments.length>0&&arguments[0]!==void 0?arguments[0]:[],[t,r]=A.useState({height:0,left:0,top:0,width:0}),a=A.useCallback(u=>{if(u!=null){var o=u.getBoundingClientRect(),s={height:o.height,left:o.left,top:o.top,width:o.width};(Math.abs(s.height-t.height)>rc||Math.abs(s.left-t.left)>rc||Math.abs(s.top-t.top)>rc||Math.abs(s.width-t.width)>rc)&&r({height:s.height,left:s.left,top:s.top,width:s.width})}},[t.width,t.height,t.top,t.left,...e]);return[t,a]}function _t(e){return`Minified Redux error #${e}; visit https://redux.js.org/Errors?code=${e} for the full message or use the non-minified dev environment for full errors. `}var lD=typeof Symbol=="function"&&Symbol.observable||"@@observable",n1=lD,Zh=()=>Math.random().toString(36).substring(7).split("").join("."),uD={INIT:`@@redux/INIT${Zh()}`,REPLACE:`@@redux/REPLACE${Zh()}`,PROBE_UNKNOWN_ACTION:()=>`@@redux/PROBE_UNKNOWN_ACTION${Zh()}`},Ac=uD;function Cm(e){if(typeof e!="object"||e===null)return!1;let t=e;for(;Object.getPrototypeOf(t)!==null;)t=Object.getPrototypeOf(t);return Object.getPrototypeOf(e)===t||Object.getPrototypeOf(e)===null}function TO(e,t,r){if(typeof e!="function")throw new Error(_t(2));if(typeof t=="function"&&typeof r=="function"||typeof r=="function"&&typeof arguments[3]=="function")throw new Error(_t(0));if(typeof t=="function"&&typeof r>"u"&&(r=t,t=void 0),typeof r<"u"){if(typeof r!="function")throw new Error(_t(1));return r(TO)(e,t)}let a=e,u=t,o=new Map,s=o,f=0,v=!1;function h(){s===o&&(s=new Map,o.forEach((O,T)=>{s.set(T,O)}))}function m(){if(v)throw new Error(_t(3));return u}function y(O){if(typeof O!="function")throw new Error(_t(4));if(v)throw new Error(_t(5));let T=!0;h();const M=f++;return s.set(M,O),function(){if(T){if(v)throw new Error(_t(6));T=!1,h(),s.delete(M),o=null}}}function g(O){if(!Cm(O))throw new Error(_t(7));if(typeof O.type>"u")throw new Error(_t(8));if(typeof O.type!="string")throw new Error(_t(17));if(v)throw new Error(_t(9));try{v=!0,u=a(u,O)}finally{v=!1}return(o=s).forEach(M=>{M()}),O}function b(O){if(typeof O!="function")throw new Error(_t(10));a=O,g({type:Ac.REPLACE})}function w(){const O=y;return{subscribe(T){if(typeof T!="object"||T===null)throw new Error(_t(11));function M(){const k=T;k.next&&k.next(m())}return M(),{unsubscribe:O(M)}},[n1](){return this}}}return g({type:Ac.INIT}),{dispatch:g,subscribe:y,getState:m,replaceReducer:b,[n1]:w}}function oD(e){Object.keys(e).forEach(t=>{const r=e[t];if(typeof r(void 0,{type:Ac.INIT})>"u")throw new Error(_t(12));if(typeof r(void 0,{type:Ac.PROBE_UNKNOWN_ACTION()})>"u")throw new Error(_t(13))})}function jO(e){const t=Object.keys(e),r={};for(let o=0;o<t.length;o++){const s=t[o];typeof e[s]=="function"&&(r[s]=e[s])}const a=Object.keys(r);let u;try{oD(r)}catch(o){u=o}return function(s={},f){if(u)throw u;let v=!1;const h={};for(let m=0;m<a.length;m++){const y=a[m],g=r[y],b=s[y],w=g(b,f);if(typeof w>"u")throw f&&f.type,new Error(_t(14));h[y]=w,v=v||w!==b}return v=v||a.length!==Object.keys(s).length,v?h:s}}function Ec(...e){return e.length===0?t=>t:e.length===1?e[0]:e.reduce((t,r)=>(...a)=>t(r(...a)))}function cD(...e){return t=>(r,a)=>{const u=t(r,a);let o=()=>{throw new Error(_t(15))};const s={getState:u.getState,dispatch:(v,...h)=>o(v,...h)},f=e.map(v=>v(s));return o=Ec(...f)(u.dispatch),{...u,dispatch:o}}}function MO(e){return Cm(e)&&"type"in e&&typeof e.type=="string"}var CO=Symbol.for("immer-nothing"),r1=Symbol.for("immer-draftable"),hn=Symbol.for("immer-state");function Ln(e,...t){throw new Error(`[Immer] minified error nr: ${e}. Full error at: https://bit.ly/3cXEKWf`)}var Bi=Object.getPrototypeOf;function La(e){return!!e&&!!e[hn]}function br(e){return e?DO(e)||Array.isArray(e)||!!e[r1]||!!e.constructor?.[r1]||Eu(e)||cs(e):!1}var sD=Object.prototype.constructor.toString();function DO(e){if(!e||typeof e!="object")return!1;const t=Bi(e);if(t===null)return!0;const r=Object.hasOwnProperty.call(t,"constructor")&&t.constructor;return r===Object?!0:typeof r=="function"&&Function.toString.call(r)===sD}function _c(e,t){os(e)===0?Reflect.ownKeys(e).forEach(r=>{t(r,e[r],e)}):e.forEach((r,a)=>t(a,r,e))}function os(e){const t=e[hn];return t?t.type_:Array.isArray(e)?1:Eu(e)?2:cs(e)?3:0}function Lv(e,t){return os(e)===2?e.has(t):Object.prototype.hasOwnProperty.call(e,t)}function NO(e,t,r){const a=os(e);a===2?e.set(t,r):a===3?e.add(r):e[t]=r}function fD(e,t){return e===t?e!==0||1/e===1/t:e!==e&&t!==t}function Eu(e){return e instanceof Map}function cs(e){return e instanceof Set}function ja(e){return e.copy_||e.base_}function Uv(e,t){if(Eu(e))return new Map(e);if(cs(e))return new Set(e);if(Array.isArray(e))return Array.prototype.slice.call(e);const r=DO(e);if(t===!0||t==="class_only"&&!r){const a=Object.getOwnPropertyDescriptors(e);delete a[hn];let u=Reflect.ownKeys(a);for(let o=0;o<u.length;o++){const s=u[o],f=a[s];f.writable===!1&&(f.writable=!0,f.configurable=!0),(f.get||f.set)&&(a[s]={configurable:!0,writable:!0,enumerable:f.enumerable,value:e[s]})}return Object.create(Bi(e),a)}else{const a=Bi(e);if(a!==null&&r)return{...e};const u=Object.create(a);return Object.assign(u,e)}}function Dm(e,t=!1){return ss(e)||La(e)||!br(e)||(os(e)>1&&Object.defineProperties(e,{set:{value:ac},add:{value:ac},clear:{value:ac},delete:{value:ac}}),Object.freeze(e),t&&Object.values(e).forEach(r=>Dm(r,!0))),e}function ac(){Ln(2)}function ss(e){return Object.isFrozen(e)}var dD={};function Ua(e){const t=dD[e];return t||Ln(0,e),t}var mu;function PO(){return mu}function hD(e,t){return{drafts_:[],parent_:e,immer_:t,canAutoFreeze_:!0,unfinalizedDrafts_:0}}function a1(e,t){t&&(Ua("Patches"),e.patches_=[],e.inversePatches_=[],e.patchListener_=t)}function qv(e){$v(e),e.drafts_.forEach(vD),e.drafts_=null}function $v(e){e===mu&&(mu=e.parent_)}function i1(e){return mu=hD(mu,e)}function vD(e){const t=e[hn];t.type_===0||t.type_===1?t.revoke_():t.revoked_=!0}function l1(e,t){t.unfinalizedDrafts_=t.drafts_.length;const r=t.drafts_[0];return e!==void 0&&e!==r?(r[hn].modified_&&(qv(t),Ln(4)),br(e)&&(e=Tc(t,e),t.parent_||jc(t,e)),t.patches_&&Ua("Patches").generateReplacementPatches_(r[hn].base_,e,t.patches_,t.inversePatches_)):e=Tc(t,r,[]),qv(t),t.patches_&&t.patchListener_(t.patches_,t.inversePatches_),e!==CO?e:void 0}function Tc(e,t,r){if(ss(t))return t;const a=t[hn];if(!a)return _c(t,(u,o)=>u1(e,a,t,u,o,r)),t;if(a.scope_!==e)return t;if(!a.modified_)return jc(e,a.base_,!0),a.base_;if(!a.finalized_){a.finalized_=!0,a.scope_.unfinalizedDrafts_--;const u=a.copy_;let o=u,s=!1;a.type_===3&&(o=new Set(u),u.clear(),s=!0),_c(o,(f,v)=>u1(e,a,u,f,v,r,s)),jc(e,u,!1),r&&e.patches_&&Ua("Patches").generatePatches_(a,r,e.patches_,e.inversePatches_)}return a.copy_}function u1(e,t,r,a,u,o,s){if(La(u)){const f=o&&t&&t.type_!==3&&!Lv(t.assigned_,a)?o.concat(a):void 0,v=Tc(e,u,f);if(NO(r,a,v),La(v))e.canAutoFreeze_=!1;else return}else s&&r.add(u);if(br(u)&&!ss(u)){if(!e.immer_.autoFreeze_&&e.unfinalizedDrafts_<1)return;Tc(e,u),(!t||!t.scope_.parent_)&&typeof a!="symbol"&&(Eu(r)?r.has(a):Object.prototype.propertyIsEnumerable.call(r,a))&&jc(e,u)}}function jc(e,t,r=!1){!e.parent_&&e.immer_.autoFreeze_&&e.canAutoFreeze_&&Dm(t,r)}function mD(e,t){const r=Array.isArray(e),a={type_:r?1:0,scope_:t?t.scope_:PO(),modified_:!1,finalized_:!1,assigned_:{},parent_:t,base_:e,draft_:null,copy_:null,revoke_:null,isManual_:!1};let u=a,o=Nm;r&&(u=[a],o=pu);const{revoke:s,proxy:f}=Proxy.revocable(u,o);return a.draft_=f,a.revoke_=s,f}var Nm={get(e,t){if(t===hn)return e;const r=ja(e);if(!Lv(r,t))return pD(e,r,t);const a=r[t];return e.finalized_||!br(a)?a:a===Qh(e.base_,t)?(Wh(e),e.copy_[t]=Yv(a,e)):a},has(e,t){return t in ja(e)},ownKeys(e){return Reflect.ownKeys(ja(e))},set(e,t,r){const a=RO(ja(e),t);if(a?.set)return a.set.call(e.draft_,r),!0;if(!e.modified_){const u=Qh(ja(e),t),o=u?.[hn];if(o&&o.base_===r)return e.copy_[t]=r,e.assigned_[t]=!1,!0;if(fD(r,u)&&(r!==void 0||Lv(e.base_,t)))return!0;Wh(e),Hv(e)}return e.copy_[t]===r&&(r!==void 0||t in e.copy_)||Number.isNaN(r)&&Number.isNaN(e.copy_[t])||(e.copy_[t]=r,e.assigned_[t]=!0),!0},deleteProperty(e,t){return Qh(e.base_,t)!==void 0||t in e.base_?(e.assigned_[t]=!1,Wh(e),Hv(e)):delete e.assigned_[t],e.copy_&&delete e.copy_[t],!0},getOwnPropertyDescriptor(e,t){const r=ja(e),a=Reflect.getOwnPropertyDescriptor(r,t);return a&&{writable:!0,configurable:e.type_!==1||t!=="length",enumerable:a.enumerable,value:r[t]}},defineProperty(){Ln(11)},getPrototypeOf(e){return Bi(e.base_)},setPrototypeOf(){Ln(12)}},pu={};_c(Nm,(e,t)=>{pu[e]=function(){return arguments[0]=arguments[0][0],t.apply(this,arguments)}});pu.deleteProperty=function(e,t){return pu.set.call(this,e,t,void 0)};pu.set=function(e,t,r){return Nm.set.call(this,e[0],t,r,e[0])};function Qh(e,t){const r=e[hn];return(r?ja(r):e)[t]}function pD(e,t,r){const a=RO(t,r);return a?"value"in a?a.value:a.get?.call(e.draft_):void 0}function RO(e,t){if(!(t in e))return;let r=Bi(e);for(;r;){const a=Object.getOwnPropertyDescriptor(r,t);if(a)return a;r=Bi(r)}}function Hv(e){e.modified_||(e.modified_=!0,e.parent_&&Hv(e.parent_))}function Wh(e){e.copy_||(e.copy_=Uv(e.base_,e.scope_.immer_.useStrictShallowCopy_))}var yD=class{constructor(e){this.autoFreeze_=!0,this.useStrictShallowCopy_=!1,this.produce=(t,r,a)=>{if(typeof t=="function"&&typeof r!="function"){const o=r;r=t;const s=this;return function(v=o,...h){return s.produce(v,m=>r.call(this,m,...h))}}typeof r!="function"&&Ln(6),a!==void 0&&typeof a!="function"&&Ln(7);let u;if(br(t)){const o=i1(this),s=Yv(t,void 0);let f=!0;try{u=r(s),f=!1}finally{f?qv(o):$v(o)}return a1(o,a),l1(u,o)}else if(!t||typeof t!="object"){if(u=r(t),u===void 0&&(u=t),u===CO&&(u=void 0),this.autoFreeze_&&Dm(u,!0),a){const o=[],s=[];Ua("Patches").generateReplacementPatches_(t,u,o,s),a(o,s)}return u}else Ln(1,t)},this.produceWithPatches=(t,r)=>{if(typeof t=="function")return(s,...f)=>this.produceWithPatches(s,v=>t(v,...f));let a,u;return[this.produce(t,r,(s,f)=>{a=s,u=f}),a,u]},typeof e?.autoFreeze=="boolean"&&this.setAutoFreeze(e.autoFreeze),typeof e?.useStrictShallowCopy=="boolean"&&this.setUseStrictShallowCopy(e.useStrictShallowCopy)}createDraft(e){br(e)||Ln(8),La(e)&&(e=yr(e));const t=i1(this),r=Yv(e,void 0);return r[hn].isManual_=!0,$v(t),r}finishDraft(e,t){const r=e&&e[hn];(!r||!r.isManual_)&&Ln(9);const{scope_:a}=r;return a1(a,t),l1(void 0,a)}setAutoFreeze(e){this.autoFreeze_=e}setUseStrictShallowCopy(e){this.useStrictShallowCopy_=e}applyPatches(e,t){let r;for(r=t.length-1;r>=0;r--){const u=t[r];if(u.path.length===0&&u.op==="replace"){e=u.value;break}}r>-1&&(t=t.slice(r+1));const a=Ua("Patches").applyPatches_;return La(e)?a(e,t):this.produce(e,u=>a(u,t))}};function Yv(e,t){const r=Eu(e)?Ua("MapSet").proxyMap_(e,t):cs(e)?Ua("MapSet").proxySet_(e,t):mD(e,t);return(t?t.scope_:PO()).drafts_.push(r),r}function yr(e){return La(e)||Ln(10,e),kO(e)}function kO(e){if(!br(e)||ss(e))return e;const t=e[hn];let r;if(t){if(!t.modified_)return t.base_;t.finalized_=!0,r=Uv(e,t.scope_.immer_.useStrictShallowCopy_)}else r=Uv(e,!0);return _c(r,(a,u)=>{NO(r,a,kO(u))}),t&&(t.finalized_=!1),r}var gD=new yD,zO=gD.produce;function BO(e){return({dispatch:r,getState:a})=>u=>o=>typeof o=="function"?o(r,a,e):u(o)}var bD=BO(),xD=BO,SD=typeof window<"u"&&window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__?window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__:function(){if(arguments.length!==0)return typeof arguments[0]=="object"?Ec:Ec.apply(null,arguments)};function Cn(e,t){function r(...a){if(t){let u=t(...a);if(!u)throw new Error(fn(0));return{type:e,payload:u.payload,..."meta"in u&&{meta:u.meta},..."error"in u&&{error:u.error}}}return{type:e,payload:a[0]}}return r.toString=()=>`${e}`,r.type=e,r.match=a=>MO(a)&&a.type===e,r}var LO=class su extends Array{constructor(...t){super(...t),Object.setPrototypeOf(this,su.prototype)}static get[Symbol.species](){return su}concat(...t){return super.concat.apply(this,t)}prepend(...t){return t.length===1&&Array.isArray(t[0])?new su(...t[0].concat(this)):new su(...t.concat(this))}};function o1(e){return br(e)?zO(e,()=>{}):e}function ic(e,t,r){return e.has(t)?e.get(t):e.set(t,r(t)).get(t)}function wD(e){return typeof e=="boolean"}var OD=()=>function(t){const{thunk:r=!0,immutableCheck:a=!0,serializableCheck:u=!0,actionCreatorCheck:o=!0}=t??{};let s=new LO;return r&&(wD(r)?s.push(bD):s.push(xD(r.extraArgument))),s},AD="RTK_autoBatch",c1=e=>t=>{setTimeout(t,e)},ED=(e={type:"raf"})=>t=>(...r)=>{const a=t(...r);let u=!0,o=!1,s=!1;const f=new Set,v=e.type==="tick"?queueMicrotask:e.type==="raf"?typeof window<"u"&&window.requestAnimationFrame?window.requestAnimationFrame:c1(10):e.type==="callback"?e.queueNotification:c1(e.timeout),h=()=>{s=!1,o&&(o=!1,f.forEach(m=>m()))};return Object.assign({},a,{subscribe(m){const y=()=>u&&m(),g=a.subscribe(y);return f.add(m),()=>{g(),f.delete(m)}},dispatch(m){try{return u=!m?.meta?.[AD],o=!u,o&&(s||(s=!0,v(h))),a.dispatch(m)}finally{u=!0}}})},_D=e=>function(r){const{autoBatch:a=!0}=r??{};let u=new LO(e);return a&&u.push(ED(typeof a=="object"?a:void 0)),u};function TD(e){const t=OD(),{reducer:r=void 0,middleware:a,devTools:u=!0,preloadedState:o=void 0,enhancers:s=void 0}=e||{};let f;if(typeof r=="function")f=r;else if(Cm(r))f=jO(r);else throw new Error(fn(1));let v;typeof a=="function"?v=a(t):v=t();let h=Ec;u&&(h=SD({trace:!1,...typeof u=="object"&&u}));const m=cD(...v),y=_D(m);let g=typeof s=="function"?s(y):y();const b=h(...g);return TO(f,o,b)}function UO(e){const t={},r=[];let a;const u={addCase(o,s){const f=typeof o=="string"?o:o.type;if(!f)throw new Error(fn(28));if(f in t)throw new Error(fn(29));return t[f]=s,u},addAsyncThunk(o,s){return s.pending&&(t[o.pending.type]=s.pending),s.rejected&&(t[o.rejected.type]=s.rejected),s.fulfilled&&(t[o.fulfilled.type]=s.fulfilled),s.settled&&r.push({matcher:o.settled,reducer:s.settled}),u},addMatcher(o,s){return r.push({matcher:o,reducer:s}),u},addDefaultCase(o){return a=o,u}};return e(u),[t,r,a]}function jD(e){return typeof e=="function"}function MD(e,t){let[r,a,u]=UO(t),o;if(jD(e))o=()=>o1(e());else{const f=o1(e);o=()=>f}function s(f=o(),v){let h=[r[v.type],...a.filter(({matcher:m})=>m(v)).map(({reducer:m})=>m)];return h.filter(m=>!!m).length===0&&(h=[u]),h.reduce((m,y)=>{if(y)if(La(m)){const b=y(m,v);return b===void 0?m:b}else{if(br(m))return zO(m,g=>y(g,v));{const g=y(m,v);if(g===void 0){if(m===null)return m;throw Error("A case reducer on a non-draftable value must not return undefined")}return g}}return m},f)}return s.getInitialState=o,s}var CD="ModuleSymbhasOwnPr-0123456789ABCDEFGHNRVfgctiUvz_KqYTJkLxpZXIjQW",DD=(e=21)=>{let t="",r=e;for(;r--;)t+=CD[Math.random()*64|0];return t},ND=Symbol.for("rtk-slice-createasyncthunk");function PD(e,t){return`${e}/${t}`}function RD({creators:e}={}){const t=e?.asyncThunk?.[ND];return function(a){const{name:u,reducerPath:o=u}=a;if(!u)throw new Error(fn(11));const s=(typeof a.reducers=="function"?a.reducers(zD()):a.reducers)||{},f=Object.keys(s),v={sliceCaseReducersByName:{},sliceCaseReducersByType:{},actionCreators:{},sliceMatchers:[]},h={addCase(j,k){const N=typeof j=="string"?j:j.type;if(!N)throw new Error(fn(12));if(N in v.sliceCaseReducersByType)throw new Error(fn(13));return v.sliceCaseReducersByType[N]=k,h},addMatcher(j,k){return v.sliceMatchers.push({matcher:j,reducer:k}),h},exposeAction(j,k){return v.actionCreators[j]=k,h},exposeCaseReducer(j,k){return v.sliceCaseReducersByName[j]=k,h}};f.forEach(j=>{const k=s[j],N={reducerName:j,type:PD(u,j),createNotation:typeof a.reducers=="function"};LD(k)?qD(N,k,h,t):BD(N,k,h)});function m(){const[j={},k=[],N=void 0]=typeof a.extraReducers=="function"?UO(a.extraReducers):[a.extraReducers],P={...j,...v.sliceCaseReducersByType};return MD(a.initialState,L=>{for(let G in P)L.addCase(G,P[G]);for(let G of v.sliceMatchers)L.addMatcher(G.matcher,G.reducer);for(let G of k)L.addMatcher(G.matcher,G.reducer);N&&L.addDefaultCase(N)})}const y=j=>j,g=new Map,b=new WeakMap;let w;function S(j,k){return w||(w=m()),w(j,k)}function O(){return w||(w=m()),w.getInitialState()}function T(j,k=!1){function N(L){let G=L[j];return typeof G>"u"&&k&&(G=ic(b,N,O)),G}function P(L=y){const G=ic(g,k,()=>new WeakMap);return ic(G,L,()=>{const ee={};for(const[ne,J]of Object.entries(a.selectors??{}))ee[ne]=kD(J,L,()=>ic(b,L,O),k);return ee})}return{reducerPath:j,getSelectors:P,get selectors(){return P(N)},selectSlice:N}}const M={name:u,reducer:S,actions:v.actionCreators,caseReducers:v.sliceCaseReducersByName,getInitialState:O,...T(o),injectInto(j,{reducerPath:k,...N}={}){const P=k??o;return j.inject({reducerPath:P,reducer:S},N),{...M,...T(P,!0)}}};return M}}function kD(e,t,r,a){function u(o,...s){let f=t(o);return typeof f>"u"&&a&&(f=r()),e(f,...s)}return u.unwrapped=e,u}var mn=RD();function zD(){function e(t,r){return{_reducerDefinitionType:"asyncThunk",payloadCreator:t,...r}}return e.withTypes=()=>e,{reducer(t){return Object.assign({[t.name](...r){return t(...r)}}[t.name],{_reducerDefinitionType:"reducer"})},preparedReducer(t,r){return{_reducerDefinitionType:"reducerWithPrepare",prepare:t,reducer:r}},asyncThunk:e}}function BD({type:e,reducerName:t,createNotation:r},a,u){let o,s;if("reducer"in a){if(r&&!UD(a))throw new Error(fn(17));o=a.reducer,s=a.prepare}else o=a;u.addCase(e,o).exposeCaseReducer(t,o).exposeAction(t,s?Cn(e,s):Cn(e))}function LD(e){return e._reducerDefinitionType==="asyncThunk"}function UD(e){return e._reducerDefinitionType==="reducerWithPrepare"}function qD({type:e,reducerName:t},r,a,u){if(!u)throw new Error(fn(18));const{payloadCreator:o,fulfilled:s,pending:f,rejected:v,settled:h,options:m}=r,y=u(e,o,m);a.exposeAction(t,y),s&&a.addCase(y.fulfilled,s),f&&a.addCase(y.pending,f),v&&a.addCase(y.rejected,v),h&&a.addMatcher(y.settled,h),a.exposeCaseReducer(t,{fulfilled:s||lc,pending:f||lc,rejected:v||lc,settled:h||lc})}function lc(){}var $D="task",qO="listener",$O="completed",Pm="cancelled",HD=`task-${Pm}`,YD=`task-${$O}`,Gv=`${qO}-${Pm}`,GD=`${qO}-${$O}`,fs=class{constructor(e){this.code=e,this.message=`${$D} ${Pm} (reason: ${e})`}name="TaskAbortError";message},Rm=(e,t)=>{if(typeof e!="function")throw new TypeError(fn(32))},Mc=()=>{},HO=(e,t=Mc)=>(e.catch(t),e),YO=(e,t)=>(e.addEventListener("abort",t,{once:!0}),()=>e.removeEventListener("abort",t)),Ra=(e,t)=>{const r=e.signal;r.aborted||("reason"in r||Object.defineProperty(r,"reason",{enumerable:!0,value:t,configurable:!0,writable:!0}),e.abort(t))},ka=e=>{if(e.aborted){const{reason:t}=e;throw new fs(t)}};function GO(e,t){let r=Mc;return new Promise((a,u)=>{const o=()=>u(new fs(e.reason));if(e.aborted){o();return}r=YO(e,o),t.finally(()=>r()).then(a,u)}).finally(()=>{r=Mc})}var KD=async(e,t)=>{try{return await Promise.resolve(),{status:"ok",value:await e()}}catch(r){return{status:r instanceof fs?"cancelled":"rejected",error:r}}finally{t?.()}},Cc=e=>t=>HO(GO(e,t).then(r=>(ka(e),r))),KO=e=>{const t=Cc(e);return r=>t(new Promise(a=>setTimeout(a,r)))},{assign:Pi}=Object,s1={},ds="listenerMiddleware",ID=(e,t)=>{const r=a=>YO(e,()=>Ra(a,e.reason));return(a,u)=>{Rm(a);const o=new AbortController;r(o);const s=KD(async()=>{ka(e),ka(o.signal);const f=await a({pause:Cc(o.signal),delay:KO(o.signal),signal:o.signal});return ka(o.signal),f},()=>Ra(o,YD));return u?.autoJoin&&t.push(s.catch(Mc)),{result:Cc(e)(s),cancel(){Ra(o,HD)}}}},XD=(e,t)=>{const r=async(a,u)=>{ka(t);let o=()=>{};const f=[new Promise((v,h)=>{let m=e({predicate:a,effect:(y,g)=>{g.unsubscribe(),v([y,g.getState(),g.getOriginalState()])}});o=()=>{m(),h()}})];u!=null&&f.push(new Promise(v=>setTimeout(v,u,null)));try{const v=await GO(t,Promise.race(f));return ka(t),v}finally{o()}};return(a,u)=>HO(r(a,u))},IO=e=>{let{type:t,actionCreator:r,matcher:a,predicate:u,effect:o}=e;if(t)u=Cn(t).match;else if(r)t=r.type,u=r.match;else if(a)u=a;else if(!u)throw new Error(fn(21));return Rm(o),{predicate:u,type:t,effect:o}},XO=Pi(e=>{const{type:t,predicate:r,effect:a}=IO(e);return{id:DD(),effect:a,type:t,predicate:r,pending:new Set,unsubscribe:()=>{throw new Error(fn(22))}}},{withTypes:()=>XO}),f1=(e,t)=>{const{type:r,effect:a,predicate:u}=IO(t);return Array.from(e.values()).find(o=>(typeof r=="string"?o.type===r:o.predicate===u)&&o.effect===a)},Kv=e=>{e.pending.forEach(t=>{Ra(t,Gv)})},VD=e=>()=>{e.forEach(Kv),e.clear()},d1=(e,t,r)=>{try{e(t,r)}catch(a){setTimeout(()=>{throw a},0)}},VO=Pi(Cn(`${ds}/add`),{withTypes:()=>VO}),ZD=Cn(`${ds}/removeAll`),ZO=Pi(Cn(`${ds}/remove`),{withTypes:()=>ZO}),QD=(...e)=>{console.error(`${ds}/error`,...e)},_u=(e={})=>{const t=new Map,{extra:r,onError:a=QD}=e;Rm(a);const u=m=>(m.unsubscribe=()=>t.delete(m.id),t.set(m.id,m),y=>{m.unsubscribe(),y?.cancelActive&&Kv(m)}),o=m=>{const y=f1(t,m)??XO(m);return u(y)};Pi(o,{withTypes:()=>o});const s=m=>{const y=f1(t,m);return y&&(y.unsubscribe(),m.cancelActive&&Kv(y)),!!y};Pi(s,{withTypes:()=>s});const f=async(m,y,g,b)=>{const w=new AbortController,S=XD(o,w.signal),O=[];try{m.pending.add(w),await Promise.resolve(m.effect(y,Pi({},g,{getOriginalState:b,condition:(T,M)=>S(T,M).then(Boolean),take:S,delay:KO(w.signal),pause:Cc(w.signal),extra:r,signal:w.signal,fork:ID(w.signal,O),unsubscribe:m.unsubscribe,subscribe:()=>{t.set(m.id,m)},cancelActiveListeners:()=>{m.pending.forEach((T,M,j)=>{T!==w&&(Ra(T,Gv),j.delete(T))})},cancel:()=>{Ra(w,Gv),m.pending.delete(w)},throwIfCancelled:()=>{ka(w.signal)}})))}catch(T){T instanceof fs||d1(a,T,{raisedBy:"effect"})}finally{await Promise.all(O),Ra(w,GD),m.pending.delete(w)}},v=VD(t);return{middleware:m=>y=>g=>{if(!MO(g))return y(g);if(VO.match(g))return o(g.payload);if(ZD.match(g)){v();return}if(ZO.match(g))return s(g.payload);let b=m.getState();const w=()=>{if(b===s1)throw new Error(fn(23));return b};let S;try{if(S=y(g),t.size>0){const O=m.getState(),T=Array.from(t.values());for(const M of T){let j=!1;try{j=M.predicate(g,O,b)}catch(k){j=!1,d1(a,k,{raisedBy:"predicate"})}j&&f(M,g,m,w)}}}finally{b=s1}return S},startListening:o,stopListening:s,clearListeners:v}};function fn(e){return`Minified Redux Toolkit error #${e}; visit https://redux-toolkit.js.org/Errors?code=${e} for the full message or use the non-minified dev environment for full errors. `}var WD={layoutType:"horizontal",width:0,height:0,margin:{top:5,right:5,bottom:5,left:5},scale:1},QO=mn({name:"chartLayout",initialState:WD,reducers:{setLayout(e,t){e.layoutType=t.payload},setChartSize(e,t){e.width=t.payload.width,e.height=t.payload.height},setMargin(e,t){var r,a,u,o;e.margin.top=(r=t.payload.top)!==null&&r!==void 0?r:0,e.margin.right=(a=t.payload.right)!==null&&a!==void 0?a:0,e.margin.bottom=(u=t.payload.bottom)!==null&&u!==void 0?u:0,e.margin.left=(o=t.payload.left)!==null&&o!==void 0?o:0},setScale(e,t){e.scale=t.payload}}}),{setMargin:FD,setLayout:JD,setChartSize:eN,setScale:tN}=QO.actions,nN=QO.reducer;function h1(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function v1(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?h1(Object(r),!0).forEach(function(a){rN(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):h1(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function rN(e,t,r){return(t=aN(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function aN(e){var t=iN(e,"string");return typeof t=="symbol"?t:t+""}function iN(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var Dc=Math.PI/180,lN=e=>e*180/Math.PI,jt=(e,t,r,a)=>({x:e+Math.cos(-Dc*a)*r,y:t+Math.sin(-Dc*a)*r}),uN=function(t,r){var a=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{top:0,right:0,bottom:0,left:0};return Math.min(Math.abs(t-(a.left||0)-(a.right||0)),Math.abs(r-(a.top||0)-(a.bottom||0)))/2},oN=(e,t)=>{var{x:r,y:a}=e,{x:u,y:o}=t;return Math.sqrt((r-u)**2+(a-o)**2)},cN=(e,t)=>{var{x:r,y:a}=e,{cx:u,cy:o}=t,s=oN({x:r,y:a},{x:u,y:o});if(s<=0)return{radius:s,angle:0};var f=(r-u)/s,v=Math.acos(f);return a>o&&(v=2*Math.PI-v),{radius:s,angle:lN(v),angleInRadian:v}},sN=e=>{var{startAngle:t,endAngle:r}=e,a=Math.floor(t/360),u=Math.floor(r/360),o=Math.min(a,u);return{startAngle:t-o*360,endAngle:r-o*360}},fN=(e,t)=>{var{startAngle:r,endAngle:a}=t,u=Math.floor(r/360),o=Math.floor(a/360),s=Math.min(u,o);return e+s*360},dN=(e,t)=>{var{x:r,y:a}=e,{radius:u,angle:o}=cN({x:r,y:a},t),{innerRadius:s,outerRadius:f}=t;if(u<s||u>f||u===0)return null;var{startAngle:v,endAngle:h}=sN(t),m=o,y;if(v<=h){for(;m>h;)m-=360;for(;m<v;)m+=360;y=m>=v&&m<=h}else{for(;m>v;)m-=360;for(;m<h;)m+=360;y=m>=h&&m<=v}return y?v1(v1({},t),{},{radius:u,angle:fN(m,t)}):null};function WO(e,t,r){return Array.isArray(e)&&e&&t+r!==0?e.slice(t,r+1):e}function m1(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function jn(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?m1(Object(r),!0).forEach(function(a){hN(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):m1(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function hN(e,t,r){return(t=vN(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function vN(e){var t=mN(e,"string");return typeof t=="symbol"?t:t+""}function mN(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function $t(e,t,r){return gt(e)||gt(t)?r:Zn(t)?ki(e,t,r):typeof t=="function"?t(e):r}var pN=(e,t,r,a,u)=>{var o,s=-1,f=(o=t?.length)!==null&&o!==void 0?o:0;if(f<=1||e==null)return 0;if(a==="angleAxis"&&u!=null&&Math.abs(Math.abs(u[1]-u[0])-360)<=1e-6)for(var v=0;v<f;v++){var h=v>0?r[v-1].coordinate:r[f-1].coordinate,m=r[v].coordinate,y=v>=f-1?r[0].coordinate:r[v+1].coordinate,g=void 0;if(Vt(m-h)!==Vt(y-m)){var b=[];if(Vt(y-m)===Vt(u[1]-u[0])){g=y;var w=m+u[1]-u[0];b[0]=Math.min(w,(w+h)/2),b[1]=Math.max(w,(w+h)/2)}else{g=h;var S=y+u[1]-u[0];b[0]=Math.min(m,(S+m)/2),b[1]=Math.max(m,(S+m)/2)}var O=[Math.min(m,(g+m)/2),Math.max(m,(g+m)/2)];if(e>O[0]&&e<=O[1]||e>=b[0]&&e<=b[1]){({index:s}=r[v]);break}}else{var T=Math.min(h,y),M=Math.max(h,y);if(e>(T+m)/2&&e<=(M+m)/2){({index:s}=r[v]);break}}}else if(t){for(var j=0;j<f;j++)if(j===0&&e<=(t[j].coordinate+t[j+1].coordinate)/2||j>0&&j<f-1&&e>(t[j].coordinate+t[j-1].coordinate)/2&&e<=(t[j].coordinate+t[j+1].coordinate)/2||j===f-1&&e>(t[j].coordinate+t[j-1].coordinate)/2){({index:s}=t[j]);break}}return s},yN=(e,t,r)=>{if(t&&r){var{width:a,height:u}=r,{align:o,verticalAlign:s,layout:f}=t;if((f==="vertical"||f==="horizontal"&&s==="middle")&&o!=="center"&&ve(e[o]))return jn(jn({},e),{},{[o]:e[o]+(a||0)});if((f==="horizontal"||f==="vertical"&&o==="center")&&s!=="middle"&&ve(e[s]))return jn(jn({},e),{},{[s]:e[s]+(u||0)})}return e},Ha=(e,t)=>e==="horizontal"&&t==="xAxis"||e==="vertical"&&t==="yAxis"||e==="centric"&&t==="angleAxis"||e==="radial"&&t==="radiusAxis",FO=(e,t,r,a)=>{if(a)return e.map(f=>f.coordinate);var u,o,s=e.map(f=>(f.coordinate===t&&(u=!0),f.coordinate===r&&(o=!0),f.coordinate));return u||s.push(t),o||s.push(r),s},JO=(e,t,r)=>{if(!e)return null;var{duplicateDomain:a,type:u,range:o,scale:s,realScaleType:f,isCategorical:v,categoricalDomain:h,tickCount:m,ticks:y,niceTicks:g,axisType:b}=e;if(!s)return null;var w=f==="scaleBand"&&s.bandwidth?s.bandwidth()/2:2,S=u==="category"&&s.bandwidth?s.bandwidth()/w:0;if(S=b==="angleAxis"&&o&&o.length>=2?Vt(o[0]-o[1])*2*S:S,y||g){var O=(y||g||[]).map((T,M)=>{var j=a?a.indexOf(T):T;return{coordinate:s(j)+S,value:T,offset:S,index:M}});return O.filter(T=>!dn(T.coordinate))}return v&&h?h.map((T,M)=>({coordinate:s(T)+S,value:T,index:M,offset:S})):s.ticks&&m!=null?s.ticks(m).map((T,M)=>({coordinate:s(T)+S,value:T,offset:S,index:M})):s.domain().map((T,M)=>({coordinate:s(T)+S,value:a?a[T]:T,index:M,offset:S}))},p1=1e-4,gN=e=>{var t=e.domain();if(!(!t||t.length<=2)){var r=t.length,a=e.range(),u=Math.min(a[0],a[1])-p1,o=Math.max(a[0],a[1])+p1,s=e(t[0]),f=e(t[r-1]);(s<u||s>o||f<u||f>o)&&e.domain([t[0],t[r-1]])}},bN=(e,t)=>{if(!t||t.length!==2||!ve(t[0])||!ve(t[1]))return e;var r=Math.min(t[0],t[1]),a=Math.max(t[0],t[1]),u=[e[0],e[1]];return(!ve(e[0])||e[0]<r)&&(u[0]=r),(!ve(e[1])||e[1]>a)&&(u[1]=a),u[0]>a&&(u[0]=a),u[1]<r&&(u[1]=r),u},xN=e=>{var t=e.length;if(!(t<=0))for(var r=0,a=e[0].length;r<a;++r)for(var u=0,o=0,s=0;s<t;++s){var f=dn(e[s][r][1])?e[s][r][0]:e[s][r][1];f>=0?(e[s][r][0]=u,e[s][r][1]=u+f,u=e[s][r][1]):(e[s][r][0]=o,e[s][r][1]=o+f,o=e[s][r][1])}},SN=e=>{var t=e.length;if(!(t<=0))for(var r=0,a=e[0].length;r<a;++r)for(var u=0,o=0;o<t;++o){var s=dn(e[o][r][1])?e[o][r][0]:e[o][r][1];s>=0?(e[o][r][0]=u,e[o][r][1]=u+s,u=e[o][r][1]):(e[o][r][0]=0,e[o][r][1]=0)}},wN={sign:xN,expand:JM,none:zi,silhouette:eC,wiggle:tC,positive:SN},ON=(e,t,r)=>{var a=wN[r],u=FM().keys(t).value((o,s)=>+$t(o,s,0)).order(zv).offset(a);return u(e)};function AN(e){return e==null?void 0:String(e)}var y1=e=>{var{axis:t,ticks:r,offset:a,bandSize:u,entry:o,index:s}=e;if(t.type==="category")return r[s]?r[s].coordinate+a:null;var f=$t(o,t.dataKey,t.scale.domain()[s]);return gt(f)?null:t.scale(f)-u/2+a},EN=e=>{var{numericAxis:t}=e,r=t.scale.domain();if(t.type==="number"){var a=Math.min(r[0],r[1]),u=Math.max(r[0],r[1]);return a<=0&&u>=0?0:u<0?u:a}return r[0]},_N=e=>{var t=e.flat(2).filter(ve);return[Math.min(...t),Math.max(...t)]},TN=e=>[e[0]===1/0?0:e[0],e[1]===-1/0?0:e[1]],jN=(e,t,r)=>{if(e!=null)return TN(Object.keys(e).reduce((a,u)=>{var o=e[u],{stackedData:s}=o,f=s.reduce((v,h)=>{var m=WO(h,t,r),y=_N(m);return[Math.min(v[0],y[0]),Math.max(v[1],y[1])]},[1/0,-1/0]);return[Math.min(f[0],a[0]),Math.max(f[1],a[1])]},[1/0,-1/0]))},g1=/^dataMin[\s]*-[\s]*([0-9]+([.]{1}[0-9]+){0,1})$/,b1=/^dataMax[\s]*\+[\s]*([0-9]+([.]{1}[0-9]+){0,1})$/,km=(e,t,r)=>{if(e&&e.scale&&e.scale.bandwidth){var a=e.scale.bandwidth();if(!r||a>0)return a}if(e&&t&&t.length>=2){for(var u=us(t,m=>m.coordinate),o=1/0,s=1,f=u.length;s<f;s++){var v=u[s],h=u[s-1];o=Math.min((v.coordinate||0)-(h.coordinate||0),o)}return o===1/0?0:o}return r?void 0:0};function x1(e){var{tooltipEntrySettings:t,dataKey:r,payload:a,value:u,name:o}=e;return jn(jn({},t),{},{dataKey:r,payload:a,value:u,name:o})}function eA(e,t){if(e)return String(e);if(typeof t=="string")return t}function MN(e,t,r,a,u){if(r==="horizontal"||r==="vertical"){var o=e>=u.left&&e<=u.left+u.width&&t>=u.top&&t<=u.top+u.height;return o?{x:e,y:t}:null}return a?dN({x:e,y:t},a):null}var CN=(e,t,r,a)=>{var u=t.find(h=>h&&h.index===r);if(u){if(e==="horizontal")return{x:u.coordinate,y:a.y};if(e==="vertical")return{x:a.x,y:u.coordinate};if(e==="centric"){var o=u.coordinate,{radius:s}=a;return jn(jn(jn({},a),jt(a.cx,a.cy,s,o)),{},{angle:o,radius:s})}var f=u.coordinate,{angle:v}=a;return jn(jn(jn({},a),jt(a.cx,a.cy,f,v)),{},{angle:v,radius:f})}return{x:0,y:0}},DN=(e,t)=>t==="horizontal"?e.x:t==="vertical"?e.y:t==="centric"?e.angle:e.radius,Or=e=>e.layout.width,Ar=e=>e.layout.height,NN=e=>e.layout.scale,tA=e=>e.layout.margin,hs=Y(e=>e.cartesianAxis.xAxis,e=>Object.values(e)),vs=Y(e=>e.cartesianAxis.yAxis,e=>Object.values(e)),PN="data-recharts-item-index",RN="data-recharts-item-data-key",Tu=60;function S1(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function uc(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?S1(Object(r),!0).forEach(function(a){kN(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):S1(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function kN(e,t,r){return(t=zN(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function zN(e){var t=BN(e,"string");return typeof t=="symbol"?t:t+""}function BN(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var LN=e=>e.brush.height;function UN(e){var t=vs(e);return t.reduce((r,a)=>{if(a.orientation==="left"&&!a.mirror&&!a.hide){var u=typeof a.width=="number"?a.width:Tu;return r+u}return r},0)}function qN(e){var t=vs(e);return t.reduce((r,a)=>{if(a.orientation==="right"&&!a.mirror&&!a.hide){var u=typeof a.width=="number"?a.width:Tu;return r+u}return r},0)}function $N(e){var t=hs(e);return t.reduce((r,a)=>a.orientation==="top"&&!a.mirror&&!a.hide?r+a.height:r,0)}function HN(e){var t=hs(e);return t.reduce((r,a)=>a.orientation==="bottom"&&!a.mirror&&!a.hide?r+a.height:r,0)}var xt=Y([Or,Ar,tA,LN,UN,qN,$N,HN,_O,rD],(e,t,r,a,u,o,s,f,v,h)=>{var m={left:(r.left||0)+u,right:(r.right||0)+o},y={top:(r.top||0)+s,bottom:(r.bottom||0)+f},g=uc(uc({},y),m),b=g.bottom;g.bottom+=a,g=yN(g,v,h);var w=e-g.left-g.right,S=t-g.top-g.bottom;return uc(uc({brushBottom:b},g),{},{width:Math.max(w,0),height:Math.max(S,0)})}),YN=Y(xt,e=>({x:e.left,y:e.top,width:e.width,height:e.height})),zm=Y(Or,Ar,(e,t)=>({x:0,y:0,width:e,height:t})),GN=A.createContext(null),Nn=()=>A.useContext(GN)!=null,ms=e=>e.brush,ps=Y([ms,xt,tA],(e,t,r)=>({height:e.height,x:ve(e.x)?e.x:t.left,y:ve(e.y)?e.y:t.top+t.height+t.brushBottom-(r?.bottom||0),width:ve(e.width)?e.width:t.width})),Bm=()=>{var e,t=Nn(),r=me(YN),a=me(ps),u=(e=me(ms))===null||e===void 0?void 0:e.padding;return!t||!a||!u?r:{width:a.width-u.left-u.right,height:a.height-u.top-u.bottom,x:u.left,y:u.top}},KN={top:0,bottom:0,left:0,right:0,width:0,height:0,brushBottom:0},nA=()=>{var e;return(e=me(xt))!==null&&e!==void 0?e:KN},rA=()=>me(Or),aA=()=>me(Ar),ke=e=>e.layout.layoutType,Lm=()=>me(ke),IN={settings:{layout:"horizontal",align:"center",verticalAlign:"middle",itemSorter:"value"},size:{width:0,height:0},payload:[]},iA=mn({name:"legend",initialState:IN,reducers:{setLegendSize(e,t){e.size.width=t.payload.width,e.size.height=t.payload.height},setLegendSettings(e,t){e.settings.align=t.payload.align,e.settings.layout=t.payload.layout,e.settings.verticalAlign=t.payload.verticalAlign,e.settings.itemSorter=t.payload.itemSorter},addLegendPayload(e,t){e.payload.push(t.payload)},removeLegendPayload(e,t){var r=yr(e).payload.indexOf(t.payload);r>-1&&e.payload.splice(r,1)}}}),{setLegendSize:_q,setLegendSettings:Tq,addLegendPayload:XN,removeLegendPayload:VN}=iA.actions,ZN=iA.reducer;function Iv(){return Iv=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},Iv.apply(null,arguments)}function w1(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function Fh(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?w1(Object(r),!0).forEach(function(a){QN(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):w1(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function QN(e,t,r){return(t=WN(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function WN(e){var t=FN(e,"string");return typeof t=="symbol"?t:t+""}function FN(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function JN(e){return Array.isArray(e)&&Zn(e[0])&&Zn(e[1])?e.join(" ~ "):e}var eP=e=>{var{separator:t=" : ",contentStyle:r={},itemStyle:a={},labelStyle:u={},payload:o,formatter:s,itemSorter:f,wrapperClassName:v,labelClassName:h,label:m,labelFormatter:y,accessibilityLayer:g=!1}=e,b=()=>{if(o&&o.length){var N={padding:0,margin:0},P=(f?us(o,f):o).map((L,G)=>{if(L.type==="none")return null;var ee=L.formatter||s||JN,{value:ne,name:J}=L,re=ne,he=J;if(ee){var ce=ee(ne,J,L,G,o);if(Array.isArray(ce))[re,he]=ce;else if(ce!=null)re=ce;else return null}var se=Fh({display:"block",paddingTop:4,paddingBottom:4,color:L.color||"#000"},a);return A.createElement("li",{className:"recharts-tooltip-item",key:"tooltip-item-".concat(G),style:se},Zn(he)?A.createElement("span",{className:"recharts-tooltip-item-name"},he):null,Zn(he)?A.createElement("span",{className:"recharts-tooltip-item-separator"},t):null,A.createElement("span",{className:"recharts-tooltip-item-value"},re),A.createElement("span",{className:"recharts-tooltip-item-unit"},L.unit||""))});return A.createElement("ul",{className:"recharts-tooltip-item-list",style:N},P)}return null},w=Fh({margin:0,padding:10,backgroundColor:"#fff",border:"1px solid #ccc",whiteSpace:"nowrap"},r),S=Fh({margin:0},u),O=!gt(m),T=O?m:"",M=Ie("recharts-default-tooltip",v),j=Ie("recharts-tooltip-label",h);O&&y&&o!==void 0&&o!==null&&(T=y(m,o));var k=g?{role:"status","aria-live":"assertive"}:{};return A.createElement("div",Iv({className:M,style:w},k),A.createElement("p",{className:j,style:S},A.isValidElement(T)?T:"".concat(T)),b())},nu="recharts-tooltip-wrapper",tP={visibility:"hidden"};function nP(e){var{coordinate:t,translateX:r,translateY:a}=e;return Ie(nu,{["".concat(nu,"-right")]:ve(r)&&t&&ve(t.x)&&r>=t.x,["".concat(nu,"-left")]:ve(r)&&t&&ve(t.x)&&r<t.x,["".concat(nu,"-bottom")]:ve(a)&&t&&ve(t.y)&&a>=t.y,["".concat(nu,"-top")]:ve(a)&&t&&ve(t.y)&&a<t.y})}function O1(e){var{allowEscapeViewBox:t,coordinate:r,key:a,offsetTopLeft:u,position:o,reverseDirection:s,tooltipDimension:f,viewBox:v,viewBoxDimension:h}=e;if(o&&ve(o[a]))return o[a];var m=r[a]-f-(u>0?u:0),y=r[a]+u;if(t[a])return s[a]?m:y;var g=v[a];if(g==null)return 0;if(s[a]){var b=m,w=g;return b<w?Math.max(y,g):Math.max(m,g)}if(h==null)return 0;var S=y+f,O=g+h;return S>O?Math.max(m,g):Math.max(y,g)}function rP(e){var{translateX:t,translateY:r,useTranslate3d:a}=e;return{transform:a?"translate3d(".concat(t,"px, ").concat(r,"px, 0)"):"translate(".concat(t,"px, ").concat(r,"px)")}}function aP(e){var{allowEscapeViewBox:t,coordinate:r,offsetTopLeft:a,position:u,reverseDirection:o,tooltipBox:s,useTranslate3d:f,viewBox:v}=e,h,m,y;return s.height>0&&s.width>0&&r?(m=O1({allowEscapeViewBox:t,coordinate:r,key:"x",offsetTopLeft:a,position:u,reverseDirection:o,tooltipDimension:s.width,viewBox:v,viewBoxDimension:v.width}),y=O1({allowEscapeViewBox:t,coordinate:r,key:"y",offsetTopLeft:a,position:u,reverseDirection:o,tooltipDimension:s.height,viewBox:v,viewBoxDimension:v.height}),h=rP({translateX:m,translateY:y,useTranslate3d:f})):h=tP,{cssProperties:h,cssClasses:nP({translateX:m,translateY:y,coordinate:r})}}function A1(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function oc(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?A1(Object(r),!0).forEach(function(a){Xv(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):A1(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function Xv(e,t,r){return(t=iP(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function iP(e){var t=lP(e,"string");return typeof t=="symbol"?t:t+""}function lP(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}class uP extends A.PureComponent{constructor(){super(...arguments),Xv(this,"state",{dismissed:!1,dismissedAtCoordinate:{x:0,y:0}}),Xv(this,"handleKeyDown",t=>{if(t.key==="Escape"){var r,a,u,o;this.setState({dismissed:!0,dismissedAtCoordinate:{x:(r=(a=this.props.coordinate)===null||a===void 0?void 0:a.x)!==null&&r!==void 0?r:0,y:(u=(o=this.props.coordinate)===null||o===void 0?void 0:o.y)!==null&&u!==void 0?u:0}})}})}componentDidMount(){document.addEventListener("keydown",this.handleKeyDown)}componentWillUnmount(){document.removeEventListener("keydown",this.handleKeyDown)}componentDidUpdate(){var t,r;this.state.dismissed&&(((t=this.props.coordinate)===null||t===void 0?void 0:t.x)!==this.state.dismissedAtCoordinate.x||((r=this.props.coordinate)===null||r===void 0?void 0:r.y)!==this.state.dismissedAtCoordinate.y)&&(this.state.dismissed=!1)}render(){var{active:t,allowEscapeViewBox:r,animationDuration:a,animationEasing:u,children:o,coordinate:s,hasPayload:f,isAnimationActive:v,offset:h,position:m,reverseDirection:y,useTranslate3d:g,viewBox:b,wrapperStyle:w,lastBoundingBox:S,innerRef:O,hasPortalFromProps:T}=this.props,{cssClasses:M,cssProperties:j}=aP({allowEscapeViewBox:r,coordinate:s,offsetTopLeft:h,position:m,reverseDirection:y,tooltipBox:{height:S.height,width:S.width},useTranslate3d:g,viewBox:b}),k=T?{}:oc(oc({transition:v&&t?"transform ".concat(a,"ms ").concat(u):void 0},j),{},{pointerEvents:"none",visibility:!this.state.dismissed&&t&&f?"visible":"hidden",position:"absolute",top:0,left:0}),N=oc(oc({},k),{},{visibility:!this.state.dismissed&&t&&f?"visible":"hidden"},w);return A.createElement("div",{xmlns:"http://www.w3.org/1999/xhtml",tabIndex:-1,className:M,style:N,ref:O},o)}}var oP=()=>!(typeof window<"u"&&window.document&&window.document.createElement&&window.setTimeout),Vi={devToolsEnabled:!1,isSsr:oP()},lA=()=>me(e=>e.rootProps.accessibilityLayer);function vn(e){return Number.isFinite(e)}function Nc(e){return typeof e=="number"&&e>0&&Number.isFinite(e)}function Vv(){return Vv=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},Vv.apply(null,arguments)}function E1(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function _1(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?E1(Object(r),!0).forEach(function(a){cP(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):E1(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function cP(e,t,r){return(t=sP(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function sP(e){var t=fP(e,"string");return typeof t=="symbol"?t:t+""}function fP(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var T1={curveBasisClosed:$M,curveBasisOpen:HM,curveBasis:qM,curveBumpX:_M,curveBumpY:TM,curveLinearClosed:YM,curveLinear:is,curveMonotoneX:GM,curveMonotoneY:KM,curveNatural:IM,curveStep:XM,curveStepAfter:ZM,curveStepBefore:VM},cc=e=>vn(e.x)&&vn(e.y),ru=e=>e.x,au=e=>e.y,dP=(e,t)=>{if(typeof e=="function")return e;var r="curve".concat(Au(e));return(r==="curveMonotone"||r==="curveBump")&&t?T1["".concat(r).concat(t==="vertical"?"Y":"X")]:T1[r]||is},hP=e=>{var{type:t="linear",points:r=[],baseLine:a,layout:u,connectNulls:o=!1}=e,s=dP(t,u),f=o?r.filter(cc):r,v;if(Array.isArray(a)){var h=o?a.filter(y=>cc(y)):a,m=f.map((y,g)=>_1(_1({},y),{},{base:h[g]}));return u==="vertical"?v=tc().y(au).x1(ru).x0(y=>y.base.x):v=tc().x(ru).y1(au).y0(y=>y.base.y),v.defined(cc).curve(s),v(m)}return u==="vertical"&&ve(a)?v=tc().y(au).x1(ru).x0(a):ve(a)?v=tc().x(ru).y1(au).y0(a):v=nO().x(ru).y(au),v.defined(cc).curve(s),v(f)},vP=e=>{var{className:t,points:r,path:a,pathRef:u}=e;if((!r||!r.length)&&!a)return null;var o=r&&r.length?hP(e):a;return A.createElement("path",Vv({},Xi(e),sM(e),{className:Ie("recharts-curve",t),d:o===null?void 0:o,ref:u}))},mP=["x","y","top","left","width","height","className"];function Zv(){return Zv=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},Zv.apply(null,arguments)}function j1(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function pP(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?j1(Object(r),!0).forEach(function(a){yP(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):j1(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function yP(e,t,r){return(t=gP(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function gP(e){var t=bP(e,"string");return typeof t=="symbol"?t:t+""}function bP(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function xP(e,t){if(e==null)return{};var r,a,u=SP(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function SP(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}var wP=(e,t,r,a,u,o)=>"M".concat(e,",").concat(u,"v").concat(a,"M").concat(o,",").concat(t,"h").concat(r),OP=e=>{var{x:t=0,y:r=0,top:a=0,left:u=0,width:o=0,height:s=0,className:f}=e,v=xP(e,mP),h=pP({x:t,y:r,top:a,left:u,width:o,height:s},v);return!ve(t)||!ve(r)||!ve(o)||!ve(s)||!ve(a)||!ve(u)?null:A.createElement("path",Zv({},Mt(h,!0),{className:Ie("recharts-cross",f),d:wP(t,r,o,s,a,u)}))};function AP(e,t,r,a){var u=a/2;return{stroke:"none",fill:"#ccc",x:e==="horizontal"?t.x-u:r.left+.5,y:e==="horizontal"?r.top+.5:t.y-u,width:e==="horizontal"?a:r.width-1,height:e==="horizontal"?r.height-1:a}}function M1(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function EP(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?M1(Object(r),!0).forEach(function(a){_P(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):M1(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function _P(e,t,r){return(t=TP(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function TP(e){var t=jP(e,"string");return typeof t=="symbol"?t:t+""}function jP(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function Pn(e,t){var r=EP({},e),a=t,u=Object.keys(t),o=u.reduce((s,f)=>(s[f]===void 0&&a[f]!==void 0&&(s[f]=a[f]),s),r);return o}function MP(){}function C1(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function D1(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?C1(Object(r),!0).forEach(function(a){CP(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):C1(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function CP(e,t,r){return(t=DP(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function DP(e){var t=NP(e,"string");return typeof t=="symbol"?t:t+""}function NP(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var PP=e=>e.replace(/([A-Z])/g,t=>"-".concat(t.toLowerCase())),uA=(e,t,r)=>e.map(a=>"".concat(PP(a)," ").concat(t,"ms ").concat(r)).join(","),RP=(e,t)=>[Object.keys(e),Object.keys(t)].reduce((r,a)=>r.filter(u=>a.includes(u))),yu=(e,t)=>Object.keys(t).reduce((r,a)=>D1(D1({},r),{},{[a]:e(a,t[a])}),{});function N1(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function Tt(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?N1(Object(r),!0).forEach(function(a){kP(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):N1(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function kP(e,t,r){return(t=zP(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function zP(e){var t=BP(e,"string");return typeof t=="symbol"?t:t+""}function BP(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var Pc=(e,t,r)=>e+(t-e)*r,Qv=e=>{var{from:t,to:r}=e;return t!==r},oA=(e,t,r)=>{var a=yu((u,o)=>{if(Qv(o)){var[s,f]=e(o.from,o.to,o.velocity);return Tt(Tt({},o),{},{from:s,velocity:f})}return o},t);return r<1?yu((u,o)=>Qv(o)?Tt(Tt({},o),{},{velocity:Pc(o.velocity,a[u].velocity,r),from:Pc(o.from,a[u].from,r)}):o,t):oA(e,a,r-1)};function LP(e,t,r,a,u,o){var s,f=a.reduce((g,b)=>Tt(Tt({},g),{},{[b]:{from:e[b],velocity:0,to:t[b]}}),{}),v=()=>yu((g,b)=>b.from,f),h=()=>!Object.values(f).filter(Qv).length,m=null,y=g=>{s||(s=g);var b=g-s,w=b/r.dt;f=oA(r,f,w),u(Tt(Tt(Tt({},e),t),v())),s=g,h()||(m=o.setTimeout(y))};return()=>(m=o.setTimeout(y),()=>{m()})}function UP(e,t,r,a,u,o,s){var f=null,v=u.reduce((y,g)=>Tt(Tt({},y),{},{[g]:[e[g],t[g]]}),{}),h,m=y=>{h||(h=y);var g=(y-h)/a,b=yu((S,O)=>Pc(...O,r(g)),v);if(o(Tt(Tt(Tt({},e),t),b)),g<1)f=s.setTimeout(m);else{var w=yu((S,O)=>Pc(...O,r(1)),v);o(Tt(Tt(Tt({},e),t),w))}};return()=>(f=s.setTimeout(m),()=>{f()})}const qP=(e,t,r,a,u,o)=>{var s=RP(e,t);return r.isStepper===!0?LP(e,t,r,s,u,o):UP(e,t,r,a,s,u,o)};var Rc=1e-4,cA=(e,t)=>[0,3*e,3*t-6*e,3*e-3*t+1],sA=(e,t)=>e.map((r,a)=>r*t**a).reduce((r,a)=>r+a),P1=(e,t)=>r=>{var a=cA(e,t);return sA(a,r)},$P=(e,t)=>r=>{var a=cA(e,t),u=[...a.map((o,s)=>o*s).slice(1),0];return sA(u,r)},R1=function(){for(var t,r,a,u,o=arguments.length,s=new Array(o),f=0;f<o;f++)s[f]=arguments[f];if(s.length===1)switch(s[0]){case"linear":[t,a,r,u]=[0,0,1,1];break;case"ease":[t,a,r,u]=[.25,.1,.25,1];break;case"ease-in":[t,a,r,u]=[.42,0,1,1];break;case"ease-out":[t,a,r,u]=[.42,0,.58,1];break;case"ease-in-out":[t,a,r,u]=[0,0,.58,1];break;default:{var v=s[0].split("(");v[0]==="cubic-bezier"&&v[1].split(")")[0].split(",").length===4&&([t,a,r,u]=v[1].split(")")[0].split(",").map(w=>parseFloat(w)))}}else s.length===4&&([t,a,r,u]=s);var h=P1(t,r),m=P1(a,u),y=$P(t,r),g=w=>w>1?1:w<0?0:w,b=w=>{for(var S=w>1?1:w,O=S,T=0;T<8;++T){var M=h(O)-S,j=y(O);if(Math.abs(M-S)<Rc||j<Rc)return m(O);O=g(O-M/j)}return m(O)};return b.isStepper=!1,b},HP=function(){var t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},{stiff:r=100,damping:a=8,dt:u=17}=t,o=(s,f,v)=>{var h=-(s-f)*r,m=v*a,y=v+(h-m)*u/1e3,g=v*u/1e3+s;return Math.abs(g-f)<Rc&&Math.abs(y)<Rc?[f,0]:[g,y]};return o.isStepper=!0,o.dt=u,o},YP=e=>{if(typeof e=="string")switch(e){case"ease":case"ease-in-out":case"ease-out":case"ease-in":case"linear":return R1(e);case"spring":return HP();default:if(e.split("(")[0]==="cubic-bezier")return R1(e)}return typeof e=="function"?e:null};function GP(e){var t,r=()=>null,a=!1,u=null,o=s=>{if(!a){if(Array.isArray(s)){if(!s.length)return;var f=s,[v,...h]=f;if(typeof v=="number"){u=e.setTimeout(o.bind(null,h),v);return}o(v),u=e.setTimeout(o.bind(null,h));return}typeof s=="string"&&(t=s,r(t)),typeof s=="object"&&(t=s,r(t)),typeof s=="function"&&s()}};return{stop:()=>{a=!0},start:s=>{a=!1,u&&(u(),u=null),o(s)},subscribe:s=>(r=s,()=>{r=()=>null}),getTimeoutController:()=>e}}class KP{setTimeout(t){var r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:0,a=performance.now(),u=null,o=s=>{s-a>=r?t(s):typeof requestAnimationFrame=="function"&&(u=requestAnimationFrame(o))};return u=requestAnimationFrame(o),()=>{cancelAnimationFrame(u)}}}function IP(){return GP(new KP)}var XP=A.createContext(IP);function VP(e,t){var r=A.useContext(XP);return A.useMemo(()=>t??r(e),[e,t,r])}var ZP={begin:0,duration:1e3,easing:"ease",isActive:!0,canBegin:!0,onAnimationEnd:()=>{},onAnimationStart:()=>{}},k1={t:0},Jh={t:1};function Um(e){var t=Pn(e,ZP),{isActive:r,canBegin:a,duration:u,easing:o,begin:s,onAnimationEnd:f,onAnimationStart:v,children:h}=t,m=VP(t.animationId,t.animationManager),[y,g]=A.useState(r?k1:Jh),b=A.useRef(null);return A.useEffect(()=>{r||g(Jh)},[r]),A.useEffect(()=>{if(!r||!a)return MP;var w=qP(k1,Jh,YP(o),u,g,m.getTimeoutController()),S=()=>{b.current=w()};return m.start([v,s,S,u,f]),()=>{m.stop(),b.current&&b.current(),f()}},[r,a,u,o,s,v,f,m]),h(y.t)}function qm(e){var t=arguments.length>1&&arguments[1]!==void 0?arguments[1]:"animation-",r=A.useRef(vu(t)),a=A.useRef(e);return a.current!==e&&(r.current=vu(t),a.current=e),r.current}function z1(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function B1(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?z1(Object(r),!0).forEach(function(a){QP(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):z1(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function QP(e,t,r){return(t=WP(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function WP(e){var t=FP(e,"string");return typeof t=="symbol"?t:t+""}function FP(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function kc(){return kc=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},kc.apply(null,arguments)}var L1=(e,t,r,a,u)=>{var o=Math.min(Math.abs(r)/2,Math.abs(a)/2),s=a>=0?1:-1,f=r>=0?1:-1,v=a>=0&&r>=0||a<0&&r<0?1:0,h;if(o>0&&u instanceof Array){for(var m=[0,0,0,0],y=0,g=4;y<g;y++)m[y]=u[y]>o?o:u[y];h="M".concat(e,",").concat(t+s*m[0]),m[0]>0&&(h+="A ".concat(m[0],",").concat(m[0],",0,0,").concat(v,",").concat(e+f*m[0],",").concat(t)),h+="L ".concat(e+r-f*m[1],",").concat(t),m[1]>0&&(h+="A ".concat(m[1],",").concat(m[1],",0,0,").concat(v,`,
        `).concat(e+r,",").concat(t+s*m[1])),h+="L ".concat(e+r,",").concat(t+a-s*m[2]),m[2]>0&&(h+="A ".concat(m[2],",").concat(m[2],",0,0,").concat(v,`,
        `).concat(e+r-f*m[2],",").concat(t+a)),h+="L ".concat(e+f*m[3],",").concat(t+a),m[3]>0&&(h+="A ".concat(m[3],",").concat(m[3],",0,0,").concat(v,`,
        `).concat(e,",").concat(t+a-s*m[3])),h+="Z"}else if(o>0&&u===+u&&u>0){var b=Math.min(o,u);h="M ".concat(e,",").concat(t+s*b,`
            A `).concat(b,",").concat(b,",0,0,").concat(v,",").concat(e+f*b,",").concat(t,`
            L `).concat(e+r-f*b,",").concat(t,`
            A `).concat(b,",").concat(b,",0,0,").concat(v,",").concat(e+r,",").concat(t+s*b,`
            L `).concat(e+r,",").concat(t+a-s*b,`
            A `).concat(b,",").concat(b,",0,0,").concat(v,",").concat(e+r-f*b,",").concat(t+a,`
            L `).concat(e+f*b,",").concat(t+a,`
            A `).concat(b,",").concat(b,",0,0,").concat(v,",").concat(e,",").concat(t+a-s*b," Z")}else h="M ".concat(e,",").concat(t," h ").concat(r," v ").concat(a," h ").concat(-r," Z");return h},JP={x:0,y:0,width:0,height:0,radius:0,isAnimationActive:!1,isUpdateAnimationActive:!1,animationBegin:0,animationDuration:1500,animationEasing:"ease"},fA=e=>{var t=Pn(e,JP),r=A.useRef(null),[a,u]=A.useState(-1);A.useEffect(()=>{if(r.current&&r.current.getTotalLength)try{var ce=r.current.getTotalLength();ce&&u(ce)}catch{}},[]);var{x:o,y:s,width:f,height:v,radius:h,className:m}=t,{animationEasing:y,animationDuration:g,animationBegin:b,isAnimationActive:w,isUpdateAnimationActive:S}=t,O=A.useRef(f),T=A.useRef(v),M=A.useRef(o),j=A.useRef(s),k=A.useMemo(()=>({x:o,y:s,width:f,height:v,radius:h}),[o,s,f,v,h]),N=qm(k,"rectangle-");if(o!==+o||s!==+s||f!==+f||v!==+v||f===0||v===0)return null;var P=Ie("recharts-rectangle",m);if(!S)return A.createElement("path",kc({},Mt(t,!0),{className:P,d:L1(o,s,f,v,h)}));var L=O.current,G=T.current,ee=M.current,ne=j.current,J="0px ".concat(a===-1?1:a,"px"),re="".concat(a,"px 0px"),he=uA(["strokeDasharray"],g,typeof y=="string"?y:void 0);return A.createElement(Um,{animationId:N,key:N,canBegin:a>0,duration:g,easing:y,isActive:S,begin:b},ce=>{var se=Lt(L,f,ce),B=Lt(G,v,ce),I=Lt(ee,o,ce),F=Lt(ne,s,ce);r.current&&(O.current=se,T.current=B,M.current=I,j.current=F);var oe;return w?ce>0?oe={transition:he,strokeDasharray:re}:oe={strokeDasharray:J}:oe={strokeDasharray:re},A.createElement("path",kc({},Mt(t,!0),{className:P,d:L1(I,F,se,B,h),ref:r,style:B1(B1({},oe),t.style)}))})};function dA(e){var{cx:t,cy:r,radius:a,startAngle:u,endAngle:o}=e,s=jt(t,r,a,u),f=jt(t,r,a,o);return{points:[s,f],cx:t,cy:r,radius:a,startAngle:u,endAngle:o}}function Wv(){return Wv=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},Wv.apply(null,arguments)}var eR=(e,t)=>{var r=Vt(t-e),a=Math.min(Math.abs(t-e),359.999);return r*a},sc=e=>{var{cx:t,cy:r,radius:a,angle:u,sign:o,isExternal:s,cornerRadius:f,cornerIsExternal:v}=e,h=f*(s?1:-1)+a,m=Math.asin(f/h)/Dc,y=v?u:u+o*m,g=jt(t,r,h,y),b=jt(t,r,a,y),w=v?u-o*m:u,S=jt(t,r,h*Math.cos(m*Dc),w);return{center:g,circleTangency:b,lineTangency:S,theta:m}},hA=e=>{var{cx:t,cy:r,innerRadius:a,outerRadius:u,startAngle:o,endAngle:s}=e,f=eR(o,s),v=o+f,h=jt(t,r,u,o),m=jt(t,r,u,v),y="M ".concat(h.x,",").concat(h.y,`
    A `).concat(u,",").concat(u,`,0,
    `).concat(+(Math.abs(f)>180),",").concat(+(o>v),`,
    `).concat(m.x,",").concat(m.y,`
  `);if(a>0){var g=jt(t,r,a,o),b=jt(t,r,a,v);y+="L ".concat(b.x,",").concat(b.y,`
            A `).concat(a,",").concat(a,`,0,
            `).concat(+(Math.abs(f)>180),",").concat(+(o<=v),`,
            `).concat(g.x,",").concat(g.y," Z")}else y+="L ".concat(t,",").concat(r," Z");return y},tR=e=>{var{cx:t,cy:r,innerRadius:a,outerRadius:u,cornerRadius:o,forceCornerRadius:s,cornerIsExternal:f,startAngle:v,endAngle:h}=e,m=Vt(h-v),{circleTangency:y,lineTangency:g,theta:b}=sc({cx:t,cy:r,radius:u,angle:v,sign:m,cornerRadius:o,cornerIsExternal:f}),{circleTangency:w,lineTangency:S,theta:O}=sc({cx:t,cy:r,radius:u,angle:h,sign:-m,cornerRadius:o,cornerIsExternal:f}),T=f?Math.abs(v-h):Math.abs(v-h)-b-O;if(T<0)return s?"M ".concat(g.x,",").concat(g.y,`
        a`).concat(o,",").concat(o,",0,0,1,").concat(o*2,`,0
        a`).concat(o,",").concat(o,",0,0,1,").concat(-o*2,`,0
      `):hA({cx:t,cy:r,innerRadius:a,outerRadius:u,startAngle:v,endAngle:h});var M="M ".concat(g.x,",").concat(g.y,`
    A`).concat(o,",").concat(o,",0,0,").concat(+(m<0),",").concat(y.x,",").concat(y.y,`
    A`).concat(u,",").concat(u,",0,").concat(+(T>180),",").concat(+(m<0),",").concat(w.x,",").concat(w.y,`
    A`).concat(o,",").concat(o,",0,0,").concat(+(m<0),",").concat(S.x,",").concat(S.y,`
  `);if(a>0){var{circleTangency:j,lineTangency:k,theta:N}=sc({cx:t,cy:r,radius:a,angle:v,sign:m,isExternal:!0,cornerRadius:o,cornerIsExternal:f}),{circleTangency:P,lineTangency:L,theta:G}=sc({cx:t,cy:r,radius:a,angle:h,sign:-m,isExternal:!0,cornerRadius:o,cornerIsExternal:f}),ee=f?Math.abs(v-h):Math.abs(v-h)-N-G;if(ee<0&&o===0)return"".concat(M,"L").concat(t,",").concat(r,"Z");M+="L".concat(L.x,",").concat(L.y,`
      A`).concat(o,",").concat(o,",0,0,").concat(+(m<0),",").concat(P.x,",").concat(P.y,`
      A`).concat(a,",").concat(a,",0,").concat(+(ee>180),",").concat(+(m>0),",").concat(j.x,",").concat(j.y,`
      A`).concat(o,",").concat(o,",0,0,").concat(+(m<0),",").concat(k.x,",").concat(k.y,"Z")}else M+="L".concat(t,",").concat(r,"Z");return M},nR={cx:0,cy:0,innerRadius:0,outerRadius:0,startAngle:0,endAngle:0,cornerRadius:0,forceCornerRadius:!1,cornerIsExternal:!1},vA=e=>{var t=Pn(e,nR),{cx:r,cy:a,innerRadius:u,outerRadius:o,cornerRadius:s,forceCornerRadius:f,cornerIsExternal:v,startAngle:h,endAngle:m,className:y}=t;if(o<u||h===m)return null;var g=Ie("recharts-sector",y),b=o-u,w=qn(s,b,0,!0),S;return w>0&&Math.abs(h-m)<360?S=tR({cx:r,cy:a,innerRadius:u,outerRadius:o,cornerRadius:Math.min(w,b/2),forceCornerRadius:f,cornerIsExternal:v,startAngle:h,endAngle:m}):S=hA({cx:r,cy:a,innerRadius:u,outerRadius:o,startAngle:h,endAngle:m}),A.createElement("path",Wv({},Mt(t,!0),{className:g,d:S}))};function rR(e,t,r){var a,u,o,s;if(e==="horizontal")a=t.x,o=a,u=r.top,s=r.top+r.height;else if(e==="vertical")u=t.y,s=u,a=r.left,o=r.left+r.width;else if(t.cx!=null&&t.cy!=null)if(e==="centric"){var{cx:f,cy:v,innerRadius:h,outerRadius:m,angle:y}=t,g=jt(f,v,h,y),b=jt(f,v,m,y);a=g.x,u=g.y,o=b.x,s=b.y}else return dA(t);return[{x:a,y:u},{x:o,y:s}]}var ev={},tv={},nv={},U1;function aR(){return U1||(U1=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=AO();function r(a){return t.isSymbol(a)?NaN:Number(a)}e.toNumber=r})(nv)),nv}var q1;function iR(){return q1||(q1=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=aR();function r(a){return a?(a=t.toNumber(a),a===1/0||a===-1/0?(a<0?-1:1)*Number.MAX_VALUE:a===a?a:0):a===0?a:0}e.toFinite=r})(tv)),tv}var $1;function lR(){return $1||($1=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=EO(),r=iR();function a(u,o,s){s&&typeof s!="number"&&t.isIterateeCall(u,o,s)&&(o=s=void 0),u=r.toFinite(u),o===void 0?(o=u,u=0):o=r.toFinite(o),s=s===void 0?u<o?1:-1:r.toFinite(s);const f=Math.max(Math.ceil((o-u)/(s||1)),0),v=new Array(f);for(let h=0;h<f;h++)v[h]=u,u+=s;return v}e.range=a})(ev)),ev}var rv,H1;function uR(){return H1||(H1=1,rv=lR().range),rv}var oR=uR();const mA=Wn(oR);function na(e,t){return e==null||t==null?NaN:e<t?-1:e>t?1:e>=t?0:NaN}function cR(e,t){return e==null||t==null?NaN:t<e?-1:t>e?1:t>=e?0:NaN}function $m(e){let t,r,a;e.length!==2?(t=na,r=(f,v)=>na(e(f),v),a=(f,v)=>e(f)-v):(t=e===na||e===cR?e:sR,r=e,a=e);function u(f,v,h=0,m=f.length){if(h<m){if(t(v,v)!==0)return m;do{const y=h+m>>>1;r(f[y],v)<0?h=y+1:m=y}while(h<m)}return h}function o(f,v,h=0,m=f.length){if(h<m){if(t(v,v)!==0)return m;do{const y=h+m>>>1;r(f[y],v)<=0?h=y+1:m=y}while(h<m)}return h}function s(f,v,h=0,m=f.length){const y=u(f,v,h,m-1);return y>h&&a(f[y-1],v)>-a(f[y],v)?y-1:y}return{left:u,center:s,right:o}}function sR(){return 0}function pA(e){return e===null?NaN:+e}function*fR(e,t){for(let r of e)r!=null&&(r=+r)>=r&&(yield r)}const dR=$m(na),ju=dR.right;$m(pA).center;class Y1 extends Map{constructor(t,r=mR){if(super(),Object.defineProperties(this,{_intern:{value:new Map},_key:{value:r}}),t!=null)for(const[a,u]of t)this.set(a,u)}get(t){return super.get(G1(this,t))}has(t){return super.has(G1(this,t))}set(t,r){return super.set(hR(this,t),r)}delete(t){return super.delete(vR(this,t))}}function G1({_intern:e,_key:t},r){const a=t(r);return e.has(a)?e.get(a):r}function hR({_intern:e,_key:t},r){const a=t(r);return e.has(a)?e.get(a):(e.set(a,r),r)}function vR({_intern:e,_key:t},r){const a=t(r);return e.has(a)&&(r=e.get(a),e.delete(a)),r}function mR(e){return e!==null&&typeof e=="object"?e.valueOf():e}function pR(e=na){if(e===na)return yA;if(typeof e!="function")throw new TypeError("compare is not a function");return(t,r)=>{const a=e(t,r);return a||a===0?a:(e(r,r)===0)-(e(t,t)===0)}}function yA(e,t){return(e==null||!(e>=e))-(t==null||!(t>=t))||(e<t?-1:e>t?1:0)}const yR=Math.sqrt(50),gR=Math.sqrt(10),bR=Math.sqrt(2);function zc(e,t,r){const a=(t-e)/Math.max(0,r),u=Math.floor(Math.log10(a)),o=a/Math.pow(10,u),s=o>=yR?10:o>=gR?5:o>=bR?2:1;let f,v,h;return u<0?(h=Math.pow(10,-u)/s,f=Math.round(e*h),v=Math.round(t*h),f/h<e&&++f,v/h>t&&--v,h=-h):(h=Math.pow(10,u)*s,f=Math.round(e/h),v=Math.round(t/h),f*h<e&&++f,v*h>t&&--v),v<f&&.5<=r&&r<2?zc(e,t,r*2):[f,v,h]}function Fv(e,t,r){if(t=+t,e=+e,r=+r,!(r>0))return[];if(e===t)return[e];const a=t<e,[u,o,s]=a?zc(t,e,r):zc(e,t,r);if(!(o>=u))return[];const f=o-u+1,v=new Array(f);if(a)if(s<0)for(let h=0;h<f;++h)v[h]=(o-h)/-s;else for(let h=0;h<f;++h)v[h]=(o-h)*s;else if(s<0)for(let h=0;h<f;++h)v[h]=(u+h)/-s;else for(let h=0;h<f;++h)v[h]=(u+h)*s;return v}function Jv(e,t,r){return t=+t,e=+e,r=+r,zc(e,t,r)[2]}function em(e,t,r){t=+t,e=+e,r=+r;const a=t<e,u=a?Jv(t,e,r):Jv(e,t,r);return(a?-1:1)*(u<0?1/-u:u)}function K1(e,t){let r;for(const a of e)a!=null&&(r<a||r===void 0&&a>=a)&&(r=a);return r}function I1(e,t){let r;for(const a of e)a!=null&&(r>a||r===void 0&&a>=a)&&(r=a);return r}function gA(e,t,r=0,a=1/0,u){if(t=Math.floor(t),r=Math.floor(Math.max(0,r)),a=Math.floor(Math.min(e.length-1,a)),!(r<=t&&t<=a))return e;for(u=u===void 0?yA:pR(u);a>r;){if(a-r>600){const v=a-r+1,h=t-r+1,m=Math.log(v),y=.5*Math.exp(2*m/3),g=.5*Math.sqrt(m*y*(v-y)/v)*(h-v/2<0?-1:1),b=Math.max(r,Math.floor(t-h*y/v+g)),w=Math.min(a,Math.floor(t+(v-h)*y/v+g));gA(e,t,b,w,u)}const o=e[t];let s=r,f=a;for(iu(e,r,t),u(e[a],o)>0&&iu(e,r,a);s<f;){for(iu(e,s,f),++s,--f;u(e[s],o)<0;)++s;for(;u(e[f],o)>0;)--f}u(e[r],o)===0?iu(e,r,f):(++f,iu(e,f,a)),f<=t&&(r=f+1),t<=f&&(a=f-1)}return e}function iu(e,t,r){const a=e[t];e[t]=e[r],e[r]=a}function xR(e,t,r){if(e=Float64Array.from(fR(e)),!(!(a=e.length)||isNaN(t=+t))){if(t<=0||a<2)return I1(e);if(t>=1)return K1(e);var a,u=(a-1)*t,o=Math.floor(u),s=K1(gA(e,o).subarray(0,o+1)),f=I1(e.subarray(o+1));return s+(f-s)*(u-o)}}function SR(e,t,r=pA){if(!(!(a=e.length)||isNaN(t=+t))){if(t<=0||a<2)return+r(e[0],0,e);if(t>=1)return+r(e[a-1],a-1,e);var a,u=(a-1)*t,o=Math.floor(u),s=+r(e[o],o,e),f=+r(e[o+1],o+1,e);return s+(f-s)*(u-o)}}function wR(e,t,r){e=+e,t=+t,r=(u=arguments.length)<2?(t=e,e=0,1):u<3?1:+r;for(var a=-1,u=Math.max(0,Math.ceil((t-e)/r))|0,o=new Array(u);++a<u;)o[a]=e+a*r;return o}function Rn(e,t){switch(arguments.length){case 0:break;case 1:this.range(e);break;default:this.range(t).domain(e);break}return this}function Er(e,t){switch(arguments.length){case 0:break;case 1:{typeof e=="function"?this.interpolator(e):this.range(e);break}default:{this.domain(e),typeof t=="function"?this.interpolator(t):this.range(t);break}}return this}const tm=Symbol("implicit");function Hm(){var e=new Y1,t=[],r=[],a=tm;function u(o){let s=e.get(o);if(s===void 0){if(a!==tm)return a;e.set(o,s=t.push(o)-1)}return r[s%r.length]}return u.domain=function(o){if(!arguments.length)return t.slice();t=[],e=new Y1;for(const s of o)e.has(s)||e.set(s,t.push(s)-1);return u},u.range=function(o){return arguments.length?(r=Array.from(o),u):r.slice()},u.unknown=function(o){return arguments.length?(a=o,u):a},u.copy=function(){return Hm(t,r).unknown(a)},Rn.apply(u,arguments),u}function Ym(){var e=Hm().unknown(void 0),t=e.domain,r=e.range,a=0,u=1,o,s,f=!1,v=0,h=0,m=.5;delete e.unknown;function y(){var g=t().length,b=u<a,w=b?u:a,S=b?a:u;o=(S-w)/Math.max(1,g-v+h*2),f&&(o=Math.floor(o)),w+=(S-w-o*(g-v))*m,s=o*(1-v),f&&(w=Math.round(w),s=Math.round(s));var O=wR(g).map(function(T){return w+o*T});return r(b?O.reverse():O)}return e.domain=function(g){return arguments.length?(t(g),y()):t()},e.range=function(g){return arguments.length?([a,u]=g,a=+a,u=+u,y()):[a,u]},e.rangeRound=function(g){return[a,u]=g,a=+a,u=+u,f=!0,y()},e.bandwidth=function(){return s},e.step=function(){return o},e.round=function(g){return arguments.length?(f=!!g,y()):f},e.padding=function(g){return arguments.length?(v=Math.min(1,h=+g),y()):v},e.paddingInner=function(g){return arguments.length?(v=Math.min(1,g),y()):v},e.paddingOuter=function(g){return arguments.length?(h=+g,y()):h},e.align=function(g){return arguments.length?(m=Math.max(0,Math.min(1,g)),y()):m},e.copy=function(){return Ym(t(),[a,u]).round(f).paddingInner(v).paddingOuter(h).align(m)},Rn.apply(y(),arguments)}function bA(e){var t=e.copy;return e.padding=e.paddingOuter,delete e.paddingInner,delete e.paddingOuter,e.copy=function(){return bA(t())},e}function OR(){return bA(Ym.apply(null,arguments).paddingInner(1))}function Gm(e,t,r){e.prototype=t.prototype=r,r.constructor=e}function xA(e,t){var r=Object.create(e.prototype);for(var a in t)r[a]=t[a];return r}function Mu(){}var gu=.7,Bc=1/gu,Ri="\\s*([+-]?\\d+)\\s*",bu="\\s*([+-]?(?:\\d*\\.)?\\d+(?:[eE][+-]?\\d+)?)\\s*",Vn="\\s*([+-]?(?:\\d*\\.)?\\d+(?:[eE][+-]?\\d+)?)%\\s*",AR=/^#([0-9a-f]{3,8})$/,ER=new RegExp(`^rgb\\(${Ri},${Ri},${Ri}\\)$`),_R=new RegExp(`^rgb\\(${Vn},${Vn},${Vn}\\)$`),TR=new RegExp(`^rgba\\(${Ri},${Ri},${Ri},${bu}\\)$`),jR=new RegExp(`^rgba\\(${Vn},${Vn},${Vn},${bu}\\)$`),MR=new RegExp(`^hsl\\(${bu},${Vn},${Vn}\\)$`),CR=new RegExp(`^hsla\\(${bu},${Vn},${Vn},${bu}\\)$`),X1={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074};Gm(Mu,xu,{copy(e){return Object.assign(new this.constructor,this,e)},displayable(){return this.rgb().displayable()},hex:V1,formatHex:V1,formatHex8:DR,formatHsl:NR,formatRgb:Z1,toString:Z1});function V1(){return this.rgb().formatHex()}function DR(){return this.rgb().formatHex8()}function NR(){return SA(this).formatHsl()}function Z1(){return this.rgb().formatRgb()}function xu(e){var t,r;return e=(e+"").trim().toLowerCase(),(t=AR.exec(e))?(r=t[1].length,t=parseInt(t[1],16),r===6?Q1(t):r===3?new Zt(t>>8&15|t>>4&240,t>>4&15|t&240,(t&15)<<4|t&15,1):r===8?fc(t>>24&255,t>>16&255,t>>8&255,(t&255)/255):r===4?fc(t>>12&15|t>>8&240,t>>8&15|t>>4&240,t>>4&15|t&240,((t&15)<<4|t&15)/255):null):(t=ER.exec(e))?new Zt(t[1],t[2],t[3],1):(t=_R.exec(e))?new Zt(t[1]*255/100,t[2]*255/100,t[3]*255/100,1):(t=TR.exec(e))?fc(t[1],t[2],t[3],t[4]):(t=jR.exec(e))?fc(t[1]*255/100,t[2]*255/100,t[3]*255/100,t[4]):(t=MR.exec(e))?J1(t[1],t[2]/100,t[3]/100,1):(t=CR.exec(e))?J1(t[1],t[2]/100,t[3]/100,t[4]):X1.hasOwnProperty(e)?Q1(X1[e]):e==="transparent"?new Zt(NaN,NaN,NaN,0):null}function Q1(e){return new Zt(e>>16&255,e>>8&255,e&255,1)}function fc(e,t,r,a){return a<=0&&(e=t=r=NaN),new Zt(e,t,r,a)}function PR(e){return e instanceof Mu||(e=xu(e)),e?(e=e.rgb(),new Zt(e.r,e.g,e.b,e.opacity)):new Zt}function nm(e,t,r,a){return arguments.length===1?PR(e):new Zt(e,t,r,a??1)}function Zt(e,t,r,a){this.r=+e,this.g=+t,this.b=+r,this.opacity=+a}Gm(Zt,nm,xA(Mu,{brighter(e){return e=e==null?Bc:Math.pow(Bc,e),new Zt(this.r*e,this.g*e,this.b*e,this.opacity)},darker(e){return e=e==null?gu:Math.pow(gu,e),new Zt(this.r*e,this.g*e,this.b*e,this.opacity)},rgb(){return this},clamp(){return new Zt(za(this.r),za(this.g),za(this.b),Lc(this.opacity))},displayable(){return-.5<=this.r&&this.r<255.5&&-.5<=this.g&&this.g<255.5&&-.5<=this.b&&this.b<255.5&&0<=this.opacity&&this.opacity<=1},hex:W1,formatHex:W1,formatHex8:RR,formatRgb:F1,toString:F1}));function W1(){return`#${Da(this.r)}${Da(this.g)}${Da(this.b)}`}function RR(){return`#${Da(this.r)}${Da(this.g)}${Da(this.b)}${Da((isNaN(this.opacity)?1:this.opacity)*255)}`}function F1(){const e=Lc(this.opacity);return`${e===1?"rgb(":"rgba("}${za(this.r)}, ${za(this.g)}, ${za(this.b)}${e===1?")":`, ${e})`}`}function Lc(e){return isNaN(e)?1:Math.max(0,Math.min(1,e))}function za(e){return Math.max(0,Math.min(255,Math.round(e)||0))}function Da(e){return e=za(e),(e<16?"0":"")+e.toString(16)}function J1(e,t,r,a){return a<=0?e=t=r=NaN:r<=0||r>=1?e=t=NaN:t<=0&&(e=NaN),new Un(e,t,r,a)}function SA(e){if(e instanceof Un)return new Un(e.h,e.s,e.l,e.opacity);if(e instanceof Mu||(e=xu(e)),!e)return new Un;if(e instanceof Un)return e;e=e.rgb();var t=e.r/255,r=e.g/255,a=e.b/255,u=Math.min(t,r,a),o=Math.max(t,r,a),s=NaN,f=o-u,v=(o+u)/2;return f?(t===o?s=(r-a)/f+(r<a)*6:r===o?s=(a-t)/f+2:s=(t-r)/f+4,f/=v<.5?o+u:2-o-u,s*=60):f=v>0&&v<1?0:s,new Un(s,f,v,e.opacity)}function kR(e,t,r,a){return arguments.length===1?SA(e):new Un(e,t,r,a??1)}function Un(e,t,r,a){this.h=+e,this.s=+t,this.l=+r,this.opacity=+a}Gm(Un,kR,xA(Mu,{brighter(e){return e=e==null?Bc:Math.pow(Bc,e),new Un(this.h,this.s,this.l*e,this.opacity)},darker(e){return e=e==null?gu:Math.pow(gu,e),new Un(this.h,this.s,this.l*e,this.opacity)},rgb(){var e=this.h%360+(this.h<0)*360,t=isNaN(e)||isNaN(this.s)?0:this.s,r=this.l,a=r+(r<.5?r:1-r)*t,u=2*r-a;return new Zt(av(e>=240?e-240:e+120,u,a),av(e,u,a),av(e<120?e+240:e-120,u,a),this.opacity)},clamp(){return new Un(eS(this.h),dc(this.s),dc(this.l),Lc(this.opacity))},displayable(){return(0<=this.s&&this.s<=1||isNaN(this.s))&&0<=this.l&&this.l<=1&&0<=this.opacity&&this.opacity<=1},formatHsl(){const e=Lc(this.opacity);return`${e===1?"hsl(":"hsla("}${eS(this.h)}, ${dc(this.s)*100}%, ${dc(this.l)*100}%${e===1?")":`, ${e})`}`}}));function eS(e){return e=(e||0)%360,e<0?e+360:e}function dc(e){return Math.max(0,Math.min(1,e||0))}function av(e,t,r){return(e<60?t+(r-t)*e/60:e<180?r:e<240?t+(r-t)*(240-e)/60:t)*255}const Km=e=>()=>e;function zR(e,t){return function(r){return e+r*t}}function BR(e,t,r){return e=Math.pow(e,r),t=Math.pow(t,r)-e,r=1/r,function(a){return Math.pow(e+a*t,r)}}function LR(e){return(e=+e)==1?wA:function(t,r){return r-t?BR(t,r,e):Km(isNaN(t)?r:t)}}function wA(e,t){var r=t-e;return r?zR(e,r):Km(isNaN(e)?t:e)}const tS=(function e(t){var r=LR(t);function a(u,o){var s=r((u=nm(u)).r,(o=nm(o)).r),f=r(u.g,o.g),v=r(u.b,o.b),h=wA(u.opacity,o.opacity);return function(m){return u.r=s(m),u.g=f(m),u.b=v(m),u.opacity=h(m),u+""}}return a.gamma=e,a})(1);function UR(e,t){t||(t=[]);var r=e?Math.min(t.length,e.length):0,a=t.slice(),u;return function(o){for(u=0;u<r;++u)a[u]=e[u]*(1-o)+t[u]*o;return a}}function qR(e){return ArrayBuffer.isView(e)&&!(e instanceof DataView)}function $R(e,t){var r=t?t.length:0,a=e?Math.min(r,e.length):0,u=new Array(a),o=new Array(r),s;for(s=0;s<a;++s)u[s]=Zi(e[s],t[s]);for(;s<r;++s)o[s]=t[s];return function(f){for(s=0;s<a;++s)o[s]=u[s](f);return o}}function HR(e,t){var r=new Date;return e=+e,t=+t,function(a){return r.setTime(e*(1-a)+t*a),r}}function Uc(e,t){return e=+e,t=+t,function(r){return e*(1-r)+t*r}}function YR(e,t){var r={},a={},u;(e===null||typeof e!="object")&&(e={}),(t===null||typeof t!="object")&&(t={});for(u in t)u in e?r[u]=Zi(e[u],t[u]):a[u]=t[u];return function(o){for(u in r)a[u]=r[u](o);return a}}var rm=/[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g,iv=new RegExp(rm.source,"g");function GR(e){return function(){return e}}function KR(e){return function(t){return e(t)+""}}function IR(e,t){var r=rm.lastIndex=iv.lastIndex=0,a,u,o,s=-1,f=[],v=[];for(e=e+"",t=t+"";(a=rm.exec(e))&&(u=iv.exec(t));)(o=u.index)>r&&(o=t.slice(r,o),f[s]?f[s]+=o:f[++s]=o),(a=a[0])===(u=u[0])?f[s]?f[s]+=u:f[++s]=u:(f[++s]=null,v.push({i:s,x:Uc(a,u)})),r=iv.lastIndex;return r<t.length&&(o=t.slice(r),f[s]?f[s]+=o:f[++s]=o),f.length<2?v[0]?KR(v[0].x):GR(t):(t=v.length,function(h){for(var m=0,y;m<t;++m)f[(y=v[m]).i]=y.x(h);return f.join("")})}function Zi(e,t){var r=typeof t,a;return t==null||r==="boolean"?Km(t):(r==="number"?Uc:r==="string"?(a=xu(t))?(t=a,tS):IR:t instanceof xu?tS:t instanceof Date?HR:qR(t)?UR:Array.isArray(t)?$R:typeof t.valueOf!="function"&&typeof t.toString!="function"||isNaN(t)?YR:Uc)(e,t)}function Im(e,t){return e=+e,t=+t,function(r){return Math.round(e*(1-r)+t*r)}}function XR(e,t){t===void 0&&(t=e,e=Zi);for(var r=0,a=t.length-1,u=t[0],o=new Array(a<0?0:a);r<a;)o[r]=e(u,u=t[++r]);return function(s){var f=Math.max(0,Math.min(a-1,Math.floor(s*=a)));return o[f](s-f)}}function VR(e){return function(){return e}}function qc(e){return+e}var nS=[0,1];function qt(e){return e}function am(e,t){return(t-=e=+e)?function(r){return(r-e)/t}:VR(isNaN(t)?NaN:.5)}function ZR(e,t){var r;return e>t&&(r=e,e=t,t=r),function(a){return Math.max(e,Math.min(t,a))}}function QR(e,t,r){var a=e[0],u=e[1],o=t[0],s=t[1];return u<a?(a=am(u,a),o=r(s,o)):(a=am(a,u),o=r(o,s)),function(f){return o(a(f))}}function WR(e,t,r){var a=Math.min(e.length,t.length)-1,u=new Array(a),o=new Array(a),s=-1;for(e[a]<e[0]&&(e=e.slice().reverse(),t=t.slice().reverse());++s<a;)u[s]=am(e[s],e[s+1]),o[s]=r(t[s],t[s+1]);return function(f){var v=ju(e,f,1,a)-1;return o[v](u[v](f))}}function Cu(e,t){return t.domain(e.domain()).range(e.range()).interpolate(e.interpolate()).clamp(e.clamp()).unknown(e.unknown())}function ys(){var e=nS,t=nS,r=Zi,a,u,o,s=qt,f,v,h;function m(){var g=Math.min(e.length,t.length);return s!==qt&&(s=ZR(e[0],e[g-1])),f=g>2?WR:QR,v=h=null,y}function y(g){return g==null||isNaN(g=+g)?o:(v||(v=f(e.map(a),t,r)))(a(s(g)))}return y.invert=function(g){return s(u((h||(h=f(t,e.map(a),Uc)))(g)))},y.domain=function(g){return arguments.length?(e=Array.from(g,qc),m()):e.slice()},y.range=function(g){return arguments.length?(t=Array.from(g),m()):t.slice()},y.rangeRound=function(g){return t=Array.from(g),r=Im,m()},y.clamp=function(g){return arguments.length?(s=g?!0:qt,m()):s!==qt},y.interpolate=function(g){return arguments.length?(r=g,m()):r},y.unknown=function(g){return arguments.length?(o=g,y):o},function(g,b){return a=g,u=b,m()}}function Xm(){return ys()(qt,qt)}function FR(e){return Math.abs(e=Math.round(e))>=1e21?e.toLocaleString("en").replace(/,/g,""):e.toString(10)}function $c(e,t){if((r=(e=t?e.toExponential(t-1):e.toExponential()).indexOf("e"))<0)return null;var r,a=e.slice(0,r);return[a.length>1?a[0]+a.slice(2):a,+e.slice(r+1)]}function Li(e){return e=$c(Math.abs(e)),e?e[1]:NaN}function JR(e,t){return function(r,a){for(var u=r.length,o=[],s=0,f=e[0],v=0;u>0&&f>0&&(v+f+1>a&&(f=Math.max(1,a-v)),o.push(r.substring(u-=f,u+f)),!((v+=f+1)>a));)f=e[s=(s+1)%e.length];return o.reverse().join(t)}}function e3(e){return function(t){return t.replace(/[0-9]/g,function(r){return e[+r]})}}var t3=/^(?:(.)?([<>=^]))?([+\-( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?(~)?([a-z%])?$/i;function Su(e){if(!(t=t3.exec(e)))throw new Error("invalid format: "+e);var t;return new Vm({fill:t[1],align:t[2],sign:t[3],symbol:t[4],zero:t[5],width:t[6],comma:t[7],precision:t[8]&&t[8].slice(1),trim:t[9],type:t[10]})}Su.prototype=Vm.prototype;function Vm(e){this.fill=e.fill===void 0?" ":e.fill+"",this.align=e.align===void 0?">":e.align+"",this.sign=e.sign===void 0?"-":e.sign+"",this.symbol=e.symbol===void 0?"":e.symbol+"",this.zero=!!e.zero,this.width=e.width===void 0?void 0:+e.width,this.comma=!!e.comma,this.precision=e.precision===void 0?void 0:+e.precision,this.trim=!!e.trim,this.type=e.type===void 0?"":e.type+""}Vm.prototype.toString=function(){return this.fill+this.align+this.sign+this.symbol+(this.zero?"0":"")+(this.width===void 0?"":Math.max(1,this.width|0))+(this.comma?",":"")+(this.precision===void 0?"":"."+Math.max(0,this.precision|0))+(this.trim?"~":"")+this.type};function n3(e){e:for(var t=e.length,r=1,a=-1,u;r<t;++r)switch(e[r]){case".":a=u=r;break;case"0":a===0&&(a=r),u=r;break;default:if(!+e[r])break e;a>0&&(a=0);break}return a>0?e.slice(0,a)+e.slice(u+1):e}var OA;function r3(e,t){var r=$c(e,t);if(!r)return e+"";var a=r[0],u=r[1],o=u-(OA=Math.max(-8,Math.min(8,Math.floor(u/3)))*3)+1,s=a.length;return o===s?a:o>s?a+new Array(o-s+1).join("0"):o>0?a.slice(0,o)+"."+a.slice(o):"0."+new Array(1-o).join("0")+$c(e,Math.max(0,t+o-1))[0]}function rS(e,t){var r=$c(e,t);if(!r)return e+"";var a=r[0],u=r[1];return u<0?"0."+new Array(-u).join("0")+a:a.length>u+1?a.slice(0,u+1)+"."+a.slice(u+1):a+new Array(u-a.length+2).join("0")}const aS={"%":(e,t)=>(e*100).toFixed(t),b:e=>Math.round(e).toString(2),c:e=>e+"",d:FR,e:(e,t)=>e.toExponential(t),f:(e,t)=>e.toFixed(t),g:(e,t)=>e.toPrecision(t),o:e=>Math.round(e).toString(8),p:(e,t)=>rS(e*100,t),r:rS,s:r3,X:e=>Math.round(e).toString(16).toUpperCase(),x:e=>Math.round(e).toString(16)};function iS(e){return e}var lS=Array.prototype.map,uS=["y","z","a","f","p","n","µ","m","","k","M","G","T","P","E","Z","Y"];function a3(e){var t=e.grouping===void 0||e.thousands===void 0?iS:JR(lS.call(e.grouping,Number),e.thousands+""),r=e.currency===void 0?"":e.currency[0]+"",a=e.currency===void 0?"":e.currency[1]+"",u=e.decimal===void 0?".":e.decimal+"",o=e.numerals===void 0?iS:e3(lS.call(e.numerals,String)),s=e.percent===void 0?"%":e.percent+"",f=e.minus===void 0?"−":e.minus+"",v=e.nan===void 0?"NaN":e.nan+"";function h(y){y=Su(y);var g=y.fill,b=y.align,w=y.sign,S=y.symbol,O=y.zero,T=y.width,M=y.comma,j=y.precision,k=y.trim,N=y.type;N==="n"?(M=!0,N="g"):aS[N]||(j===void 0&&(j=12),k=!0,N="g"),(O||g==="0"&&b==="=")&&(O=!0,g="0",b="=");var P=S==="$"?r:S==="#"&&/[boxX]/.test(N)?"0"+N.toLowerCase():"",L=S==="$"?a:/[%p]/.test(N)?s:"",G=aS[N],ee=/[defgprs%]/.test(N);j=j===void 0?6:/[gprs]/.test(N)?Math.max(1,Math.min(21,j)):Math.max(0,Math.min(20,j));function ne(J){var re=P,he=L,ce,se,B;if(N==="c")he=G(J)+he,J="";else{J=+J;var I=J<0||1/J<0;if(J=isNaN(J)?v:G(Math.abs(J),j),k&&(J=n3(J)),I&&+J==0&&w!=="+"&&(I=!1),re=(I?w==="("?w:f:w==="-"||w==="("?"":w)+re,he=(N==="s"?uS[8+OA/3]:"")+he+(I&&w==="("?")":""),ee){for(ce=-1,se=J.length;++ce<se;)if(B=J.charCodeAt(ce),48>B||B>57){he=(B===46?u+J.slice(ce+1):J.slice(ce))+he,J=J.slice(0,ce);break}}}M&&!O&&(J=t(J,1/0));var F=re.length+J.length+he.length,oe=F<T?new Array(T-F+1).join(g):"";switch(M&&O&&(J=t(oe+J,oe.length?T-he.length:1/0),oe=""),b){case"<":J=re+J+he+oe;break;case"=":J=re+oe+J+he;break;case"^":J=oe.slice(0,F=oe.length>>1)+re+J+he+oe.slice(F);break;default:J=oe+re+J+he;break}return o(J)}return ne.toString=function(){return y+""},ne}function m(y,g){var b=h((y=Su(y),y.type="f",y)),w=Math.max(-8,Math.min(8,Math.floor(Li(g)/3)))*3,S=Math.pow(10,-w),O=uS[8+w/3];return function(T){return b(S*T)+O}}return{format:h,formatPrefix:m}}var hc,Zm,AA;i3({thousands:",",grouping:[3],currency:["$",""]});function i3(e){return hc=a3(e),Zm=hc.format,AA=hc.formatPrefix,hc}function l3(e){return Math.max(0,-Li(Math.abs(e)))}function u3(e,t){return Math.max(0,Math.max(-8,Math.min(8,Math.floor(Li(t)/3)))*3-Li(Math.abs(e)))}function o3(e,t){return e=Math.abs(e),t=Math.abs(t)-e,Math.max(0,Li(t)-Li(e))+1}function EA(e,t,r,a){var u=em(e,t,r),o;switch(a=Su(a??",f"),a.type){case"s":{var s=Math.max(Math.abs(e),Math.abs(t));return a.precision==null&&!isNaN(o=u3(u,s))&&(a.precision=o),AA(a,s)}case"":case"e":case"g":case"p":case"r":{a.precision==null&&!isNaN(o=o3(u,Math.max(Math.abs(e),Math.abs(t))))&&(a.precision=o-(a.type==="e"));break}case"f":case"%":{a.precision==null&&!isNaN(o=l3(u))&&(a.precision=o-(a.type==="%")*2);break}}return Zm(a)}function ia(e){var t=e.domain;return e.ticks=function(r){var a=t();return Fv(a[0],a[a.length-1],r??10)},e.tickFormat=function(r,a){var u=t();return EA(u[0],u[u.length-1],r??10,a)},e.nice=function(r){r==null&&(r=10);var a=t(),u=0,o=a.length-1,s=a[u],f=a[o],v,h,m=10;for(f<s&&(h=s,s=f,f=h,h=u,u=o,o=h);m-- >0;){if(h=Jv(s,f,r),h===v)return a[u]=s,a[o]=f,t(a);if(h>0)s=Math.floor(s/h)*h,f=Math.ceil(f/h)*h;else if(h<0)s=Math.ceil(s*h)/h,f=Math.floor(f*h)/h;else break;v=h}return e},e}function _A(){var e=Xm();return e.copy=function(){return Cu(e,_A())},Rn.apply(e,arguments),ia(e)}function TA(e){var t;function r(a){return a==null||isNaN(a=+a)?t:a}return r.invert=r,r.domain=r.range=function(a){return arguments.length?(e=Array.from(a,qc),r):e.slice()},r.unknown=function(a){return arguments.length?(t=a,r):t},r.copy=function(){return TA(e).unknown(t)},e=arguments.length?Array.from(e,qc):[0,1],ia(r)}function jA(e,t){e=e.slice();var r=0,a=e.length-1,u=e[r],o=e[a],s;return o<u&&(s=r,r=a,a=s,s=u,u=o,o=s),e[r]=t.floor(u),e[a]=t.ceil(o),e}function oS(e){return Math.log(e)}function cS(e){return Math.exp(e)}function c3(e){return-Math.log(-e)}function s3(e){return-Math.exp(-e)}function f3(e){return isFinite(e)?+("1e"+e):e<0?0:e}function d3(e){return e===10?f3:e===Math.E?Math.exp:t=>Math.pow(e,t)}function h3(e){return e===Math.E?Math.log:e===10&&Math.log10||e===2&&Math.log2||(e=Math.log(e),t=>Math.log(t)/e)}function sS(e){return(t,r)=>-e(-t,r)}function Qm(e){const t=e(oS,cS),r=t.domain;let a=10,u,o;function s(){return u=h3(a),o=d3(a),r()[0]<0?(u=sS(u),o=sS(o),e(c3,s3)):e(oS,cS),t}return t.base=function(f){return arguments.length?(a=+f,s()):a},t.domain=function(f){return arguments.length?(r(f),s()):r()},t.ticks=f=>{const v=r();let h=v[0],m=v[v.length-1];const y=m<h;y&&([h,m]=[m,h]);let g=u(h),b=u(m),w,S;const O=f==null?10:+f;let T=[];if(!(a%1)&&b-g<O){if(g=Math.floor(g),b=Math.ceil(b),h>0){for(;g<=b;++g)for(w=1;w<a;++w)if(S=g<0?w/o(-g):w*o(g),!(S<h)){if(S>m)break;T.push(S)}}else for(;g<=b;++g)for(w=a-1;w>=1;--w)if(S=g>0?w/o(-g):w*o(g),!(S<h)){if(S>m)break;T.push(S)}T.length*2<O&&(T=Fv(h,m,O))}else T=Fv(g,b,Math.min(b-g,O)).map(o);return y?T.reverse():T},t.tickFormat=(f,v)=>{if(f==null&&(f=10),v==null&&(v=a===10?"s":","),typeof v!="function"&&(!(a%1)&&(v=Su(v)).precision==null&&(v.trim=!0),v=Zm(v)),f===1/0)return v;const h=Math.max(1,a*f/t.ticks().length);return m=>{let y=m/o(Math.round(u(m)));return y*a<a-.5&&(y*=a),y<=h?v(m):""}},t.nice=()=>r(jA(r(),{floor:f=>o(Math.floor(u(f))),ceil:f=>o(Math.ceil(u(f)))})),t}function MA(){const e=Qm(ys()).domain([1,10]);return e.copy=()=>Cu(e,MA()).base(e.base()),Rn.apply(e,arguments),e}function fS(e){return function(t){return Math.sign(t)*Math.log1p(Math.abs(t/e))}}function dS(e){return function(t){return Math.sign(t)*Math.expm1(Math.abs(t))*e}}function Wm(e){var t=1,r=e(fS(t),dS(t));return r.constant=function(a){return arguments.length?e(fS(t=+a),dS(t)):t},ia(r)}function CA(){var e=Wm(ys());return e.copy=function(){return Cu(e,CA()).constant(e.constant())},Rn.apply(e,arguments)}function hS(e){return function(t){return t<0?-Math.pow(-t,e):Math.pow(t,e)}}function v3(e){return e<0?-Math.sqrt(-e):Math.sqrt(e)}function m3(e){return e<0?-e*e:e*e}function Fm(e){var t=e(qt,qt),r=1;function a(){return r===1?e(qt,qt):r===.5?e(v3,m3):e(hS(r),hS(1/r))}return t.exponent=function(u){return arguments.length?(r=+u,a()):r},ia(t)}function Jm(){var e=Fm(ys());return e.copy=function(){return Cu(e,Jm()).exponent(e.exponent())},Rn.apply(e,arguments),e}function p3(){return Jm.apply(null,arguments).exponent(.5)}function vS(e){return Math.sign(e)*e*e}function y3(e){return Math.sign(e)*Math.sqrt(Math.abs(e))}function DA(){var e=Xm(),t=[0,1],r=!1,a;function u(o){var s=y3(e(o));return isNaN(s)?a:r?Math.round(s):s}return u.invert=function(o){return e.invert(vS(o))},u.domain=function(o){return arguments.length?(e.domain(o),u):e.domain()},u.range=function(o){return arguments.length?(e.range((t=Array.from(o,qc)).map(vS)),u):t.slice()},u.rangeRound=function(o){return u.range(o).round(!0)},u.round=function(o){return arguments.length?(r=!!o,u):r},u.clamp=function(o){return arguments.length?(e.clamp(o),u):e.clamp()},u.unknown=function(o){return arguments.length?(a=o,u):a},u.copy=function(){return DA(e.domain(),t).round(r).clamp(e.clamp()).unknown(a)},Rn.apply(u,arguments),ia(u)}function NA(){var e=[],t=[],r=[],a;function u(){var s=0,f=Math.max(1,t.length);for(r=new Array(f-1);++s<f;)r[s-1]=SR(e,s/f);return o}function o(s){return s==null||isNaN(s=+s)?a:t[ju(r,s)]}return o.invertExtent=function(s){var f=t.indexOf(s);return f<0?[NaN,NaN]:[f>0?r[f-1]:e[0],f<r.length?r[f]:e[e.length-1]]},o.domain=function(s){if(!arguments.length)return e.slice();e=[];for(let f of s)f!=null&&!isNaN(f=+f)&&e.push(f);return e.sort(na),u()},o.range=function(s){return arguments.length?(t=Array.from(s),u()):t.slice()},o.unknown=function(s){return arguments.length?(a=s,o):a},o.quantiles=function(){return r.slice()},o.copy=function(){return NA().domain(e).range(t).unknown(a)},Rn.apply(o,arguments)}function PA(){var e=0,t=1,r=1,a=[.5],u=[0,1],o;function s(v){return v!=null&&v<=v?u[ju(a,v,0,r)]:o}function f(){var v=-1;for(a=new Array(r);++v<r;)a[v]=((v+1)*t-(v-r)*e)/(r+1);return s}return s.domain=function(v){return arguments.length?([e,t]=v,e=+e,t=+t,f()):[e,t]},s.range=function(v){return arguments.length?(r=(u=Array.from(v)).length-1,f()):u.slice()},s.invertExtent=function(v){var h=u.indexOf(v);return h<0?[NaN,NaN]:h<1?[e,a[0]]:h>=r?[a[r-1],t]:[a[h-1],a[h]]},s.unknown=function(v){return arguments.length&&(o=v),s},s.thresholds=function(){return a.slice()},s.copy=function(){return PA().domain([e,t]).range(u).unknown(o)},Rn.apply(ia(s),arguments)}function RA(){var e=[.5],t=[0,1],r,a=1;function u(o){return o!=null&&o<=o?t[ju(e,o,0,a)]:r}return u.domain=function(o){return arguments.length?(e=Array.from(o),a=Math.min(e.length,t.length-1),u):e.slice()},u.range=function(o){return arguments.length?(t=Array.from(o),a=Math.min(e.length,t.length-1),u):t.slice()},u.invertExtent=function(o){var s=t.indexOf(o);return[e[s-1],e[s]]},u.unknown=function(o){return arguments.length?(r=o,u):r},u.copy=function(){return RA().domain(e).range(t).unknown(r)},Rn.apply(u,arguments)}const lv=new Date,uv=new Date;function ct(e,t,r,a){function u(o){return e(o=arguments.length===0?new Date:new Date(+o)),o}return u.floor=o=>(e(o=new Date(+o)),o),u.ceil=o=>(e(o=new Date(o-1)),t(o,1),e(o),o),u.round=o=>{const s=u(o),f=u.ceil(o);return o-s<f-o?s:f},u.offset=(o,s)=>(t(o=new Date(+o),s==null?1:Math.floor(s)),o),u.range=(o,s,f)=>{const v=[];if(o=u.ceil(o),f=f==null?1:Math.floor(f),!(o<s)||!(f>0))return v;let h;do v.push(h=new Date(+o)),t(o,f),e(o);while(h<o&&o<s);return v},u.filter=o=>ct(s=>{if(s>=s)for(;e(s),!o(s);)s.setTime(s-1)},(s,f)=>{if(s>=s)if(f<0)for(;++f<=0;)for(;t(s,-1),!o(s););else for(;--f>=0;)for(;t(s,1),!o(s););}),r&&(u.count=(o,s)=>(lv.setTime(+o),uv.setTime(+s),e(lv),e(uv),Math.floor(r(lv,uv))),u.every=o=>(o=Math.floor(o),!isFinite(o)||!(o>0)?null:o>1?u.filter(a?s=>a(s)%o===0:s=>u.count(0,s)%o===0):u)),u}const Hc=ct(()=>{},(e,t)=>{e.setTime(+e+t)},(e,t)=>t-e);Hc.every=e=>(e=Math.floor(e),!isFinite(e)||!(e>0)?null:e>1?ct(t=>{t.setTime(Math.floor(t/e)*e)},(t,r)=>{t.setTime(+t+r*e)},(t,r)=>(r-t)/e):Hc);Hc.range;const mr=1e3,Mn=mr*60,pr=Mn*60,xr=pr*24,ep=xr*7,mS=xr*30,ov=xr*365,Na=ct(e=>{e.setTime(e-e.getMilliseconds())},(e,t)=>{e.setTime(+e+t*mr)},(e,t)=>(t-e)/mr,e=>e.getUTCSeconds());Na.range;const tp=ct(e=>{e.setTime(e-e.getMilliseconds()-e.getSeconds()*mr)},(e,t)=>{e.setTime(+e+t*Mn)},(e,t)=>(t-e)/Mn,e=>e.getMinutes());tp.range;const np=ct(e=>{e.setUTCSeconds(0,0)},(e,t)=>{e.setTime(+e+t*Mn)},(e,t)=>(t-e)/Mn,e=>e.getUTCMinutes());np.range;const rp=ct(e=>{e.setTime(e-e.getMilliseconds()-e.getSeconds()*mr-e.getMinutes()*Mn)},(e,t)=>{e.setTime(+e+t*pr)},(e,t)=>(t-e)/pr,e=>e.getHours());rp.range;const ap=ct(e=>{e.setUTCMinutes(0,0,0)},(e,t)=>{e.setTime(+e+t*pr)},(e,t)=>(t-e)/pr,e=>e.getUTCHours());ap.range;const Du=ct(e=>e.setHours(0,0,0,0),(e,t)=>e.setDate(e.getDate()+t),(e,t)=>(t-e-(t.getTimezoneOffset()-e.getTimezoneOffset())*Mn)/xr,e=>e.getDate()-1);Du.range;const gs=ct(e=>{e.setUTCHours(0,0,0,0)},(e,t)=>{e.setUTCDate(e.getUTCDate()+t)},(e,t)=>(t-e)/xr,e=>e.getUTCDate()-1);gs.range;const kA=ct(e=>{e.setUTCHours(0,0,0,0)},(e,t)=>{e.setUTCDate(e.getUTCDate()+t)},(e,t)=>(t-e)/xr,e=>Math.floor(e/xr));kA.range;function Ya(e){return ct(t=>{t.setDate(t.getDate()-(t.getDay()+7-e)%7),t.setHours(0,0,0,0)},(t,r)=>{t.setDate(t.getDate()+r*7)},(t,r)=>(r-t-(r.getTimezoneOffset()-t.getTimezoneOffset())*Mn)/ep)}const bs=Ya(0),Yc=Ya(1),g3=Ya(2),b3=Ya(3),Ui=Ya(4),x3=Ya(5),S3=Ya(6);bs.range;Yc.range;g3.range;b3.range;Ui.range;x3.range;S3.range;function Ga(e){return ct(t=>{t.setUTCDate(t.getUTCDate()-(t.getUTCDay()+7-e)%7),t.setUTCHours(0,0,0,0)},(t,r)=>{t.setUTCDate(t.getUTCDate()+r*7)},(t,r)=>(r-t)/ep)}const xs=Ga(0),Gc=Ga(1),w3=Ga(2),O3=Ga(3),qi=Ga(4),A3=Ga(5),E3=Ga(6);xs.range;Gc.range;w3.range;O3.range;qi.range;A3.range;E3.range;const ip=ct(e=>{e.setDate(1),e.setHours(0,0,0,0)},(e,t)=>{e.setMonth(e.getMonth()+t)},(e,t)=>t.getMonth()-e.getMonth()+(t.getFullYear()-e.getFullYear())*12,e=>e.getMonth());ip.range;const lp=ct(e=>{e.setUTCDate(1),e.setUTCHours(0,0,0,0)},(e,t)=>{e.setUTCMonth(e.getUTCMonth()+t)},(e,t)=>t.getUTCMonth()-e.getUTCMonth()+(t.getUTCFullYear()-e.getUTCFullYear())*12,e=>e.getUTCMonth());lp.range;const Sr=ct(e=>{e.setMonth(0,1),e.setHours(0,0,0,0)},(e,t)=>{e.setFullYear(e.getFullYear()+t)},(e,t)=>t.getFullYear()-e.getFullYear(),e=>e.getFullYear());Sr.every=e=>!isFinite(e=Math.floor(e))||!(e>0)?null:ct(t=>{t.setFullYear(Math.floor(t.getFullYear()/e)*e),t.setMonth(0,1),t.setHours(0,0,0,0)},(t,r)=>{t.setFullYear(t.getFullYear()+r*e)});Sr.range;const wr=ct(e=>{e.setUTCMonth(0,1),e.setUTCHours(0,0,0,0)},(e,t)=>{e.setUTCFullYear(e.getUTCFullYear()+t)},(e,t)=>t.getUTCFullYear()-e.getUTCFullYear(),e=>e.getUTCFullYear());wr.every=e=>!isFinite(e=Math.floor(e))||!(e>0)?null:ct(t=>{t.setUTCFullYear(Math.floor(t.getUTCFullYear()/e)*e),t.setUTCMonth(0,1),t.setUTCHours(0,0,0,0)},(t,r)=>{t.setUTCFullYear(t.getUTCFullYear()+r*e)});wr.range;function zA(e,t,r,a,u,o){const s=[[Na,1,mr],[Na,5,5*mr],[Na,15,15*mr],[Na,30,30*mr],[o,1,Mn],[o,5,5*Mn],[o,15,15*Mn],[o,30,30*Mn],[u,1,pr],[u,3,3*pr],[u,6,6*pr],[u,12,12*pr],[a,1,xr],[a,2,2*xr],[r,1,ep],[t,1,mS],[t,3,3*mS],[e,1,ov]];function f(h,m,y){const g=m<h;g&&([h,m]=[m,h]);const b=y&&typeof y.range=="function"?y:v(h,m,y),w=b?b.range(h,+m+1):[];return g?w.reverse():w}function v(h,m,y){const g=Math.abs(m-h)/y,b=$m(([,,O])=>O).right(s,g);if(b===s.length)return e.every(em(h/ov,m/ov,y));if(b===0)return Hc.every(Math.max(em(h,m,y),1));const[w,S]=s[g/s[b-1][2]<s[b][2]/g?b-1:b];return w.every(S)}return[f,v]}const[_3,T3]=zA(wr,lp,xs,kA,ap,np),[j3,M3]=zA(Sr,ip,bs,Du,rp,tp);function cv(e){if(0<=e.y&&e.y<100){var t=new Date(-1,e.m,e.d,e.H,e.M,e.S,e.L);return t.setFullYear(e.y),t}return new Date(e.y,e.m,e.d,e.H,e.M,e.S,e.L)}function sv(e){if(0<=e.y&&e.y<100){var t=new Date(Date.UTC(-1,e.m,e.d,e.H,e.M,e.S,e.L));return t.setUTCFullYear(e.y),t}return new Date(Date.UTC(e.y,e.m,e.d,e.H,e.M,e.S,e.L))}function lu(e,t,r){return{y:e,m:t,d:r,H:0,M:0,S:0,L:0}}function C3(e){var t=e.dateTime,r=e.date,a=e.time,u=e.periods,o=e.days,s=e.shortDays,f=e.months,v=e.shortMonths,h=uu(u),m=ou(u),y=uu(o),g=ou(o),b=uu(s),w=ou(s),S=uu(f),O=ou(f),T=uu(v),M=ou(v),j={a:I,A:F,b:oe,B:D,c:null,d:SS,e:SS,f:J3,g:ck,G:fk,H:Q3,I:W3,j:F3,L:BA,m:ek,M:tk,p:W,q:Z,Q:AS,s:ES,S:nk,u:rk,U:ak,V:ik,w:lk,W:uk,x:null,X:null,y:ok,Y:sk,Z:dk,"%":OS},k={a:ae,A:fe,b:Se,B:ye,c:null,d:wS,e:wS,f:pk,g:_k,G:jk,H:hk,I:vk,j:mk,L:UA,m:yk,M:gk,p:ft,q:ze,Q:AS,s:ES,S:bk,u:xk,U:Sk,V:wk,w:Ok,W:Ak,x:null,X:null,y:Ek,Y:Tk,Z:Mk,"%":OS},N={a:ne,A:J,b:re,B:he,c:ce,d:bS,e:bS,f:I3,g:gS,G:yS,H:xS,I:xS,j:H3,L:K3,m:$3,M:Y3,p:ee,q:q3,Q:V3,s:Z3,S:G3,u:k3,U:z3,V:B3,w:R3,W:L3,x:se,X:B,y:gS,Y:yS,Z:U3,"%":X3};j.x=P(r,j),j.X=P(a,j),j.c=P(t,j),k.x=P(r,k),k.X=P(a,k),k.c=P(t,k);function P(V,ue){return function(ge){var te=[],Fe=-1,je=0,kt=V.length,zt,dt,Uu;for(ge instanceof Date||(ge=new Date(+ge));++Fe<kt;)V.charCodeAt(Fe)===37&&(te.push(V.slice(je,Fe)),(dt=pS[zt=V.charAt(++Fe)])!=null?zt=V.charAt(++Fe):dt=zt==="e"?" ":"0",(Uu=ue[zt])&&(zt=Uu(ge,dt)),te.push(zt),je=Fe+1);return te.push(V.slice(je,Fe)),te.join("")}}function L(V,ue){return function(ge){var te=lu(1900,void 0,1),Fe=G(te,V,ge+="",0),je,kt;if(Fe!=ge.length)return null;if("Q"in te)return new Date(te.Q);if("s"in te)return new Date(te.s*1e3+("L"in te?te.L:0));if(ue&&!("Z"in te)&&(te.Z=0),"p"in te&&(te.H=te.H%12+te.p*12),te.m===void 0&&(te.m="q"in te?te.q:0),"V"in te){if(te.V<1||te.V>53)return null;"w"in te||(te.w=1),"Z"in te?(je=sv(lu(te.y,0,1)),kt=je.getUTCDay(),je=kt>4||kt===0?Gc.ceil(je):Gc(je),je=gs.offset(je,(te.V-1)*7),te.y=je.getUTCFullYear(),te.m=je.getUTCMonth(),te.d=je.getUTCDate()+(te.w+6)%7):(je=cv(lu(te.y,0,1)),kt=je.getDay(),je=kt>4||kt===0?Yc.ceil(je):Yc(je),je=Du.offset(je,(te.V-1)*7),te.y=je.getFullYear(),te.m=je.getMonth(),te.d=je.getDate()+(te.w+6)%7)}else("W"in te||"U"in te)&&("w"in te||(te.w="u"in te?te.u%7:"W"in te?1:0),kt="Z"in te?sv(lu(te.y,0,1)).getUTCDay():cv(lu(te.y,0,1)).getDay(),te.m=0,te.d="W"in te?(te.w+6)%7+te.W*7-(kt+5)%7:te.w+te.U*7-(kt+6)%7);return"Z"in te?(te.H+=te.Z/100|0,te.M+=te.Z%100,sv(te)):cv(te)}}function G(V,ue,ge,te){for(var Fe=0,je=ue.length,kt=ge.length,zt,dt;Fe<je;){if(te>=kt)return-1;if(zt=ue.charCodeAt(Fe++),zt===37){if(zt=ue.charAt(Fe++),dt=N[zt in pS?ue.charAt(Fe++):zt],!dt||(te=dt(V,ge,te))<0)return-1}else if(zt!=ge.charCodeAt(te++))return-1}return te}function ee(V,ue,ge){var te=h.exec(ue.slice(ge));return te?(V.p=m.get(te[0].toLowerCase()),ge+te[0].length):-1}function ne(V,ue,ge){var te=b.exec(ue.slice(ge));return te?(V.w=w.get(te[0].toLowerCase()),ge+te[0].length):-1}function J(V,ue,ge){var te=y.exec(ue.slice(ge));return te?(V.w=g.get(te[0].toLowerCase()),ge+te[0].length):-1}function re(V,ue,ge){var te=T.exec(ue.slice(ge));return te?(V.m=M.get(te[0].toLowerCase()),ge+te[0].length):-1}function he(V,ue,ge){var te=S.exec(ue.slice(ge));return te?(V.m=O.get(te[0].toLowerCase()),ge+te[0].length):-1}function ce(V,ue,ge){return G(V,t,ue,ge)}function se(V,ue,ge){return G(V,r,ue,ge)}function B(V,ue,ge){return G(V,a,ue,ge)}function I(V){return s[V.getDay()]}function F(V){return o[V.getDay()]}function oe(V){return v[V.getMonth()]}function D(V){return f[V.getMonth()]}function W(V){return u[+(V.getHours()>=12)]}function Z(V){return 1+~~(V.getMonth()/3)}function ae(V){return s[V.getUTCDay()]}function fe(V){return o[V.getUTCDay()]}function Se(V){return v[V.getUTCMonth()]}function ye(V){return f[V.getUTCMonth()]}function ft(V){return u[+(V.getUTCHours()>=12)]}function ze(V){return 1+~~(V.getUTCMonth()/3)}return{format:function(V){var ue=P(V+="",j);return ue.toString=function(){return V},ue},parse:function(V){var ue=L(V+="",!1);return ue.toString=function(){return V},ue},utcFormat:function(V){var ue=P(V+="",k);return ue.toString=function(){return V},ue},utcParse:function(V){var ue=L(V+="",!0);return ue.toString=function(){return V},ue}}}var pS={"-":"",_:" ",0:"0"},St=/^\s*\d+/,D3=/^%/,N3=/[\\^$*+?|[\]().{}]/g;function Ce(e,t,r){var a=e<0?"-":"",u=(a?-e:e)+"",o=u.length;return a+(o<r?new Array(r-o+1).join(t)+u:u)}function P3(e){return e.replace(N3,"\\$&")}function uu(e){return new RegExp("^(?:"+e.map(P3).join("|")+")","i")}function ou(e){return new Map(e.map((t,r)=>[t.toLowerCase(),r]))}function R3(e,t,r){var a=St.exec(t.slice(r,r+1));return a?(e.w=+a[0],r+a[0].length):-1}function k3(e,t,r){var a=St.exec(t.slice(r,r+1));return a?(e.u=+a[0],r+a[0].length):-1}function z3(e,t,r){var a=St.exec(t.slice(r,r+2));return a?(e.U=+a[0],r+a[0].length):-1}function B3(e,t,r){var a=St.exec(t.slice(r,r+2));return a?(e.V=+a[0],r+a[0].length):-1}function L3(e,t,r){var a=St.exec(t.slice(r,r+2));return a?(e.W=+a[0],r+a[0].length):-1}function yS(e,t,r){var a=St.exec(t.slice(r,r+4));return a?(e.y=+a[0],r+a[0].length):-1}function gS(e,t,r){var a=St.exec(t.slice(r,r+2));return a?(e.y=+a[0]+(+a[0]>68?1900:2e3),r+a[0].length):-1}function U3(e,t,r){var a=/^(Z)|([+-]\d\d)(?::?(\d\d))?/.exec(t.slice(r,r+6));return a?(e.Z=a[1]?0:-(a[2]+(a[3]||"00")),r+a[0].length):-1}function q3(e,t,r){var a=St.exec(t.slice(r,r+1));return a?(e.q=a[0]*3-3,r+a[0].length):-1}function $3(e,t,r){var a=St.exec(t.slice(r,r+2));return a?(e.m=a[0]-1,r+a[0].length):-1}function bS(e,t,r){var a=St.exec(t.slice(r,r+2));return a?(e.d=+a[0],r+a[0].length):-1}function H3(e,t,r){var a=St.exec(t.slice(r,r+3));return a?(e.m=0,e.d=+a[0],r+a[0].length):-1}function xS(e,t,r){var a=St.exec(t.slice(r,r+2));return a?(e.H=+a[0],r+a[0].length):-1}function Y3(e,t,r){var a=St.exec(t.slice(r,r+2));return a?(e.M=+a[0],r+a[0].length):-1}function G3(e,t,r){var a=St.exec(t.slice(r,r+2));return a?(e.S=+a[0],r+a[0].length):-1}function K3(e,t,r){var a=St.exec(t.slice(r,r+3));return a?(e.L=+a[0],r+a[0].length):-1}function I3(e,t,r){var a=St.exec(t.slice(r,r+6));return a?(e.L=Math.floor(a[0]/1e3),r+a[0].length):-1}function X3(e,t,r){var a=D3.exec(t.slice(r,r+1));return a?r+a[0].length:-1}function V3(e,t,r){var a=St.exec(t.slice(r));return a?(e.Q=+a[0],r+a[0].length):-1}function Z3(e,t,r){var a=St.exec(t.slice(r));return a?(e.s=+a[0],r+a[0].length):-1}function SS(e,t){return Ce(e.getDate(),t,2)}function Q3(e,t){return Ce(e.getHours(),t,2)}function W3(e,t){return Ce(e.getHours()%12||12,t,2)}function F3(e,t){return Ce(1+Du.count(Sr(e),e),t,3)}function BA(e,t){return Ce(e.getMilliseconds(),t,3)}function J3(e,t){return BA(e,t)+"000"}function ek(e,t){return Ce(e.getMonth()+1,t,2)}function tk(e,t){return Ce(e.getMinutes(),t,2)}function nk(e,t){return Ce(e.getSeconds(),t,2)}function rk(e){var t=e.getDay();return t===0?7:t}function ak(e,t){return Ce(bs.count(Sr(e)-1,e),t,2)}function LA(e){var t=e.getDay();return t>=4||t===0?Ui(e):Ui.ceil(e)}function ik(e,t){return e=LA(e),Ce(Ui.count(Sr(e),e)+(Sr(e).getDay()===4),t,2)}function lk(e){return e.getDay()}function uk(e,t){return Ce(Yc.count(Sr(e)-1,e),t,2)}function ok(e,t){return Ce(e.getFullYear()%100,t,2)}function ck(e,t){return e=LA(e),Ce(e.getFullYear()%100,t,2)}function sk(e,t){return Ce(e.getFullYear()%1e4,t,4)}function fk(e,t){var r=e.getDay();return e=r>=4||r===0?Ui(e):Ui.ceil(e),Ce(e.getFullYear()%1e4,t,4)}function dk(e){var t=e.getTimezoneOffset();return(t>0?"-":(t*=-1,"+"))+Ce(t/60|0,"0",2)+Ce(t%60,"0",2)}function wS(e,t){return Ce(e.getUTCDate(),t,2)}function hk(e,t){return Ce(e.getUTCHours(),t,2)}function vk(e,t){return Ce(e.getUTCHours()%12||12,t,2)}function mk(e,t){return Ce(1+gs.count(wr(e),e),t,3)}function UA(e,t){return Ce(e.getUTCMilliseconds(),t,3)}function pk(e,t){return UA(e,t)+"000"}function yk(e,t){return Ce(e.getUTCMonth()+1,t,2)}function gk(e,t){return Ce(e.getUTCMinutes(),t,2)}function bk(e,t){return Ce(e.getUTCSeconds(),t,2)}function xk(e){var t=e.getUTCDay();return t===0?7:t}function Sk(e,t){return Ce(xs.count(wr(e)-1,e),t,2)}function qA(e){var t=e.getUTCDay();return t>=4||t===0?qi(e):qi.ceil(e)}function wk(e,t){return e=qA(e),Ce(qi.count(wr(e),e)+(wr(e).getUTCDay()===4),t,2)}function Ok(e){return e.getUTCDay()}function Ak(e,t){return Ce(Gc.count(wr(e)-1,e),t,2)}function Ek(e,t){return Ce(e.getUTCFullYear()%100,t,2)}function _k(e,t){return e=qA(e),Ce(e.getUTCFullYear()%100,t,2)}function Tk(e,t){return Ce(e.getUTCFullYear()%1e4,t,4)}function jk(e,t){var r=e.getUTCDay();return e=r>=4||r===0?qi(e):qi.ceil(e),Ce(e.getUTCFullYear()%1e4,t,4)}function Mk(){return"+0000"}function OS(){return"%"}function AS(e){return+e}function ES(e){return Math.floor(+e/1e3)}var Di,$A,HA;Ck({dateTime:"%x, %X",date:"%-m/%-d/%Y",time:"%-I:%M:%S %p",periods:["AM","PM"],days:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],shortDays:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],months:["January","February","March","April","May","June","July","August","September","October","November","December"],shortMonths:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"]});function Ck(e){return Di=C3(e),$A=Di.format,Di.parse,HA=Di.utcFormat,Di.utcParse,Di}function Dk(e){return new Date(e)}function Nk(e){return e instanceof Date?+e:+new Date(+e)}function up(e,t,r,a,u,o,s,f,v,h){var m=Xm(),y=m.invert,g=m.domain,b=h(".%L"),w=h(":%S"),S=h("%I:%M"),O=h("%I %p"),T=h("%a %d"),M=h("%b %d"),j=h("%B"),k=h("%Y");function N(P){return(v(P)<P?b:f(P)<P?w:s(P)<P?S:o(P)<P?O:a(P)<P?u(P)<P?T:M:r(P)<P?j:k)(P)}return m.invert=function(P){return new Date(y(P))},m.domain=function(P){return arguments.length?g(Array.from(P,Nk)):g().map(Dk)},m.ticks=function(P){var L=g();return e(L[0],L[L.length-1],P??10)},m.tickFormat=function(P,L){return L==null?N:h(L)},m.nice=function(P){var L=g();return(!P||typeof P.range!="function")&&(P=t(L[0],L[L.length-1],P??10)),P?g(jA(L,P)):m},m.copy=function(){return Cu(m,up(e,t,r,a,u,o,s,f,v,h))},m}function Pk(){return Rn.apply(up(j3,M3,Sr,ip,bs,Du,rp,tp,Na,$A).domain([new Date(2e3,0,1),new Date(2e3,0,2)]),arguments)}function Rk(){return Rn.apply(up(_3,T3,wr,lp,xs,gs,ap,np,Na,HA).domain([Date.UTC(2e3,0,1),Date.UTC(2e3,0,2)]),arguments)}function Ss(){var e=0,t=1,r,a,u,o,s=qt,f=!1,v;function h(y){return y==null||isNaN(y=+y)?v:s(u===0?.5:(y=(o(y)-r)*u,f?Math.max(0,Math.min(1,y)):y))}h.domain=function(y){return arguments.length?([e,t]=y,r=o(e=+e),a=o(t=+t),u=r===a?0:1/(a-r),h):[e,t]},h.clamp=function(y){return arguments.length?(f=!!y,h):f},h.interpolator=function(y){return arguments.length?(s=y,h):s};function m(y){return function(g){var b,w;return arguments.length?([b,w]=g,s=y(b,w),h):[s(0),s(1)]}}return h.range=m(Zi),h.rangeRound=m(Im),h.unknown=function(y){return arguments.length?(v=y,h):v},function(y){return o=y,r=y(e),a=y(t),u=r===a?0:1/(a-r),h}}function la(e,t){return t.domain(e.domain()).interpolator(e.interpolator()).clamp(e.clamp()).unknown(e.unknown())}function YA(){var e=ia(Ss()(qt));return e.copy=function(){return la(e,YA())},Er.apply(e,arguments)}function GA(){var e=Qm(Ss()).domain([1,10]);return e.copy=function(){return la(e,GA()).base(e.base())},Er.apply(e,arguments)}function KA(){var e=Wm(Ss());return e.copy=function(){return la(e,KA()).constant(e.constant())},Er.apply(e,arguments)}function op(){var e=Fm(Ss());return e.copy=function(){return la(e,op()).exponent(e.exponent())},Er.apply(e,arguments)}function kk(){return op.apply(null,arguments).exponent(.5)}function IA(){var e=[],t=qt;function r(a){if(a!=null&&!isNaN(a=+a))return t((ju(e,a,1)-1)/(e.length-1))}return r.domain=function(a){if(!arguments.length)return e.slice();e=[];for(let u of a)u!=null&&!isNaN(u=+u)&&e.push(u);return e.sort(na),r},r.interpolator=function(a){return arguments.length?(t=a,r):t},r.range=function(){return e.map((a,u)=>t(u/(e.length-1)))},r.quantiles=function(a){return Array.from({length:a+1},(u,o)=>xR(e,o/a))},r.copy=function(){return IA(t).domain(e)},Er.apply(r,arguments)}function ws(){var e=0,t=.5,r=1,a=1,u,o,s,f,v,h=qt,m,y=!1,g;function b(S){return isNaN(S=+S)?g:(S=.5+((S=+m(S))-o)*(a*S<a*o?f:v),h(y?Math.max(0,Math.min(1,S)):S))}b.domain=function(S){return arguments.length?([e,t,r]=S,u=m(e=+e),o=m(t=+t),s=m(r=+r),f=u===o?0:.5/(o-u),v=o===s?0:.5/(s-o),a=o<u?-1:1,b):[e,t,r]},b.clamp=function(S){return arguments.length?(y=!!S,b):y},b.interpolator=function(S){return arguments.length?(h=S,b):h};function w(S){return function(O){var T,M,j;return arguments.length?([T,M,j]=O,h=XR(S,[T,M,j]),b):[h(0),h(.5),h(1)]}}return b.range=w(Zi),b.rangeRound=w(Im),b.unknown=function(S){return arguments.length?(g=S,b):g},function(S){return m=S,u=S(e),o=S(t),s=S(r),f=u===o?0:.5/(o-u),v=o===s?0:.5/(s-o),a=o<u?-1:1,b}}function XA(){var e=ia(ws()(qt));return e.copy=function(){return la(e,XA())},Er.apply(e,arguments)}function VA(){var e=Qm(ws()).domain([.1,1,10]);return e.copy=function(){return la(e,VA()).base(e.base())},Er.apply(e,arguments)}function ZA(){var e=Wm(ws());return e.copy=function(){return la(e,ZA()).constant(e.constant())},Er.apply(e,arguments)}function cp(){var e=Fm(ws());return e.copy=function(){return la(e,cp()).exponent(e.exponent())},Er.apply(e,arguments)}function zk(){return cp.apply(null,arguments).exponent(.5)}const fu=Object.freeze(Object.defineProperty({__proto__:null,scaleBand:Ym,scaleDiverging:XA,scaleDivergingLog:VA,scaleDivergingPow:cp,scaleDivergingSqrt:zk,scaleDivergingSymlog:ZA,scaleIdentity:TA,scaleImplicit:tm,scaleLinear:_A,scaleLog:MA,scaleOrdinal:Hm,scalePoint:OR,scalePow:Jm,scaleQuantile:NA,scaleQuantize:PA,scaleRadial:DA,scaleSequential:YA,scaleSequentialLog:GA,scaleSequentialPow:op,scaleSequentialQuantile:IA,scaleSequentialSqrt:kk,scaleSequentialSymlog:KA,scaleSqrt:p3,scaleSymlog:CA,scaleThreshold:RA,scaleTime:Pk,scaleUtc:Rk,tickFormat:EA},Symbol.toStringTag,{value:"Module"}));var ua=e=>e.chartData,Bk=Y([ua],e=>{var t=e.chartData!=null?e.chartData.length-1:0;return{chartData:e.chartData,computedData:e.computedData,dataEndIndex:t,dataStartIndex:0}}),sp=(e,t,r,a)=>a?Bk(e):ua(e);function $i(e){if(Array.isArray(e)&&e.length===2){var[t,r]=e;if(vn(t)&&vn(r))return!0}return!1}function _S(e,t,r){return r?e:[Math.min(e[0],t[0]),Math.max(e[1],t[1])]}function Lk(e,t){if(t&&typeof e!="function"&&Array.isArray(e)&&e.length===2){var[r,a]=e,u,o;if(vn(r))u=r;else if(typeof r=="function")return;if(vn(a))o=a;else if(typeof a=="function")return;var s=[u,o];if($i(s))return s}}function Uk(e,t,r){if(!(!r&&t==null)){if(typeof e=="function"&&t!=null)try{var a=e(t,r);if($i(a))return _S(a,t,r)}catch{}if(Array.isArray(e)&&e.length===2){var[u,o]=e,s,f;if(u==="auto")t!=null&&(s=Math.min(...t));else if(ve(u))s=u;else if(typeof u=="function")try{t!=null&&(s=u(t?.[0]))}catch{}else if(typeof u=="string"&&g1.test(u)){var v=g1.exec(u);if(v==null||t==null)s=void 0;else{var h=+v[1];s=t[0]-h}}else s=t?.[0];if(o==="auto")t!=null&&(f=Math.max(...t));else if(ve(o))f=o;else if(typeof o=="function")try{t!=null&&(f=o(t?.[1]))}catch{}else if(typeof o=="string"&&b1.test(o)){var m=b1.exec(o);if(m==null||t==null)f=void 0;else{var y=+m[1];f=t[1]+y}}else f=t?.[1];var g=[s,f];if($i(g))return t==null?g:_S(g,t,r)}}}var Qi=1e9,qk={precision:20,rounding:4,toExpNeg:-7,toExpPos:21,LN10:"2.302585092994045684017991454684364207601101488628772976033327900967572609677352480235997205089598298341967784042286"},dp,Ve=!0,Dn="[DecimalError] ",Ba=Dn+"Invalid argument: ",fp=Dn+"Exponent out of range: ",Wi=Math.floor,Ma=Math.pow,$k=/^(\d+(\.\d*)?|\.\d+)(e[+-]?\d+)?$/i,sn,yt=1e7,Ke=7,QA=9007199254740991,Kc=Wi(QA/Ke),ie={};ie.absoluteValue=ie.abs=function(){var e=new this.constructor(this);return e.s&&(e.s=1),e};ie.comparedTo=ie.cmp=function(e){var t,r,a,u,o=this;if(e=new o.constructor(e),o.s!==e.s)return o.s||-e.s;if(o.e!==e.e)return o.e>e.e^o.s<0?1:-1;for(a=o.d.length,u=e.d.length,t=0,r=a<u?a:u;t<r;++t)if(o.d[t]!==e.d[t])return o.d[t]>e.d[t]^o.s<0?1:-1;return a===u?0:a>u^o.s<0?1:-1};ie.decimalPlaces=ie.dp=function(){var e=this,t=e.d.length-1,r=(t-e.e)*Ke;if(t=e.d[t],t)for(;t%10==0;t/=10)r--;return r<0?0:r};ie.dividedBy=ie.div=function(e){return gr(this,new this.constructor(e))};ie.dividedToIntegerBy=ie.idiv=function(e){var t=this,r=t.constructor;return He(gr(t,new r(e),0,1),r.precision)};ie.equals=ie.eq=function(e){return!this.cmp(e)};ie.exponent=function(){return at(this)};ie.greaterThan=ie.gt=function(e){return this.cmp(e)>0};ie.greaterThanOrEqualTo=ie.gte=function(e){return this.cmp(e)>=0};ie.isInteger=ie.isint=function(){return this.e>this.d.length-2};ie.isNegative=ie.isneg=function(){return this.s<0};ie.isPositive=ie.ispos=function(){return this.s>0};ie.isZero=function(){return this.s===0};ie.lessThan=ie.lt=function(e){return this.cmp(e)<0};ie.lessThanOrEqualTo=ie.lte=function(e){return this.cmp(e)<1};ie.logarithm=ie.log=function(e){var t,r=this,a=r.constructor,u=a.precision,o=u+5;if(e===void 0)e=new a(10);else if(e=new a(e),e.s<1||e.eq(sn))throw Error(Dn+"NaN");if(r.s<1)throw Error(Dn+(r.s?"NaN":"-Infinity"));return r.eq(sn)?new a(0):(Ve=!1,t=gr(wu(r,o),wu(e,o),o),Ve=!0,He(t,u))};ie.minus=ie.sub=function(e){var t=this;return e=new t.constructor(e),t.s==e.s?JA(t,e):WA(t,(e.s=-e.s,e))};ie.modulo=ie.mod=function(e){var t,r=this,a=r.constructor,u=a.precision;if(e=new a(e),!e.s)throw Error(Dn+"NaN");return r.s?(Ve=!1,t=gr(r,e,0,1).times(e),Ve=!0,r.minus(t)):He(new a(r),u)};ie.naturalExponential=ie.exp=function(){return FA(this)};ie.naturalLogarithm=ie.ln=function(){return wu(this)};ie.negated=ie.neg=function(){var e=new this.constructor(this);return e.s=-e.s||0,e};ie.plus=ie.add=function(e){var t=this;return e=new t.constructor(e),t.s==e.s?WA(t,e):JA(t,(e.s=-e.s,e))};ie.precision=ie.sd=function(e){var t,r,a,u=this;if(e!==void 0&&e!==!!e&&e!==1&&e!==0)throw Error(Ba+e);if(t=at(u)+1,a=u.d.length-1,r=a*Ke+1,a=u.d[a],a){for(;a%10==0;a/=10)r--;for(a=u.d[0];a>=10;a/=10)r++}return e&&t>r?t:r};ie.squareRoot=ie.sqrt=function(){var e,t,r,a,u,o,s,f=this,v=f.constructor;if(f.s<1){if(!f.s)return new v(0);throw Error(Dn+"NaN")}for(e=at(f),Ve=!1,u=Math.sqrt(+f),u==0||u==1/0?(t=Xn(f.d),(t.length+e)%2==0&&(t+="0"),u=Math.sqrt(t),e=Wi((e+1)/2)-(e<0||e%2),u==1/0?t="5e"+e:(t=u.toExponential(),t=t.slice(0,t.indexOf("e")+1)+e),a=new v(t)):a=new v(u.toString()),r=v.precision,u=s=r+3;;)if(o=a,a=o.plus(gr(f,o,s+2)).times(.5),Xn(o.d).slice(0,s)===(t=Xn(a.d)).slice(0,s)){if(t=t.slice(s-3,s+1),u==s&&t=="4999"){if(He(o,r+1,0),o.times(o).eq(f)){a=o;break}}else if(t!="9999")break;s+=4}return Ve=!0,He(a,r)};ie.times=ie.mul=function(e){var t,r,a,u,o,s,f,v,h,m=this,y=m.constructor,g=m.d,b=(e=new y(e)).d;if(!m.s||!e.s)return new y(0);for(e.s*=m.s,r=m.e+e.e,v=g.length,h=b.length,v<h&&(o=g,g=b,b=o,s=v,v=h,h=s),o=[],s=v+h,a=s;a--;)o.push(0);for(a=h;--a>=0;){for(t=0,u=v+a;u>a;)f=o[u]+b[a]*g[u-a-1]+t,o[u--]=f%yt|0,t=f/yt|0;o[u]=(o[u]+t)%yt|0}for(;!o[--s];)o.pop();return t?++r:o.shift(),e.d=o,e.e=r,Ve?He(e,y.precision):e};ie.toDecimalPlaces=ie.todp=function(e,t){var r=this,a=r.constructor;return r=new a(r),e===void 0?r:(Qn(e,0,Qi),t===void 0?t=a.rounding:Qn(t,0,8),He(r,e+at(r)+1,t))};ie.toExponential=function(e,t){var r,a=this,u=a.constructor;return e===void 0?r=qa(a,!0):(Qn(e,0,Qi),t===void 0?t=u.rounding:Qn(t,0,8),a=He(new u(a),e+1,t),r=qa(a,!0,e+1)),r};ie.toFixed=function(e,t){var r,a,u=this,o=u.constructor;return e===void 0?qa(u):(Qn(e,0,Qi),t===void 0?t=o.rounding:Qn(t,0,8),a=He(new o(u),e+at(u)+1,t),r=qa(a.abs(),!1,e+at(a)+1),u.isneg()&&!u.isZero()?"-"+r:r)};ie.toInteger=ie.toint=function(){var e=this,t=e.constructor;return He(new t(e),at(e)+1,t.rounding)};ie.toNumber=function(){return+this};ie.toPower=ie.pow=function(e){var t,r,a,u,o,s,f=this,v=f.constructor,h=12,m=+(e=new v(e));if(!e.s)return new v(sn);if(f=new v(f),!f.s){if(e.s<1)throw Error(Dn+"Infinity");return f}if(f.eq(sn))return f;if(a=v.precision,e.eq(sn))return He(f,a);if(t=e.e,r=e.d.length-1,s=t>=r,o=f.s,s){if((r=m<0?-m:m)<=QA){for(u=new v(sn),t=Math.ceil(a/Ke+4),Ve=!1;r%2&&(u=u.times(f),jS(u.d,t)),r=Wi(r/2),r!==0;)f=f.times(f),jS(f.d,t);return Ve=!0,e.s<0?new v(sn).div(u):He(u,a)}}else if(o<0)throw Error(Dn+"NaN");return o=o<0&&e.d[Math.max(t,r)]&1?-1:1,f.s=1,Ve=!1,u=e.times(wu(f,a+h)),Ve=!0,u=FA(u),u.s=o,u};ie.toPrecision=function(e,t){var r,a,u=this,o=u.constructor;return e===void 0?(r=at(u),a=qa(u,r<=o.toExpNeg||r>=o.toExpPos)):(Qn(e,1,Qi),t===void 0?t=o.rounding:Qn(t,0,8),u=He(new o(u),e,t),r=at(u),a=qa(u,e<=r||r<=o.toExpNeg,e)),a};ie.toSignificantDigits=ie.tosd=function(e,t){var r=this,a=r.constructor;return e===void 0?(e=a.precision,t=a.rounding):(Qn(e,1,Qi),t===void 0?t=a.rounding:Qn(t,0,8)),He(new a(r),e,t)};ie.toString=ie.valueOf=ie.val=ie.toJSON=ie[Symbol.for("nodejs.util.inspect.custom")]=function(){var e=this,t=at(e),r=e.constructor;return qa(e,t<=r.toExpNeg||t>=r.toExpPos)};function WA(e,t){var r,a,u,o,s,f,v,h,m=e.constructor,y=m.precision;if(!e.s||!t.s)return t.s||(t=new m(e)),Ve?He(t,y):t;if(v=e.d,h=t.d,s=e.e,u=t.e,v=v.slice(),o=s-u,o){for(o<0?(a=v,o=-o,f=h.length):(a=h,u=s,f=v.length),s=Math.ceil(y/Ke),f=s>f?s+1:f+1,o>f&&(o=f,a.length=1),a.reverse();o--;)a.push(0);a.reverse()}for(f=v.length,o=h.length,f-o<0&&(o=f,a=h,h=v,v=a),r=0;o;)r=(v[--o]=v[o]+h[o]+r)/yt|0,v[o]%=yt;for(r&&(v.unshift(r),++u),f=v.length;v[--f]==0;)v.pop();return t.d=v,t.e=u,Ve?He(t,y):t}function Qn(e,t,r){if(e!==~~e||e<t||e>r)throw Error(Ba+e)}function Xn(e){var t,r,a,u=e.length-1,o="",s=e[0];if(u>0){for(o+=s,t=1;t<u;t++)a=e[t]+"",r=Ke-a.length,r&&(o+=Jr(r)),o+=a;s=e[t],a=s+"",r=Ke-a.length,r&&(o+=Jr(r))}else if(s===0)return"0";for(;s%10===0;)s/=10;return o+s}var gr=(function(){function e(a,u){var o,s=0,f=a.length;for(a=a.slice();f--;)o=a[f]*u+s,a[f]=o%yt|0,s=o/yt|0;return s&&a.unshift(s),a}function t(a,u,o,s){var f,v;if(o!=s)v=o>s?1:-1;else for(f=v=0;f<o;f++)if(a[f]!=u[f]){v=a[f]>u[f]?1:-1;break}return v}function r(a,u,o){for(var s=0;o--;)a[o]-=s,s=a[o]<u[o]?1:0,a[o]=s*yt+a[o]-u[o];for(;!a[0]&&a.length>1;)a.shift()}return function(a,u,o,s){var f,v,h,m,y,g,b,w,S,O,T,M,j,k,N,P,L,G,ee=a.constructor,ne=a.s==u.s?1:-1,J=a.d,re=u.d;if(!a.s)return new ee(a);if(!u.s)throw Error(Dn+"Division by zero");for(v=a.e-u.e,L=re.length,N=J.length,b=new ee(ne),w=b.d=[],h=0;re[h]==(J[h]||0);)++h;if(re[h]>(J[h]||0)&&--v,o==null?M=o=ee.precision:s?M=o+(at(a)-at(u))+1:M=o,M<0)return new ee(0);if(M=M/Ke+2|0,h=0,L==1)for(m=0,re=re[0],M++;(h<N||m)&&M--;h++)j=m*yt+(J[h]||0),w[h]=j/re|0,m=j%re|0;else{for(m=yt/(re[0]+1)|0,m>1&&(re=e(re,m),J=e(J,m),L=re.length,N=J.length),k=L,S=J.slice(0,L),O=S.length;O<L;)S[O++]=0;G=re.slice(),G.unshift(0),P=re[0],re[1]>=yt/2&&++P;do m=0,f=t(re,S,L,O),f<0?(T=S[0],L!=O&&(T=T*yt+(S[1]||0)),m=T/P|0,m>1?(m>=yt&&(m=yt-1),y=e(re,m),g=y.length,O=S.length,f=t(y,S,g,O),f==1&&(m--,r(y,L<g?G:re,g))):(m==0&&(f=m=1),y=re.slice()),g=y.length,g<O&&y.unshift(0),r(S,y,O),f==-1&&(O=S.length,f=t(re,S,L,O),f<1&&(m++,r(S,L<O?G:re,O))),O=S.length):f===0&&(m++,S=[0]),w[h++]=m,f&&S[0]?S[O++]=J[k]||0:(S=[J[k]],O=1);while((k++<N||S[0]!==void 0)&&M--)}return w[0]||w.shift(),b.e=v,He(b,s?o+at(b)+1:o)}})();function FA(e,t){var r,a,u,o,s,f,v=0,h=0,m=e.constructor,y=m.precision;if(at(e)>16)throw Error(fp+at(e));if(!e.s)return new m(sn);for(Ve=!1,f=y,s=new m(.03125);e.abs().gte(.1);)e=e.times(s),h+=5;for(a=Math.log(Ma(2,h))/Math.LN10*2+5|0,f+=a,r=u=o=new m(sn),m.precision=f;;){if(u=He(u.times(e),f),r=r.times(++v),s=o.plus(gr(u,r,f)),Xn(s.d).slice(0,f)===Xn(o.d).slice(0,f)){for(;h--;)o=He(o.times(o),f);return m.precision=y,t==null?(Ve=!0,He(o,y)):o}o=s}}function at(e){for(var t=e.e*Ke,r=e.d[0];r>=10;r/=10)t++;return t}function fv(e,t,r){if(t>e.LN10.sd())throw Ve=!0,r&&(e.precision=r),Error(Dn+"LN10 precision limit exceeded");return He(new e(e.LN10),t)}function Jr(e){for(var t="";e--;)t+="0";return t}function wu(e,t){var r,a,u,o,s,f,v,h,m,y=1,g=10,b=e,w=b.d,S=b.constructor,O=S.precision;if(b.s<1)throw Error(Dn+(b.s?"NaN":"-Infinity"));if(b.eq(sn))return new S(0);if(t==null?(Ve=!1,h=O):h=t,b.eq(10))return t==null&&(Ve=!0),fv(S,h);if(h+=g,S.precision=h,r=Xn(w),a=r.charAt(0),o=at(b),Math.abs(o)<15e14){for(;a<7&&a!=1||a==1&&r.charAt(1)>3;)b=b.times(e),r=Xn(b.d),a=r.charAt(0),y++;o=at(b),a>1?(b=new S("0."+r),o++):b=new S(a+"."+r.slice(1))}else return v=fv(S,h+2,O).times(o+""),b=wu(new S(a+"."+r.slice(1)),h-g).plus(v),S.precision=O,t==null?(Ve=!0,He(b,O)):b;for(f=s=b=gr(b.minus(sn),b.plus(sn),h),m=He(b.times(b),h),u=3;;){if(s=He(s.times(m),h),v=f.plus(gr(s,new S(u),h)),Xn(v.d).slice(0,h)===Xn(f.d).slice(0,h))return f=f.times(2),o!==0&&(f=f.plus(fv(S,h+2,O).times(o+""))),f=gr(f,new S(y),h),S.precision=O,t==null?(Ve=!0,He(f,O)):f;f=v,u+=2}}function TS(e,t){var r,a,u;for((r=t.indexOf("."))>-1&&(t=t.replace(".","")),(a=t.search(/e/i))>0?(r<0&&(r=a),r+=+t.slice(a+1),t=t.substring(0,a)):r<0&&(r=t.length),a=0;t.charCodeAt(a)===48;)++a;for(u=t.length;t.charCodeAt(u-1)===48;)--u;if(t=t.slice(a,u),t){if(u-=a,r=r-a-1,e.e=Wi(r/Ke),e.d=[],a=(r+1)%Ke,r<0&&(a+=Ke),a<u){for(a&&e.d.push(+t.slice(0,a)),u-=Ke;a<u;)e.d.push(+t.slice(a,a+=Ke));t=t.slice(a),a=Ke-t.length}else a-=u;for(;a--;)t+="0";if(e.d.push(+t),Ve&&(e.e>Kc||e.e<-Kc))throw Error(fp+r)}else e.s=0,e.e=0,e.d=[0];return e}function He(e,t,r){var a,u,o,s,f,v,h,m,y=e.d;for(s=1,o=y[0];o>=10;o/=10)s++;if(a=t-s,a<0)a+=Ke,u=t,h=y[m=0];else{if(m=Math.ceil((a+1)/Ke),o=y.length,m>=o)return e;for(h=o=y[m],s=1;o>=10;o/=10)s++;a%=Ke,u=a-Ke+s}if(r!==void 0&&(o=Ma(10,s-u-1),f=h/o%10|0,v=t<0||y[m+1]!==void 0||h%o,v=r<4?(f||v)&&(r==0||r==(e.s<0?3:2)):f>5||f==5&&(r==4||v||r==6&&(a>0?u>0?h/Ma(10,s-u):0:y[m-1])%10&1||r==(e.s<0?8:7))),t<1||!y[0])return v?(o=at(e),y.length=1,t=t-o-1,y[0]=Ma(10,(Ke-t%Ke)%Ke),e.e=Wi(-t/Ke)||0):(y.length=1,y[0]=e.e=e.s=0),e;if(a==0?(y.length=m,o=1,m--):(y.length=m+1,o=Ma(10,Ke-a),y[m]=u>0?(h/Ma(10,s-u)%Ma(10,u)|0)*o:0),v)for(;;)if(m==0){(y[0]+=o)==yt&&(y[0]=1,++e.e);break}else{if(y[m]+=o,y[m]!=yt)break;y[m--]=0,o=1}for(a=y.length;y[--a]===0;)y.pop();if(Ve&&(e.e>Kc||e.e<-Kc))throw Error(fp+at(e));return e}function JA(e,t){var r,a,u,o,s,f,v,h,m,y,g=e.constructor,b=g.precision;if(!e.s||!t.s)return t.s?t.s=-t.s:t=new g(e),Ve?He(t,b):t;if(v=e.d,y=t.d,a=t.e,h=e.e,v=v.slice(),s=h-a,s){for(m=s<0,m?(r=v,s=-s,f=y.length):(r=y,a=h,f=v.length),u=Math.max(Math.ceil(b/Ke),f)+2,s>u&&(s=u,r.length=1),r.reverse(),u=s;u--;)r.push(0);r.reverse()}else{for(u=v.length,f=y.length,m=u<f,m&&(f=u),u=0;u<f;u++)if(v[u]!=y[u]){m=v[u]<y[u];break}s=0}for(m&&(r=v,v=y,y=r,t.s=-t.s),f=v.length,u=y.length-f;u>0;--u)v[f++]=0;for(u=y.length;u>s;){if(v[--u]<y[u]){for(o=u;o&&v[--o]===0;)v[o]=yt-1;--v[o],v[u]+=yt}v[u]-=y[u]}for(;v[--f]===0;)v.pop();for(;v[0]===0;v.shift())--a;return v[0]?(t.d=v,t.e=a,Ve?He(t,b):t):new g(0)}function qa(e,t,r){var a,u=at(e),o=Xn(e.d),s=o.length;return t?(r&&(a=r-s)>0?o=o.charAt(0)+"."+o.slice(1)+Jr(a):s>1&&(o=o.charAt(0)+"."+o.slice(1)),o=o+(u<0?"e":"e+")+u):u<0?(o="0."+Jr(-u-1)+o,r&&(a=r-s)>0&&(o+=Jr(a))):u>=s?(o+=Jr(u+1-s),r&&(a=r-u-1)>0&&(o=o+"."+Jr(a))):((a=u+1)<s&&(o=o.slice(0,a)+"."+o.slice(a)),r&&(a=r-s)>0&&(u+1===s&&(o+="."),o+=Jr(a))),e.s<0?"-"+o:o}function jS(e,t){if(e.length>t)return e.length=t,!0}function eE(e){var t,r,a;function u(o){var s=this;if(!(s instanceof u))return new u(o);if(s.constructor=u,o instanceof u){s.s=o.s,s.e=o.e,s.d=(o=o.d)?o.slice():o;return}if(typeof o=="number"){if(o*0!==0)throw Error(Ba+o);if(o>0)s.s=1;else if(o<0)o=-o,s.s=-1;else{s.s=0,s.e=0,s.d=[0];return}if(o===~~o&&o<1e7){s.e=0,s.d=[o];return}return TS(s,o.toString())}else if(typeof o!="string")throw Error(Ba+o);if(o.charCodeAt(0)===45?(o=o.slice(1),s.s=-1):s.s=1,$k.test(o))TS(s,o);else throw Error(Ba+o)}if(u.prototype=ie,u.ROUND_UP=0,u.ROUND_DOWN=1,u.ROUND_CEIL=2,u.ROUND_FLOOR=3,u.ROUND_HALF_UP=4,u.ROUND_HALF_DOWN=5,u.ROUND_HALF_EVEN=6,u.ROUND_HALF_CEIL=7,u.ROUND_HALF_FLOOR=8,u.clone=eE,u.config=u.set=Hk,e===void 0&&(e={}),e)for(a=["precision","rounding","toExpNeg","toExpPos","LN10"],t=0;t<a.length;)e.hasOwnProperty(r=a[t++])||(e[r]=this[r]);return u.config(e),u}function Hk(e){if(!e||typeof e!="object")throw Error(Dn+"Object expected");var t,r,a,u=["precision",1,Qi,"rounding",0,8,"toExpNeg",-1/0,0,"toExpPos",0,1/0];for(t=0;t<u.length;t+=3)if((a=e[r=u[t]])!==void 0)if(Wi(a)===a&&a>=u[t+1]&&a<=u[t+2])this[r]=a;else throw Error(Ba+r+": "+a);if((a=e[r="LN10"])!==void 0)if(a==Math.LN10)this[r]=new this(a);else throw Error(Ba+r+": "+a);return this}var dp=eE(qk);sn=new dp(1);const Re=dp;var Yk=e=>e,tE={},nE=e=>e===tE,MS=e=>function t(){return arguments.length===0||arguments.length===1&&nE(arguments.length<=0?void 0:arguments[0])?t:e(...arguments)},rE=(e,t)=>e===1?t:MS(function(){for(var r=arguments.length,a=new Array(r),u=0;u<r;u++)a[u]=arguments[u];var o=a.filter(s=>s!==tE).length;return o>=e?t(...a):rE(e-o,MS(function(){for(var s=arguments.length,f=new Array(s),v=0;v<s;v++)f[v]=arguments[v];var h=a.map(m=>nE(m)?f.shift():m);return t(...h,...f)}))}),Os=e=>rE(e.length,e),im=(e,t)=>{for(var r=[],a=e;a<t;++a)r[a-e]=a;return r},Gk=Os((e,t)=>Array.isArray(t)?t.map(e):Object.keys(t).map(r=>t[r]).map(e)),Kk=function(){for(var t=arguments.length,r=new Array(t),a=0;a<t;a++)r[a]=arguments[a];if(!r.length)return Yk;var u=r.reverse(),o=u[0],s=u.slice(1);return function(){return s.reduce((f,v)=>v(f),o(...arguments))}},lm=e=>Array.isArray(e)?e.reverse():e.split("").reverse().join(""),aE=e=>{var t=null,r=null;return function(){for(var a=arguments.length,u=new Array(a),o=0;o<a;o++)u[o]=arguments[o];return t&&u.every((s,f)=>{var v;return s===((v=t)===null||v===void 0?void 0:v[f])})||(t=u,r=e(...u)),r}};function iE(e){var t;return e===0?t=1:t=Math.floor(new Re(e).abs().log(10).toNumber())+1,t}function lE(e,t,r){for(var a=new Re(e),u=0,o=[];a.lt(t)&&u<1e5;)o.push(a.toNumber()),a=a.add(r),u++;return o}Os((e,t,r)=>{var a=+e,u=+t;return a+r*(u-a)});Os((e,t,r)=>{var a=t-+e;return a=a||1/0,(r-e)/a});Os((e,t,r)=>{var a=t-+e;return a=a||1/0,Math.max(0,Math.min(1,(r-e)/a))});var uE=e=>{var[t,r]=e,[a,u]=[t,r];return t>r&&([a,u]=[r,t]),[a,u]},oE=(e,t,r)=>{if(e.lte(0))return new Re(0);var a=iE(e.toNumber()),u=new Re(10).pow(a),o=e.div(u),s=a!==1?.05:.1,f=new Re(Math.ceil(o.div(s).toNumber())).add(r).mul(s),v=f.mul(u);return t?new Re(v.toNumber()):new Re(Math.ceil(v.toNumber()))},Ik=(e,t,r)=>{var a=new Re(1),u=new Re(e);if(!u.isint()&&r){var o=Math.abs(e);o<1?(a=new Re(10).pow(iE(e)-1),u=new Re(Math.floor(u.div(a).toNumber())).mul(a)):o>1&&(u=new Re(Math.floor(e)))}else e===0?u=new Re(Math.floor((t-1)/2)):r||(u=new Re(Math.floor(e)));var s=Math.floor((t-1)/2),f=Kk(Gk(v=>u.add(new Re(v-s).mul(a)).toNumber()),im);return f(0,t)},cE=function(t,r,a,u){var o=arguments.length>4&&arguments[4]!==void 0?arguments[4]:0;if(!Number.isFinite((r-t)/(a-1)))return{step:new Re(0),tickMin:new Re(0),tickMax:new Re(0)};var s=oE(new Re(r).sub(t).div(a-1),u,o),f;t<=0&&r>=0?f=new Re(0):(f=new Re(t).add(r).div(2),f=f.sub(new Re(f).mod(s)));var v=Math.ceil(f.sub(t).div(s).toNumber()),h=Math.ceil(new Re(r).sub(f).div(s).toNumber()),m=v+h+1;return m>a?cE(t,r,a,u,o+1):(m<a&&(h=r>0?h+(a-m):h,v=r>0?v:v+(a-m)),{step:s,tickMin:f.sub(new Re(v).mul(s)),tickMax:f.add(new Re(h).mul(s))})};function Xk(e){var[t,r]=e,a=arguments.length>1&&arguments[1]!==void 0?arguments[1]:6,u=arguments.length>2&&arguments[2]!==void 0?arguments[2]:!0,o=Math.max(a,2),[s,f]=uE([t,r]);if(s===-1/0||f===1/0){var v=f===1/0?[s,...im(0,a-1).map(()=>1/0)]:[...im(0,a-1).map(()=>-1/0),f];return t>r?lm(v):v}if(s===f)return Ik(s,a,u);var{step:h,tickMin:m,tickMax:y}=cE(s,f,o,u,0),g=lE(m,y.add(new Re(.1).mul(h)),h);return t>r?lm(g):g}function Vk(e,t){var[r,a]=e,u=arguments.length>2&&arguments[2]!==void 0?arguments[2]:!0,[o,s]=uE([r,a]);if(o===-1/0||s===1/0)return[r,a];if(o===s)return[o];var f=Math.max(t,2),v=oE(new Re(s).sub(o).div(f-1),u,0),h=[...lE(new Re(o),new Re(s),v),s];return u===!1&&(h=h.map(m=>Math.round(m))),r>a?lm(h):h}var Zk=aE(Xk),Qk=aE(Vk),sE=e=>e.rootProps.maxBarSize,Wk=e=>e.rootProps.barGap,fE=e=>e.rootProps.barCategoryGap,Fk=e=>e.rootProps.barSize,As=e=>e.rootProps.stackOffset,hp=e=>e.options.chartName,vp=e=>e.rootProps.syncId,dE=e=>e.rootProps.syncMethod,mp=e=>e.options.eventEmitter,hr={allowDuplicatedCategory:!0,angleAxisId:0,reversed:!1,scale:"auto",tick:!0,type:"category"},cn={allowDataOverflow:!1,allowDuplicatedCategory:!0,radiusAxisId:0,scale:"auto",tick:!0,tickCount:5,type:"number"},Es=(e,t)=>{if(!(!e||!t))return e!=null&&e.reversed?[t[1],t[0]]:t},Jk={allowDataOverflow:!1,allowDecimals:!1,allowDuplicatedCategory:!1,dataKey:void 0,domain:void 0,id:hr.angleAxisId,includeHidden:!1,name:void 0,reversed:hr.reversed,scale:hr.scale,tick:hr.tick,tickCount:void 0,ticks:void 0,type:hr.type,unit:void 0},e4={allowDataOverflow:cn.allowDataOverflow,allowDecimals:!1,allowDuplicatedCategory:cn.allowDuplicatedCategory,dataKey:void 0,domain:void 0,id:cn.radiusAxisId,includeHidden:!1,name:void 0,reversed:!1,scale:cn.scale,tick:cn.tick,tickCount:cn.tickCount,ticks:void 0,type:cn.type,unit:void 0},t4={allowDataOverflow:!1,allowDecimals:!1,allowDuplicatedCategory:hr.allowDuplicatedCategory,dataKey:void 0,domain:void 0,id:hr.angleAxisId,includeHidden:!1,name:void 0,reversed:!1,scale:hr.scale,tick:hr.tick,tickCount:void 0,ticks:void 0,type:"number",unit:void 0},n4={allowDataOverflow:cn.allowDataOverflow,allowDecimals:!1,allowDuplicatedCategory:cn.allowDuplicatedCategory,dataKey:void 0,domain:void 0,id:cn.radiusAxisId,includeHidden:!1,name:void 0,reversed:!1,scale:cn.scale,tick:cn.tick,tickCount:cn.tickCount,ticks:void 0,type:"category",unit:void 0},pp=(e,t)=>e.polarAxis.angleAxis[t]!=null?e.polarAxis.angleAxis[t]:e.layout.layoutType==="radial"?t4:Jk,yp=(e,t)=>e.polarAxis.radiusAxis[t]!=null?e.polarAxis.radiusAxis[t]:e.layout.layoutType==="radial"?n4:e4,_s=e=>e.polarOptions,gp=Y([Or,Ar,xt],uN),hE=Y([_s,gp],(e,t)=>{if(e!=null)return qn(e.innerRadius,t,0)}),vE=Y([_s,gp],(e,t)=>{if(e!=null)return qn(e.outerRadius,t,t*.8)}),r4=e=>{if(e==null)return[0,0];var{startAngle:t,endAngle:r}=e;return[t,r]},mE=Y([_s],r4);Y([pp,mE],Es);var pE=Y([gp,hE,vE],(e,t,r)=>{if(!(e==null||t==null||r==null))return[t,r]});Y([yp,pE],Es);var yE=Y([ke,_s,hE,vE,Or,Ar],(e,t,r,a,u,o)=>{if(!(e!=="centric"&&e!=="radial"||t==null||r==null||a==null)){var{cx:s,cy:f,startAngle:v,endAngle:h}=t;return{cx:qn(s,u,u/2),cy:qn(f,o,o/2),innerRadius:r,outerRadius:a,startAngle:v,endAngle:h,clockWise:!1}}}),it=(e,t)=>t,Ts=(e,t,r)=>r;function bp(e){return e?.id}var st=e=>{var t=ke(e);return t==="horizontal"?"xAxis":t==="vertical"?"yAxis":t==="centric"?"angleAxis":"radiusAxis"},Fi=e=>e.tooltip.settings.axisId,wt=e=>{var t=st(e),r=Fi(e);return Nu(e,t,r)},gE=Y([wt],e=>e?.dataKey);function bE(e,t,r){var{chartData:a=[]}=t,{allowDuplicatedCategory:u,dataKey:o}=r,s=new Map;return e.forEach(f=>{var v,h=(v=f.data)!==null&&v!==void 0?v:a;if(!(h==null||h.length===0)){var m=bp(f);h.forEach((y,g)=>{var b=o==null||u?g:String($t(y,o,null)),w=$t(y,f.dataKey,0),S;s.has(b)?S=s.get(b):S={},Object.assign(S,{[m]:w}),s.set(b,S)})}}),Array.from(s.values())}function js(e){return e.stackId!=null&&e.dataKey!=null}function CS(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function Ic(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?CS(Object(r),!0).forEach(function(a){a4(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):CS(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function a4(e,t,r){return(t=i4(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function i4(e){var t=l4(e,"string");return typeof t=="symbol"?t:t+""}function l4(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var um=[0,"auto"],un={allowDataOverflow:!1,allowDecimals:!0,allowDuplicatedCategory:!0,angle:0,dataKey:void 0,domain:void 0,height:30,hide:!0,id:0,includeHidden:!1,interval:"preserveEnd",minTickGap:5,mirror:!1,name:void 0,orientation:"bottom",padding:{left:0,right:0},reversed:!1,scale:"auto",tick:!0,tickCount:5,tickFormatter:void 0,ticks:void 0,type:"category",unit:void 0},xE=(e,t)=>e.cartesianAxis.xAxis[t],_r=(e,t)=>{var r=xE(e,t);return r??un},on={allowDataOverflow:!1,allowDecimals:!0,allowDuplicatedCategory:!0,angle:0,dataKey:void 0,domain:um,hide:!0,id:0,includeHidden:!1,interval:"preserveEnd",minTickGap:5,mirror:!1,name:void 0,orientation:"left",padding:{top:0,bottom:0},reversed:!1,scale:"auto",tick:!0,tickCount:5,tickFormatter:void 0,ticks:void 0,type:"number",unit:void 0,width:Tu},SE=(e,t)=>e.cartesianAxis.yAxis[t],Tr=(e,t)=>{var r=SE(e,t);return r??on},u4={domain:[0,"auto"],includeHidden:!1,reversed:!1,allowDataOverflow:!1,allowDuplicatedCategory:!1,dataKey:void 0,id:0,name:"",range:[64,64],scale:"auto",type:"number",unit:""},xp=(e,t)=>{var r=e.cartesianAxis.zAxis[t];return r??u4},Qt=(e,t,r)=>{switch(t){case"xAxis":return _r(e,r);case"yAxis":return Tr(e,r);case"zAxis":return xp(e,r);case"angleAxis":return pp(e,r);case"radiusAxis":return yp(e,r);default:throw new Error("Unexpected axis type: ".concat(t))}},o4=(e,t,r)=>{switch(t){case"xAxis":return _r(e,r);case"yAxis":return Tr(e,r);default:throw new Error("Unexpected axis type: ".concat(t))}},Nu=(e,t,r)=>{switch(t){case"xAxis":return _r(e,r);case"yAxis":return Tr(e,r);case"angleAxis":return pp(e,r);case"radiusAxis":return yp(e,r);default:throw new Error("Unexpected axis type: ".concat(t))}},wE=e=>e.graphicalItems.cartesianItems.some(t=>t.type==="bar")||e.graphicalItems.polarItems.some(t=>t.type==="radialBar");function OE(e,t){return r=>{switch(e){case"xAxis":return"xAxisId"in r&&r.xAxisId===t;case"yAxis":return"yAxisId"in r&&r.yAxisId===t;case"zAxis":return"zAxisId"in r&&r.zAxisId===t;case"angleAxis":return"angleAxisId"in r&&r.angleAxisId===t;case"radiusAxis":return"radiusAxisId"in r&&r.radiusAxisId===t;default:return!1}}}var Sp=e=>e.graphicalItems.cartesianItems,c4=Y([it,Ts],OE),AE=(e,t,r)=>e.filter(r).filter(a=>t?.includeHidden===!0?!0:!a.hide),Pu=Y([Sp,Qt,c4],AE),EE=Y([Pu],e=>e.filter(t=>t.type==="area"||t.type==="bar").filter(js)),_E=e=>e.filter(t=>!("stackId"in t)||t.stackId===void 0),TE=Y([Pu],_E),jE=e=>e.map(t=>t.data).filter(Boolean).flat(1),s4=Y([Pu],jE),ME=(e,t)=>{var{chartData:r=[],dataStartIndex:a,dataEndIndex:u}=t;return e.length>0?e:r.slice(a,u+1)},wp=Y([s4,sp],ME),CE=(e,t,r)=>t?.dataKey!=null?e.map(a=>({value:$t(a,t.dataKey)})):r.length>0?r.map(a=>a.dataKey).flatMap(a=>e.map(u=>({value:$t(u,a)}))):e.map(a=>({value:a})),Ms=Y([wp,Qt,Pu],CE);function DE(e,t){switch(e){case"xAxis":return t.direction==="x";case"yAxis":return t.direction==="y";default:return!1}}function Ka(e){return e.filter(t=>Zn(t)||t instanceof Date).map(Number).filter(t=>dn(t)===!1)}function f4(e,t,r){return!r||typeof t!="number"||dn(t)?[]:r.length?Ka(r.flatMap(a=>{var u=$t(e,a.dataKey),o,s;if(Array.isArray(u)?[o,s]=u:o=s=u,!(!vn(o)||!vn(s)))return[t-o,t+s]})):[]}var d4=Y([EE,sp,wt],bE),NE=(e,t,r)=>{var a={},u=t.reduce((o,s)=>(s.stackId==null||(o[s.stackId]==null&&(o[s.stackId]=[]),o[s.stackId].push(s)),o),a);return Object.fromEntries(Object.entries(u).map(o=>{var[s,f]=o,v=f.map(bp);return[s,{stackedData:ON(e,v,r),graphicalItems:f}]}))},om=Y([d4,EE,As],NE),PE=(e,t,r)=>{var{dataStartIndex:a,dataEndIndex:u}=t;if(r!=="zAxis"){var o=jN(e,a,u);if(!(o!=null&&o[0]===0&&o[1]===0))return o}},h4=Y([om,ua,it],PE),RE=(e,t,r,a,u)=>r.length>0?e.flatMap(o=>r.flatMap(s=>{var f,v,h=(f=a[s.id])===null||f===void 0?void 0:f.filter(y=>DE(u,y)),m=$t(o,(v=t.dataKey)!==null&&v!==void 0?v:s.dataKey);return{value:m,errorDomain:f4(o,m,h)}})).filter(Boolean):t?.dataKey!=null?e.map(o=>({value:$t(o,t.dataKey),errorDomain:[]})):e.map(o=>({value:o,errorDomain:[]})),Cs=e=>e.errorBars,kE=(e,t,r)=>e.flatMap(a=>t[a.id]).filter(Boolean).filter(a=>DE(r,a));Y([TE,Cs,it],kE);var v4=Y([wp,Qt,TE,Cs,it],RE);function m4(e){var{value:t}=e;if(Zn(t)||t instanceof Date)return t}var DS=e=>{var t=e.flatMap(a=>[a.value,a.errorDomain]).flat(1),r=Ka(t);if(r.length!==0)return[Math.min(...r),Math.max(...r)]},p4=(e,t,r)=>{var a=e.map(m4).filter(u=>u!=null);return r&&(t.dataKey==null||t.allowDuplicatedCategory&&Xw(a))?mA(0,e.length):t.allowDuplicatedCategory?a:Array.from(new Set(a))},Op=e=>{var t;if(e==null||!("domain"in e))return um;if(e.domain!=null)return e.domain;if(e.ticks!=null){if(e.type==="number"){var r=Ka(e.ticks);return[Math.min(...r),Math.max(...r)]}if(e.type==="category")return e.ticks.map(String)}return(t=e?.domain)!==null&&t!==void 0?t:um},Xc=function(){for(var t=arguments.length,r=new Array(t),a=0;a<t;a++)r[a]=arguments[a];var u=r.filter(Boolean);if(u.length!==0){var o=u.flat(),s=Math.min(...o),f=Math.max(...o);return[s,f]}},zE=e=>e.referenceElements.dots,Ji=(e,t,r)=>e.filter(a=>a.ifOverflow==="extendDomain").filter(a=>t==="xAxis"?a.xAxisId===r:a.yAxisId===r),y4=Y([zE,it,Ts],Ji),BE=e=>e.referenceElements.areas,g4=Y([BE,it,Ts],Ji),LE=e=>e.referenceElements.lines,b4=Y([LE,it,Ts],Ji),UE=(e,t)=>{var r=Ka(e.map(a=>t==="xAxis"?a.x:a.y));if(r.length!==0)return[Math.min(...r),Math.max(...r)]},x4=Y(y4,it,UE),qE=(e,t)=>{var r=Ka(e.flatMap(a=>[t==="xAxis"?a.x1:a.y1,t==="xAxis"?a.x2:a.y2]));if(r.length!==0)return[Math.min(...r),Math.max(...r)]},S4=Y([g4,it],qE),$E=(e,t)=>{var r=Ka(e.map(a=>t==="xAxis"?a.x:a.y));if(r.length!==0)return[Math.min(...r),Math.max(...r)]},w4=Y(b4,it,$E),O4=Y(x4,w4,S4,(e,t,r)=>Xc(e,r,t)),A4=Y([Qt],Op),HE=(e,t,r,a,u,o,s)=>{var f=Lk(t,e.allowDataOverflow);if(f!=null)return f;var v=o==="vertical"&&s==="xAxis"||o==="horizontal"&&s==="yAxis",h=v?Xc(r,u,DS(a)):Xc(u,DS(a));return Uk(t,h,e.allowDataOverflow)},E4=Y([Qt,A4,h4,v4,O4,ke,it],HE),_4=[0,1],YE=(e,t,r,a,u,o,s)=>{if(!((e==null||r==null||r.length===0)&&s===void 0)){var{dataKey:f,type:v}=e,h=Ha(t,o);return h&&f==null?mA(0,r.length):v==="category"?p4(a,e,h):u==="expand"?_4:s}},Ap=Y([Qt,ke,wp,Ms,As,it,E4],YE),GE=(e,t,r,a,u)=>{if(e!=null){var{scale:o,type:s}=e;if(o==="auto")return t==="radial"&&u==="radiusAxis"?"band":t==="radial"&&u==="angleAxis"?"linear":s==="category"&&a&&(a.indexOf("LineChart")>=0||a.indexOf("AreaChart")>=0||a.indexOf("ComposedChart")>=0&&!r)?"point":s==="category"?"band":"linear";if(typeof o=="string"){var f="scale".concat(Au(o));return f in fu?f:"point"}}},Ru=Y([Qt,ke,wE,hp,it],GE);function T4(e){if(e!=null){if(e in fu)return fu[e]();var t="scale".concat(Au(e));if(t in fu)return fu[t]()}}function Ep(e,t,r,a){if(!(r==null||a==null)){if(typeof e.scale=="function")return e.scale.copy().domain(r).range(a);var u=T4(t);if(u!=null){var o=u.domain(r).range(a);return gN(o),o}}}var KE=(e,t,r)=>{var a=Op(t);if(!(r!=="auto"&&r!=="linear")){if(t!=null&&t.tickCount&&Array.isArray(a)&&(a[0]==="auto"||a[1]==="auto")&&$i(e))return Zk(e,t.tickCount,t.allowDecimals);if(t!=null&&t.tickCount&&t.type==="number"&&$i(e))return Qk(e,t.tickCount,t.allowDecimals)}},_p=Y([Ap,Nu,Ru],KE),IE=(e,t,r,a)=>{if(a!=="angleAxis"&&e?.type==="number"&&$i(t)&&Array.isArray(r)&&r.length>0){var u=t[0],o=r[0],s=t[1],f=r[r.length-1];return[Math.min(u,o),Math.max(s,f)]}return t},j4=Y([Qt,Ap,_p,it],IE),M4=Y(Ms,Qt,(e,t)=>{if(!(!t||t.type!=="number")){var r=1/0,a=Array.from(Ka(e.map(f=>f.value))).sort((f,v)=>f-v);if(a.length<2)return 1/0;var u=a[a.length-1]-a[0];if(u===0)return 1/0;for(var o=0;o<a.length-1;o++){var s=a[o+1]-a[o];r=Math.min(r,s)}return r/u}}),XE=Y(M4,ke,fE,xt,(e,t,r,a)=>a,(e,t,r,a,u)=>{if(!vn(e))return 0;var o=t==="vertical"?a.height:a.width;if(u==="gap")return e*o/2;if(u==="no-gap"){var s=qn(r,e*o),f=e*o/2;return f-s-(f-s)/o*s}return 0}),C4=(e,t)=>{var r=_r(e,t);return r==null||typeof r.padding!="string"?0:XE(e,"xAxis",t,r.padding)},D4=(e,t)=>{var r=Tr(e,t);return r==null||typeof r.padding!="string"?0:XE(e,"yAxis",t,r.padding)},N4=Y(_r,C4,(e,t)=>{var r,a;if(e==null)return{left:0,right:0};var{padding:u}=e;return typeof u=="string"?{left:t,right:t}:{left:((r=u.left)!==null&&r!==void 0?r:0)+t,right:((a=u.right)!==null&&a!==void 0?a:0)+t}}),P4=Y(Tr,D4,(e,t)=>{var r,a;if(e==null)return{top:0,bottom:0};var{padding:u}=e;return typeof u=="string"?{top:t,bottom:t}:{top:((r=u.top)!==null&&r!==void 0?r:0)+t,bottom:((a=u.bottom)!==null&&a!==void 0?a:0)+t}}),R4=Y([xt,N4,ps,ms,(e,t,r)=>r],(e,t,r,a,u)=>{var{padding:o}=a;return u?[o.left,r.width-o.right]:[e.left+t.left,e.left+e.width-t.right]}),k4=Y([xt,ke,P4,ps,ms,(e,t,r)=>r],(e,t,r,a,u,o)=>{var{padding:s}=u;return o?[a.height-s.bottom,s.top]:t==="horizontal"?[e.top+e.height-r.bottom,e.top+r.top]:[e.top+r.top,e.top+e.height-r.bottom]}),ku=(e,t,r,a)=>{var u;switch(t){case"xAxis":return R4(e,r,a);case"yAxis":return k4(e,r,a);case"zAxis":return(u=xp(e,r))===null||u===void 0?void 0:u.range;case"angleAxis":return mE(e);case"radiusAxis":return pE(e,r);default:return}},VE=Y([Qt,ku],Es),el=Y([Qt,Ru,j4,VE],Ep);Y([Pu,Cs,it],kE);function ZE(e,t){return e.id<t.id?-1:e.id>t.id?1:0}var Ds=(e,t)=>t,Ns=(e,t,r)=>r,z4=Y(hs,Ds,Ns,(e,t,r)=>e.filter(a=>a.orientation===t).filter(a=>a.mirror===r).sort(ZE)),B4=Y(vs,Ds,Ns,(e,t,r)=>e.filter(a=>a.orientation===t).filter(a=>a.mirror===r).sort(ZE)),QE=(e,t)=>({width:e.width,height:t.height}),L4=(e,t)=>{var r=typeof t.width=="number"?t.width:Tu;return{width:r,height:e.height}},WE=Y(xt,_r,QE),U4=(e,t,r)=>{switch(t){case"top":return e.top;case"bottom":return r-e.bottom;default:return 0}},q4=(e,t,r)=>{switch(t){case"left":return e.left;case"right":return r-e.right;default:return 0}},$4=Y(Ar,xt,z4,Ds,Ns,(e,t,r,a,u)=>{var o={},s;return r.forEach(f=>{var v=QE(t,f);s==null&&(s=U4(t,a,e));var h=a==="top"&&!u||a==="bottom"&&u;o[f.id]=s-Number(h)*v.height,s+=(h?-1:1)*v.height}),o}),H4=Y(Or,xt,B4,Ds,Ns,(e,t,r,a,u)=>{var o={},s;return r.forEach(f=>{var v=L4(t,f);s==null&&(s=q4(t,a,e));var h=a==="left"&&!u||a==="right"&&u;o[f.id]=s-Number(h)*v.width,s+=(h?-1:1)*v.width}),o}),Y4=(e,t)=>{var r=_r(e,t);if(r!=null)return $4(e,r.orientation,r.mirror)},G4=Y([xt,_r,Y4,(e,t)=>t],(e,t,r,a)=>{if(t!=null){var u=r?.[a];return u==null?{x:e.left,y:0}:{x:e.left,y:u}}}),K4=(e,t)=>{var r=Tr(e,t);if(r!=null)return H4(e,r.orientation,r.mirror)},I4=Y([xt,Tr,K4,(e,t)=>t],(e,t,r,a)=>{if(t!=null){var u=r?.[a];return u==null?{x:0,y:e.top}:{x:u,y:e.top}}}),FE=Y(xt,Tr,(e,t)=>{var r=typeof t.width=="number"?t.width:Tu;return{width:r,height:e.height}}),NS=(e,t,r)=>{switch(t){case"xAxis":return WE(e,r).width;case"yAxis":return FE(e,r).height;default:return}},JE=(e,t,r,a)=>{if(r!=null){var{allowDuplicatedCategory:u,type:o,dataKey:s}=r,f=Ha(e,a),v=t.map(h=>h.value);if(s&&f&&o==="category"&&u&&Xw(v))return v}},Tp=Y([ke,Ms,Qt,it],JE),e2=(e,t,r,a)=>{if(!(r==null||r.dataKey==null)){var{type:u,scale:o}=r,s=Ha(e,a);if(s&&(u==="number"||o!=="auto"))return t.map(f=>f.value)}},jp=Y([ke,Ms,Nu,it],e2),PS=Y([ke,o4,Ru,el,Tp,jp,ku,_p,it],(e,t,r,a,u,o,s,f,v)=>{if(t==null)return null;var h=Ha(e,v);return{angle:t.angle,interval:t.interval,minTickGap:t.minTickGap,orientation:t.orientation,tick:t.tick,tickCount:t.tickCount,tickFormatter:t.tickFormatter,ticks:t.ticks,type:t.type,unit:t.unit,axisType:v,categoricalDomain:o,duplicateDomain:u,isCategorical:h,niceTicks:f,range:s,realScaleType:r,scale:a}}),X4=(e,t,r,a,u,o,s,f,v)=>{if(!(t==null||a==null)){var h=Ha(e,v),{type:m,ticks:y,tickCount:g}=t,b=r==="scaleBand"&&typeof a.bandwidth=="function"?a.bandwidth()/2:2,w=m==="category"&&a.bandwidth?a.bandwidth()/b:0;w=v==="angleAxis"&&o!=null&&o.length>=2?Vt(o[0]-o[1])*2*w:w;var S=y||u;if(S){var O=S.map((T,M)=>{var j=s?s.indexOf(T):T;return{index:M,coordinate:a(j)+w,value:T,offset:w}});return O.filter(T=>!dn(T.coordinate))}return h&&f?f.map((T,M)=>({coordinate:a(T)+w,value:T,index:M,offset:w})):a.ticks?a.ticks(g).map(T=>({coordinate:a(T)+w,value:T,offset:w})):a.domain().map((T,M)=>({coordinate:a(T)+w,value:s?s[T]:T,index:M,offset:w}))}},t2=Y([ke,Nu,Ru,el,_p,ku,Tp,jp,it],X4),V4=(e,t,r,a,u,o,s)=>{if(!(t==null||r==null||a==null||a[0]===a[1])){var f=Ha(e,s),{tickCount:v}=t,h=0;return h=s==="angleAxis"&&a?.length>=2?Vt(a[0]-a[1])*2*h:h,f&&o?o.map((m,y)=>({coordinate:r(m)+h,value:m,index:y,offset:h})):r.ticks?r.ticks(v).map(m=>({coordinate:r(m)+h,value:m,offset:h})):r.domain().map((m,y)=>({coordinate:r(m)+h,value:u?u[m]:m,index:y,offset:h}))}},Hi=Y([ke,Nu,el,ku,Tp,jp,it],V4),Yi=Y(Qt,el,(e,t)=>{if(!(e==null||t==null))return Ic(Ic({},e),{},{scale:t})}),Z4=Y([Qt,Ru,Ap,VE],Ep);Y((e,t,r)=>xp(e,r),Z4,(e,t)=>{if(!(e==null||t==null))return Ic(Ic({},e),{},{scale:t})});var Q4=Y([ke,hs,vs],(e,t,r)=>{switch(e){case"horizontal":return t.some(a=>a.reversed)?"right-to-left":"left-to-right";case"vertical":return r.some(a=>a.reversed)?"bottom-to-top":"top-to-bottom";case"centric":case"radial":return"left-to-right";default:return}}),n2=e=>e.options.defaultTooltipEventType,r2=e=>e.options.validateTooltipEventTypes;function a2(e,t,r){if(e==null)return t;var a=e?"axis":"item";return r==null?t:r.includes(a)?a:t}function Mp(e,t){var r=n2(e),a=r2(e);return a2(t,r,a)}function W4(e){return me(t=>Mp(t,e))}var i2=(e,t)=>{var r,a=Number(t);if(!(dn(a)||t==null))return a>=0?e==null||(r=e[a])===null||r===void 0?void 0:r.value:void 0},F4=e=>e.tooltip.settings,ta={active:!1,index:null,dataKey:void 0,coordinate:void 0},J4={itemInteraction:{click:ta,hover:ta},axisInteraction:{click:ta,hover:ta},keyboardInteraction:ta,syncInteraction:{active:!1,index:null,dataKey:void 0,label:void 0,coordinate:void 0},tooltipItemPayloads:[],settings:{shared:void 0,trigger:"hover",axisId:0,active:!1,defaultIndex:void 0}},l2=mn({name:"tooltip",initialState:J4,reducers:{addTooltipEntrySettings(e,t){e.tooltipItemPayloads.push(t.payload)},removeTooltipEntrySettings(e,t){var r=yr(e).tooltipItemPayloads.indexOf(t.payload);r>-1&&e.tooltipItemPayloads.splice(r,1)},setTooltipSettingsState(e,t){e.settings=t.payload},setActiveMouseOverItemIndex(e,t){e.syncInteraction.active=!1,e.keyboardInteraction.active=!1,e.itemInteraction.hover.active=!0,e.itemInteraction.hover.index=t.payload.activeIndex,e.itemInteraction.hover.dataKey=t.payload.activeDataKey,e.itemInteraction.hover.coordinate=t.payload.activeCoordinate},mouseLeaveChart(e){e.itemInteraction.hover.active=!1,e.axisInteraction.hover.active=!1},mouseLeaveItem(e){e.itemInteraction.hover.active=!1},setActiveClickItemIndex(e,t){e.syncInteraction.active=!1,e.itemInteraction.click.active=!0,e.keyboardInteraction.active=!1,e.itemInteraction.click.index=t.payload.activeIndex,e.itemInteraction.click.dataKey=t.payload.activeDataKey,e.itemInteraction.click.coordinate=t.payload.activeCoordinate},setMouseOverAxisIndex(e,t){e.syncInteraction.active=!1,e.axisInteraction.hover.active=!0,e.keyboardInteraction.active=!1,e.axisInteraction.hover.index=t.payload.activeIndex,e.axisInteraction.hover.dataKey=t.payload.activeDataKey,e.axisInteraction.hover.coordinate=t.payload.activeCoordinate},setMouseClickAxisIndex(e,t){e.syncInteraction.active=!1,e.keyboardInteraction.active=!1,e.axisInteraction.click.active=!0,e.axisInteraction.click.index=t.payload.activeIndex,e.axisInteraction.click.dataKey=t.payload.activeDataKey,e.axisInteraction.click.coordinate=t.payload.activeCoordinate},setSyncInteraction(e,t){e.syncInteraction=t.payload},setKeyboardInteraction(e,t){e.keyboardInteraction.active=t.payload.active,e.keyboardInteraction.index=t.payload.activeIndex,e.keyboardInteraction.coordinate=t.payload.activeCoordinate,e.keyboardInteraction.dataKey=t.payload.activeDataKey}}}),{addTooltipEntrySettings:e6,removeTooltipEntrySettings:t6,setTooltipSettingsState:n6,setActiveMouseOverItemIndex:u2,mouseLeaveItem:r6,mouseLeaveChart:o2,setActiveClickItemIndex:a6,setMouseOverAxisIndex:c2,setMouseClickAxisIndex:i6,setSyncInteraction:cm,setKeyboardInteraction:sm}=l2.actions,l6=l2.reducer;function RS(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function vc(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?RS(Object(r),!0).forEach(function(a){u6(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):RS(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function u6(e,t,r){return(t=o6(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function o6(e){var t=c6(e,"string");return typeof t=="symbol"?t:t+""}function c6(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function s6(e,t,r){return t==="axis"?r==="click"?e.axisInteraction.click:e.axisInteraction.hover:r==="click"?e.itemInteraction.click:e.itemInteraction.hover}function f6(e){return e.index!=null}var s2=(e,t,r,a)=>{if(t==null)return ta;var u=s6(e,t,r);if(u==null)return ta;if(u.active)return u;if(e.keyboardInteraction.active)return e.keyboardInteraction;if(e.syncInteraction.active&&e.syncInteraction.index!=null)return e.syncInteraction;var o=e.settings.active===!0;if(f6(u)){if(o)return vc(vc({},u),{},{active:!0})}else if(a!=null)return{active:!0,coordinate:void 0,dataKey:void 0,index:a};return vc(vc({},ta),{},{coordinate:u.coordinate})},Cp=(e,t)=>{var r=e?.index;if(r==null)return null;var a=Number(r);if(!vn(a))return r;var u=0,o=1/0;return t.length>0&&(o=t.length-1),String(Math.max(u,Math.min(a,o)))},f2=(e,t,r,a,u,o,s,f)=>{if(!(o==null||f==null)){var v=s[0],h=v==null?void 0:f(v.positions,o);if(h!=null)return h;var m=u?.[Number(o)];if(m)switch(r){case"horizontal":return{x:m.coordinate,y:(a.top+t)/2};default:return{x:(a.left+e)/2,y:m.coordinate}}}},d2=(e,t,r,a)=>{if(t==="axis")return e.tooltipItemPayloads;if(e.tooltipItemPayloads.length===0)return[];var u;return r==="hover"?u=e.itemInteraction.hover.dataKey:u=e.itemInteraction.click.dataKey,u==null&&a!=null?[e.tooltipItemPayloads[0]]:e.tooltipItemPayloads.filter(o=>{var s;return((s=o.settings)===null||s===void 0?void 0:s.dataKey)===u})},zu=e=>e.options.tooltipPayloadSearcher,tl=e=>e.tooltip;function kS(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function zS(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?kS(Object(r),!0).forEach(function(a){d6(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):kS(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function d6(e,t,r){return(t=h6(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function h6(e){var t=v6(e,"string");return typeof t=="symbol"?t:t+""}function v6(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function m6(e,t){return e??t}var h2=(e,t,r,a,u,o,s)=>{if(!(t==null||o==null)){var{chartData:f,computedData:v,dataStartIndex:h,dataEndIndex:m}=r,y=[];return e.reduce((g,b)=>{var w,{dataDefinedOnItem:S,settings:O}=b,T=m6(S,f),M=Array.isArray(T)?WO(T,h,m):T,j=(w=O?.dataKey)!==null&&w!==void 0?w:a,k=O?.nameKey,N;if(a&&Array.isArray(M)&&!Array.isArray(M[0])&&s==="axis"?N=uM(M,a,u):N=o(M,t,v,k),Array.isArray(N))N.forEach(L=>{var G=zS(zS({},O),{},{name:L.name,unit:L.unit,color:void 0,fill:void 0});g.push(x1({tooltipEntrySettings:G,dataKey:L.dataKey,payload:L.payload,value:$t(L.payload,L.dataKey),name:L.name}))});else{var P;g.push(x1({tooltipEntrySettings:O,dataKey:j,payload:N,value:$t(N,j),name:(P=$t(N,k))!==null&&P!==void 0?P:O?.name}))}return g},y)}},Dp=Y([wt,ke,wE,hp,st],GE),p6=Y([e=>e.graphicalItems.cartesianItems,e=>e.graphicalItems.polarItems],(e,t)=>[...e,...t]),y6=Y([st,Fi],OE),Bu=Y([p6,wt,y6],AE),g6=Y([Bu],e=>e.filter(js)),b6=Y([Bu],jE),nl=Y([b6,ua],ME),x6=Y([g6,ua,wt],bE),Np=Y([nl,wt,Bu],CE),S6=Y([wt],Op),w6=Y([Bu],e=>e.filter(js)),O6=Y([x6,w6,As],NE),A6=Y([O6,ua,st],PE),E6=Y([Bu],_E),_6=Y([nl,wt,E6,Cs,st],RE),T6=Y([zE,st,Fi],Ji),j6=Y([T6,st],UE),M6=Y([BE,st,Fi],Ji),C6=Y([M6,st],qE),D6=Y([LE,st,Fi],Ji),N6=Y([D6,st],$E),P6=Y([j6,N6,C6],Xc),R6=Y([wt,S6,A6,_6,P6,ke,st],HE),v2=Y([wt,ke,nl,Np,As,st,R6],YE),k6=Y([v2,wt,Dp],KE),z6=Y([wt,v2,k6,st],IE),m2=e=>{var t=st(e),r=Fi(e),a=!1;return ku(e,t,r,a)},p2=Y([wt,m2],Es),y2=Y([wt,Dp,z6,p2],Ep),B6=Y([ke,Np,wt,st],JE),L6=Y([ke,Np,wt,st],e2),U6=(e,t,r,a,u,o,s,f)=>{if(t){var{type:v}=t,h=Ha(e,f);if(a){var m=r==="scaleBand"&&a.bandwidth?a.bandwidth()/2:2,y=v==="category"&&a.bandwidth?a.bandwidth()/m:0;return y=f==="angleAxis"&&u!=null&&u?.length>=2?Vt(u[0]-u[1])*2*y:y,h&&s?s.map((g,b)=>({coordinate:a(g)+y,value:g,index:b,offset:y})):a.domain().map((g,b)=>({coordinate:a(g)+y,value:o?o[g]:g,index:b,offset:y}))}}},jr=Y([ke,wt,Dp,y2,m2,B6,L6,st],U6),Pp=Y([n2,r2,F4],(e,t,r)=>a2(r.shared,e,t)),g2=e=>e.tooltip.settings.trigger,Rp=e=>e.tooltip.settings.defaultIndex,Ps=Y([tl,Pp,g2,Rp],s2),Gi=Y([Ps,nl],Cp),b2=Y([jr,Gi],i2),x2=Y([Ps],e=>{if(e)return e.dataKey}),S2=Y([tl,Pp,g2,Rp],d2),q6=Y([Or,Ar,ke,xt,jr,Rp,S2,zu],f2),$6=Y([Ps,q6],(e,t)=>e!=null&&e.coordinate?e.coordinate:t),H6=Y([Ps],e=>e.active),Y6=Y([S2,Gi,ua,gE,b2,zu,Pp],h2);Y([Y6],e=>{if(e!=null){var t=e.map(r=>r.payload).filter(r=>r!=null);return Array.from(new Set(t))}});function BS(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function LS(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?BS(Object(r),!0).forEach(function(a){G6(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):BS(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function G6(e,t,r){return(t=K6(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function K6(e){var t=I6(e,"string");return typeof t=="symbol"?t:t+""}function I6(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var X6=()=>me(wt),V6=()=>{var e=X6(),t=me(jr),r=me(y2);return km(LS(LS({},e),{},{scale:r}),t)},Z6=()=>me(hp),kp=(e,t)=>t,w2=(e,t,r)=>r,zp=(e,t,r,a)=>a,Q6=Y(jr,e=>us(e,t=>t.coordinate)),Bp=Y([tl,kp,w2,zp],s2),O2=Y([Bp,nl],Cp),W6=(e,t,r)=>{if(t!=null){var a=tl(e);return t==="axis"?r==="hover"?a.axisInteraction.hover.dataKey:a.axisInteraction.click.dataKey:r==="hover"?a.itemInteraction.hover.dataKey:a.itemInteraction.click.dataKey}},A2=Y([tl,kp,w2,zp],d2),Vc=Y([Or,Ar,ke,xt,jr,zp,A2,zu],f2),F6=Y([Bp,Vc],(e,t)=>{var r;return(r=e.coordinate)!==null&&r!==void 0?r:t}),E2=Y(jr,O2,i2),J6=Y([A2,O2,ua,gE,E2,zu,kp],h2),ez=Y([Bp],e=>({isActive:e.active,activeIndex:e.index})),tz=(e,t,r,a,u,o,s,f)=>{if(!(!e||!t||!a||!u||!o)){var v=MN(e.chartX,e.chartY,t,r,f);if(v){var h=DN(v,t),m=pN(h,s,o,a,u),y=CN(t,o,m,v);return{activeIndex:String(m),activeCoordinate:y}}}};function fm(){return fm=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},fm.apply(null,arguments)}function US(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function mc(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?US(Object(r),!0).forEach(function(a){nz(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):US(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function nz(e,t,r){return(t=rz(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function rz(e){var t=az(e,"string");return typeof t=="symbol"?t:t+""}function az(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function iz(e){var{coordinate:t,payload:r,index:a,offset:u,tooltipAxisBandSize:o,layout:s,cursor:f,tooltipEventType:v,chartName:h}=e,m=t,y=r,g=a;if(!f||!m||h!=="ScatterChart"&&v!=="axis")return null;var b,w;if(h==="ScatterChart")b=m,w=OP;else if(h==="BarChart")b=AP(s,m,u,o),w=fA;else if(s==="radial"){var{cx:S,cy:O,radius:T,startAngle:M,endAngle:j}=dA(m);b={cx:S,cy:O,startAngle:M,endAngle:j,innerRadius:T,outerRadius:T},w=vA}else b={points:rR(s,m,u)},w=vP;var k=typeof f=="object"&&"className"in f?f.className:void 0,N=mc(mc(mc(mc({stroke:"#ccc",pointerEvents:"none"},u),b),Mt(f,!1)),{},{payload:y,payloadIndex:g,className:Ie("recharts-tooltip-cursor",k)});return A.isValidElement(f)?A.cloneElement(f,N):A.createElement(w,N)}function lz(e){var t=V6(),r=nA(),a=Lm(),u=Z6();return A.createElement(iz,fm({},e,{coordinate:e.coordinate,index:e.index,payload:e.payload,offset:r,layout:a,tooltipAxisBandSize:t,chartName:u}))}var _2=A.createContext(null),uz=()=>A.useContext(_2),dv={exports:{}},qS;function oz(){return qS||(qS=1,(function(e){var t=Object.prototype.hasOwnProperty,r="~";function a(){}Object.create&&(a.prototype=Object.create(null),new a().__proto__||(r=!1));function u(v,h,m){this.fn=v,this.context=h,this.once=m||!1}function o(v,h,m,y,g){if(typeof m!="function")throw new TypeError("The listener must be a function");var b=new u(m,y||v,g),w=r?r+h:h;return v._events[w]?v._events[w].fn?v._events[w]=[v._events[w],b]:v._events[w].push(b):(v._events[w]=b,v._eventsCount++),v}function s(v,h){--v._eventsCount===0?v._events=new a:delete v._events[h]}function f(){this._events=new a,this._eventsCount=0}f.prototype.eventNames=function(){var h=[],m,y;if(this._eventsCount===0)return h;for(y in m=this._events)t.call(m,y)&&h.push(r?y.slice(1):y);return Object.getOwnPropertySymbols?h.concat(Object.getOwnPropertySymbols(m)):h},f.prototype.listeners=function(h){var m=r?r+h:h,y=this._events[m];if(!y)return[];if(y.fn)return[y.fn];for(var g=0,b=y.length,w=new Array(b);g<b;g++)w[g]=y[g].fn;return w},f.prototype.listenerCount=function(h){var m=r?r+h:h,y=this._events[m];return y?y.fn?1:y.length:0},f.prototype.emit=function(h,m,y,g,b,w){var S=r?r+h:h;if(!this._events[S])return!1;var O=this._events[S],T=arguments.length,M,j;if(O.fn){switch(O.once&&this.removeListener(h,O.fn,void 0,!0),T){case 1:return O.fn.call(O.context),!0;case 2:return O.fn.call(O.context,m),!0;case 3:return O.fn.call(O.context,m,y),!0;case 4:return O.fn.call(O.context,m,y,g),!0;case 5:return O.fn.call(O.context,m,y,g,b),!0;case 6:return O.fn.call(O.context,m,y,g,b,w),!0}for(j=1,M=new Array(T-1);j<T;j++)M[j-1]=arguments[j];O.fn.apply(O.context,M)}else{var k=O.length,N;for(j=0;j<k;j++)switch(O[j].once&&this.removeListener(h,O[j].fn,void 0,!0),T){case 1:O[j].fn.call(O[j].context);break;case 2:O[j].fn.call(O[j].context,m);break;case 3:O[j].fn.call(O[j].context,m,y);break;case 4:O[j].fn.call(O[j].context,m,y,g);break;default:if(!M)for(N=1,M=new Array(T-1);N<T;N++)M[N-1]=arguments[N];O[j].fn.apply(O[j].context,M)}}return!0},f.prototype.on=function(h,m,y){return o(this,h,m,y,!1)},f.prototype.once=function(h,m,y){return o(this,h,m,y,!0)},f.prototype.removeListener=function(h,m,y,g){var b=r?r+h:h;if(!this._events[b])return this;if(!m)return s(this,b),this;var w=this._events[b];if(w.fn)w.fn===m&&(!g||w.once)&&(!y||w.context===y)&&s(this,b);else{for(var S=0,O=[],T=w.length;S<T;S++)(w[S].fn!==m||g&&!w[S].once||y&&w[S].context!==y)&&O.push(w[S]);O.length?this._events[b]=O.length===1?O[0]:O:s(this,b)}return this},f.prototype.removeAllListeners=function(h){var m;return h?(m=r?r+h:h,this._events[m]&&s(this,m)):(this._events=new a,this._eventsCount=0),this},f.prototype.off=f.prototype.removeListener,f.prototype.addListener=f.prototype.on,f.prefixed=r,f.EventEmitter=f,e.exports=f})(dv)),dv.exports}var cz=oz();const sz=Wn(cz);var Ou=new sz,dm="recharts.syncEvent.tooltip",$S="recharts.syncEvent.brush";function fz(e,t){if(t){var r=Number.parseInt(t,10);if(!dn(r))return e?.[r]}}var dz={chartName:"",tooltipPayloadSearcher:void 0,eventEmitter:void 0,defaultTooltipEventType:"axis"},T2=mn({name:"options",initialState:dz,reducers:{createEventEmitter:e=>{e.eventEmitter==null&&(e.eventEmitter=Symbol("rechartsEventEmitter"))}}}),hz=T2.reducer,{createEventEmitter:vz}=T2.actions;function mz(e){return e.tooltip.syncInteraction}var pz={chartData:void 0,computedData:void 0,dataStartIndex:0,dataEndIndex:0},j2=mn({name:"chartData",initialState:pz,reducers:{setChartData(e,t){if(e.chartData=t.payload,t.payload==null){e.dataStartIndex=0,e.dataEndIndex=0;return}t.payload.length>0&&e.dataEndIndex!==t.payload.length-1&&(e.dataEndIndex=t.payload.length-1)},setComputedData(e,t){e.computedData=t.payload},setDataStartEndIndexes(e,t){var{startIndex:r,endIndex:a}=t.payload;r!=null&&(e.dataStartIndex=r),a!=null&&(e.dataEndIndex=a)}}}),{setChartData:HS,setDataStartEndIndexes:yz,setComputedData:jq}=j2.actions,gz=j2.reducer,bz=["x","y"];function YS(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function _a(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?YS(Object(r),!0).forEach(function(a){xz(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):YS(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function xz(e,t,r){return(t=Sz(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function Sz(e){var t=wz(e,"string");return typeof t=="symbol"?t:t+""}function wz(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function Oz(e,t){if(e==null)return{};var r,a,u=Az(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function Az(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}var M2=()=>{};function Ez(){var e=me(vp),t=me(mp),r=bt(),a=me(dE),u=me(jr),o=Lm(),s=Bm(),f=me(v=>v.rootProps.className);A.useEffect(()=>{if(e==null)return M2;var v=(h,m,y)=>{if(t!==y&&e===h){if(a==="index"){var g;if(s&&m!==null&&m!==void 0&&(g=m.payload)!==null&&g!==void 0&&g.coordinate){var b=m.payload.coordinate,{x:w,y:S}=b,O=Oz(b,bz),T=_a(_a(_a({},O),typeof w=="number"&&{x:Math.max(s.x,Math.min(w,s.x+s.width))}),typeof S=="number"&&{y:Math.max(s.y,Math.min(S,s.y+s.height))}),M=_a(_a({},m),{},{payload:_a(_a({},m.payload),{},{coordinate:T})});r(M)}else r(m);return}if(u!=null){var j;if(typeof a=="function"){var k={activeTooltipIndex:m.payload.index==null?void 0:Number(m.payload.index),isTooltipActive:m.payload.active,activeIndex:m.payload.index==null?void 0:Number(m.payload.index),activeLabel:m.payload.label,activeDataKey:m.payload.dataKey,activeCoordinate:m.payload.coordinate},N=a(u,k);j=u[N]}else a==="value"&&(j=u.find(he=>String(he.value)===m.payload.label));var{coordinate:P}=m.payload;if(j==null||m.payload.active===!1||P==null||s==null){r(cm({active:!1,coordinate:void 0,dataKey:void 0,index:null,label:void 0}));return}var{x:L,y:G}=P,ee=Math.min(L,s.x+s.width),ne=Math.min(G,s.y+s.height),J={x:o==="horizontal"?j.coordinate:ee,y:o==="horizontal"?ne:j.coordinate},re=cm({active:m.payload.active,coordinate:J,dataKey:m.payload.dataKey,index:String(j.index),label:m.payload.label});r(re)}}};return Ou.on(dm,v),()=>{Ou.off(dm,v)}},[f,r,t,e,a,u,o,s])}function _z(){var e=me(vp),t=me(mp),r=bt();A.useEffect(()=>{if(e==null)return M2;var a=(u,o,s)=>{t!==s&&e===u&&r(yz(o))};return Ou.on($S,a),()=>{Ou.off($S,a)}},[r,t,e])}function Tz(){var e=bt();A.useEffect(()=>{e(vz())},[e]),Ez(),_z()}function jz(e,t,r,a,u,o){var s=me(g=>W6(g,e,t)),f=me(mp),v=me(vp),h=me(dE),m=me(mz),y=m?.active;A.useEffect(()=>{if(!y&&v!=null&&f!=null){var g=cm({active:o,coordinate:r,dataKey:s,index:u,label:typeof a=="number"?String(a):a});Ou.emit(dm,v,g,f)}},[y,r,s,u,a,f,v,h,o])}function GS(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function KS(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?GS(Object(r),!0).forEach(function(a){Mz(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):GS(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function Mz(e,t,r){return(t=Cz(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function Cz(e){var t=Dz(e,"string");return typeof t=="symbol"?t:t+""}function Dz(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function Nz(e){return e.dataKey}function Pz(e,t){return A.isValidElement(e)?A.cloneElement(e,t):typeof e=="function"?A.createElement(e,t):A.createElement(eP,t)}var IS=[],Rz={allowEscapeViewBox:{x:!1,y:!1},animationDuration:400,animationEasing:"ease",axisId:0,contentStyle:{},cursor:!0,filterNull:!0,isAnimationActive:!Vi.isSsr,itemSorter:"name",itemStyle:{},labelStyle:{},offset:10,reverseDirection:{x:!1,y:!1},separator:" : ",trigger:"hover",useTranslate3d:!1,wrapperStyle:{}};function kz(e){var t=Pn(e,Rz),{active:r,allowEscapeViewBox:a,animationDuration:u,animationEasing:o,content:s,filterNull:f,isAnimationActive:v,offset:h,payloadUniqBy:m,position:y,reverseDirection:g,useTranslate3d:b,wrapperStyle:w,cursor:S,shared:O,trigger:T,defaultIndex:M,portal:j,axisId:k}=t,N=bt(),P=typeof M=="number"?String(M):M;A.useEffect(()=>{N(n6({shared:O,trigger:T,axisId:k,active:r,defaultIndex:P}))},[N,O,T,k,r,P]);var L=Bm(),G=lA(),ee=W4(O),{activeIndex:ne,isActive:J}=me(Se=>ez(Se,ee,T,P)),re=me(Se=>J6(Se,ee,T,P)),he=me(Se=>E2(Se,ee,T,P)),ce=me(Se=>F6(Se,ee,T,P)),se=re,B=uz(),I=r??J,[F,oe]=iD([se,I]),D=ee==="axis"?he:void 0;jz(ee,T,ce,D,ne,I);var W=j??B;if(W==null)return null;var Z=se??IS;I||(Z=IS),f&&Z.length&&(Z=DC(se.filter(Se=>Se.value!=null&&(Se.hide!==!0||t.includeHidden)),m,Nz));var ae=Z.length>0,fe=A.createElement(uP,{allowEscapeViewBox:a,animationDuration:u,animationEasing:o,isAnimationActive:v,active:I,coordinate:ce,hasPayload:ae,offset:h,position:y,reverseDirection:g,useTranslate3d:b,viewBox:L,wrapperStyle:w,lastBoundingBox:F,innerRef:oe,hasPortalFromProps:!!j},Pz(s,KS(KS({},t),{},{payload:Z,label:D,active:I,coordinate:ce,accessibilityLayer:G})));return A.createElement(A.Fragment,null,SM.createPortal(fe,W),I&&A.createElement(lz,{cursor:S,tooltipEventType:ee,coordinate:ce,payload:se,index:ne}))}var hv={},vv={},mv={},XS;function zz(){return XS||(XS=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r,a,{signal:u,edges:o}={}){let s,f=null;const v=o!=null&&o.includes("leading"),h=o==null||o.includes("trailing"),m=()=>{f!==null&&(r.apply(s,f),s=void 0,f=null)},y=()=>{h&&m(),S()};let g=null;const b=()=>{g!=null&&clearTimeout(g),g=setTimeout(()=>{g=null,y()},a)},w=()=>{g!==null&&(clearTimeout(g),g=null)},S=()=>{w(),s=void 0,f=null},O=()=>{m()},T=function(...M){if(u?.aborted)return;s=this,f=M;const j=g==null;b(),v&&j&&m()};return T.schedule=b,T.cancel=S,T.flush=O,u?.addEventListener("abort",S,{once:!0}),T}e.debounce=t})(mv)),mv}var VS;function Bz(){return VS||(VS=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=zz();function r(a,u=0,o={}){typeof o!="object"&&(o={});const{leading:s=!1,trailing:f=!0,maxWait:v}=o,h=Array(2);s&&(h[0]="leading"),f&&(h[1]="trailing");let m,y=null;const g=t.debounce(function(...S){m=a.apply(this,S),y=null},u,{edges:h}),b=function(...S){return v!=null&&(y===null&&(y=Date.now()),Date.now()-y>=v)?(m=a.apply(this,S),y=Date.now(),g.cancel(),g.schedule(),m):(g.apply(this,S),m)},w=()=>(g.flush(),m);return b.cancel=g.cancel,b.flush=w,b}e.debounce=r})(vv)),vv}var ZS;function Lz(){return ZS||(ZS=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=Bz();function r(a,u=0,o={}){const{leading:s=!0,trailing:f=!0}=o;return t.debounce(a,u,{leading:s,maxWait:u,trailing:f})}e.throttle=r})(hv)),hv}var pv,QS;function Uz(){return QS||(QS=1,pv=Lz().throttle),pv}var qz=Uz();const $z=Wn(qz);var du=function(t,r){for(var a=arguments.length,u=new Array(a>2?a-2:0),o=2;o<a;o++)u[o-2]=arguments[o]};function WS(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function yv(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?WS(Object(r),!0).forEach(function(a){Hz(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):WS(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function Hz(e,t,r){return(t=Yz(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function Yz(e){var t=Gz(e,"string");return typeof t=="symbol"?t:t+""}function Gz(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var Kz=A.forwardRef((e,t)=>{var{aspect:r,initialDimension:a={width:-1,height:-1},width:u="100%",height:o="100%",minWidth:s=0,minHeight:f,maxHeight:v,children:h,debounce:m=0,id:y,className:g,onResize:b,style:w={}}=e,S=A.useRef(null),O=A.useRef();O.current=b,A.useImperativeHandle(t,()=>S.current);var[T,M]=A.useState({containerWidth:a.width,containerHeight:a.height}),j=A.useCallback((N,P)=>{M(L=>{var G=Math.round(N),ee=Math.round(P);return L.containerWidth===G&&L.containerHeight===ee?L:{containerWidth:G,containerHeight:ee}})},[]);A.useEffect(()=>{var N=ee=>{var ne,{width:J,height:re}=ee[0].contentRect;j(J,re),(ne=O.current)===null||ne===void 0||ne.call(O,J,re)};m>0&&(N=$z(N,m,{trailing:!0,leading:!1}));var P=new ResizeObserver(N),{width:L,height:G}=S.current.getBoundingClientRect();return j(L,G),P.observe(S.current),()=>{P.disconnect()}},[j,m]);var k=A.useMemo(()=>{var{containerWidth:N,containerHeight:P}=T;if(N<0||P<0)return null;du(Ca(u)||Ca(o),`The width(%s) and height(%s) are both fixed numbers,
       maybe you don't need to use a ResponsiveContainer.`,u,o),du(!r||r>0,"The aspect(%s) must be greater than zero.",r);var L=Ca(u)?N:u,G=Ca(o)?P:o;return r&&r>0&&(L?G=L/r:G&&(L=G*r),v&&G>v&&(G=v)),du(L>0||G>0,`The width(%s) and height(%s) of chart should be greater than 0,
       please check the style of container, or the props width(%s) and height(%s),
       or add a minWidth(%s) or minHeight(%s) or use aspect(%s) to control the
       height and width.`,L,G,u,o,s,f,r),A.Children.map(h,ee=>A.cloneElement(ee,{width:L,height:G,style:yv({width:L,height:G},ee.props.style)}))},[r,h,o,v,f,s,T,u]);return A.createElement("div",{id:y?"".concat(y):void 0,className:Ie("recharts-responsive-container",g),style:yv(yv({},w),{},{width:u,height:o,minWidth:s,minHeight:f,maxHeight:v}),ref:S},A.createElement("div",{style:{width:0,height:0,overflow:"visible"}},k))}),C2=e=>null;C2.displayName="Cell";function Iz(e,t,r){return(t=Xz(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function Xz(e){var t=Vz(e,"string");return typeof t=="symbol"?t:t+""}function Vz(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}class Zz{constructor(t){Iz(this,"cache",new Map),this.maxSize=t}get(t){var r=this.cache.get(t);return r!==void 0&&(this.cache.delete(t),this.cache.set(t,r)),r}set(t,r){if(this.cache.has(t))this.cache.delete(t);else if(this.cache.size>=this.maxSize){var a=this.cache.keys().next().value;this.cache.delete(a)}this.cache.set(t,r)}clear(){this.cache.clear()}size(){return this.cache.size}}function FS(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function Qz(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?FS(Object(r),!0).forEach(function(a){Wz(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):FS(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function Wz(e,t,r){return(t=Fz(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function Fz(e){var t=Jz(e,"string");return typeof t=="symbol"?t:t+""}function Jz(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var e5={cacheSize:2e3,enableCache:!0},D2=Qz({},e5),JS=new Zz(D2.cacheSize),t5={position:"absolute",top:"-20000px",left:0,padding:0,margin:0,border:"none",whiteSpace:"pre"},ew="recharts_measurement_span";function n5(e,t){var r=t.fontSize||"",a=t.fontFamily||"",u=t.fontWeight||"",o=t.fontStyle||"",s=t.letterSpacing||"",f=t.textTransform||"";return"".concat(e,"|").concat(r,"|").concat(a,"|").concat(u,"|").concat(o,"|").concat(s,"|").concat(f)}var tw=(e,t)=>{try{var r=document.getElementById(ew);r||(r=document.createElement("span"),r.setAttribute("id",ew),r.setAttribute("aria-hidden","true"),document.body.appendChild(r)),Object.assign(r.style,t5,t),r.textContent="".concat(e);var a=r.getBoundingClientRect();return{width:a.width,height:a.height}}catch{return{width:0,height:0}}},hu=function(t){var r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};if(t==null||Vi.isSsr)return{width:0,height:0};if(!D2.enableCache)return tw(t,r);var a=n5(t,r),u=JS.get(a);if(u)return u;var o=tw(t,r);return JS.set(a,o),o},nw=/(-?\d+(?:\.\d+)?[a-zA-Z%]*)([*/])(-?\d+(?:\.\d+)?[a-zA-Z%]*)/,rw=/(-?\d+(?:\.\d+)?[a-zA-Z%]*)([+-])(-?\d+(?:\.\d+)?[a-zA-Z%]*)/,r5=/^px|cm|vh|vw|em|rem|%|mm|in|pt|pc|ex|ch|vmin|vmax|Q$/,a5=/(-?\d+(?:\.\d+)?)([a-zA-Z%]+)?/,N2={cm:96/2.54,mm:96/25.4,pt:96/72,pc:96/6,in:96,Q:96/(2.54*40),px:1},i5=Object.keys(N2),Ni="NaN";function l5(e,t){return e*N2[t]}class Xt{static parse(t){var r,[,a,u]=(r=a5.exec(t))!==null&&r!==void 0?r:[];return new Xt(parseFloat(a),u??"")}constructor(t,r){this.num=t,this.unit=r,this.num=t,this.unit=r,dn(t)&&(this.unit=""),r!==""&&!r5.test(r)&&(this.num=NaN,this.unit=""),i5.includes(r)&&(this.num=l5(t,r),this.unit="px")}add(t){return this.unit!==t.unit?new Xt(NaN,""):new Xt(this.num+t.num,this.unit)}subtract(t){return this.unit!==t.unit?new Xt(NaN,""):new Xt(this.num-t.num,this.unit)}multiply(t){return this.unit!==""&&t.unit!==""&&this.unit!==t.unit?new Xt(NaN,""):new Xt(this.num*t.num,this.unit||t.unit)}divide(t){return this.unit!==""&&t.unit!==""&&this.unit!==t.unit?new Xt(NaN,""):new Xt(this.num/t.num,this.unit||t.unit)}toString(){return"".concat(this.num).concat(this.unit)}isNaN(){return dn(this.num)}}function P2(e){if(e.includes(Ni))return Ni;for(var t=e;t.includes("*")||t.includes("/");){var r,[,a,u,o]=(r=nw.exec(t))!==null&&r!==void 0?r:[],s=Xt.parse(a??""),f=Xt.parse(o??""),v=u==="*"?s.multiply(f):s.divide(f);if(v.isNaN())return Ni;t=t.replace(nw,v.toString())}for(;t.includes("+")||/.-\d+(?:\.\d+)?/.test(t);){var h,[,m,y,g]=(h=rw.exec(t))!==null&&h!==void 0?h:[],b=Xt.parse(m??""),w=Xt.parse(g??""),S=y==="+"?b.add(w):b.subtract(w);if(S.isNaN())return Ni;t=t.replace(rw,S.toString())}return t}var aw=/\(([^()]*)\)/;function u5(e){for(var t=e,r;(r=aw.exec(t))!=null;){var[,a]=r;t=t.replace(aw,P2(a))}return t}function o5(e){var t=e.replace(/\s+/g,"");return t=u5(t),t=P2(t),t}function c5(e){try{return o5(e)}catch{return Ni}}function gv(e){var t=c5(e.slice(5,-1));return t===Ni?"":t}var s5=["x","y","lineHeight","capHeight","scaleToFit","textAnchor","verticalAnchor","fill"],f5=["dx","dy","angle","className","breakAll"];function hm(){return hm=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},hm.apply(null,arguments)}function iw(e,t){if(e==null)return{};var r,a,u=d5(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function d5(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}var R2=/[ \f\n\r\t\v\u2028\u2029]+/,k2=e=>{var{children:t,breakAll:r,style:a}=e;try{var u=[];gt(t)||(r?u=t.toString().split(""):u=t.toString().split(R2));var o=u.map(f=>({word:f,width:hu(f,a).width})),s=r?0:hu(" ",a).width;return{wordsWithComputedWidth:o,spaceWidth:s}}catch{return null}},h5=(e,t,r,a,u)=>{var{maxLines:o,children:s,style:f,breakAll:v}=e,h=ve(o),m=s,y=function(){var J=arguments.length>0&&arguments[0]!==void 0?arguments[0]:[];return J.reduce((re,he)=>{var{word:ce,width:se}=he,B=re[re.length-1];if(B&&(a==null||u||B.width+se+r<Number(a)))B.words.push(ce),B.width+=se+r;else{var I={words:[ce],width:se};re.push(I)}return re},[])},g=y(t),b=ne=>ne.reduce((J,re)=>J.width>re.width?J:re);if(!h||u)return g;var w=g.length>o||b(g).width>Number(a);if(!w)return g;for(var S="…",O=ne=>{var J=m.slice(0,ne),re=k2({breakAll:v,style:f,children:J+S}).wordsWithComputedWidth,he=y(re),ce=he.length>o||b(he).width>Number(a);return[ce,he]},T=0,M=m.length-1,j=0,k;T<=M&&j<=m.length-1;){var N=Math.floor((T+M)/2),P=N-1,[L,G]=O(P),[ee]=O(N);if(!L&&!ee&&(T=N+1),L&&ee&&(M=N-1),!L&&ee){k=G;break}j++}return k||g},lw=e=>{var t=gt(e)?[]:e.toString().split(R2);return[{words:t}]},v5=e=>{var{width:t,scaleToFit:r,children:a,style:u,breakAll:o,maxLines:s}=e;if((t||r)&&!Vi.isSsr){var f,v,h=k2({breakAll:o,children:a,style:u});if(h){var{wordsWithComputedWidth:m,spaceWidth:y}=h;f=m,v=y}else return lw(a);return h5({breakAll:o,children:a,maxLines:s,style:u},f,v,t,r)}return lw(a)},uw="#808080",Lp=A.forwardRef((e,t)=>{var{x:r=0,y:a=0,lineHeight:u="1em",capHeight:o="0.71em",scaleToFit:s=!1,textAnchor:f="start",verticalAnchor:v="end",fill:h=uw}=e,m=iw(e,s5),y=A.useMemo(()=>v5({breakAll:m.breakAll,children:m.children,maxLines:m.maxLines,scaleToFit:s,style:m.style,width:m.width}),[m.breakAll,m.children,m.maxLines,s,m.style,m.width]),{dx:g,dy:b,angle:w,className:S,breakAll:O}=m,T=iw(m,f5);if(!Zn(r)||!Zn(a)||y.length===0)return null;var M=r+(ve(g)?g:0),j=a+(ve(b)?b:0),k;switch(v){case"start":k=gv("calc(".concat(o,")"));break;case"middle":k=gv("calc(".concat((y.length-1)/2," * -").concat(u," + (").concat(o," / 2))"));break;default:k=gv("calc(".concat(y.length-1," * -").concat(u,")"));break}var N=[];if(s){var P=y[0].width,{width:L}=m;N.push("scale(".concat(ve(L)?L/P:1,")"))}return w&&N.push("rotate(".concat(w,", ").concat(M,", ").concat(j,")")),N.length&&(T.transform=N.join(" ")),A.createElement("text",hm({},Mt(T,!0),{ref:t,x:M,y:j,className:Ie("recharts-text",S),textAnchor:f,fill:h.includes("url")?uw:h}),y.map((G,ee)=>{var ne=G.words.join(O?"":" ");return A.createElement("tspan",{x:M,dy:ee===0?k:u,key:"".concat(ne,"-").concat(ee)},ne)}))});Lp.displayName="Text";var m5=["labelRef"];function p5(e,t){if(e==null)return{};var r,a,u=y5(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function y5(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}function ow(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function tt(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?ow(Object(r),!0).forEach(function(a){g5(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):ow(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function g5(e,t,r){return(t=b5(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function b5(e){var t=x5(e,"string");return typeof t=="symbol"?t:t+""}function x5(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function vr(){return vr=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},vr.apply(null,arguments)}var z2=A.createContext(null),S5=e=>{var{x:t,y:r,width:a,height:u,children:o}=e,s=A.useMemo(()=>({x:t,y:r,width:a,height:u}),[t,r,a,u]);return A.createElement(z2.Provider,{value:s},o)},B2=()=>{var e=A.useContext(z2),t=Bm();return e||t},w5=A.createContext(null),O5=()=>{var e=A.useContext(w5),t=me(yE);return e||t},A5=e=>{var{value:t,formatter:r}=e,a=gt(e.children)?t:e.children;return typeof r=="function"?r(a):a},Up=e=>e!=null&&typeof e=="function",E5=(e,t)=>{var r=Vt(t-e),a=Math.min(Math.abs(t-e),360);return r*a},_5=(e,t,r,a,u)=>{var{offset:o,className:s}=e,{cx:f,cy:v,innerRadius:h,outerRadius:m,startAngle:y,endAngle:g,clockWise:b}=u,w=(h+m)/2,S=E5(y,g),O=S>=0?1:-1,T,M;switch(t){case"insideStart":T=y+O*o,M=b;break;case"insideEnd":T=g-O*o,M=!b;break;case"end":T=g+O*o,M=b;break;default:throw new Error("Unsupported position ".concat(t))}M=S<=0?M:!M;var j=jt(f,v,w,T),k=jt(f,v,w,T+(M?1:-1)*359),N="M".concat(j.x,",").concat(j.y,`
    A`).concat(w,",").concat(w,",0,1,").concat(M?0:1,`,
    `).concat(k.x,",").concat(k.y),P=gt(e.id)?vu("recharts-radial-line-"):e.id;return A.createElement("text",vr({},a,{dominantBaseline:"central",className:Ie("recharts-radial-bar-label",s)}),A.createElement("defs",null,A.createElement("path",{id:P,d:N})),A.createElement("textPath",{xlinkHref:"#".concat(P)},r))},T5=(e,t,r)=>{var{cx:a,cy:u,innerRadius:o,outerRadius:s,startAngle:f,endAngle:v}=e,h=(f+v)/2;if(r==="outside"){var{x:m,y}=jt(a,u,s+t,h);return{x:m,y,textAnchor:m>=a?"start":"end",verticalAnchor:"middle"}}if(r==="center")return{x:a,y:u,textAnchor:"middle",verticalAnchor:"middle"};if(r==="centerTop")return{x:a,y:u,textAnchor:"middle",verticalAnchor:"start"};if(r==="centerBottom")return{x:a,y:u,textAnchor:"middle",verticalAnchor:"end"};var g=(o+s)/2,{x:b,y:w}=jt(a,u,g,h);return{x:b,y:w,textAnchor:"middle",verticalAnchor:"middle"}},L2=e=>"cx"in e&&ve(e.cx),j5=(e,t)=>{var{parentViewBox:r,offset:a,position:u}=e,o;r!=null&&!L2(r)&&(o=r);var{x:s,y:f,width:v,height:h}=t,m=h>=0?1:-1,y=m*a,g=m>0?"end":"start",b=m>0?"start":"end",w=v>=0?1:-1,S=w*a,O=w>0?"end":"start",T=w>0?"start":"end";if(u==="top"){var M={x:s+v/2,y:f-m*a,textAnchor:"middle",verticalAnchor:g};return tt(tt({},M),o?{height:Math.max(f-o.y,0),width:v}:{})}if(u==="bottom"){var j={x:s+v/2,y:f+h+y,textAnchor:"middle",verticalAnchor:b};return tt(tt({},j),o?{height:Math.max(o.y+o.height-(f+h),0),width:v}:{})}if(u==="left"){var k={x:s-S,y:f+h/2,textAnchor:O,verticalAnchor:"middle"};return tt(tt({},k),o?{width:Math.max(k.x-o.x,0),height:h}:{})}if(u==="right"){var N={x:s+v+S,y:f+h/2,textAnchor:T,verticalAnchor:"middle"};return tt(tt({},N),o?{width:Math.max(o.x+o.width-N.x,0),height:h}:{})}var P=o?{width:v,height:h}:{};return u==="insideLeft"?tt({x:s+S,y:f+h/2,textAnchor:T,verticalAnchor:"middle"},P):u==="insideRight"?tt({x:s+v-S,y:f+h/2,textAnchor:O,verticalAnchor:"middle"},P):u==="insideTop"?tt({x:s+v/2,y:f+y,textAnchor:"middle",verticalAnchor:b},P):u==="insideBottom"?tt({x:s+v/2,y:f+h-y,textAnchor:"middle",verticalAnchor:g},P):u==="insideTopLeft"?tt({x:s+S,y:f+y,textAnchor:T,verticalAnchor:b},P):u==="insideTopRight"?tt({x:s+v-S,y:f+y,textAnchor:O,verticalAnchor:b},P):u==="insideBottomLeft"?tt({x:s+S,y:f+h-y,textAnchor:T,verticalAnchor:g},P):u==="insideBottomRight"?tt({x:s+v-S,y:f+h-y,textAnchor:O,verticalAnchor:g},P):u&&typeof u=="object"&&(ve(u.x)||Ca(u.x))&&(ve(u.y)||Ca(u.y))?tt({x:s+qn(u.x,v),y:f+qn(u.y,h),textAnchor:"end",verticalAnchor:"end"},P):tt({x:s+v/2,y:f+h/2,textAnchor:"middle",verticalAnchor:"middle"},P)},M5={offset:5};function ea(e){var t=Pn(e,M5),{viewBox:r,position:a,value:u,children:o,content:s,className:f="",textBreakAll:v,labelRef:h}=t,m=O5(),y=B2(),g=a==="center"?y:m??y,b=r||g;if(!b||gt(u)&&gt(o)&&!A.isValidElement(s)&&typeof s!="function")return null;var w=tt(tt({},t),{},{viewBox:b});if(A.isValidElement(s)){var{labelRef:S}=w,O=p5(w,m5);return A.cloneElement(s,O)}var T;if(typeof s=="function"){if(T=A.createElement(s,w),A.isValidElement(T))return T}else T=A5(t);var M=L2(b),j=Mt(t,!0);if(M&&(a==="insideStart"||a==="insideEnd"||a==="end"))return _5(t,a,T,j,b);var k=M?T5(b,t.offset,t.position):j5(t,b);return A.createElement(Lp,vr({ref:h,className:Ie("recharts-label",f)},j,k,{breakAll:v}),T)}ea.displayName="Label";var C5=(e,t,r)=>{if(!e)return null;var a={viewBox:t,labelRef:r};return e===!0?A.createElement(ea,vr({key:"label-implicit"},a)):Zn(e)?A.createElement(ea,vr({key:"label-implicit",value:e},a)):A.isValidElement(e)?e.type===ea?A.cloneElement(e,tt({key:"label-implicit"},a)):A.createElement(ea,vr({key:"label-implicit",content:e},a)):Up(e)?A.createElement(ea,vr({key:"label-implicit",content:e},a)):e&&typeof e=="object"?A.createElement(ea,vr({},e,{key:"label-implicit"},a)):null};function D5(e){var{label:t}=e,r=B2();return C5(t,r)||null}var bv={},xv={},cw;function N5(){return cw||(cw=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return r[r.length-1]}e.last=t})(xv)),xv}var Sv={},sw;function P5(){return sw||(sw=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){return Array.isArray(r)?r:Array.from(r)}e.toArray=t})(Sv)),Sv}var fw;function R5(){return fw||(fw=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});const t=N5(),r=P5(),a=Tm();function u(o){if(a.isArrayLike(o))return t.last(r.toArray(o))}e.last=u})(bv)),bv}var wv,dw;function k5(){return dw||(dw=1,wv=R5().last),wv}var z5=k5();const B5=Wn(z5);var L5=["valueAccessor"],U5=["dataKey","clockWise","id","textBreakAll"];function Zc(){return Zc=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},Zc.apply(null,arguments)}function hw(e,t){if(e==null)return{};var r,a,u=q5(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function q5(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}var $5=e=>Array.isArray(e.value)?B5(e.value):e.value,U2=A.createContext(void 0),H5=U2.Provider,q2=A.createContext(void 0);q2.Provider;function Y5(){return A.useContext(U2)}function G5(){return A.useContext(q2)}function gc(e){var{valueAccessor:t=$5}=e,r=hw(e,L5),{dataKey:a,clockWise:u,id:o,textBreakAll:s}=r,f=hw(r,U5),v=Y5(),h=G5(),m=v||h;return!m||!m.length?null:A.createElement(ra,{className:"recharts-label-list"},m.map((y,g)=>{var b,w=gt(a)?t(y,g):$t(y&&y.payload,a),S=gt(o)?{}:{id:"".concat(o,"-").concat(g)};return A.createElement(ea,Zc({},Mt(y,!0),f,S,{fill:(b=r.fill)!==null&&b!==void 0?b:y.fill,parentViewBox:y.parentViewBox,value:w,textBreakAll:s,viewBox:y.viewBox,key:"label-".concat(g),index:g}))}))}gc.displayName="LabelList";function K5(e){var{label:t}=e;return t?t===!0?A.createElement(gc,{key:"labelList-implicit"}):A.isValidElement(t)||Up(t)?A.createElement(gc,{key:"labelList-implicit",content:t}):typeof t=="object"?A.createElement(gc,Zc({key:"labelList-implicit"},t,{type:String(t.type)})):null:null}var I5={radiusAxis:{},angleAxis:{}},$2=mn({name:"polarAxis",initialState:I5,reducers:{addRadiusAxis(e,t){e.radiusAxis[t.payload.id]=t.payload},removeRadiusAxis(e,t){delete e.radiusAxis[t.payload.id]},addAngleAxis(e,t){e.angleAxis[t.payload.id]=t.payload},removeAngleAxis(e,t){delete e.angleAxis[t.payload.id]}}}),{addRadiusAxis:Mq,removeRadiusAxis:Cq,addAngleAxis:Dq,removeAngleAxis:Nq}=$2.actions,X5=$2.reducer,Ov={},vw;function V5(){return vw||(vw=1,(function(e){Object.defineProperty(e,Symbol.toStringTag,{value:"Module"});function t(r){if(typeof r!="object"||r==null)return!1;if(Object.getPrototypeOf(r)===null)return!0;if(Object.prototype.toString.call(r)!=="[object Object]"){const u=r[Symbol.toStringTag];return u==null||!Object.getOwnPropertyDescriptor(r,Symbol.toStringTag)?.writable?!1:r.toString()===`[object ${u}]`}let a=r;for(;Object.getPrototypeOf(a)!==null;)a=Object.getPrototypeOf(a);return Object.getPrototypeOf(r)===a}e.isPlainObject=t})(Ov)),Ov}var Av,mw;function Z5(){return mw||(mw=1,Av=V5().isPlainObject),Av}var Q5=Z5();const W5=Wn(Q5);function pw(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function yw(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?pw(Object(r),!0).forEach(function(a){F5(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):pw(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function F5(e,t,r){return(t=J5(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function J5(e){var t=eB(e,"string");return typeof t=="symbol"?t:t+""}function eB(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function Qc(){return Qc=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},Qc.apply(null,arguments)}var gw=(e,t,r,a,u)=>{var o=r-a,s;return s="M ".concat(e,",").concat(t),s+="L ".concat(e+r,",").concat(t),s+="L ".concat(e+r-o/2,",").concat(t+u),s+="L ".concat(e+r-o/2-a,",").concat(t+u),s+="L ".concat(e,",").concat(t," Z"),s},tB={x:0,y:0,upperWidth:0,lowerWidth:0,height:0,isUpdateAnimationActive:!1,animationBegin:0,animationDuration:1500,animationEasing:"ease"},nB=e=>{var t=Pn(e,tB),{x:r,y:a,upperWidth:u,lowerWidth:o,height:s,className:f}=t,{animationEasing:v,animationDuration:h,animationBegin:m,isUpdateAnimationActive:y}=t,g=A.useRef(),[b,w]=A.useState(-1),S=A.useRef(u),O=A.useRef(o),T=A.useRef(s),M=A.useRef(r),j=A.useRef(a),k=qm(e,"trapezoid-");if(A.useEffect(()=>{if(g.current&&g.current.getTotalLength)try{var ce=g.current.getTotalLength();ce&&w(ce)}catch{}},[]),r!==+r||a!==+a||u!==+u||o!==+o||s!==+s||u===0&&o===0||s===0)return null;var N=Ie("recharts-trapezoid",f);if(!y)return A.createElement("g",null,A.createElement("path",Qc({},Mt(t,!0),{className:N,d:gw(r,a,u,o,s)})));var P=S.current,L=O.current,G=T.current,ee=M.current,ne=j.current,J="0px ".concat(b===-1?1:b,"px"),re="".concat(b,"px 0px"),he=uA(["strokeDasharray"],h,v);return A.createElement(Um,{animationId:k,key:k,canBegin:b>0,duration:h,easing:v,isActive:y,begin:m},ce=>{var se=Lt(P,u,ce),B=Lt(L,o,ce),I=Lt(G,s,ce),F=Lt(ee,r,ce),oe=Lt(ne,a,ce);g.current&&(S.current=se,O.current=B,T.current=I,M.current=F,j.current=oe);var D=ce>0?{transition:he,strokeDasharray:re}:{strokeDasharray:J};return A.createElement("path",Qc({},Mt(t,!0),{className:N,d:gw(F,oe,se,B,I),ref:g,style:yw(yw({},D),t.style)}))})},rB=["option","shapeType","propTransformer","activeClassName","isActive"];function aB(e,t){if(e==null)return{};var r,a,u=iB(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function iB(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}function bw(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function Wc(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?bw(Object(r),!0).forEach(function(a){lB(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):bw(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function lB(e,t,r){return(t=uB(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function uB(e){var t=oB(e,"string");return typeof t=="symbol"?t:t+""}function oB(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function cB(e,t){return Wc(Wc({},t),e)}function sB(e,t){return e==="symbols"}function xw(e){var{shapeType:t,elementProps:r}=e;switch(t){case"rectangle":return A.createElement(fA,r);case"trapezoid":return A.createElement(nB,r);case"sector":return A.createElement(vA,r);case"symbols":if(sB(t))return A.createElement(vO,r);break;default:return null}}function fB(e){return A.isValidElement(e)?e.props:e}function dB(e){var{option:t,shapeType:r,propTransformer:a=cB,activeClassName:u="recharts-active-shape",isActive:o}=e,s=aB(e,rB),f;if(A.isValidElement(t))f=A.cloneElement(t,Wc(Wc({},s),fB(t)));else if(typeof t=="function")f=t(s);else if(W5(t)&&typeof t!="boolean"){var v=a(t,s);f=A.createElement(xw,{shapeType:r,elementProps:v})}else{var h=s;f=A.createElement(xw,{shapeType:r,elementProps:h})}return o?A.createElement(ra,{className:u},f):f}var H2=(e,t)=>{var r=bt();return(a,u)=>o=>{e?.(a,u,o),r(u2({activeIndex:String(u),activeDataKey:t,activeCoordinate:a.tooltipPosition}))}},Y2=e=>{var t=bt();return(r,a)=>u=>{e?.(r,a,u),t(r6())}},G2=(e,t)=>{var r=bt();return(a,u)=>o=>{e?.(a,u,o),r(a6({activeIndex:String(u),activeDataKey:t,activeCoordinate:a.tooltipPosition}))}};function hB(e){var{fn:t,args:r}=e,a=bt(),u=Nn();return A.useEffect(()=>{if(!u){var o=t(r);return a(e6(o)),()=>{a(t6(o))}}},[t,r,a,u]),null}var vB=()=>{};function mB(e){var{legendPayload:t}=e,r=bt(),a=Nn();return A.useEffect(()=>a?vB:(r(XN(t)),()=>{r(VN(t))}),[r,a,t]),null}var Ev,pB=()=>{var[e]=A.useState(()=>vu("uid-"));return e},yB=(Ev=qj.useId)!==null&&Ev!==void 0?Ev:pB;function gB(e,t){var r=yB();return t||(e?"".concat(e,"-").concat(r):r)}var bB=A.createContext(void 0),xB=e=>{var{id:t,type:r,children:a}=e,u=gB("recharts-".concat(r),t);return A.createElement(bB.Provider,{value:u},a(u))},SB={cartesianItems:[],polarItems:[]},K2=mn({name:"graphicalItems",initialState:SB,reducers:{addCartesianGraphicalItem(e,t){e.cartesianItems.push(t.payload)},replaceCartesianGraphicalItem(e,t){var{prev:r,next:a}=t.payload,u=yr(e).cartesianItems.indexOf(r);u>-1&&(e.cartesianItems[u]=a)},removeCartesianGraphicalItem(e,t){var r=yr(e).cartesianItems.indexOf(t.payload);r>-1&&e.cartesianItems.splice(r,1)},addPolarGraphicalItem(e,t){e.polarItems.push(t.payload)},removePolarGraphicalItem(e,t){var r=yr(e).polarItems.indexOf(t.payload);r>-1&&e.polarItems.splice(r,1)}}}),{addCartesianGraphicalItem:wB,replaceCartesianGraphicalItem:OB,removeCartesianGraphicalItem:AB,addPolarGraphicalItem:Pq,removePolarGraphicalItem:Rq}=K2.actions,EB=K2.reducer;function _B(e){var t=bt(),r=A.useRef(null);return A.useEffect(()=>{r.current===null?t(wB(e)):r.current!==e&&t(OB({prev:r.current,next:e})),r.current=e},[t,e]),A.useEffect(()=>()=>{r.current&&(t(AB(r.current)),r.current=null)},[t]),null}function Sw(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function ww(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?Sw(Object(r),!0).forEach(function(a){TB(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):Sw(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function TB(e,t,r){return(t=jB(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function jB(e){var t=MB(e,"string");return typeof t=="symbol"?t:t+""}function MB(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var CB={xAxis:{},yAxis:{},zAxis:{}},I2=mn({name:"cartesianAxis",initialState:CB,reducers:{addXAxis(e,t){e.xAxis[t.payload.id]=t.payload},removeXAxis(e,t){delete e.xAxis[t.payload.id]},addYAxis(e,t){e.yAxis[t.payload.id]=t.payload},removeYAxis(e,t){delete e.yAxis[t.payload.id]},addZAxis(e,t){e.zAxis[t.payload.id]=t.payload},removeZAxis(e,t){delete e.zAxis[t.payload.id]},updateYAxisWidth(e,t){var{id:r,width:a}=t.payload;e.yAxis[r]&&(e.yAxis[r]=ww(ww({},e.yAxis[r]),{},{width:a}))}}}),{addXAxis:DB,removeXAxis:NB,addYAxis:PB,removeYAxis:RB,addZAxis:kq,removeZAxis:zq,updateYAxisWidth:kB}=I2.actions,zB=I2.reducer,BB=Y([xt],e=>{if(e)return{top:e.top,bottom:e.bottom,left:e.left,right:e.right}}),LB=Y([BB,Or,Ar],(e,t,r)=>{if(!(!e||t==null||r==null))return{x:e.left,y:e.top,width:Math.max(0,t-e.left-e.right),height:Math.max(0,r-e.top-e.bottom)}}),X2=()=>me(LB),UB="Invariant failed";function qB(e,t){throw new Error(UB)}var $B=["x","y"];function vm(){return vm=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},vm.apply(null,arguments)}function Ow(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function cu(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?Ow(Object(r),!0).forEach(function(a){HB(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):Ow(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function HB(e,t,r){return(t=YB(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function YB(e){var t=GB(e,"string");return typeof t=="symbol"?t:t+""}function GB(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function KB(e,t){if(e==null)return{};var r,a,u=IB(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function IB(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}function XB(e,t){var{x:r,y:a}=e,u=KB(e,$B),o="".concat(r),s=parseInt(o,10),f="".concat(a),v=parseInt(f,10),h="".concat(t.height||u.height),m=parseInt(h,10),y="".concat(t.width||u.width),g=parseInt(y,10);return cu(cu(cu(cu(cu({},t),u),s?{x:s}:{}),v?{y:v}:{}),{},{height:m,width:g,name:t.name,radius:t.radius})}function qp(e){return A.createElement(dB,vm({shapeType:"rectangle",propTransformer:XB,activeClassName:"recharts-active-bar"},e))}var VB=function(t){var r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:0;return(a,u)=>{if(ve(t))return t;var o=ve(a)||gt(a);return o?t(a,u):(o||qB(),r)}},ZB={},V2=mn({name:"errorBars",initialState:ZB,reducers:{addErrorBar:(e,t)=>{var{itemId:r,errorBar:a}=t.payload;e[r]||(e[r]=[]),e[r].push(a)},replaceErrorBar:(e,t)=>{var{itemId:r,prev:a,next:u}=t.payload;e[r]&&(e[r]=e[r].map(o=>o.dataKey===a.dataKey&&o.direction===a.direction?u:o))},removeErrorBar:(e,t)=>{var{itemId:r,errorBar:a}=t.payload;e[r]&&(e[r]=e[r].filter(u=>u.dataKey!==a.dataKey||u.direction!==a.direction))}}}),{addErrorBar:Bq,replaceErrorBar:Lq,removeErrorBar:Uq}=V2.actions,QB=V2.reducer,WB=["children"];function FB(e,t){if(e==null)return{};var r,a,u=JB(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function JB(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}var e8={data:[],xAxisId:"xAxis-0",yAxisId:"yAxis-0",dataPointFormatter:()=>({x:0,y:0,value:0}),errorBarOffset:0},t8=A.createContext(e8);function n8(e){var{children:t}=e,r=FB(e,WB);return A.createElement(t8.Provider,{value:r},t)}function Z2(e,t){var r,a,u=me(h=>_r(h,e)),o=me(h=>Tr(h,t)),s=(r=u?.allowDataOverflow)!==null&&r!==void 0?r:un.allowDataOverflow,f=(a=o?.allowDataOverflow)!==null&&a!==void 0?a:on.allowDataOverflow,v=s||f;return{needClip:v,needClipX:s,needClipY:f}}function r8(e){var{xAxisId:t,yAxisId:r,clipPathId:a}=e,u=X2(),{needClipX:o,needClipY:s,needClip:f}=Z2(t,r);if(!f)return null;var{x:v,y:h,width:m,height:y}=u;return A.createElement("clipPath",{id:"clipPath-".concat(a)},A.createElement("rect",{x:o?v:v-m/2,y:s?h:h-y/2,width:o?m:m*2,height:s?y:y*2}))}var a8=["onMouseEnter","onMouseLeave","onClick"],i8=["value","background","tooltipPosition"],l8=["id"],u8=["onMouseEnter","onClick","onMouseLeave"];function aa(){return aa=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},aa.apply(null,arguments)}function Aw(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function Ut(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?Aw(Object(r),!0).forEach(function(a){o8(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):Aw(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function o8(e,t,r){return(t=c8(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function c8(e){var t=s8(e,"string");return typeof t=="symbol"?t:t+""}function s8(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function Fc(e,t){if(e==null)return{};var r,a,u=f8(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function f8(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}var d8=e=>{var{dataKey:t,name:r,fill:a,legendType:u,hide:o}=e;return[{inactive:o,dataKey:t,type:u,color:a,value:eA(r,t),payload:e}]};function h8(e){var{dataKey:t,stroke:r,strokeWidth:a,fill:u,name:o,hide:s,unit:f}=e;return{dataDefinedOnItem:void 0,positions:void 0,settings:{stroke:r,strokeWidth:a,fill:u,dataKey:t,nameKey:void 0,name:eA(o,t),hide:s,type:e.tooltipType,color:e.fill,unit:f}}}function v8(e){var t=me(Gi),{data:r,dataKey:a,background:u,allOtherBarProps:o}=e,{onMouseEnter:s,onMouseLeave:f,onClick:v}=o,h=Fc(o,a8),m=H2(s,a),y=Y2(f),g=G2(v,a);if(!u||r==null)return null;var b=Mt(u,!1);return A.createElement(A.Fragment,null,r.map((w,S)=>{var{value:O,background:T,tooltipPosition:M}=w,j=Fc(w,i8);if(!T)return null;var k=m(w,S),N=y(w,S),P=g(w,S),L=Ut(Ut(Ut(Ut(Ut({option:u,isActive:String(S)===t},j),{},{fill:"#eee"},T),b),Om(h,w,S)),{},{onMouseEnter:k,onMouseLeave:N,onClick:P,dataKey:a,index:S,className:"recharts-bar-background-rectangle"});return A.createElement(qp,aa({key:"background-bar-".concat(L.index)},L))}))}function m8(e){var{showLabels:t,children:r,rects:a}=e,u=a?.map(o=>{var s={x:o.x,y:o.y,width:o.width,height:o.height};return Ut(Ut({},s),{},{value:o.value,payload:o.payload,parentViewBox:o.parentViewBox,viewBox:s,fill:o.fill})});return A.createElement(H5,{value:t?u:void 0},r)}function p8(e){var{shape:t,activeBar:r,baseProps:a,entry:u,index:o,dataKey:s}=e,f=me(Gi),v=me(x2),h=r&&String(o)===f&&(v==null||s===v),m=h?r:t;return A.createElement(qp,aa({},a,u,{isActive:h,option:m,index:o,dataKey:s}))}function y8(e){var{shape:t,baseProps:r,entry:a,index:u,dataKey:o}=e;return A.createElement(qp,aa({},r,a,{isActive:!1,option:t,index:u,dataKey:o}))}function g8(e){var{data:t,props:r}=e,a=Xi(r),{id:u}=a,o=Fc(a,l8),{shape:s,dataKey:f,activeBar:v}=r,{onMouseEnter:h,onClick:m,onMouseLeave:y}=r,g=Fc(r,u8),b=H2(h,f),w=Y2(y),S=G2(m,f);return t?A.createElement(A.Fragment,null,t.map((O,T)=>A.createElement(ra,aa({className:"recharts-bar-rectangle"},Om(g,O,T),{onMouseEnter:b(O,T),onMouseLeave:w(O,T),onClick:S(O,T),key:"rectangle-".concat(O?.x,"-").concat(O?.y,"-").concat(O?.value,"-").concat(T)}),v?A.createElement(p8,{shape:s,activeBar:v,baseProps:o,entry:O,index:T,dataKey:f}):A.createElement(y8,{shape:s,baseProps:o,entry:O,index:T,dataKey:f})))):null}function b8(e){var{props:t,previousRectanglesRef:r}=e,{data:a,layout:u,isAnimationActive:o,animationBegin:s,animationDuration:f,animationEasing:v,onAnimationEnd:h,onAnimationStart:m}=t,y=r.current,g=qm(t,"recharts-bar-"),[b,w]=A.useState(!1),S=!b,O=A.useCallback(()=>{typeof h=="function"&&h(),w(!1)},[h]),T=A.useCallback(()=>{typeof m=="function"&&m(),w(!0)},[m]);return A.createElement(m8,{showLabels:S,rects:a},A.createElement(Um,{animationId:g,begin:s,duration:f,isActive:o,easing:v,onAnimationEnd:O,onAnimationStart:T,key:g},M=>{var j=M===1?a:a?.map((k,N)=>{var P=y&&y[N];if(P)return Ut(Ut({},k),{},{x:Lt(P.x,k.x,M),y:Lt(P.y,k.y,M),width:Lt(P.width,k.width,M),height:Lt(P.height,k.height,M)});if(u==="horizontal"){var L=Lt(0,k.height,M);return Ut(Ut({},k),{},{y:k.y+k.height-L,height:L})}var G=Lt(0,k.width,M);return Ut(Ut({},k),{},{width:G})});return M>0&&(r.current=j??null),j==null?null:A.createElement(ra,null,A.createElement(g8,{props:t,data:j}))}),A.createElement(K5,{label:t.label}),t.children)}function x8(e){var t=A.useRef(null);return A.createElement(b8,{previousRectanglesRef:t,props:e})}var Q2=0,S8=(e,t)=>{var r=Array.isArray(e.value)?e.value[1]:e.value;return{x:e.x,y:e.y,value:r,errorVal:$t(e,t)}};class w8 extends A.PureComponent{render(){var{hide:t,data:r,dataKey:a,className:u,xAxisId:o,yAxisId:s,needClip:f,background:v,id:h}=this.props;if(t||r==null)return null;var m=Ie("recharts-bar",u),y=h;return A.createElement(ra,{className:m,id:h},f&&A.createElement("defs",null,A.createElement(r8,{clipPathId:y,xAxisId:o,yAxisId:s})),A.createElement(ra,{className:"recharts-bar-rectangles",clipPath:f?"url(#clipPath-".concat(y,")"):void 0},A.createElement(v8,{data:r,dataKey:a,background:v,allOtherBarProps:this.props}),A.createElement(x8,this.props)))}}var O8={activeBar:!1,animationBegin:0,animationDuration:400,animationEasing:"ease",hide:!1,isAnimationActive:!Vi.isSsr,legendType:"rect",minPointSize:Q2,xAxisId:0,yAxisId:0};function A8(e){var{xAxisId:t,yAxisId:r,hide:a,legendType:u,minPointSize:o,activeBar:s,animationBegin:f,animationDuration:v,animationEasing:h,isAnimationActive:m}=e,{needClip:y}=Z2(t,r),g=Lm(),b=Nn(),w=hM(e.children,C2),S=me(M=>F8(M,t,r,b,e.id,w));if(g!=="vertical"&&g!=="horizontal")return null;var O,T=S?.[0];return T==null||T.height==null||T.width==null?O=0:O=g==="vertical"?T.height/2:T.width/2,A.createElement(n8,{xAxisId:t,yAxisId:r,data:S,dataPointFormatter:S8,errorBarOffset:O},A.createElement(w8,aa({},e,{layout:g,needClip:y,data:S,xAxisId:t,yAxisId:r,hide:a,legendType:u,minPointSize:o,activeBar:s,animationBegin:f,animationDuration:v,animationEasing:h,isAnimationActive:m})))}function E8(e){var{layout:t,barSettings:{dataKey:r,minPointSize:a},pos:u,bandSize:o,xAxis:s,yAxis:f,xAxisTicks:v,yAxisTicks:h,stackedData:m,displayedData:y,offset:g,cells:b,parentViewBox:w}=e,S=t==="horizontal"?f:s,O=m?S.scale.domain():null,T=EN({numericAxis:S});return y.map((M,j)=>{var k,N,P,L,G,ee;m?k=bN(m[j],O):(k=$t(M,r),Array.isArray(k)||(k=[T,k]));var ne=VB(a,Q2)(k[1],j);if(t==="horizontal"){var J,[re,he]=[f.scale(k[0]),f.scale(k[1])];N=y1({axis:s,ticks:v,bandSize:o,offset:u.offset,entry:M,index:j}),P=(J=he??re)!==null&&J!==void 0?J:void 0,L=u.size;var ce=re-he;if(G=dn(ce)?0:ce,ee={x:N,y:g.top,width:L,height:g.height},Math.abs(ne)>0&&Math.abs(G)<Math.abs(ne)){var se=Vt(G||ne)*(Math.abs(ne)-Math.abs(G));P-=se,G+=se}}else{var[B,I]=[s.scale(k[0]),s.scale(k[1])];if(N=B,P=y1({axis:f,ticks:h,bandSize:o,offset:u.offset,entry:M,index:j}),L=I-B,G=u.size,ee={x:g.left,y:P,width:g.width,height:G},Math.abs(ne)>0&&Math.abs(L)<Math.abs(ne)){var F=Vt(L||ne)*(Math.abs(ne)-Math.abs(L));L+=F}}if(N==null||P==null||L==null||G==null)return null;var oe=Ut(Ut({},M),{},{x:N,y:P,width:L,height:G,value:m?k:k[1],payload:M,background:ee,tooltipPosition:{x:N+L/2,y:P+G/2},parentViewBox:w},b&&b[j]&&b[j].props);return oe}).filter(Boolean)}function _8(e){var t=Pn(e,O8),r=Nn();return A.createElement(xB,{id:t.id,type:"bar"},a=>A.createElement(A.Fragment,null,A.createElement(mB,{legendPayload:d8(t)}),A.createElement(hB,{fn:h8,args:t}),A.createElement(_B,{type:"bar",id:a,data:void 0,xAxisId:t.xAxisId,yAxisId:t.yAxisId,zAxisId:0,dataKey:t.dataKey,stackId:AN(t.stackId),hide:t.hide,barSize:t.barSize,minPointSize:t.minPointSize,maxBarSize:t.maxBarSize,isPanorama:r}),A.createElement(A8,aa({},t,{id:a}))))}var W2=A.memo(_8);W2.displayName="Bar";function Ew(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function pc(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?Ew(Object(r),!0).forEach(function(a){T8(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):Ew(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function T8(e,t,r){return(t=j8(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function j8(e){var t=M8(e,"string");return typeof t=="symbol"?t:t+""}function M8(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var C8=(e,t)=>t,D8=(e,t,r)=>r,N8=(e,t,r,a)=>a,P8=(e,t,r,a,u)=>u,Lu=Y([Sp,P8],(e,t)=>e.filter(r=>r.type==="bar").find(r=>r.id===t)),R8=Y([Lu],e=>e?.maxBarSize),k8=(e,t,r,a,u,o)=>o,_w=(e,t,r)=>{var a=r??e;if(!gt(a))return qn(a,t,0)},z8=Y([ke,Sp,C8,D8,N8],(e,t,r,a,u)=>t.filter(o=>e==="horizontal"?o.xAxisId===r:o.yAxisId===a).filter(o=>o.isPanorama===u).filter(o=>o.hide===!1).filter(o=>o.type==="bar")),B8=(e,t,r,a)=>{var u=ke(e);return u==="horizontal"?om(e,"yAxis",r,a):om(e,"xAxis",t,a)},L8=(e,t,r)=>{var a=ke(e);return a==="horizontal"?NS(e,"xAxis",t):NS(e,"yAxis",r)},U8=(e,t,r)=>{var a={},u=e.filter(js),o=e.filter(h=>h.stackId==null),s=u.reduce((h,m)=>(h[m.stackId]||(h[m.stackId]=[]),h[m.stackId].push(m),h),a),f=Object.entries(s).map(h=>{var[m,y]=h,g=y.map(w=>w.dataKey),b=_w(t,r,y[0].barSize);return{stackId:m,dataKeys:g,barSize:b}}),v=o.map(h=>{var m=[h.dataKey].filter(g=>g!=null),y=_w(t,r,h.barSize);return{stackId:void 0,dataKeys:m,barSize:y}});return[...f,...v]},q8=Y([z8,Fk,L8],U8),$8=(e,t,r,a,u)=>{var o,s,f=Lu(e,t,r,a,u);if(f!=null){var v=ke(e),h=sE(e),{maxBarSize:m}=f,y=gt(m)?h:m,g,b;return v==="horizontal"?(g=Yi(e,"xAxis",t,a),b=Hi(e,"xAxis",t,a)):(g=Yi(e,"yAxis",r,a),b=Hi(e,"yAxis",r,a)),(o=(s=km(g,b,!0))!==null&&s!==void 0?s:y)!==null&&o!==void 0?o:0}},F2=(e,t,r,a)=>{var u=ke(e),o,s;return u==="horizontal"?(o=Yi(e,"xAxis",t,a),s=Hi(e,"xAxis",t,a)):(o=Yi(e,"yAxis",r,a),s=Hi(e,"yAxis",r,a)),km(o,s)};function H8(e,t,r,a,u){var o=a.length;if(!(o<1)){var s=qn(e,r,0,!0),f,v=[];if(vn(a[0].barSize)){var h=!1,m=r/o,y=a.reduce((T,M)=>T+(M.barSize||0),0);y+=(o-1)*s,y>=r&&(y-=(o-1)*s,s=0),y>=r&&m>0&&(h=!0,m*=.9,y=o*m);var g=(r-y)/2>>0,b={offset:g-s,size:0};f=a.reduce((T,M)=>{var j,k={stackId:M.stackId,dataKeys:M.dataKeys,position:{offset:b.offset+b.size+s,size:h?m:(j=M.barSize)!==null&&j!==void 0?j:0}},N=[...T,k];return b=N[N.length-1].position,N},v)}else{var w=qn(t,r,0,!0);r-2*w-(o-1)*s<=0&&(s=0);var S=(r-2*w-(o-1)*s)/o;S>1&&(S>>=0);var O=vn(u)?Math.min(S,u):S;f=a.reduce((T,M,j)=>[...T,{stackId:M.stackId,dataKeys:M.dataKeys,position:{offset:w+(S+s)*j+(S-O)/2,size:O}}],v)}return f}}var Y8=(e,t,r,a,u,o,s)=>{var f=gt(s)?t:s,v=H8(r,a,u!==o?u:o,e,f);return u!==o&&v!=null&&(v=v.map(h=>pc(pc({},h),{},{position:pc(pc({},h.position),{},{offset:h.position.offset-u/2})}))),v},G8=Y([q8,sE,Wk,fE,$8,F2,R8],Y8),K8=(e,t,r,a)=>Yi(e,"xAxis",t,a),I8=(e,t,r,a)=>Yi(e,"yAxis",r,a),X8=(e,t,r,a)=>Hi(e,"xAxis",t,a),V8=(e,t,r,a)=>Hi(e,"yAxis",r,a),Z8=Y([G8,Lu],(e,t)=>{if(!(e==null||t==null)){var r=e.find(a=>a.stackId===t.stackId&&t.dataKey!=null&&a.dataKeys.includes(t.dataKey));if(r!=null)return r.position}}),Q8=(e,t)=>{var r=bp(t);if(!(!e||r==null||t==null)){var{stackId:a}=t;if(a!=null){var u=e[a];if(u){var{stackedData:o}=u;if(o)return o.find(s=>s.key===r)}}}},W8=Y([B8,Lu],Q8),F8=Y([xt,zm,K8,I8,X8,V8,Z8,ke,sp,F2,W8,Lu,k8],(e,t,r,a,u,o,s,f,v,h,m,y,g)=>{var{chartData:b,dataStartIndex:w,dataEndIndex:S}=v;if(!(y==null||s==null||t==null||f!=="horizontal"&&f!=="vertical"||r==null||a==null||u==null||o==null||h==null)){var{data:O}=y,T;if(O!=null&&O.length>0?T=O:T=b?.slice(w,S+1),T!=null)return E8({layout:f,barSettings:y,pos:s,parentViewBox:t,bandSize:h,xAxis:r,yAxis:a,xAxisTicks:u,yAxisTicks:o,stackedData:m,displayedData:T,offset:e,cells:g})}}),J8=e=>{var{chartData:t}=e,r=bt(),a=Nn();return A.useEffect(()=>a?()=>{}:(r(HS(t)),()=>{r(HS(void 0))}),[t,r,a]),null},Tw={x:0,y:0,width:0,height:0,padding:{top:0,right:0,bottom:0,left:0}},J2=mn({name:"brush",initialState:Tw,reducers:{setBrushSettings(e,t){return t.payload==null?Tw:t.payload}}}),{setBrushSettings:qq}=J2.actions,eL=J2.reducer;function tL(e,t,r){return(t=nL(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function nL(e){var t=rL(e,"string");return typeof t=="symbol"?t:t+""}function rL(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}class $p{static create(t){return new $p(t)}constructor(t){this.scale=t}get domain(){return this.scale.domain}get range(){return this.scale.range}get rangeMin(){return this.range()[0]}get rangeMax(){return this.range()[1]}get bandwidth(){return this.scale.bandwidth}apply(t){var{bandAware:r,position:a}=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};if(t!==void 0){if(a)switch(a){case"start":return this.scale(t);case"middle":{var u=this.bandwidth?this.bandwidth()/2:0;return this.scale(t)+u}case"end":{var o=this.bandwidth?this.bandwidth():0;return this.scale(t)+o}default:return this.scale(t)}if(r){var s=this.bandwidth?this.bandwidth()/2:0;return this.scale(t)+s}return this.scale(t)}}isInRange(t){var r=this.range(),a=r[0],u=r[r.length-1];return a<=u?t>=a&&t<=u:t>=u&&t<=a}}tL($p,"EPS",1e-4);function aL(e){return(e%180+180)%180}var iL=function(t){var{width:r,height:a}=t,u=arguments.length>1&&arguments[1]!==void 0?arguments[1]:0,o=aL(u),s=o*Math.PI/180,f=Math.atan(a/r),v=s>f&&s<Math.PI-f?a/Math.sin(s):r/Math.cos(s);return Math.abs(v)},lL={dots:[],areas:[],lines:[]},e_=mn({name:"referenceElements",initialState:lL,reducers:{addDot:(e,t)=>{e.dots.push(t.payload)},removeDot:(e,t)=>{var r=yr(e).dots.findIndex(a=>a===t.payload);r!==-1&&e.dots.splice(r,1)},addArea:(e,t)=>{e.areas.push(t.payload)},removeArea:(e,t)=>{var r=yr(e).areas.findIndex(a=>a===t.payload);r!==-1&&e.areas.splice(r,1)},addLine:(e,t)=>{e.lines.push(t.payload)},removeLine:(e,t)=>{var r=yr(e).lines.findIndex(a=>a===t.payload);r!==-1&&e.lines.splice(r,1)}}}),{addDot:$q,removeDot:Hq,addArea:Yq,removeArea:Gq,addLine:Kq,removeLine:Iq}=e_.actions,uL=e_.reducer,oL=A.createContext(void 0),cL=e=>{var{children:t}=e,[r]=A.useState("".concat(vu("recharts"),"-clip")),a=X2();if(a==null)return null;var{x:u,y:o,width:s,height:f}=a;return A.createElement(oL.Provider,{value:r},A.createElement("defs",null,A.createElement("clipPath",{id:r},A.createElement("rect",{x:u,y:o,height:f,width:s}))),t)};function Ki(e,t){for(var r in e)if({}.hasOwnProperty.call(e,r)&&(!{}.hasOwnProperty.call(t,r)||e[r]!==t[r]))return!1;for(var a in t)if({}.hasOwnProperty.call(t,a)&&!{}.hasOwnProperty.call(e,a))return!1;return!0}function t_(e,t,r){if(t<1)return[];if(t===1&&r===void 0)return e;for(var a=[],u=0;u<e.length;u+=t)a.push(e[u]);return a}function sL(e,t,r){var a={width:e.width+t.width,height:e.height+t.height};return iL(a,r)}function fL(e,t,r){var a=r==="width",{x:u,y:o,width:s,height:f}=e;return t===1?{start:a?u:o,end:a?u+s:o+f}:{start:a?u+s:o+f,end:a?u:o}}function Jc(e,t,r,a,u){if(e*t<e*a||e*t>e*u)return!1;var o=r();return e*(t-e*o/2-a)>=0&&e*(t+e*o/2-u)<=0}function dL(e,t){return t_(e,t+1)}function hL(e,t,r,a,u){for(var o=(a||[]).slice(),{start:s,end:f}=t,v=0,h=1,m=s,y=function(){var w=a?.[v];if(w===void 0)return{v:t_(a,h)};var S=v,O,T=()=>(O===void 0&&(O=r(w,S)),O),M=w.coordinate,j=v===0||Jc(e,M,T,m,f);j||(v=0,m=s,h+=1),j&&(m=M+e*(T()/2+u),v+=h)},g;h<=o.length;)if(g=y(),g)return g.v;return[]}function jw(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function Pt(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?jw(Object(r),!0).forEach(function(a){vL(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):jw(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function vL(e,t,r){return(t=mL(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function mL(e){var t=pL(e,"string");return typeof t=="symbol"?t:t+""}function pL(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function yL(e,t,r,a,u){for(var o=(a||[]).slice(),s=o.length,{start:f}=t,{end:v}=t,h=function(g){var b=o[g],w,S=()=>(w===void 0&&(w=r(b,g)),w);if(g===s-1){var O=e*(b.coordinate+e*S()/2-v);o[g]=b=Pt(Pt({},b),{},{tickCoord:O>0?b.coordinate-O*e:b.coordinate})}else o[g]=b=Pt(Pt({},b),{},{tickCoord:b.coordinate});var T=Jc(e,b.tickCoord,S,f,v);T&&(v=b.tickCoord-e*(S()/2+u),o[g]=Pt(Pt({},b),{},{isShow:!0}))},m=s-1;m>=0;m--)h(m);return o}function gL(e,t,r,a,u,o){var s=(a||[]).slice(),f=s.length,{start:v,end:h}=t;if(o){var m=a[f-1],y=r(m,f-1),g=e*(m.coordinate+e*y/2-h);s[f-1]=m=Pt(Pt({},m),{},{tickCoord:g>0?m.coordinate-g*e:m.coordinate});var b=Jc(e,m.tickCoord,()=>y,v,h);b&&(h=m.tickCoord-e*(y/2+u),s[f-1]=Pt(Pt({},m),{},{isShow:!0}))}for(var w=o?f-1:f,S=function(M){var j=s[M],k,N=()=>(k===void 0&&(k=r(j,M)),k);if(M===0){var P=e*(j.coordinate-e*N()/2-v);s[M]=j=Pt(Pt({},j),{},{tickCoord:P<0?j.coordinate-P*e:j.coordinate})}else s[M]=j=Pt(Pt({},j),{},{tickCoord:j.coordinate});var L=Jc(e,j.tickCoord,N,v,h);L&&(v=j.tickCoord+e*(N()/2+u),s[M]=Pt(Pt({},j),{},{isShow:!0}))},O=0;O<w;O++)S(O);return s}function Hp(e,t,r){var{tick:a,ticks:u,viewBox:o,minTickGap:s,orientation:f,interval:v,tickFormatter:h,unit:m,angle:y}=e;if(!u||!u.length||!a)return[];if(ve(v)||Vi.isSsr){var g;return(g=dL(u,ve(v)?v:0))!==null&&g!==void 0?g:[]}var b=[],w=f==="top"||f==="bottom"?"width":"height",S=m&&w==="width"?hu(m,{fontSize:t,letterSpacing:r}):{width:0,height:0},O=(j,k)=>{var N=typeof h=="function"?h(j.value,k):j.value;return w==="width"?sL(hu(N,{fontSize:t,letterSpacing:r}),S,y):hu(N,{fontSize:t,letterSpacing:r})[w]},T=u.length>=2?Vt(u[1].coordinate-u[0].coordinate):1,M=fL(o,T,w);return v==="equidistantPreserveStart"?hL(T,M,O,u,s):(v==="preserveStart"||v==="preserveStartEnd"?b=gL(T,M,O,u,s,v==="preserveStartEnd"):b=yL(T,M,O,u,s),b.filter(j=>j.isShow))}var bL=e=>{var{ticks:t,label:r,labelGapWithTick:a=5,tickSize:u=0,tickMargin:o=0}=e,s=0;if(t){t.forEach(m=>{if(m){var y=m.getBoundingClientRect();y.width>s&&(s=y.width)}});var f=r?r.getBoundingClientRect().width:0,v=u+o,h=s+v+f+(r?a:0);return Math.round(h)}return 0},xL=["axisLine","width","height","className","hide","ticks"],SL=["viewBox"],wL=["viewBox"];function mm(e,t){if(e==null)return{};var r,a,u=OL(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function OL(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}function $a(){return $a=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},$a.apply(null,arguments)}function Mw(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function ot(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?Mw(Object(r),!0).forEach(function(a){AL(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):Mw(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function AL(e,t,r){return(t=EL(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function EL(e){var t=_L(e,"string");return typeof t=="symbol"?t:t+""}function _L(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var Yp={x:0,y:0,width:0,height:0,viewBox:{x:0,y:0,width:0,height:0},orientation:"bottom",ticks:[],stroke:"#666",tickLine:!0,axisLine:!0,tick:!0,mirror:!1,minTickGap:5,tickSize:6,tickMargin:2,interval:"preserveEnd"};function TL(e){var{x:t,y:r,width:a,height:u,orientation:o,mirror:s,axisLine:f,otherSvgProps:v}=e;if(!f)return null;var h=ot(ot(ot({},v),Mt(f,!1)),{},{fill:"none"});if(o==="top"||o==="bottom"){var m=+(o==="top"&&!s||o==="bottom"&&s);h=ot(ot({},h),{},{x1:t,y1:r+m*u,x2:t+a,y2:r+m*u})}else{var y=+(o==="left"&&!s||o==="right"&&s);h=ot(ot({},h),{},{x1:t+y*a,y1:r,x2:t+y*a,y2:r+u})}return A.createElement("line",$a({},h,{className:Ie("recharts-cartesian-axis-line",ki(f,"className"))}))}function jL(e,t,r,a,u,o,s,f,v){var h,m,y,g,b,w,S=f?-1:1,O=e.tickSize||s,T=ve(e.tickCoord)?e.tickCoord:e.coordinate;switch(o){case"top":h=m=e.coordinate,g=r+ +!f*u,y=g-S*O,w=y-S*v,b=T;break;case"left":y=g=e.coordinate,m=t+ +!f*a,h=m-S*O,b=h-S*v,w=T;break;case"right":y=g=e.coordinate,m=t+ +f*a,h=m+S*O,b=h+S*v,w=T;break;default:h=m=e.coordinate,g=r+ +f*u,y=g+S*O,w=y+S*v,b=T;break}return{line:{x1:h,y1:y,x2:m,y2:g},tick:{x:b,y:w}}}function ML(e,t){switch(e){case"left":return t?"start":"end";case"right":return t?"end":"start";default:return"middle"}}function CL(e,t){switch(e){case"left":case"right":return"middle";case"top":return t?"start":"end";default:return t?"end":"start"}}function DL(e){var{option:t,tickProps:r,value:a}=e,u,o=Ie(r.className,"recharts-cartesian-axis-tick-value");if(A.isValidElement(t))u=A.cloneElement(t,ot(ot({},r),{},{className:o}));else if(typeof t=="function")u=t(ot(ot({},r),{},{className:o}));else{var s="recharts-cartesian-axis-tick-value";typeof t!="boolean"&&(s=Ie(s,t?.className)),u=A.createElement(Lp,$a({},r,{className:s}),a)}return u}function NL(e){var{ticks:t=[],tick:r,tickLine:a,stroke:u,tickFormatter:o,unit:s,padding:f,tickTextProps:v,orientation:h,mirror:m,x:y,y:g,width:b,height:w,tickSize:S,tickMargin:O,fontSize:T,letterSpacing:M,getTicksConfig:j,events:k}=e,N=Hp(ot(ot({},j),{},{ticks:t}),T,M),P=ML(h,m),L=CL(h,m),G=Xi(j),ee=Mt(r,!1),ne=ot(ot({},G),{},{fill:"none"},Mt(a,!1)),J=N.map((re,he)=>{var{line:ce,tick:se}=jL(re,y,g,b,w,h,S,m,O),B=ot(ot(ot(ot({textAnchor:P,verticalAnchor:L},G),{},{stroke:"none",fill:u},ee),se),{},{index:he,payload:re,visibleTicksCount:N.length,tickFormatter:o,padding:f},v);return A.createElement(ra,$a({className:"recharts-cartesian-axis-tick",key:"tick-".concat(re.value,"-").concat(re.coordinate,"-").concat(re.tickCoord)},Om(k,re,he)),a&&A.createElement("line",$a({},ne,ce,{className:Ie("recharts-cartesian-axis-tick-line",ki(a,"className"))})),r&&A.createElement(DL,{option:r,tickProps:B,value:"".concat(typeof o=="function"?o(re.value,he):re.value).concat(s||"")}))});return J.length>0?A.createElement("g",{className:"recharts-cartesian-axis-ticks"},J):null}var PL=A.forwardRef((e,t)=>{var{axisLine:r,width:a,height:u,className:o,hide:s,ticks:f}=e,v=mm(e,xL),[h,m]=A.useState(""),[y,g]=A.useState(""),b=A.useRef([]);A.useImperativeHandle(t,()=>({getCalculatedWidth:()=>{var S;return bL({ticks:b.current,label:(S=e.labelRef)===null||S===void 0?void 0:S.current,labelGapWithTick:5,tickSize:e.tickSize,tickMargin:e.tickMargin})}}));var w=A.useCallback(S=>{if(S){var O=S.getElementsByClassName("recharts-cartesian-axis-tick-value");b.current=Array.from(O);var T=O[0];if(T){var M=window.getComputedStyle(T),j=M.fontSize,k=M.letterSpacing;(j!==h||k!==y)&&(m(j),g(k))}}},[h,y]);return s||a!=null&&a<=0||u!=null&&u<=0?null:A.createElement(ra,{className:Ie("recharts-cartesian-axis",o),ref:w},A.createElement(TL,{x:e.x,y:e.y,width:a,height:u,orientation:e.orientation,mirror:e.mirror,axisLine:r,otherSvgProps:Xi(e)}),A.createElement(NL,{ticks:f,tick:e.tick,tickLine:e.tickLine,stroke:e.stroke,tickFormatter:e.tickFormatter,unit:e.unit,padding:e.padding,tickTextProps:e.tickTextProps,orientation:e.orientation,mirror:e.mirror,x:e.x,y:e.y,width:e.width,height:e.height,tickSize:e.tickSize,tickMargin:e.tickMargin,fontSize:h,letterSpacing:y,getTicksConfig:e,events:v}),A.createElement(S5,{x:e.x,y:e.y,width:e.width,height:e.height},A.createElement(D5,{label:e.label}),e.children))}),RL=A.memo(PL,(e,t)=>{var{viewBox:r}=e,a=mm(e,SL),{viewBox:u}=t,o=mm(t,wL);return Ki(r,u)&&Ki(a,o)}),Gp=A.forwardRef((e,t)=>{var r=Pn(e,Yp);return A.createElement(RL,$a({},r,{ref:t}))});Gp.displayName="CartesianAxis";var kL=["x1","y1","x2","y2","key"],zL=["offset"],BL=["xAxisId","yAxisId"],LL=["xAxisId","yAxisId"];function Cw(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function Rt(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?Cw(Object(r),!0).forEach(function(a){UL(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):Cw(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function UL(e,t,r){return(t=qL(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function qL(e){var t=$L(e,"string");return typeof t=="symbol"?t:t+""}function $L(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}function Pa(){return Pa=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},Pa.apply(null,arguments)}function es(e,t){if(e==null)return{};var r,a,u=HL(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function HL(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}var YL=e=>{var{fill:t}=e;if(!t||t==="none")return null;var{fillOpacity:r,x:a,y:u,width:o,height:s,ry:f}=e;return A.createElement("rect",{x:a,y:u,ry:f,width:o,height:s,stroke:"none",fill:t,fillOpacity:r,className:"recharts-cartesian-grid-bg"})};function n_(e,t){var r;if(A.isValidElement(e))r=A.cloneElement(e,t);else if(typeof e=="function")r=e(t);else{var{x1:a,y1:u,x2:o,y2:s,key:f}=t,v=es(t,kL),h=Xi(v),{offset:m}=h,y=es(h,zL);r=A.createElement("line",Pa({},y,{x1:a,y1:u,x2:o,y2:s,fill:"none",key:f}))}return r}function GL(e){var{x:t,width:r,horizontal:a=!0,horizontalPoints:u}=e;if(!a||!u||!u.length)return null;var{xAxisId:o,yAxisId:s}=e,f=es(e,BL),v=u.map((h,m)=>{var y=Rt(Rt({},f),{},{x1:t,y1:h,x2:t+r,y2:h,key:"line-".concat(m),index:m});return n_(a,y)});return A.createElement("g",{className:"recharts-cartesian-grid-horizontal"},v)}function KL(e){var{y:t,height:r,vertical:a=!0,verticalPoints:u}=e;if(!a||!u||!u.length)return null;var{xAxisId:o,yAxisId:s}=e,f=es(e,LL),v=u.map((h,m)=>{var y=Rt(Rt({},f),{},{x1:h,y1:t,x2:h,y2:t+r,key:"line-".concat(m),index:m});return n_(a,y)});return A.createElement("g",{className:"recharts-cartesian-grid-vertical"},v)}function IL(e){var{horizontalFill:t,fillOpacity:r,x:a,y:u,width:o,height:s,horizontalPoints:f,horizontal:v=!0}=e;if(!v||!t||!t.length)return null;var h=f.map(y=>Math.round(y+u-u)).sort((y,g)=>y-g);u!==h[0]&&h.unshift(0);var m=h.map((y,g)=>{var b=!h[g+1],w=b?u+s-y:h[g+1]-y;if(w<=0)return null;var S=g%t.length;return A.createElement("rect",{key:"react-".concat(g),y,x:a,height:w,width:o,stroke:"none",fill:t[S],fillOpacity:r,className:"recharts-cartesian-grid-bg"})});return A.createElement("g",{className:"recharts-cartesian-gridstripes-horizontal"},m)}function XL(e){var{vertical:t=!0,verticalFill:r,fillOpacity:a,x:u,y:o,width:s,height:f,verticalPoints:v}=e;if(!t||!r||!r.length)return null;var h=v.map(y=>Math.round(y+u-u)).sort((y,g)=>y-g);u!==h[0]&&h.unshift(0);var m=h.map((y,g)=>{var b=!h[g+1],w=b?u+s-y:h[g+1]-y;if(w<=0)return null;var S=g%r.length;return A.createElement("rect",{key:"react-".concat(g),x:y,y:o,width:w,height:f,stroke:"none",fill:r[S],fillOpacity:a,className:"recharts-cartesian-grid-bg"})});return A.createElement("g",{className:"recharts-cartesian-gridstripes-vertical"},m)}var VL=(e,t)=>{var{xAxis:r,width:a,height:u,offset:o}=e;return FO(Hp(Rt(Rt(Rt({},Yp),r),{},{ticks:JO(r),viewBox:{x:0,y:0,width:a,height:u}})),o.left,o.left+o.width,t)},ZL=(e,t)=>{var{yAxis:r,width:a,height:u,offset:o}=e;return FO(Hp(Rt(Rt(Rt({},Yp),r),{},{ticks:JO(r),viewBox:{x:0,y:0,width:a,height:u}})),o.top,o.top+o.height,t)},QL={horizontal:!0,vertical:!0,horizontalPoints:[],verticalPoints:[],stroke:"#ccc",fill:"none",verticalFill:[],horizontalFill:[],xAxisId:0,yAxisId:0};function r_(e){var t=rA(),r=aA(),a=nA(),u=Rt(Rt({},Pn(e,QL)),{},{x:ve(e.x)?e.x:a.left,y:ve(e.y)?e.y:a.top,width:ve(e.width)?e.width:a.width,height:ve(e.height)?e.height:a.height}),{xAxisId:o,yAxisId:s,x:f,y:v,width:h,height:m,syncWithTicks:y,horizontalValues:g,verticalValues:b}=u,w=Nn(),S=me(ee=>PS(ee,"xAxis",o,w)),O=me(ee=>PS(ee,"yAxis",s,w));if(!ve(h)||h<=0||!ve(m)||m<=0||!ve(f)||f!==+f||!ve(v)||v!==+v)return null;var T=u.verticalCoordinatesGenerator||VL,M=u.horizontalCoordinatesGenerator||ZL,{horizontalPoints:j,verticalPoints:k}=u;if((!j||!j.length)&&typeof M=="function"){var N=g&&g.length,P=M({yAxis:O?Rt(Rt({},O),{},{ticks:N?g:O.ticks}):void 0,width:t,height:r,offset:a},N?!0:y);du(Array.isArray(P),"horizontalCoordinatesGenerator should return Array but instead it returned [".concat(typeof P,"]")),Array.isArray(P)&&(j=P)}if((!k||!k.length)&&typeof T=="function"){var L=b&&b.length,G=T({xAxis:S?Rt(Rt({},S),{},{ticks:L?b:S.ticks}):void 0,width:t,height:r,offset:a},L?!0:y);du(Array.isArray(G),"verticalCoordinatesGenerator should return Array but instead it returned [".concat(typeof G,"]")),Array.isArray(G)&&(k=G)}return A.createElement("g",{className:"recharts-cartesian-grid"},A.createElement(YL,{fill:u.fill,fillOpacity:u.fillOpacity,x:u.x,y:u.y,width:u.width,height:u.height,ry:u.ry}),A.createElement(IL,Pa({},u,{horizontalPoints:j})),A.createElement(XL,Pa({},u,{verticalPoints:k})),A.createElement(GL,Pa({},u,{offset:a,horizontalPoints:j,xAxis:S,yAxis:O})),A.createElement(KL,Pa({},u,{offset:a,verticalPoints:k,xAxis:S,yAxis:O})))}r_.displayName="CartesianGrid";var WL=["dangerouslySetInnerHTML","ticks"],FL=["id"],JL=["domain"],eU=["domain"];function pm(){return pm=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},pm.apply(null,arguments)}function ts(e,t){if(e==null)return{};var r,a,u=tU(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function tU(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}function nU(e){var t=bt();return A.useEffect(()=>(t(DB(e)),()=>{t(NB(e))}),[e,t]),null}var rU=e=>{var{xAxisId:t,className:r}=e,a=me(zm),u=Nn(),o="xAxis",s=me(O=>el(O,o,t,u)),f=me(O=>t2(O,o,t,u)),v=me(O=>WE(O,t)),h=me(O=>G4(O,t)),m=me(O=>xE(O,t));if(v==null||h==null||m==null)return null;var{dangerouslySetInnerHTML:y,ticks:g}=e,b=ts(e,WL),{id:w}=m,S=ts(m,FL);return A.createElement(Gp,pm({},b,S,{scale:s,x:h.x,y:h.y,width:v.width,height:v.height,className:Ie("recharts-".concat(o," ").concat(o),r),viewBox:a,ticks:f}))},aU={allowDataOverflow:un.allowDataOverflow,allowDecimals:un.allowDecimals,allowDuplicatedCategory:un.allowDuplicatedCategory,height:un.height,hide:!1,mirror:un.mirror,orientation:un.orientation,padding:un.padding,reversed:un.reversed,scale:un.scale,tickCount:un.tickCount,type:un.type,xAxisId:0},iU=e=>{var t,r,a,u,o,s=Pn(e,aU);return A.createElement(A.Fragment,null,A.createElement(nU,{interval:(t=s.interval)!==null&&t!==void 0?t:"preserveEnd",id:s.xAxisId,scale:s.scale,type:s.type,padding:s.padding,allowDataOverflow:s.allowDataOverflow,domain:s.domain,dataKey:s.dataKey,allowDuplicatedCategory:s.allowDuplicatedCategory,allowDecimals:s.allowDecimals,tickCount:s.tickCount,includeHidden:(r=s.includeHidden)!==null&&r!==void 0?r:!1,reversed:s.reversed,ticks:s.ticks,height:s.height,orientation:s.orientation,mirror:s.mirror,hide:s.hide,unit:s.unit,name:s.name,angle:(a=s.angle)!==null&&a!==void 0?a:0,minTickGap:(u=s.minTickGap)!==null&&u!==void 0?u:5,tick:(o=s.tick)!==null&&o!==void 0?o:!0,tickFormatter:s.tickFormatter}),A.createElement(rU,s))},lU=(e,t)=>{var{domain:r}=e,a=ts(e,JL),{domain:u}=t,o=ts(t,eU);return Ki(a,o)?Array.isArray(r)&&r.length===2&&Array.isArray(u)&&u.length===2?r[0]===u[0]&&r[1]===u[1]:Ki({domain:r},{domain:u}):!1},a_=A.memo(iU,lU);a_.displayName="XAxis";var uU=["dangerouslySetInnerHTML","ticks"],oU=["id"],cU=["domain"],sU=["domain"];function ym(){return ym=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},ym.apply(null,arguments)}function ns(e,t){if(e==null)return{};var r,a,u=fU(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function fU(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}function dU(e){var t=bt();return A.useEffect(()=>(t(PB(e)),()=>{t(RB(e))}),[e,t]),null}var hU=e=>{var{yAxisId:t,className:r,width:a,label:u}=e,o=A.useRef(null),s=A.useRef(null),f=me(zm),v=Nn(),h=bt(),m="yAxis",y=me(N=>el(N,m,t,v)),g=me(N=>FE(N,t)),b=me(N=>I4(N,t)),w=me(N=>t2(N,m,t,v)),S=me(N=>SE(N,t));if(A.useLayoutEffect(()=>{if(!(a!=="auto"||!g||Up(u)||A.isValidElement(u)||S==null)){var N=o.current;if(N){var P=N.getCalculatedWidth();Math.round(g.width)!==Math.round(P)&&h(kB({id:t,width:P}))}}},[w,g,h,u,t,a,S]),g==null||b==null||S==null)return null;var{dangerouslySetInnerHTML:O,ticks:T}=e,M=ns(e,uU),{id:j}=S,k=ns(S,oU);return A.createElement(Gp,ym({},M,k,{ref:o,labelRef:s,scale:y,x:b.x,y:b.y,tickTextProps:a==="auto"?{width:void 0}:{width:a},width:g.width,height:g.height,className:Ie("recharts-".concat(m," ").concat(m),r),viewBox:f,ticks:w}))},vU={allowDataOverflow:on.allowDataOverflow,allowDecimals:on.allowDecimals,allowDuplicatedCategory:on.allowDuplicatedCategory,hide:!1,mirror:on.mirror,orientation:on.orientation,padding:on.padding,reversed:on.reversed,scale:on.scale,tickCount:on.tickCount,type:on.type,width:on.width,yAxisId:0},mU=e=>{var t,r,a,u,o,s=Pn(e,vU);return A.createElement(A.Fragment,null,A.createElement(dU,{interval:(t=s.interval)!==null&&t!==void 0?t:"preserveEnd",id:s.yAxisId,scale:s.scale,type:s.type,domain:s.domain,allowDataOverflow:s.allowDataOverflow,dataKey:s.dataKey,allowDuplicatedCategory:s.allowDuplicatedCategory,allowDecimals:s.allowDecimals,tickCount:s.tickCount,padding:s.padding,includeHidden:(r=s.includeHidden)!==null&&r!==void 0?r:!1,reversed:s.reversed,ticks:s.ticks,width:s.width,orientation:s.orientation,mirror:s.mirror,hide:s.hide,unit:s.unit,name:s.name,angle:(a=s.angle)!==null&&a!==void 0?a:0,minTickGap:(u=s.minTickGap)!==null&&u!==void 0?u:5,tick:(o=s.tick)!==null&&o!==void 0?o:!0,tickFormatter:s.tickFormatter}),A.createElement(hU,s))},pU=(e,t)=>{var{domain:r}=e,a=ns(e,cU),{domain:u}=t,o=ns(t,sU);return Ki(a,o)?Array.isArray(r)&&r.length===2&&Array.isArray(u)&&u.length===2?r[0]===u[0]&&r[1]===u[1]:Ki({domain:r},{domain:u}):!1},i_=A.memo(mU,pU);i_.displayName="YAxis";var _v={exports:{}},Tv={};/**
 * @license React
 * use-sync-external-store-with-selector.production.js
 *
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var Dw;function yU(){if(Dw)return Tv;Dw=1;var e=Ii();function t(v,h){return v===h&&(v!==0||1/v===1/h)||v!==v&&h!==h}var r=typeof Object.is=="function"?Object.is:t,a=e.useSyncExternalStore,u=e.useRef,o=e.useEffect,s=e.useMemo,f=e.useDebugValue;return Tv.useSyncExternalStoreWithSelector=function(v,h,m,y,g){var b=u(null);if(b.current===null){var w={hasValue:!1,value:null};b.current=w}else w=b.current;b=s(function(){function O(N){if(!T){if(T=!0,M=N,N=y(N),g!==void 0&&w.hasValue){var P=w.value;if(g(P,N))return j=P}return j=N}if(P=j,r(M,N))return P;var L=y(N);return g!==void 0&&g(P,L)?(M=N,P):(M=N,j=L)}var T=!1,M,j,k=m===void 0?null:m;return[function(){return O(h())},k===null?void 0:function(){return O(k())}]},[h,m,y,g]);var S=a(v,b[0],b[1]);return o(function(){w.hasValue=!0,w.value=S},[S]),f(S),S},Tv}var Nw;function gU(){return Nw||(Nw=1,_v.exports=yU()),_v.exports}gU();function bU(e){e()}function xU(){let e=null,t=null;return{clear(){e=null,t=null},notify(){bU(()=>{let r=e;for(;r;)r.callback(),r=r.next})},get(){const r=[];let a=e;for(;a;)r.push(a),a=a.next;return r},subscribe(r){let a=!0;const u=t={callback:r,next:null,prev:t};return u.prev?u.prev.next=u:e=u,function(){!a||e===null||(a=!1,u.next?u.next.prev=u.prev:t=u.prev,u.prev?u.prev.next=u.next:e=u.next)}}}}var Pw={notify(){},get:()=>[]};function SU(e,t){let r,a=Pw,u=0,o=!1;function s(S){m();const O=a.subscribe(S);let T=!1;return()=>{T||(T=!0,O(),y())}}function f(){a.notify()}function v(){w.onStateChange&&w.onStateChange()}function h(){return o}function m(){u++,r||(r=e.subscribe(v),a=xU())}function y(){u--,r&&u===0&&(r(),r=void 0,a.clear(),a=Pw)}function g(){o||(o=!0,m())}function b(){o&&(o=!1,y())}const w={addNestedSub:s,notifyNestedSubs:f,handleChangeWrapper:v,isSubscribed:h,trySubscribe:g,tryUnsubscribe:b,getListeners:()=>a};return w}var wU=()=>typeof window<"u"&&typeof window.document<"u"&&typeof window.document.createElement<"u",OU=wU(),AU=()=>typeof navigator<"u"&&navigator.product==="ReactNative",EU=AU(),_U=()=>OU||EU?A.useLayoutEffect:A.useEffect,TU=_U(),jU=Symbol.for("react-redux-context"),MU=typeof globalThis<"u"?globalThis:{};function CU(){if(!A.createContext)return{};const e=MU[jU]??=new Map;let t=e.get(A.createContext);return t||(t=A.createContext(null),e.set(A.createContext,t)),t}var DU=CU();function NU(e){const{children:t,context:r,serverState:a,store:u}=e,o=A.useMemo(()=>{const v=SU(u);return{store:u,subscription:v,getServerState:a?()=>a:void 0}},[u,a]),s=A.useMemo(()=>u.getState(),[u]);TU(()=>{const{subscription:v}=o;return v.onStateChange=v.notifyNestedSubs,v.trySubscribe(),s!==u.getState()&&v.notifyNestedSubs(),()=>{v.tryUnsubscribe(),v.onStateChange=void 0}},[o,s]);const f=r||DU;return A.createElement(f.Provider,{value:o},t)}var PU=NU,RU=(e,t)=>t,Kp=Y([RU,ke,yE,st,p2,jr,Q6,xt],tz),Ip=e=>{var t=e.currentTarget.getBoundingClientRect(),r=t.width/e.currentTarget.offsetWidth,a=t.height/e.currentTarget.offsetHeight;return{chartX:Math.round((e.clientX-t.left)/r),chartY:Math.round((e.clientY-t.top)/a)}},l_=Cn("mouseClick"),u_=_u();u_.startListening({actionCreator:l_,effect:(e,t)=>{var r=e.payload,a=Kp(t.getState(),Ip(r));a?.activeIndex!=null&&t.dispatch(i6({activeIndex:a.activeIndex,activeDataKey:void 0,activeCoordinate:a.activeCoordinate}))}});var gm=Cn("mouseMove"),o_=_u();o_.startListening({actionCreator:gm,effect:(e,t)=>{var r=e.payload,a=t.getState(),u=Mp(a,a.tooltip.settings.shared),o=Kp(a,Ip(r));u==="axis"&&(o?.activeIndex!=null?t.dispatch(c2({activeIndex:o.activeIndex,activeDataKey:void 0,activeCoordinate:o.activeCoordinate})):t.dispatch(o2()))}});var Rw={accessibilityLayer:!0,barCategoryGap:"10%",barGap:4,barSize:void 0,className:void 0,maxBarSize:void 0,stackOffset:"none",syncId:void 0,syncMethod:"index"},c_=mn({name:"rootProps",initialState:Rw,reducers:{updateOptions:(e,t)=>{var r;e.accessibilityLayer=t.payload.accessibilityLayer,e.barCategoryGap=t.payload.barCategoryGap,e.barGap=(r=t.payload.barGap)!==null&&r!==void 0?r:Rw.barGap,e.barSize=t.payload.barSize,e.maxBarSize=t.payload.maxBarSize,e.stackOffset=t.payload.stackOffset,e.syncId=t.payload.syncId,e.syncMethod=t.payload.syncMethod,e.className=t.payload.className}}}),kU=c_.reducer,{updateOptions:zU}=c_.actions,s_=mn({name:"polarOptions",initialState:null,reducers:{updatePolarOptions:(e,t)=>t.payload}}),{updatePolarOptions:Xq}=s_.actions,BU=s_.reducer,f_=Cn("keyDown"),d_=Cn("focus"),Xp=_u();Xp.startListening({actionCreator:f_,effect:(e,t)=>{var r=t.getState(),a=r.rootProps.accessibilityLayer!==!1;if(a){var{keyboardInteraction:u}=r.tooltip,o=e.payload;if(!(o!=="ArrowRight"&&o!=="ArrowLeft"&&o!=="Enter")){var s=Number(Cp(u,nl(r))),f=jr(r);if(o==="Enter"){var v=Vc(r,"axis","hover",String(u.index));t.dispatch(sm({active:!u.active,activeIndex:u.index,activeDataKey:u.dataKey,activeCoordinate:v}));return}var h=Q4(r),m=h==="left-to-right"?1:-1,y=o==="ArrowRight"?1:-1,g=s+y*m;if(!(f==null||g>=f.length||g<0)){var b=Vc(r,"axis","hover",String(g));t.dispatch(sm({active:!0,activeIndex:g.toString(),activeDataKey:void 0,activeCoordinate:b}))}}}}});Xp.startListening({actionCreator:d_,effect:(e,t)=>{var r=t.getState(),a=r.rootProps.accessibilityLayer!==!1;if(a){var{keyboardInteraction:u}=r.tooltip;if(!u.active&&u.index==null){var o="0",s=Vc(r,"axis","hover",String(o));t.dispatch(sm({activeDataKey:void 0,active:!0,activeIndex:o,activeCoordinate:s}))}}}});var Tn=Cn("externalEvent"),h_=_u();h_.startListening({actionCreator:Tn,effect:(e,t)=>{if(e.payload.handler!=null){var r=t.getState(),a={activeCoordinate:$6(r),activeDataKey:x2(r),activeIndex:Gi(r),activeLabel:b2(r),activeTooltipIndex:Gi(r),isTooltipActive:H6(r)};e.payload.handler(a,e.payload.reactEvent)}}});var LU=Y([tl],e=>e.tooltipItemPayloads),UU=Y([LU,zu,(e,t,r)=>t,(e,t,r)=>r],(e,t,r,a)=>{var u=e.find(f=>f.settings.dataKey===a);if(u!=null){var{positions:o}=u;if(o!=null){var s=t(o,r);return s}}}),v_=Cn("touchMove"),m_=_u();m_.startListening({actionCreator:v_,effect:(e,t)=>{var r=e.payload,a=t.getState(),u=Mp(a,a.tooltip.settings.shared);if(u==="axis"){var o=Kp(a,Ip({clientX:r.touches[0].clientX,clientY:r.touches[0].clientY,currentTarget:r.currentTarget}));o?.activeIndex!=null&&t.dispatch(c2({activeIndex:o.activeIndex,activeDataKey:void 0,activeCoordinate:o.activeCoordinate}))}else if(u==="item"){var s,f=r.touches[0],v=document.elementFromPoint(f.clientX,f.clientY);if(!v||!v.getAttribute)return;var h=v.getAttribute(PN),m=(s=v.getAttribute(RN))!==null&&s!==void 0?s:void 0,y=UU(t.getState(),h,m);t.dispatch(u2({activeDataKey:m,activeIndex:h,activeCoordinate:y}))}}});var qU=jO({brush:eL,cartesianAxis:zB,chartData:gz,errorBars:QB,graphicalItems:EB,layout:nN,legend:ZN,options:hz,polarAxis:X5,polarOptions:BU,referenceElements:uL,rootProps:kU,tooltip:l6}),$U=function(t){return TD({reducer:qU,preloadedState:t,middleware:r=>r({serializableCheck:!1}).concat([u_.middleware,o_.middleware,Xp.middleware,h_.middleware,m_.middleware]),devTools:Vi.devToolsEnabled})};function HU(e){var{preloadedState:t,children:r,reduxStoreName:a}=e,u=Nn(),o=A.useRef(null);if(u)return r;o.current==null&&(o.current=$U(t));var s=Mm;return A.createElement(PU,{context:s,store:o.current},r)}function YU(e){var{layout:t,width:r,height:a,margin:u}=e,o=bt(),s=Nn();return A.useEffect(()=>{s||(o(JD(t)),o(eN({width:r,height:a})),o(FD(u)))},[o,s,t,r,a,u]),null}function GU(e){var t=bt();return A.useEffect(()=>{t(zU(e))},[t,e]),null}var KU=["children"];function IU(e,t){if(e==null)return{};var r,a,u=XU(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function XU(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}function rs(){return rs=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},rs.apply(null,arguments)}var VU={width:"100%",height:"100%"},ZU=A.forwardRef((e,t)=>{var r=rA(),a=aA(),u=lA();if(!Nc(r)||!Nc(a))return null;var{children:o,otherAttributes:s,title:f,desc:v}=e,h,m;return typeof s.tabIndex=="number"?h=s.tabIndex:h=u?0:void 0,typeof s.role=="string"?m=s.role:m=u?"application":void 0,A.createElement(Qw,rs({},s,{title:f,desc:v,role:m,tabIndex:h,width:r,height:a,style:VU,ref:t}),o)}),QU=e=>{var{children:t}=e,r=me(ps);if(!r)return null;var{width:a,height:u,y:o,x:s}=r;return A.createElement(Qw,{width:a,height:u,x:s,y:o},t)},kw=A.forwardRef((e,t)=>{var{children:r}=e,a=IU(e,KU),u=Nn();return u?A.createElement(QU,null,r):A.createElement(ZU,rs({ref:t},a),r)});function WU(){var e=bt(),[t,r]=A.useState(null),a=me(NN);return A.useEffect(()=>{if(t!=null){var u=t.getBoundingClientRect(),o=u.width/t.offsetWidth;vn(o)&&o!==a&&e(tN(o))}},[t,e,a]),r}function zw(e,t){var r=Object.keys(e);if(Object.getOwnPropertySymbols){var a=Object.getOwnPropertySymbols(e);t&&(a=a.filter(function(u){return Object.getOwnPropertyDescriptor(e,u).enumerable})),r.push.apply(r,a)}return r}function FU(e){for(var t=1;t<arguments.length;t++){var r=arguments[t]!=null?arguments[t]:{};t%2?zw(Object(r),!0).forEach(function(a){JU(e,a,r[a])}):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(r)):zw(Object(r)).forEach(function(a){Object.defineProperty(e,a,Object.getOwnPropertyDescriptor(r,a))})}return e}function JU(e,t,r){return(t=eq(t))in e?Object.defineProperty(e,t,{value:r,enumerable:!0,configurable:!0,writable:!0}):e[t]=r,e}function eq(e){var t=tq(e,"string");return typeof t=="symbol"?t:t+""}function tq(e,t){if(typeof e!="object"||!e)return e;var r=e[Symbol.toPrimitive];if(r!==void 0){var a=r.call(e,t);if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(t==="string"?String:Number)(e)}var nq=A.forwardRef((e,t)=>{var{children:r,className:a,height:u,onClick:o,onContextMenu:s,onDoubleClick:f,onMouseDown:v,onMouseEnter:h,onMouseLeave:m,onMouseMove:y,onMouseUp:g,onTouchEnd:b,onTouchMove:w,onTouchStart:S,style:O,width:T}=e,M=bt(),[j,k]=A.useState(null),[N,P]=A.useState(null);Tz();var L=WU(),G=A.useCallback(Z=>{L(Z),typeof t=="function"&&t(Z),k(Z),P(Z)},[L,t,k,P]),ee=A.useCallback(Z=>{M(l_(Z)),M(Tn({handler:o,reactEvent:Z}))},[M,o]),ne=A.useCallback(Z=>{M(gm(Z)),M(Tn({handler:h,reactEvent:Z}))},[M,h]),J=A.useCallback(Z=>{M(o2()),M(Tn({handler:m,reactEvent:Z}))},[M,m]),re=A.useCallback(Z=>{M(gm(Z)),M(Tn({handler:y,reactEvent:Z}))},[M,y]),he=A.useCallback(()=>{M(d_())},[M]),ce=A.useCallback(Z=>{M(f_(Z.key))},[M]),se=A.useCallback(Z=>{M(Tn({handler:s,reactEvent:Z}))},[M,s]),B=A.useCallback(Z=>{M(Tn({handler:f,reactEvent:Z}))},[M,f]),I=A.useCallback(Z=>{M(Tn({handler:v,reactEvent:Z}))},[M,v]),F=A.useCallback(Z=>{M(Tn({handler:g,reactEvent:Z}))},[M,g]),oe=A.useCallback(Z=>{M(Tn({handler:S,reactEvent:Z}))},[M,S]),D=A.useCallback(Z=>{M(v_(Z)),M(Tn({handler:w,reactEvent:Z}))},[M,w]),W=A.useCallback(Z=>{M(Tn({handler:b,reactEvent:Z}))},[M,b]);return A.createElement(_2.Provider,{value:j},A.createElement(wM.Provider,{value:N},A.createElement("div",{className:Ie("recharts-wrapper",a),style:FU({position:"relative",cursor:"default",width:T,height:u},O),onClick:ee,onContextMenu:se,onDoubleClick:B,onFocus:he,onKeyDown:ce,onMouseDown:I,onMouseEnter:ne,onMouseLeave:J,onMouseMove:re,onMouseUp:F,onTouchEnd:W,onTouchMove:D,onTouchStart:oe,ref:G},r)))}),rq=["children","className","width","height","style","compact","title","desc"];function aq(e,t){if(e==null)return{};var r,a,u=iq(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function iq(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}var lq=A.forwardRef((e,t)=>{var{children:r,className:a,width:u,height:o,style:s,compact:f,title:v,desc:h}=e,m=aq(e,rq),y=Xi(m);return f?A.createElement(kw,{otherAttributes:y,title:v,desc:h},r):A.createElement(nq,{className:a,style:s,width:u,height:o,onClick:e.onClick,onMouseLeave:e.onMouseLeave,onMouseEnter:e.onMouseEnter,onMouseMove:e.onMouseMove,onMouseDown:e.onMouseDown,onMouseUp:e.onMouseUp,onContextMenu:e.onContextMenu,onDoubleClick:e.onDoubleClick,onTouchStart:e.onTouchStart,onTouchMove:e.onTouchMove,onTouchEnd:e.onTouchEnd},A.createElement(kw,{otherAttributes:y,title:v,desc:h,ref:t},A.createElement(cL,null,r)))}),uq=["width","height"];function bm(){return bm=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var r=arguments[t];for(var a in r)({}).hasOwnProperty.call(r,a)&&(e[a]=r[a])}return e},bm.apply(null,arguments)}function oq(e,t){if(e==null)return{};var r,a,u=cq(e,t);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(e);for(a=0;a<o.length;a++)r=o[a],t.indexOf(r)===-1&&{}.propertyIsEnumerable.call(e,r)&&(u[r]=e[r])}return u}function cq(e,t){if(e==null)return{};var r={};for(var a in e)if({}.hasOwnProperty.call(e,a)){if(t.indexOf(a)!==-1)continue;r[a]=e[a]}return r}var sq={top:5,right:5,bottom:5,left:5},fq={accessibilityLayer:!0,layout:"horizontal",stackOffset:"none",barCategoryGap:"10%",barGap:4,margin:sq,reverseStackOrder:!1,syncMethod:"index"},dq=A.forwardRef(function(t,r){var a,u=Pn(t.categoricalChartProps,fq),{width:o,height:s}=u,f=oq(u,uq);if(!Nc(o)||!Nc(s))return null;var{chartName:v,defaultTooltipEventType:h,validateTooltipEventTypes:m,tooltipPayloadSearcher:y,categoricalChartProps:g}=t,b={chartName:v,defaultTooltipEventType:h,validateTooltipEventTypes:m,tooltipPayloadSearcher:y,eventEmitter:void 0};return A.createElement(HU,{preloadedState:{options:b},reduxStoreName:(a=g.id)!==null&&a!==void 0?a:v},A.createElement(J8,{chartData:g.data}),A.createElement(YU,{width:o,height:s,layout:u.layout,margin:u.margin}),A.createElement(GU,{accessibilityLayer:u.accessibilityLayer,barCategoryGap:u.barCategoryGap,maxBarSize:u.maxBarSize,stackOffset:u.stackOffset,barGap:u.barGap,barSize:u.barSize,syncId:u.syncId,syncMethod:u.syncMethod,className:u.className}),A.createElement(lq,bm({},f,{width:o,height:s,ref:r})))}),hq=["axis","item"],vq=A.forwardRef((e,t)=>A.createElement(dq,{chartName:"BarChart",defaultTooltipEventType:"axis",validateTooltipEventTypes:hq,tooltipPayloadSearcher:fz,categoricalChartProps:e,ref:t}));function mq({data:e,title:t,height:r=400,className:a=""}){return!e||e.length===0?_.jsx("div",{className:`flex items-center justify-center ${a}`,style:{height:r/4},children:_.jsx("p",{className:"text-gray-500",children:"No data available"})}):_.jsxs("div",{className:`w-full ${a}`,children:[t&&_.jsx("h3",{className:"text-lg font-semibold mb-4 text-gray-800",children:t}),_.jsx(Kz,{width:"100%",height:r,children:_.jsxs(vq,{data:e,margin:{top:20,right:30,left:20,bottom:60},children:[_.jsx(r_,{strokeDasharray:"3 3",className:"opacity-30"}),_.jsx(a_,{dataKey:"name",angle:-45,textAnchor:"end",height:60,fontSize:12,className:"fill-gray-600"}),_.jsx(i_,{fontSize:12,className:"fill-gray-600"}),_.jsx(kz,{contentStyle:{backgroundColor:"white",border:"1px solid #e2e8f0",borderRadius:"6px",boxShadow:"0 4px 6px -1px rgba(0, 0, 0, 0.1)",fontSize:"14px"},formatter:u=>[u.toFixed(2),"Score"],labelStyle:{color:"#374151",fontWeight:"bold"}}),_.jsx(W2,{dataKey:"value",fill:"#3b82f6",radius:[2,2,0,0],className:"hover:opacity-80 transition-opacity"})]})})]})}function Bw(e){return Object.entries(e).map(([t,r])=>({name:t,value:r})).sort((t,r)=>r.value-t.value)}function pq({stats:e,className:t=""}){const{overall_scores:r,metadata:a}=e,u=Bw(r),o=Object.keys(r).length>0;return _.jsxs("div",{className:`space-y-6 ${t}`,children:[_.jsxs("div",{className:"bg-white rounded-lg shadow-sm border border-gray-200 p-6",children:[_.jsx("h2",{className:"text-xl font-bold text-gray-900 mb-4",children:"Summary Statistics"}),_.jsxs("div",{className:"grid grid-cols-2 md:grid-cols-4 gap-4",children:[_.jsxs("div",{className:"text-center",children:[_.jsx("div",{className:"text-2xl font-bold text-blue-600",children:a.num_players}),_.jsx("div",{className:"text-sm text-gray-600",children:"Players"})]}),_.jsxs("div",{className:"text-center",children:[_.jsx("div",{className:"text-2xl font-bold text-green-600",children:a.num_games}),_.jsx("div",{className:"text-sm text-gray-600",children:"Games"})]}),_.jsxs("div",{className:"text-center",children:[_.jsx("div",{className:"text-2xl font-bold text-purple-600",children:a.num_maps}),_.jsx("div",{className:"text-sm text-gray-600",children:"Maps"})]}),_.jsxs("div",{className:"text-center",children:[_.jsx("div",{className:"text-2xl font-bold text-orange-600",children:Object.keys(r).length}),_.jsx("div",{className:"text-sm text-gray-600",children:"Active Players"})]})]})]}),o?_.jsxs(_.Fragment,{children:[_.jsxs("div",{className:"grid grid-cols-1 lg:grid-cols-2 gap-6",children:[_.jsx("div",{className:"bg-white rounded-lg shadow-sm border border-gray-200 p-6",children:_.jsx(Jj,{data:u,title:"Rankings"})}),_.jsx("div",{className:"bg-white rounded-lg shadow-sm border border-gray-200 p-6",children:_.jsx(mq,{data:u,title:"Score Distribution",height:300})})]}),Object.keys(e.per_game_scores).length>0&&_.jsxs("div",{className:"bg-white rounded-lg shadow-sm border border-gray-200 p-6",children:[_.jsx("h3",{className:"text-lg font-semibold text-gray-900 mb-4",children:"Per-Game Breakdown"}),_.jsx("div",{className:"grid grid-cols-1 md:grid-cols-2 lg:grid-cols-3 gap-4",children:Object.entries(e.per_game_scores).map(([s,f])=>{const h=Bw(f)[0],m=Object.keys(f).length;return _.jsxs("div",{className:"border border-gray-200 rounded-lg p-4",children:[_.jsx("h4",{className:"font-medium text-gray-900 mb-2",children:s}),_.jsxs("div",{className:"space-y-2",children:[_.jsxs("div",{className:"flex justify-between text-sm",children:[_.jsx("span",{className:"text-gray-600",children:"Top Player:"}),_.jsx("span",{className:"font-medium",children:h?.name||"N/A"})]}),_.jsxs("div",{className:"flex justify-between text-sm",children:[_.jsx("span",{className:"text-gray-600",children:"Top Score:"}),_.jsx("span",{className:"font-mono",children:h?.value.toFixed(2)||"N/A"})]}),_.jsxs("div",{className:"flex justify-between text-sm",children:[_.jsx("span",{className:"text-gray-600",children:"Players:"}),_.jsx("span",{children:m})]})]})]},s)})})]})]}):_.jsx("div",{className:"bg-white rounded-lg shadow-sm border border-gray-200 p-12",children:_.jsxs("div",{className:"text-center",children:[_.jsx("div",{className:"w-16 h-16 mx-auto mb-4 text-gray-400",children:_.jsx("svg",{fill:"none",stroke:"currentColor",viewBox:"0 0 24 24",children:_.jsx("path",{strokeLinecap:"round",strokeLinejoin:"round",strokeWidth:1.5,d:"M9 19v-6a2 2 0 00-2-2H5a2 2 0 00-2 2v6a2 2 0 002 2h2a2 2 0 002-2zm0 0V9a2 2 0 012-2h2a2 2 0 012 2v10m-6 0a2 2 0 002 2h2a2 2 0 002-2m0 0V5a2 2 0 012-2h2a2 2 0 012 2v14a2 2 0 01-2 2h-2a2 2 0 01-2-2z"})})}),_.jsx("h3",{className:"text-lg font-medium text-gray-900 mb-2",children:"No Results Yet"}),_.jsx("p",{className:"text-gray-500 max-w-sm mx-auto",children:e.status==="ready"?'This benchmark is ready to run. Click "Start Benchmark" to begin execution.':e.status==="running"?"The benchmark is currently running. Results will appear as they become available.":"No benchmark results found. Make sure the benchmark has been executed."})]})})]})}function yq({isOpen:e,onClose:t,title:r,children:a,maxWidth:u="600px"}){return A.useEffect(()=>{const o=s=>{s.key==="Escape"&&e&&t()};return e&&(document.addEventListener("keydown",o),document.body.style.overflow="hidden"),()=>{document.removeEventListener("keydown",o),document.body.style.overflow="unset"}},[e,t]),e?_.jsx(_.Fragment,{children:_.jsx("div",{onClick:t,style:{position:"fixed",top:0,left:0,right:0,bottom:0,backgroundColor:"rgba(0, 0, 0, 0.5)",zIndex:1e3,display:"flex",alignItems:"center",justifyContent:"center",padding:"20px"},children:_.jsxs("div",{onClick:o=>o.stopPropagation(),style:{backgroundColor:"white",borderRadius:"8px",maxWidth:u,width:"100%",maxHeight:"90vh",display:"flex",flexDirection:"column",boxShadow:"0 20px 25px -5px rgba(0, 0, 0, 0.1), 0 10px 10px -5px rgba(0, 0, 0, 0.04)"},children:[_.jsxs("div",{style:{padding:"20px",borderBottom:"1px solid #e5e7eb",display:"flex",alignItems:"center",justifyContent:"space-between"},children:[_.jsx("h2",{style:{margin:0,fontSize:"20px",fontWeight:600,color:"#111827"},children:r}),_.jsx("button",{onClick:t,style:{background:"none",border:"none",cursor:"pointer",padding:"4px",display:"flex",alignItems:"center",justifyContent:"center",borderRadius:"4px",transition:"background-color 0.2s"},onMouseEnter:o=>o.currentTarget.style.backgroundColor="#f3f4f6",onMouseLeave:o=>o.currentTarget.style.backgroundColor="transparent","aria-label":"Close modal",children:_.jsx("svg",{width:"20",height:"20",fill:"none",stroke:"currentColor",strokeWidth:"2",viewBox:"0 0 24 24",children:_.jsx("path",{strokeLinecap:"round",strokeLinejoin:"round",d:"M6 18L18 6M6 6l12 12"})})})]}),_.jsx("div",{style:{padding:"20px",overflowY:"auto",flex:1},children:a})]})})}):null}function jv(e){if(!e)return{model:"",params:{}};const t=e.indexOf("?");if(t===-1)return{model:e,params:{}};const r=e.slice(0,t),a=e.slice(t+1),u={},o=new URLSearchParams(a);for(const[s,f]of o.entries())if(!(s in u)){let v=f;try{v=JSON.parse(f)}catch{}u[s]=v}return{model:r,params:u}}function Lw(e){const t=e.toLowerCase();return t.includes("gpt")||t.includes("o3")||t.includes("o4")?"openai":t.includes("claude")?"anthropic":t.includes("gemini")?"gemini":t.includes("grok")?"grok":t.includes("deepseek")?"deepseek":e.includes("/")&&!e.startsWith("ollama/")?"openrouter":e.startsWith("ollama/")?"ollama":"openai"}function p_({onSubmit:e,onCancel:t,existingPlayerIds:r=[],submitLabel:a="Add Players",embedded:u=!1,value:o,onChange:s}){const f=()=>o&&o.length>0?o[0].player_type==="human"?[]:o.map(P=>P.id):["gpt-4o-mini"],v=()=>!!(o&&o.length>0&&o[0].player_type==="human"),[h,m]=A.useState(f()),[y,g]=A.useState(v()),[b,w]=A.useState(null),S=(P,L)=>L?[{name:"black",id:"human",player_type:"human",options:{}}]:P.map(G=>{const{model:ee,params:ne}=jv(G),J={name:"black",id:ee,player_type:"default",options:{model:ee,provider:Lw(ee)}};return ne&&Object.keys(ne).length>0&&(J.options.request_params=ne),J}),O=(P,L)=>{if(u&&s){const G=S(P,L);s(G)}},T=()=>{const P=[...h,"gpt-4o-mini"];m(P),O(P,y)},M=P=>{const L=h.filter((G,ee)=>ee!==P);m(L),O(L,y)},j=(P,L)=>{const G=h.map((ee,ne)=>ne===P?L:ee);m(G),w(null),O(G,y)},k=P=>{g(P),w(null),O(h,P)},N=P=>{if(P.preventDefault(),u)return;let L;if(y){if(r.includes("human")){w("A human player already exists in this benchmark");return}L=S(h,!0)}else{const G=h.map(ee=>jv(ee).model).filter(ee=>r.includes(ee));if(G.length>0){w(`These models already exist in the benchmark: ${G.join(", ")}`);return}if(h.some(ee=>!ee.trim())){w("All model names must be filled in");return}L=S(h,!1)}e&&e(L)};return _.jsxs("form",{onSubmit:N,children:[b&&_.jsx("div",{style:{marginBottom:"15px",padding:"10px",backgroundColor:"#fee",border:"1px solid #fcc",borderRadius:"4px",color:"#c00"},children:b}),_.jsxs("div",{style:{marginBottom:"20px"},children:[_.jsxs("label",{style:{display:"flex",alignItems:"center",marginBottom:"15px"},children:[_.jsx("input",{type:"checkbox",checked:y,onChange:P=>k(P.target.checked),style:{marginRight:"10px"}}),_.jsx("span",{children:u?"Human Player":"Add Human Player"})]}),y?_.jsxs("div",{style:{padding:"15px",backgroundColor:"#e8f4fd",border:"1px solid #b3d9ff",borderRadius:"4px"},children:[_.jsx("strong",{children:"Human Player Mode"}),_.jsx("p",{style:{margin:"5px 0",fontSize:"14px"},children:"A human player will be added for interactive testing."})]}):_.jsxs("div",{children:[_.jsx("h4",{style:{marginBottom:"10px"},children:"Model Players"}),h.map((P,L)=>_.jsx("div",{style:{marginBottom:"10px",padding:"10px",backgroundColor:"#f5f5f5",borderRadius:"4px"},children:_.jsxs("div",{style:{display:"flex",gap:"10px",alignItems:"center"},children:[_.jsx("input",{type:"text",placeholder:"Model name (e.g., gpt-4o, claude-3-sonnet)",value:P,onChange:G=>j(L,G.target.value),style:{flex:1,padding:"8px",border:"1px solid #ddd",borderRadius:"4px"},required:!0}),_.jsx("span",{style:{fontSize:"12px",color:"#666",minWidth:"80px"},children:Lw(jv(P).model)}),h.length>1&&_.jsx("button",{type:"button",onClick:()=>M(L),style:{padding:"6px 12px",backgroundColor:"#ff4444",color:"white",border:"none",borderRadius:"4px",cursor:"pointer"},children:"Remove"})]})},L)),_.jsxs("div",{style:{marginTop:"6px",fontSize:"12px",color:"#666"},children:["Tip: add query params like ",_.jsx("code",{children:'?temperature=0.7&reasoning_effort="high"'})]}),_.jsx("button",{type:"button",onClick:T,style:{padding:"8px 16px",backgroundColor:"#4CAF50",color:"white",border:"none",borderRadius:"4px",cursor:"pointer",marginTop:"10px"},children:"+ Add Another Model"})]})]}),r.length>0&&_.jsxs("div",{style:{marginBottom:"20px",padding:"10px",backgroundColor:"#f8f9fa",borderRadius:"4px"},children:[_.jsx("strong",{children:"Existing Players:"}),_.jsx("div",{style:{marginTop:"5px"},children:r.map(P=>_.jsx("span",{style:{display:"inline-block",marginRight:"8px",marginTop:"4px",padding:"4px 8px",backgroundColor:"#e9ecef",borderRadius:"3px",fontSize:"12px"},children:P},P))})]}),!u&&_.jsxs("div",{style:{display:"flex",gap:"10px",justifyContent:"flex-end"},children:[_.jsx("button",{type:"button",onClick:t,style:{padding:"10px 20px",backgroundColor:"#6c757d",color:"white",border:"none",borderRadius:"4px",cursor:"pointer"},children:"Cancel"}),_.jsx("button",{type:"submit",style:{padding:"10px 20px",backgroundColor:"#2196F3",color:"white",border:"none",borderRadius:"4px",cursor:"pointer"},children:a})]})]})}function gq({benchmarkId:e,onBack:t}){const[r,a]=A.useState(null),[u,o]=A.useState(!0),[s,f]=A.useState(null),[v,h]=A.useState("overall"),[m,y]=A.useState(!1),[g,b]=A.useState(!1),[w,S]=A.useState(!1),[O,T]=A.useState(!1),[M,j]=A.useState(!1),k=async()=>{try{o(!0),f(null);const ne=await Vj(e);a(ne)}catch(ne){f(ne instanceof Error?ne.message:"Failed to load benchmark stats")}finally{o(!1)}},N=async()=>{if(!g)try{b(!0),await Qj(e),setTimeout(k,5e3)}catch(ne){alert(`Error starting benchmark: ${ne instanceof Error?ne.message:"Unknown error"}`)}finally{b(!1)}},P=async()=>{if(!w){S(!0),setTimeout(()=>S(!1),3e3);return}try{await Wj(e),await k(),S(!1)}catch(ne){alert(`Error deleting results: ${ne instanceof Error?ne.message:"Unknown error"}`)}},L=async ne=>{try{j(!0),await Fj(e,ne),T(!1),await k()}catch(J){alert(`Error adding players: ${J instanceof Error?J.message:"Unknown error"}`)}finally{j(!1)}},G=()=>r?.config?.players?r.config.players.map(ne=>ne.id):[];A.useEffect(()=>{k()},[e]),A.useEffect(()=>{if(!r||r.status!=="running")return;const ne=setInterval(k,5e3);return()=>clearInterval(ne)},[r?.status]);const ee=r?Object.keys(r.per_game_scores):[];return u&&!r?_.jsx("div",{className:"min-h-screen bg-gray-50 flex items-center justify-center",children:_.jsxs("div",{className:"text-center",children:[_.jsx("div",{className:"animate-spin rounded-full h-12 w-12 border-b-2 border-blue-600 mx-auto mb-4"}),_.jsx("p",{className:"text-gray-600",children:"Loading benchmark data..."})]})}):s&&!r?_.jsx("div",{className:"min-h-screen bg-gray-50 flex items-center justify-center",children:_.jsxs("div",{className:"text-center max-w-md mx-auto",children:[_.jsx("div",{className:"w-16 h-16 mx-auto mb-4 text-red-400",children:_.jsx("svg",{fill:"none",stroke:"currentColor",viewBox:"0 0 24 24",children:_.jsx("path",{strokeLinecap:"round",strokeLinejoin:"round",strokeWidth:1.5,d:"M12 9v2m0 4h.01m-6.938 4h13.856c1.54 0 2.502-1.667 1.732-2.5L13.732 4c-.77-.833-1.964-.833-2.732 0L3.732 16.5c-.77.833.192 2.5 1.732 2.5z"})})}),_.jsx("h3",{className:"text-lg font-medium text-gray-900 mb-2",children:"Error Loading Benchmark"}),_.jsx("p",{className:"text-gray-600 mb-4",children:s}),_.jsx("button",{onClick:k,className:"px-4 py-2 bg-blue-600 text-white rounded-md hover:bg-blue-700 transition-colors mr-2",children:"Retry"}),_.jsx("button",{onClick:t,className:"px-4 py-2 bg-gray-300 text-gray-700 rounded-md hover:bg-gray-400 transition-colors",children:"Back to List"})]})}):r?(r.metadata.actual_results>0,_.jsxs("div",{className:"min-h-screen bg-gray-50",children:[_.jsx("div",{className:"bg-white shadow-sm border-b border-gray-200",children:_.jsx("div",{className:"max-w-7xl mx-auto px-4 sm:px-6 lg:px-8",children:_.jsxs("div",{className:"flex items-center justify-between h-16",children:[_.jsxs("div",{className:"flex items-center space-x-4",children:[_.jsxs("button",{onClick:t,className:"inline-flex items-center px-3 py-2 border border-gray-300 rounded-md text-sm font-medium text-gray-700 bg-white hover:bg-gray-50 transition-colors",children:[_.jsx("svg",{className:"w-4 h-4 mr-2",fill:"none",stroke:"currentColor",viewBox:"0 0 24 24",children:_.jsx("path",{strokeLinecap:"round",strokeLinejoin:"round",strokeWidth:2,d:"M15 19l-7-7 7-7"})}),"Back to List"]}),_.jsxs("div",{children:[_.jsx("h1",{className:"text-xl font-semibold text-gray-900",children:"Benchmark Dashboard"}),_.jsx("p",{className:"text-sm text-gray-600",children:e})]})]}),_.jsxs("div",{className:"flex items-center space-x-4",children:[r.status==="running"&&_.jsxs("span",{className:"inline-flex items-center text-sm text-green-600",children:[_.jsx("span",{className:"animate-pulse w-2 h-2 bg-green-500 rounded-full mr-2"}),"Live Updates"]}),_.jsxs("button",{onClick:k,disabled:u,className:"inline-flex items-center px-3 py-2 border border-gray-300 rounded-md text-sm font-medium text-gray-700 bg-white hover:bg-gray-50 transition-colors disabled:opacity-50",children:[_.jsx("svg",{className:`w-4 h-4 mr-2 ${u?"animate-spin":""}`,fill:"none",stroke:"currentColor",viewBox:"0 0 24 24",children:_.jsx("path",{strokeLinecap:"round",strokeLinejoin:"round",strokeWidth:2,d:"M4 4v5h.582m15.356 2A8.001 8.001 0 004.582 9m0 0H9m11 11v-5h-.581m0 0a8.003 8.003 0 01-15.357-2m15.357 2H15"})}),"Refresh"]})]})]})})}),_.jsx("div",{className:"bg-white shadow-sm border-b border-gray-200",children:_.jsx("div",{className:"max-w-7xl mx-auto px-4 sm:px-6 lg:px-8 py-4",children:_.jsxs("div",{className:"flex flex-col lg:flex-row lg:items-center lg:justify-between gap-4",children:[_.jsxs("div",{className:"flex-1",children:[_.jsxs("div",{className:"flex items-center gap-4 mb-2",children:[_.jsx("span",{className:`inline-flex items-center px-3 py-1 rounded-full text-sm font-medium ${r.status==="completed"?"bg-green-100 text-green-800":r.status==="running"?"bg-yellow-100 text-yellow-800":"bg-blue-100 text-blue-800"}`,children:r.status==="completed"?"✓ Completed":r.status==="running"?"⏳ Running":"🚀 Ready to Start"}),_.jsxs("span",{className:"text-sm text-gray-600",children:[r.metadata.actual_results," / ",r.metadata.expected_results," results"]})]}),_.jsx("div",{className:"w-full bg-gray-200 rounded-full h-2",children:_.jsx("div",{className:`h-2 rounded-full transition-all duration-500 ${r.status==="completed"?"bg-green-600":r.status==="running"?"bg-blue-600":"bg-gray-400"}`,style:{width:`${Math.min(100,r.metadata.actual_results/r.metadata.expected_results*100)}%`}})})]}),_.jsxs("div",{className:"flex gap-2",children:[_.jsx("button",{onClick:N,disabled:g||r.status==="running",className:`px-4 py-2 rounded-md text-sm font-medium transition-colors ${g||r.status==="running"||r.status==="completed"?"bg-gray-300 text-gray-500 cursor-not-allowed":"bg-blue-600 text-white hover:bg-blue-700"}`,children:g?"⏳ Starting...":r.status==="ready"?"🚀 Start Benchmark":r.status==="running"?"⏳ Running...":r.status==="completed"?"✅ Finished":"🚀 Continue Benchmark"}),_.jsx("button",{onClick:()=>T(!0),disabled:r.status==="running",className:`px-4 py-2 rounded-md text-sm font-medium transition-colors ${r.status==="running"?"bg-gray-300 text-gray-500 cursor-not-allowed":"bg-green-600 text-white hover:bg-green-700"}`,children:"➕ Add Player"}),_.jsx("button",{onClick:P,className:`px-4 py-2 rounded-md text-sm font-medium transition-colors ${w?"bg-red-600 text-white hover:bg-red-700":"bg-gray-300 text-gray-700 hover:bg-gray-400"}`,children:w?"⚠️ Confirm Delete?":"🗑️ Delete Results"})]})]})})}),_.jsx("div",{className:"bg-white shadow-sm border-b border-gray-200",children:_.jsxs("div",{className:"max-w-7xl mx-auto px-4 sm:px-6 lg:px-8",children:[_.jsxs("button",{onClick:()=>y(!m),className:"w-full py-4 flex items-center justify-between text-left hover:bg-gray-50 transition-colors",children:[_.jsxs("div",{className:"flex items-center gap-4",children:[_.jsx("h3",{className:"text-lg font-medium text-gray-900",children:"Configuration"}),_.jsxs("div",{className:"flex gap-2 text-sm text-gray-600",children:[_.jsxs("span",{className:"px-2 py-1 bg-gray-100 rounded",children:[r.metadata.num_players," Player",r.metadata.num_players!==1?"s":""]}),_.jsxs("span",{className:"px-2 py-1 bg-gray-100 rounded",children:[r.metadata.num_games," Game",r.metadata.num_games!==1?"s":""]}),_.jsxs("span",{className:"px-2 py-1 bg-gray-100 rounded",children:["Judge: ",r.config?.metadata?.judge_model||"N/A"]})]})]}),_.jsx("svg",{className:`w-5 h-5 text-gray-500 transition-transform ${m?"rotate-180":""}`,fill:"none",stroke:"currentColor",viewBox:"0 0 24 24",children:_.jsx("path",{strokeLinecap:"round",strokeLinejoin:"round",strokeWidth:2,d:"M19 9l-7 7-7-7"})})]}),m&&_.jsx("div",{className:"pb-4",children:_.jsx("div",{className:"bg-gray-50 rounded-lg p-4 max-h-96 overflow-y-auto",children:_.jsx("pre",{className:"text-xs font-mono text-gray-700 whitespace-pre-wrap",children:JSON.stringify(r.config,null,2)})})})]})}),_.jsx("div",{className:"bg-white border-b border-gray-200",children:_.jsx("div",{className:"max-w-7xl mx-auto px-4 sm:px-6 lg:px-8",children:_.jsxs("nav",{className:"flex space-x-8","aria-label":"Tabs",children:[_.jsx("button",{onClick:()=>h("overall"),className:`py-4 px-1 border-b-2 font-medium text-sm transition-colors ${v==="overall"?"border-blue-500 text-blue-600":"border-transparent text-gray-500 hover:text-gray-700 hover:border-gray-300"}`,children:"Overall"}),ee.map(ne=>_.jsx("button",{onClick:()=>h(ne),className:`py-4 px-1 border-b-2 font-medium text-sm transition-colors ${v===ne?"border-blue-500 text-blue-600":"border-transparent text-gray-500 hover:text-gray-700 hover:border-gray-300"}`,children:ne},ne))]})})}),_.jsx("div",{className:"max-w-7xl mx-auto px-4 sm:px-6 lg:px-8 py-8",children:v==="overall"?_.jsx(pq,{stats:r}):_.jsxs("div",{className:"bg-white rounded-lg shadow-sm border border-gray-200 p-6",children:[_.jsxs("h3",{className:"text-lg font-semibold text-gray-900 mb-4",children:[v," - Detailed Analysis"]}),_.jsx("p",{className:"text-gray-600",children:"Game-specific charts and analysis will be implemented in the next phase."})]})}),_.jsx(yq,{isOpen:O,onClose:()=>T(!1),title:"Add Players to Benchmark",children:_.jsx(p_,{onSubmit:L,onCancel:()=>T(!1),existingPlayerIds:G(),submitLabel:M?"Adding...":"Add Players"})})]})):null}function bq({varName:e,maxLen:t,onSubmit:r}){const[a,u]=A.useState(""),[o,s]=A.useState(!1),f=()=>{!o&&a.trim()&&(r(a),s(!0))},v=h=>{h.key==="Enter"&&!h.shiftKey&&(h.preventDefault(),f())};return o?_.jsxs("div",{style:{marginBottom:"15px",padding:"10px",backgroundColor:"#e8f5e9",borderRadius:"5px",border:"1px solid #4CAF50"},children:[_.jsxs("strong",{children:["Input for ",e,":"]})," ",a]}):_.jsxs("div",{style:{marginBottom:"15px",padding:"15px",backgroundColor:"#fff3cd",borderRadius:"5px",border:"1px solid #ffc107"},children:[_.jsxs("div",{style:{marginBottom:"10px"},children:[_.jsxs("strong",{children:["Input requested for variable: ",e]}),_.jsxs("span",{style:{marginLeft:"10px",fontSize:"12px",color:"#666"},children:["(max ",t," tokens)"]})]}),_.jsxs("div",{style:{display:"flex",gap:"10px"},children:[_.jsx("input",{type:"text",value:a,onChange:h=>u(h.target.value),onKeyPress:v,placeholder:`Enter your response for ${e}...`,autoFocus:!0,style:{flex:1,padding:"8px",border:"1px solid #ddd",borderRadius:"4px",fontSize:"14px"}}),_.jsx("button",{onClick:f,disabled:!a.trim(),style:{padding:"8px 16px",backgroundColor:a.trim()?"#4CAF50":"#ccc",color:"white",border:"none",borderRadius:"4px",cursor:a.trim()?"pointer":"not-allowed"},children:"Submit"})]})]})}function xq({code:e,currentLine:t,registers:r,currentRound:a=0,totalRounds:u=0}){return _.jsxs("div",{style:{backgroundColor:"#f8f9fa",padding:"15px",borderRadius:"5px"},children:[_.jsxs("div",{style:{display:"flex",justifyContent:"space-between",alignItems:"center",marginBottom:"15px"},children:[_.jsx("h3",{style:{margin:0},children:"Game State"}),a>0&&_.jsxs("div",{style:{fontSize:"14px",color:"#666",backgroundColor:"#e3f2fd",padding:"4px 8px",borderRadius:"4px",border:"1px solid #bbdefb"},children:["Round: ",a," | Completed: ",u]})]}),_.jsxs("div",{style:{marginBottom:"15px"},children:[_.jsx("h4",{style:{fontSize:"14px",marginBottom:"10px"},children:"Code Execution"}),_.jsx("div",{style:{backgroundColor:"white",border:"1px solid #ddd",borderRadius:"3px",padding:"10px",maxHeight:"150px",overflowY:"auto",fontFamily:"monospace",fontSize:"12px"},children:e.length===0?_.jsx("div",{style:{color:"#666"},children:"No code loaded"}):e.map((o,s)=>_.jsxs("div",{style:{padding:"2px 5px",backgroundColor:t===s?"#fffbdd":"transparent",borderLeft:t===s?"3px solid #ffc107":"3px solid transparent",whiteSpace:"pre-wrap",wordBreak:"break-word"},children:[_.jsx("span",{style:{color:"#666",marginRight:"10px",minWidth:"20px",display:"inline-block"},children:s+1}),o]},s))})]}),_.jsxs("div",{children:[_.jsx("h4",{style:{fontSize:"14px",marginBottom:"10px"},children:"Registers"}),_.jsx("div",{style:{backgroundColor:"white",border:"1px solid #ddd",borderRadius:"3px",padding:"10px",minHeight:"40px"},children:(()=>{const o=Object.entries(r).filter(([s,f])=>f&&(typeof f=="string"&&f.trim()!==""||Array.isArray(f)&&f.length>0));return o.length===0?_.jsx("div",{style:{color:"#666",fontStyle:"italic"},children:"No registers set"}):_.jsx("table",{style:{width:"100%",fontSize:"12px"},children:_.jsx("tbody",{children:o.map(([s,f])=>_.jsxs("tr",{children:[_.jsxs("td",{style:{fontWeight:"bold",paddingRight:"15px",color:"#2196F3",verticalAlign:"top",width:"60px"},children:[s,":"]}),_.jsx("td",{style:{fontFamily:"monospace",wordBreak:"break-word",whiteSpace:"pre-wrap"},children:f})]},s))})})})()})]})]})}function Uw({error:e,isWarning:t=!1}){return _.jsxs("div",{style:{marginBottom:"15px",padding:"12px",backgroundColor:t?"#fff3cd":"#f8d7da",borderRadius:"5px",border:`1px solid ${t?"#ffc107":"#f5c6cb"}`},children:[_.jsxs("strong",{style:{color:t?"#856404":"#721c24"},children:[t?"⚠️ Warning":"❌ Error",":"]}),_.jsx("span",{style:{marginLeft:"8px",color:t?"#856404":"#721c24"},children:e})]})}const Mv=e=>e.toFixed(4);function Sq({perTokenXent:e}){const[t,r]=A.useState(null),a=e.pairs,u=a.reduce((s,[f,v])=>s+v,0),o=s=>{let f=Math.tanh(s/5);const v=Math.abs(f);return s>=0?{backgroundColor:`rgba(37, 99, 235, ${v})`}:{backgroundColor:`rgba(239, 68, 68, ${v})`}};return!a||a.length===0?_.jsx("div",{style:{marginBottom:"15px"},children:_.jsxs("span",{style:{fontSize:"12px",color:"#666"},children:["Reward: ",Mv(u)]})}):_.jsxs("div",{style:{marginBottom:"15px"},children:[_.jsxs("span",{style:{fontSize:"12px",color:"#666",marginBottom:"8px",display:"block"},children:["Reward: ",Mv(u)]}),_.jsx("div",{style:{display:"flex",flexWrap:"wrap",gap:"4px",position:"relative"},children:a.map(([s,f],v)=>_.jsxs("span",{style:{fontFamily:"monospace",fontSize:"14px",padding:"2px 4px",borderRadius:"3px",display:"inline-block",position:"relative",cursor:"default",...o(f)},onMouseEnter:()=>r(v),onMouseLeave:()=>r(null),children:[_.jsx("span",{style:{mixBlendMode:"difference",color:"white"},children:s||"⎵"}),t===v&&_.jsxs("div",{style:{position:"absolute",bottom:"100%",left:"50%",transform:"translateX(-50%)",marginBottom:"8px",padding:"4px 8px",backgroundColor:"#374151",color:"white",fontSize:"12px",borderRadius:"4px",whiteSpace:"nowrap",zIndex:1e3,boxShadow:"0 2px 4px rgba(0,0,0,0.1)"},children:["Score: ",Mv(f),_.jsx("div",{style:{position:"absolute",top:"100%",left:"50%",transform:"translateX(-50%)",width:0,height:0,borderLeft:"4px solid transparent",borderRight:"4px solid transparent",borderTop:"4px solid #374151"}})]})]},v))})]})}const wq=`assign(s="Once upon a time, there was a brave knight who fought dragons and saved kingdoms.")
reveal(s)
elicit(t, 10)
reveal(t)
assign(t1=remove_common_words(t, s))
reveal(t1)
reward(xed(s | t1))`;function Oq({onBack:e}){const t=A.useRef(null),[r,a]=A.useState(wq),[u,o]=A.useState([]),[s,f]=A.useState(-1),[v,h]=A.useState({}),[m,y]=A.useState([]),[g,b]=A.useState("disconnected"),[w,S]=A.useState(!1),[O,T]=A.useState(!1),[M,j]=A.useState(null),[k,N]=A.useState(0),[P,L]=A.useState([]),G=A.useRef(0),ee=A.useRef(null),ne=B=>{y(I=>[...I,B])},J=B=>{t.current&&t.current.readyState===WebSocket.OPEN?(t.current.send(JSON.stringify(B)),console.log("Sent message:",B)):console.error("WebSocket is not open. Cannot send message.")},re=B=>{J({type:"xent_input",input:B})},he=()=>{J({type:"xent_control",command:"start",code:r}),S(!0),T(!1),j(null),N(0),L([]),y([]),h({}),f(-1)},ce=A.useCallback(B=>{try{const I=JSON.parse(B.data);if(console.log("Received message:",I),I.type==="xent_error"){console.error("Error from server:",I.error),ne(_.jsx(Uw,{error:I.error},`output-${++G.current}`)),S(!1);return}if(I.type!=="xent_event"){console.warn("Received non-event message:",I);return}const F=I.event;switch(console.log("Received event:",F),F.line_num!==void 0&&f(F.line_num),F.type){case"round_started":N(F.round_index+1),ne(_.jsx("div",{style:{marginBottom:"10px",padding:"8px 12px",backgroundColor:"#e3f2fd",borderRadius:"4px",borderLeft:"4px solid #2196F3",fontSize:"14px",color:"#1976D2"},children:_.jsxs("strong",{children:["Round ",F.round_index+1," started"]})},`output-${++G.current}`));break;case"round_finished":ne(_.jsx("div",{style:{marginBottom:"15px",padding:"8px 12px",backgroundColor:"#f3e5f5",borderRadius:"4px",borderLeft:"4px solid #9C27B0",fontSize:"14px",color:"#7B1FA2"},children:_.jsxs("strong",{children:["Round ",F.round_index+1," completed"]})},`output-${++G.current}`));break;case"elicit_request":F.registers&&(console.log("Updating registers from elicit_request:",F.registers),h(F.registers)),ne(_.jsx(bq,{varName:F.var_name,maxLen:F.max_len,onSubmit:D=>{re(D)}},`output-${++G.current}`));break;case"elicit_response":ne(_.jsxs("div",{style:{marginBottom:"10px"},children:[_.jsx("strong",{children:"Response:"})," ",F.response]},`output-${++G.current}`));break;case"reveal":ne(_.jsxs("div",{style:{marginBottom:"10px"},children:[_.jsx("strong",{children:"Revealed:"})," ",JSON.stringify(F.values)]},`output-${++G.current}`));break;case"reward":let oe=0;typeof F.value=="number"?oe=F.value:F.value&&typeof F.value=="object"&&(oe=F.value.pairs.reduce((D,[W,Z])=>D+Z,0)),L(D=>[...D,{round:k,score:oe}]),j(oe),ne(_.jsx(Sq,{perTokenXent:F.value},`output-${++G.current}`)),ne(_.jsxs("div",{style:{marginBottom:"15px",padding:"10px 12px",backgroundColor:"#e8f5e9",borderRadius:"4px",borderLeft:"4px solid #4CAF50",fontSize:"14px",display:"flex",justifyContent:"space-between",alignItems:"center"},children:[_.jsxs("span",{children:[_.jsxs("strong",{children:["Round ",k," Score:"]})," ",oe.toFixed(2)]}),P.length>0&&_.jsxs("span",{style:{fontSize:"12px",color:"#666"},children:["Total rounds: ",P.length+1]})]},`output-${++G.current}`));break;case"failed_ensure":ne(_.jsx(Uw,{error:`Failed ensure. Moving execution to ${F.beacon} - Ensure Results: ${F.ensure_results?.join(", ")}`,isWarning:!0},`output-${++G.current}`));break;default:console.warn("Unknown event type:",F.type),ne(_.jsxs("div",{style:{marginBottom:"10px"},children:["Unknown event: ",JSON.stringify(F)]},`output-${++G.current}`))}}catch(I){console.error("Failed to parse message:",I)}},[]),se=A.useCallback(()=>{if(t.current?.readyState===WebSocket.OPEN)return;b("connecting");const B=new WebSocket("ws://localhost:8000/ws");t.current=B,B.onopen=()=>{console.log("WebSocket connection established"),b("connected"),ee.current&&(clearTimeout(ee.current),ee.current=null),B.readyState===WebSocket.OPEN&&B.send(JSON.stringify({type:"xent_configure",code:r}))},B.onmessage=I=>{ce(I)},B.onclose=I=>{b("disconnected"),S(!1),I.wasClean?console.log(`Connection closed cleanly, code=${I.code} reason=${I.reason}`):(console.error("Connection died unexpectedly"),ee.current||(ee.current=setTimeout(()=>{console.log("Attempting to reconnect..."),se()},3e3)))},B.onerror=I=>{console.error("WebSocket error:",I),b("disconnected")}},[r,ce]);return A.useEffect(()=>(se(),()=>{ee.current&&clearTimeout(ee.current),t.current&&t.current.close(1e3,"User navigated away")}),[se]),A.useEffect(()=>{const B=()=>{t.current&&t.current.readyState===WebSocket.OPEN&&t.current.close(1e3,"Page unloading")};return window.addEventListener("beforeunload",B),()=>window.removeEventListener("beforeunload",B)},[]),A.useEffect(()=>{o(r.split(`
`))},[r]),_.jsxs("div",{style:{padding:"20px",fontFamily:"system-ui, sans-serif",maxWidth:"1200px",margin:"0 auto"},children:[_.jsxs("div",{style:{display:"flex",justifyContent:"space-between",alignItems:"center",marginBottom:"20px"},children:[_.jsx("h1",{children:"XENT Interactive Play"}),_.jsxs("div",{style:{display:"flex",gap:"10px",alignItems:"center"},children:[g==="disconnected"&&_.jsxs(_.Fragment,{children:[_.jsx("span",{style:{padding:"5px 10px",borderRadius:"5px",backgroundColor:"#f44336",color:"white",fontSize:"12px"},children:"● Disconnected"}),_.jsx("button",{onClick:se,style:{padding:"6px 12px",backgroundColor:"#2196F3",color:"white",border:"none",borderRadius:"4px",cursor:"pointer",fontSize:"12px"},children:"Reconnect"})]}),_.jsx("button",{onClick:e,style:{padding:"8px 16px",backgroundColor:"#6c757d",color:"white",border:"none",borderRadius:"4px",cursor:"pointer"},children:"Back to Dashboard"})]})]}),_.jsxs("div",{style:{display:"grid",gridTemplateColumns:"1fr 1fr",gap:"20px",marginBottom:"20px"},children:[_.jsxs("div",{style:{backgroundColor:"#f8f9fa",padding:"15px",borderRadius:"5px"},children:[_.jsx("h3",{children:"Game Code"}),_.jsx("textarea",{value:r,onChange:B=>a(B.target.value),style:{width:"100%",minHeight:"200px",fontFamily:"monospace",fontSize:"12px",padding:"10px",border:"1px solid #ddd",borderRadius:"3px"},disabled:w}),_.jsx("div",{style:{marginTop:"10px",display:"flex",gap:"10px"},children:_.jsx("button",{onClick:he,disabled:g!=="connected"||w,style:{padding:"8px 16px",backgroundColor:"#4CAF50",color:"white",border:"none",borderRadius:"4px",cursor:g==="connected"&&!w?"pointer":"not-allowed",opacity:g==="connected"&&!w?1:.6},children:w?"Game Running...":"Start Game"})})]}),_.jsx(xq,{code:u,currentLine:s,registers:v,currentRound:k,totalRounds:P.length})]}),_.jsxs("div",{style:{backgroundColor:"#f8f9fa",padding:"15px",borderRadius:"5px"},children:[_.jsx("div",{style:{marginBottom:"10px"},children:_.jsx("h3",{style:{margin:0},children:"Game Output"})}),_.jsx("div",{style:{backgroundColor:"white",border:"1px solid #ddd",borderRadius:"3px",padding:"15px",minHeight:"200px",maxHeight:"400px",overflowY:"auto"},children:m.length===0?_.jsx("div",{style:{color:"#666",fontStyle:"italic"},children:w?"Game is running...":'Click "Start Game" to begin'}):m})]})]})}const Cv=`assign(s=story())
reveal(s)
elicit(x, 10)
assign(x1=remove_common_words(x, s)) # Remove any words in story from input text
reward(xed(s | x1))`,Aq=`from typing import Any
from xent.presentation.sdk import (
    ChatBuilder,
    format_elicit_request,
    format_elicit_response,
    format_reveal,
    format_reward,
    format_failed_ensure,
)

def present_turn(state, since_events, metadata, full_history=None, ctx=None):
    if ctx is None:
        ctx = {}

    b = ChatBuilder()

    # Send a one-time introduction/instructions
    if not ctx.get('intro_sent', False):
        b.user(
            "You are playing a text game. Provide your move inside <move></move> tags. Any other text will be ignored."
        )
        ctx['intro_sent'] = True

    # Summarize what happened since the last elicit
    for event in since_events:
        t = event.get('type')
        if t == 'elicit_request':
            b.user(format_elicit_request(event))
        elif t == 'elicit_response':
            b.user(format_elicit_response(event))
        elif t == 'reveal':
            b.user(format_reveal(event))
        elif t == 'reward':
            formatted, _ = format_reward(event)
            b.user("Score update:\\n" + str(formatted))
        elif t == 'failed_ensure':
            b.user(format_failed_ensure(event))
        else:
            b.user("Event: " + str(event))

    # Close with a concise instruction cue
    b.user("Now provide your next move inside <move></move> tags.")

    return b.render(), ctx`,qw=`from xent.presentation.sdk import (
    ChatBuilder,
    PresentationBuilder,
    extract_reveals,
    extract_rewards,
    format_reward,
    get_max_score,
    split_rounds,
)


def present_turn(state, since_events, metadata, full_history=None, ctx=None):
    if ctx is None:
        ctx = {}

    history = full_history if full_history is not None else since_events
    rounds = split_rounds(history)
    all_rewards = extract_rewards(history)
    best_score, _ = get_max_score(all_rewards) if all_rewards else (None, None)

    b = ChatBuilder()

    # One-time header and story
    if not ctx.get("intro_sent", False):
        builder = PresentationBuilder()
        header = f"""You are playing a text game that uses the cross-entropy function of an LLM ({metadata["judge_model"]}) to compute score.

<gameCode>
assign(s=story())
elicit(x, 10)
assign(x1=remove_common_words(x, s))
reveal(x1)
reward(xed(s | x1))
</gameCode>

You will be given a short story, and your task is to provide a prefix that makes that story as predictable as possible. Think of it as providing context that helps predict what comes next.

The scoring measures how much information your prefix provides about the story. Specifically, your score is the difference between the story's baseline cross-entropy and its cross-entropy given your prefix: \`Score = crossEntropy(story) - crossEntropy(story | prefix)\`. Your goal is to maximize this score. So you want to find a prefix that will most help an LLM predict the story.

After each attempt, you'll see your score and a per-token breakdown of the cross entropy difference. The game automatically restarts after each successful attempt, allowing you to continuously optimize your approach. Use the feedback from previous rounds to improve your score.  Your final score is the maximum score you obtain over {metadata["num_rounds_per_game"]} rounds, so you should not worry about decreasing temporarily your score.

You cannot use any words that appear in the story itself (regardless of case or punctuation). Your prefix is limited to 10 tokens maximum.

Provide your prefix in <move></move> tags. Any other text in your response will be ignored."""

        builder.add_header(header)
        builder.add_line("")
        builder.add_line("The story:")
        builder.start_section("story")
        builder.add_line(str(state.get("s", "")))
        builder.end_section()
        if len(rounds) == 1:
            builder.add_line("First round starting.")
            builder.add_line("")
        builder.add_line("Provide your prefix in <move></move> tags.")
        b.user(builder.render())
        ctx["intro_sent"] = True
        return b.render(), ctx

    # Subsequent turns: summarize last completed round and prompt
    builder = PresentationBuilder()

    # Find the most recent completed round (with a reward)
    last_completed_index = None
    for i in range(len(rounds) - 1, -1, -1):
        if extract_rewards(rounds[i]):
            last_completed_index = i
            break

    if last_completed_index is not None:
        round_events = rounds[last_completed_index]
        rewards = extract_rewards(round_events)

        builder.add_line(f"Round {last_completed_index}:")

        # Show only the normalized prefix revealed
        reveals = extract_reveals(round_events)
        if reveals:
            prefix = reveals[0]["values"].get("x1")
            if prefix is not None:
                builder.add_line(f"<prefix>{prefix}</prefix>")

        # Score
        if rewards:
            builder.start_section("score")
            builder.add_lines(format_reward(rewards[0])[0])
            builder.end_section()

        builder.add_line("")

    if best_score is not None:
        builder.add_line(f"Best score achieved: {best_score:.3f}")

    builder.add_line("")
    builder.add_line("Remember: You want to maximize your score. Higher is better!")
    builder.add_line("")
    builder.add_line("Provide your prefix in <move></move> tags.")

    b.user(builder.render())
    return b.render(), ctx
  `;function $w(){const e=new Date,t=e.getFullYear()+"-"+String(e.getMonth()+1).padStart(2,"0")+"-"+String(e.getDate()).padStart(2,"0")+"-"+String(e.getHours()).padStart(2,"0")+":"+String(e.getMinutes()).padStart(2,"0")+":"+String(e.getSeconds()).padStart(2,"0"),r=Math.abs(e.getTime()).toString(16).slice(-6);return`${t}-${r}`}function Eq(){const[e,t]=A.useState([{name:"black",id:"gpt-4o",player_type:"default",options:{model:"gpt-4o",provider:"openai"}}]),[r,a]=A.useState("gpt2"),[u,o]=A.useState($w()),[s,f]=A.useState("notrandom"),[v,h]=A.useState(5),[m,y]=A.useState(1),[g,b]=A.useState(!1),[w,S]=A.useState([]),[O,T]=A.useState([]),[M,j]=A.useState([]),[k,N]=A.useState(!0),[P,L]=A.useState("list"),[G,ee]=A.useState(null),[ne,J]=A.useState(null),[re,he]=A.useState(!1),[ce,se]=A.useState(!1),[B,I]=A.useState(!1),[F,oe]=A.useState(!1);A.useEffect(()=>{D()},[]),A.useEffect(()=>{(async()=>{try{const ue=await Zj();Array.isArray(ue)&&T(ue)}catch(ue){console.warn("Unable to load packaged games:",ue)}})()},[]);const D=async()=>{try{N(!0);const V=await fetch("/api/benchmarks");if(V.ok){const ue=await V.json();j(ue)}else console.error("Failed to fetch benchmarks")}catch(V){console.error("Error fetching benchmarks:",V)}finally{N(!1)}},Z=(()=>{let V;return g?V=w.length>0?w:[{name:"simple_game",code:Cv,presentation_function:qw}]:V=O&&O.length>0?O:[{name:"simple_game",code:Cv,presentation_function:qw}],{config_type:"condensed_xent_config",metadata:{benchmark_id:u,xent_version:"0.3.0",judge_model:r,num_rounds_per_game:v,seed:s},expansion_config:{num_maps_per_game:m,text_generation_config:{generator_type:"JUDGE",generator_config:{},max_length:50}},players:e,games:V}})(),ae=async V=>{if(V.preventDefault(),!F){oe(!0);try{const ue=await fetch("/api/config",{method:"POST",headers:{"Content-Type":"application/json"},body:JSON.stringify({config:Z})});if(ue.ok){const ge=await ue.json();ft(ge.benchmark_id),D()}else{const ge=await ue.json();alert(`Error: ${ge.detail||"Failed to store configuration"}`)}}catch(ue){console.error("Network error:",ue),alert("Network error: Could not connect to server")}finally{oe(!1)}}},fe=()=>{S(V=>[...V,{name:"New Game",code:Cv,presentation_function:Aq}])},Se=V=>{S(ue=>ue.filter((ge,te)=>te!==V))},ye=(V,ue,ge)=>{S(te=>te.map((Fe,je)=>je===V?{...Fe,[ue]:ge}:Fe))},ft=V=>{ee(V),L("dashboard"),J(null),se(!1),I(!1)},ze=()=>{L("list"),ee(null),J(null),se(!1),I(!1)};return A.useEffect(()=>{const V=()=>{B&&I(!1)};return document.addEventListener("click",V),()=>document.removeEventListener("click",V)},[B]),P==="play"?_.jsx(Oq,{onBack:()=>L("list")}):P==="dashboard"&&G?_.jsx(gq,{benchmarkId:G,onBack:ze}):_.jsxs("div",{style:{padding:"20px",fontFamily:"system-ui, sans-serif",maxWidth:"800px",margin:"0 auto"},children:[_.jsx("style",{children:`
        @keyframes spin { from { transform: rotate(0deg); } to { transform: rotate(360deg); } }
        `}),_.jsxs("div",{style:{display:"flex",justifyContent:"space-between",alignItems:"center",marginBottom:"20px"},children:[_.jsx("h1",{children:"XENT Benchmarks"}),_.jsx("button",{onClick:()=>L("play"),style:{padding:"10px 20px",backgroundColor:"#9C27B0",color:"white",border:"none",borderRadius:"4px",cursor:"pointer",fontSize:"14px",fontWeight:"bold"},children:"🎮 Play Game"})]}),_.jsxs("div",{style:{marginBottom:"30px",padding:"15px",backgroundColor:"#f8f9fa",borderRadius:"5px"},children:[_.jsx("h3",{children:"Saved Configurations"}),k?_.jsx("p",{style:{color:"#666"},children:"Loading benchmarks..."}):M.length===0?_.jsx("p",{style:{color:"#666"},children:"No saved benchmarks yet. Create one below!"}):_.jsx("ul",{style:{listStyle:"none",padding:0},children:M.map(V=>_.jsxs("li",{style:{marginBottom:"8px",padding:"8px",backgroundColor:"white",borderRadius:"3px",display:"flex",justifyContent:"space-between",alignItems:"center"},children:[_.jsx("span",{style:{fontFamily:"monospace",fontSize:"14px"},children:V}),_.jsx("button",{type:"button",onClick:()=>ft(V),style:{padding:"4px 12px",backgroundColor:"#3b82f6",color:"white",border:"none",borderRadius:"3px",cursor:"pointer",fontSize:"12px"},children:"Dashboard"})]},V))})]}),_.jsx("hr",{style:{margin:"30px 0",border:"none",borderTop:"1px solid #dee2e6"}}),_.jsx("h1",{children:"Create New Benchmark"}),_.jsxs("form",{onSubmit:ae,style:{marginTop:"20px"},children:[_.jsxs("fieldset",{style:{marginBottom:"20px",padding:"15px",border:"1px solid #ccc"},children:[_.jsx("legend",{children:_.jsx("strong",{children:"Basic Settings"})}),_.jsxs("div",{style:{marginBottom:"10px"},children:[_.jsx("label",{style:{display:"block",marginBottom:"5px"},children:"Judge Model:"}),_.jsx("input",{type:"text",value:r,onChange:V=>a(V.target.value),style:{width:"100%",padding:"5px"},placeholder:"gpt2"})]}),_.jsxs("div",{style:{marginBottom:"10px"},children:[_.jsx("label",{style:{display:"block",marginBottom:"5px"},children:"Rounds per Game:"}),_.jsx("input",{type:"number",min:"1",value:v,onChange:V=>h(parseInt(V.target.value)||1),style:{width:"100%",padding:"5px"}})]}),_.jsxs("div",{style:{marginBottom:"10px"},children:[_.jsx("label",{style:{display:"block",marginBottom:"5px"},children:"Seed:"}),_.jsx("input",{type:"text",value:s,onChange:V=>f(V.target.value||"notrandom"),style:{width:"100%",padding:"5px"},placeholder:"notrandom"})]}),_.jsxs("div",{style:{marginBottom:"10px"},children:[_.jsx("label",{style:{display:"block",marginBottom:"5px"},children:"Benchmark ID:"}),_.jsx("input",{type:"text",value:u,onChange:V=>o(V.target.value),style:{width:"100%",padding:"5px"}}),_.jsx("button",{type:"button",onClick:()=>o($w()),style:{marginTop:"5px",padding:"5px 10px",fontSize:"12px"},children:"Generate New ID"})]}),_.jsxs("div",{style:{marginBottom:"10px"},children:[_.jsx("label",{style:{display:"block",marginBottom:"5px"},children:"Maps per Game:"}),_.jsx("input",{type:"number",min:"1",value:m,onChange:V=>y(parseInt(V.target.value)||1),style:{width:"100%",padding:"5px"}})]})]}),_.jsxs("fieldset",{style:{marginBottom:"20px",padding:"15px",border:"1px solid #ccc"},children:[_.jsx("legend",{children:_.jsx("strong",{children:"Players"})}),_.jsx(p_,{embedded:!0,value:e,onChange:t})]}),_.jsxs("fieldset",{style:{marginBottom:"20px",padding:"15px",border:"1px solid #ccc"},children:[_.jsx("legend",{children:_.jsx("strong",{children:"Games"})}),_.jsxs("div",{style:{marginBottom:"15px"},children:[_.jsxs("label",{children:[_.jsx("input",{type:"radio",name:"gameType",checked:!g,onChange:()=>b(!1),style:{marginRight:"10px"}}),"Use Packaged Games"]}),!g&&_.jsx("div",{style:{marginLeft:"25px",fontSize:"12px",color:"#666",marginTop:"5px"},children:O.length>0?_.jsxs("div",{children:["Found ",O.length," packaged game",O.length!==1?"s":"",": ",O.map(V=>V.name).join(", ")]}):_.jsx("div",{children:"No packaged games found. Falling back to the default simple game."})})]}),_.jsx("div",{style:{marginBottom:"15px"},children:_.jsxs("label",{children:[_.jsx("input",{type:"radio",name:"gameType",checked:g,onChange:()=>b(!0),style:{marginRight:"10px"}}),"Define Custom Games"]})}),g&&_.jsxs("div",{children:[_.jsx("h4",{children:"Custom Games:"}),w.map((V,ue)=>_.jsxs("div",{style:{marginBottom:"15px",padding:"15px",backgroundColor:"#f5f5f5",border:"1px solid #ddd"},children:[_.jsxs("div",{style:{marginBottom:"10px"},children:[_.jsx("label",{style:{display:"block",marginBottom:"5px",fontWeight:"bold"},children:"Game Name:"}),_.jsx("input",{type:"text",value:V.name,onChange:ge=>ye(ue,"name",ge.target.value),style:{width:"100%",padding:"5px"},placeholder:"e.g., my_custom_game"})]}),_.jsxs("div",{style:{marginBottom:"10px"},children:[_.jsx("label",{style:{display:"block",marginBottom:"5px",fontWeight:"bold"},children:"Game Code (.xent DSL):"}),_.jsx("textarea",{value:V.code,onChange:ge=>ye(ue,"code",ge.target.value),style:{width:"100%",padding:"10px",minHeight:"150px",fontFamily:"monospace"},placeholder:"Enter game logic using XENT DSL..."})]}),_.jsxs("div",{style:{marginBottom:"10px"},children:[_.jsx("label",{style:{display:"block",marginBottom:"5px",fontWeight:"bold"},children:"Presentation Function (Python):"}),_.jsx("textarea",{value:V.presentation_function,onChange:ge=>ye(ue,"presentation_function",ge.target.value),style:{width:"100%",padding:"10px",minHeight:"150px",fontFamily:"monospace"},placeholder:"def present(state, history, metadata):..."})]}),_.jsx("button",{type:"button",onClick:()=>Se(ue),style:{padding:"5px 15px",backgroundColor:"#ff4444",color:"white",border:"none",cursor:"pointer"},children:"Remove Game"})]},ue)),_.jsx("button",{type:"button",onClick:fe,style:{padding:"5px 15px",backgroundColor:"#4CAF50",color:"white",border:"none",cursor:"pointer"},children:"Add Custom Game"}),w.length===0&&_.jsxs("div",{style:{padding:"10px",backgroundColor:"#fff3cd",border:"1px solid #ffeaa7",marginTop:"10px"},children:[_.jsx("strong",{children:"Note:"})," No custom games defined. The default simple game will be used instead."]})]})]}),_.jsx("div",{style:{display:"flex",gap:"10px"},children:_.jsxs("button",{type:"submit",disabled:F,style:{padding:"10px 20px",backgroundColor:F?"#7aaef0":"#2196F3",color:"white",border:"none",cursor:F?"default":"pointer",fontSize:"16px",display:"inline-flex",alignItems:"center",gap:"8px"},children:[F&&_.jsx("span",{style:{width:"14px",height:"14px",border:"2px solid rgba(255,255,255,0.6)",borderTopColor:"white",borderRadius:"50%",display:"inline-block",animation:"spin 1s linear infinite"}}),F?"Saving…":"Create Benchmark Configuration"]})})]}),_.jsxs("div",{style:{marginTop:"30px",padding:"15px",backgroundColor:"#f0f0f0",borderRadius:"5px"},children:[_.jsx("h3",{children:"Configuration Preview (JSON)"}),_.jsxs("p",{style:{fontSize:"12px",color:"#666",marginBottom:"10px"},children:["This is the condensed configuration that will be sent to the backend. It matches the output of ",_.jsx("code",{children:"xent configure"})," CLI command."]}),_.jsx("pre",{style:{backgroundColor:"white",padding:"10px",overflow:"auto",maxHeight:"400px",fontSize:"12px"},children:JSON.stringify(Z,null,2)})]})]})}Xj.createRoot(document.getElementById("root")).render(_.jsx(Hw.StrictMode,{children:_.jsx(Eq,{})}));
